#ifndef __SCE_H_
#define __SCE_H_

#include <stdint.h>

#include "sce_crypto.h"

#define SCE_LINK_END_ADDRESS (0)
#define SCE_LINK_END_SIZE (0xffffffffULL)

typedef struct {
	uint32_t size;
	uint32_t reserved;
	uint64_t address;
} sce_link_list_item_t;

#define CMD_SUCCESS (0x10000)
typedef struct {
	uint32_t state;
	uint32_t reserved;
	uint64_t tag;
} sce_rsp_t;

typedef void *sce_queue_handle;

typedef struct {
	uint8_t alg;
	uint8_t mode;
	uint8_t dma_mode;
	uint8_t key_length;
	uint32_t in_length;
	uint32_t out_length;
	uint64_t in;
	uint64_t out;
	uint64_t key;
	uint64_t tag;
} sce_hash_t;

typedef struct {
	uint8_t alg;
	uint8_t mode; // ECB CTR ...
	uint8_t op_mode; // ENC DEC
	uint8_t dma_mode;
	uint32_t in_length;
	uint32_t out_length;
	uint32_t tag_size;
	uint32_t aad_size;
	uint64_t in;
	uint64_t out;
	uint64_t key; // 16
	uint64_t iv;
	uint64_t tag;
} sce_sca_t;

typedef struct {
	uint8_t reserved[4];
	uint8_t mode;
	uint32_t length;
	uint64_t buf; // data buf, physical address.
	uint64_t tag;
} sce_trng_t;

typedef struct {
	uint64_t op_type; // SCE_HASH SCA, see SCE_ALGO_TYPE
	union {
		sce_hash_t hash;
		sce_sca_t sca;
		sce_trng_t trng;
	};
} sce_op_data_t;

/* memory VA to DMA address map */
typedef void *(*sce_map_io)(void *usr, void *va);

/* memory from user, it is given at lib init, used to io_map */
struct sce_mm_map {
	sce_map_io iova_map; /* get iova from user space VA */
	void *usr; /* data for the above operations */
};

#define SCE_MEM_MODE_IOVA 0
#define SCE_MEM_MODE_VA 1

typedef struct {
	uint32_t op_addr_mode;
	uint32_t reserved;
	struct sce_mm_map iomap;
} sce_lib_cfg_t;

int sce_request_queue(int type, sce_queue_handle *queue_handle);
int sce_release_queue(sce_queue_handle queue_handle);
int sce_init_queue(sce_queue_handle queue_handle);
int sce_stop_queue(sce_queue_handle queue_handle);

int sce_requset_dmabuf(uint64_t size, uint8_t **buf);
int sce_release_dmabuf(void *buf);
uint64_t sce_virt_to_io(void *vaddr);

int sce_enqueue(sce_queue_handle queue, void **op_data, int op_num);
int sce_dequeue(sce_queue_handle queue, sce_rsp_t *rsp, int max_num);

int sce_init_lib(sce_lib_cfg_t *cfg);
void sce_exit_lib(void);

int sce_get_rand_bytes(void *buf, uint32_t len, uint32_t mode);

int sce_get_rand_bytes_dma(void *buf, uint32_t len, uint32_t mode);

int sce_sm2_sign(uint8_t *e, const uint32_t e_len, uint8_t *priv_key,
		 const uint32_t priv_key_len, uint8_t *der_code_out,
		 uint32_t *der_code_out_len);
int sce_sm2_verify(uint8_t *e, const uint32_t e_len, uint8_t *pub_key,
		   const uint32_t pub_key_len, uint8_t *sign_result,
		   uint32_t sign_result_len);
int sce_sm2_dec(uint8_t *src, const uint32_t src_len, uint8_t *priv_key,
		const uint32_t priv_key_len, uint8_t *dst, uint32_t *dst_len);
int sce_sm2_enc(uint8_t *src, const uint32_t src_len, uint8_t *pub_key,
		const uint32_t pub_key_len, uint8_t *dst, uint32_t *dst_len);

void sce_sm2signature_encode(uint8_t *input, uint8_t *sig, uint32_t *sig_len);

int sce_sm2signature_decode(uint8_t *sig, uint32_t sig_len, uint8_t *out);

void sce_sm2encrypt_encode(uint8_t *input, uint32_t inlen, uint8_t *output,
			   uint32_t *outlen);

int sce_sm2decrypt_decode(uint8_t *input, uint32_t inlen, uint8_t *output,
			  uint32_t *outlen);

#endif
