#include <linux/module.h>
#include <linux/irq.h>
#include <linux/platform_device.h>
#include <linux/dma-mapping.h>
#include <linux/cdev.h>
#include <linux/types.h>
#include <linux/mutex.h>
#include <linux/acpi.h>
#include <crypto/internal/hash.h>
#include <linux/crypto.h>
#include <crypto/hash.h>
#include <crypto/internal/skcipher.h>
#include <crypto/internal/akcipher.h>
#include <linux/ioport.h>
#include <linux/io.h>
#include <linux/scatterlist.h>
#include <linux/timex.h>
#include <linux/kthread.h>
#include <linux/delay.h>
#include <linux/sched/signal.h>

#include "sm3_test.h"
#include "sm3_software.h"
#include "test.h"

#define TVMEMSIZE 4

#define MAX_DIGEST_SIZE 64

static char *tvmem[TVMEMSIZE];
static unsigned int klen;
static void test_hash_sg_init(struct scatterlist *sg)
{
	int i;

	sg_init_table(sg, TVMEMSIZE);
	for (i = 0; i < TVMEMSIZE; i++) {
		sg_set_buf(sg + i, tvmem[i], PAGE_SIZE);
		memset(tvmem[i], 0xff, PAGE_SIZE);
	}
}

static inline int do_one_ahash_op(struct ahash_request *req, int ret)
{
	struct crypto_wait *wait = req->base.data;

	return crypto_wait_req(ret, wait);
}

static int test_ahash_jiffies_digest(struct ahash_request *req, int blen,
				     char *out, int secs)
{
	unsigned long start, end;
	int bcount;
	int ret;

	for (start = jiffies, end = start + secs * HZ, bcount = 0;
	     time_before(jiffies, end); bcount++) {
		ret = do_one_ahash_op(req, crypto_ahash_digest(req));
		if (ret)
			return ret;
	}
	printk("%6u opers/sec, %9lu bytes/sec %lu Mbps\n", bcount / secs,
	       ((long)bcount * blen) / secs,
	       ((long)bcount * blen * 8) / (secs * 1000000));

	return 0;
}

static int test_ahash_cycles_digest(struct ahash_request *req, int blen,
				    char *out)
{
	unsigned long cycles = 0;
	int ret, i;

	/* Warm-up run. */
	for (i = 0; i < 4; i++) {
		ret = do_one_ahash_op(req, crypto_ahash_digest(req));
		if (ret)
			goto out;
	}

	/* The real thing. */
	for (i = 0; i < 8; i++) {
		cycles_t start, end;

		start = get_cycles();

		ret = do_one_ahash_op(req, crypto_ahash_digest(req));
		if (ret)
			goto out;

		end = get_cycles();

		cycles += end - start;
	}

out:
	if (ret)
		return ret;

	pr_cont("%6lu cycles/operation, %4lu cycles/byte\n", cycles / 8,
		cycles / (8 * blen));

	return 0;
}

static int test_ahash_jiffies(struct ahash_request *req, int blen, int plen,
			      char *out, int secs)
{
	(void)plen;
	return test_ahash_jiffies_digest(req, blen, out, secs);
}

static int test_ahash_cycles(struct ahash_request *req, int blen, int plen,
			     char *out)
{
	(void)plen;

	return test_ahash_cycles_digest(req, blen, out);
}

static void test_ahash_speed_common(const char *algo, unsigned int secs,
				    struct hash_speed *speed, unsigned mask)
{
	struct scatterlist sg[TVMEMSIZE];
	struct crypto_wait wait;
	struct ahash_request *req;
	struct crypto_ahash *tfm;
	char *output;
	int i, ret;

	tfm = crypto_alloc_ahash(algo, 0, mask);
	if (IS_ERR(tfm)) {
		pr_err("failed to load transform for %s: %ld\n", algo,
		       PTR_ERR(tfm));
		return;
	}

	printk(KERN_INFO "\ntesting speed of async %s (%s)\n", algo,
	       get_driver_name(crypto_ahash, tfm));

	if (crypto_ahash_digestsize(tfm) > MAX_DIGEST_SIZE) {
		pr_err("digestsize(%u) > %d\n", crypto_ahash_digestsize(tfm),
		       MAX_DIGEST_SIZE);
		goto out;
	}

	test_hash_sg_init(sg);
	req = ahash_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_err("ahash request allocation failure\n");
		goto out;
	}

	crypto_init_wait(&wait);
	ahash_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG,
				   crypto_req_done, &wait);

	output = kmalloc(MAX_DIGEST_SIZE, GFP_KERNEL);
	if (!output)
		goto out_nomem;

	for (i = 0; speed[i].blen != 0; i++) {
		if (speed[i].blen > TVMEMSIZE * PAGE_SIZE) {
			pr_err("template (%u) too big for tvmem (%lu)\n",
			       speed[i].blen, TVMEMSIZE * PAGE_SIZE);
			break;
		}

		if (klen)
			crypto_ahash_setkey(tfm, tvmem[0], klen);

		pr_info("test%3u "
			"(%5u byte blocks,%5u bytes per update,%4u updates): ",
			i, speed[i].blen, speed[i].plen,
			speed[i].blen / speed[i].plen);

		ahash_request_set_crypt(req, sg, output, speed[i].plen);

		if (secs) {
			ret = test_ahash_jiffies(req, speed[i].blen,
						 speed[i].plen, output, secs);
			cond_resched();
		} else {
			ret = test_ahash_cycles(req, speed[i].blen,
						speed[i].plen, output);
		}

		if (ret) {
			pr_err("hashing failed ret=%d\n", ret);
			break;
		}
	}

	kfree(output);

out_nomem:
	ahash_request_free(req);

out:
	crypto_free_ahash(tfm);
}

void sm3_speed(u32 secs)
{
	int i;
	for (i = 0; i < TVMEMSIZE; i++) {
		tvmem[i] = (void *)__get_free_page(GFP_KERNEL);
		if (!tvmem[i])
			goto err_free_tv;
	}

	test_ahash_speed_common(alg, secs, generic_hash_speed_template, 0);

err_free_tv:
	for (i = 0; i < TVMEMSIZE && tvmem[i]; i++)
		free_page((unsigned long)tvmem[i]);
}

void hmac_sm3_speed(u32 secs)
{
	int i;
	klen = 16;

	for (i = 0; i < TVMEMSIZE; i++) {
		tvmem[i] = (void *)__get_free_page(GFP_KERNEL);
		if (!tvmem[i])
			goto err_free_tv;
	}

	test_ahash_speed_common(alg, secs, generic_hash_speed_template, 0);

err_free_tv:
	for (i = 0; i < TVMEMSIZE && tvmem[i]; i++)
		free_page((unsigned long)tvmem[i]);
}

static int sm3_dgst_test(struct crypto_ahash *tfm, const u8 *data,
			 unsigned int len, u8 *out)
{
	struct scatterlist sg;
	struct crypto_wait wait;
	struct ahash_request *req;
	int ret = -1;
	static u32 flag = 0;

	if (flag++ == 0)
		pr_info("\ntesting of async %s (%s)\n", alg,
			get_driver_name(crypto_ahash, tfm));

	sg_init_table(&sg, 1);
	sg_set_buf(&sg, data, len);

	req = ahash_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_err("ahash request allocation failure\n");
		goto fail;
	}

	crypto_init_wait(&wait);
	ahash_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG,
				   crypto_req_done, &wait);

	ahash_request_set_crypt(req, &sg, out, len);

	ret = crypto_wait_req(crypto_ahash_digest(req), &wait);
	if (ret) {
		pr_err("hashing failed ret=%d\n", ret);
	}

	ahash_request_free(req);
	ret = 0;
fail:

	return ret;
}

static int sm3_step_test(struct crypto_ahash *tfm, const u8 *data,
			 unsigned int len, u8 *out)
{
	struct scatterlist sg;
	struct crypto_wait wait;
	struct ahash_request *req;
	int ret = -1;
	unsigned int offset, step_len;
	static u32 flag = 0;

	if (flag++ == 0)
		pr_info("\ntesting of async %s (%s)\n", alg,
			get_driver_name(crypto_ahash, tfm));

	req = ahash_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_err("ahash request allocation failure\n");
		goto fail;
	}

	crypto_init_wait(&wait);
	ahash_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG,
				   crypto_req_done, &wait);
	ret = crypto_wait_req(crypto_ahash_init(req), &wait);
	if (ret) {
		pr_err("hashing failed ret=%d\n", ret);
		goto fail1;
	}

	for (offset = 0; offset < len;) {
		step_len = 0xFF;
		if (offset + step_len > len)
			step_len = len - offset;

		crypto_init_wait(&wait);
		ahash_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG,
					   crypto_req_done, &wait);

		sg_init_table(&sg, 1);
		sg_set_buf(&sg, data + offset, step_len);
		ahash_request_set_crypt(req, &sg, out, step_len);

		ret = crypto_wait_req(crypto_ahash_update(req), &wait);
		if (ret) {
			pr_err("crypto_ahash_update failed ret=%d\n", ret);
			break;
		}
		offset += step_len;
	}

	crypto_init_wait(&wait);
	ahash_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG,
				   crypto_req_done, &wait);

	sg_init_table(&sg, 1);
	sg_set_buf(&sg, data, 0);
	ahash_request_set_crypt(req, &sg, out, 0);
	ret = crypto_wait_req(crypto_ahash_final(req), &wait);
	if (ret) {
		pr_err("crypto_ahash_final failed ret=%d\n", ret);
	}

fail1:
	ahash_request_free(req);
fail:

	return ret;
}

int hmac_sm3_test(struct crypto_ahash *tfm, const u8 *data, unsigned int len,
		  u8 *out)
{
	struct scatterlist sg;
	struct crypto_wait wait;
	struct ahash_request *req;
	int ret = -1;
	static u32 flag = 0;

	if (IS_ERR(tfm)) {
		pr_err("failed to load transform for %s: %ld\n", alg,
		       PTR_ERR(tfm));
		return ret;
	}

	if (flag++ == 0)
		pr_info("\ntesting of async %s (%s)\n", alg,
			get_driver_name(crypto_ahash, tfm));

	sg_init_table(&sg, 1);
	sg_set_buf(&sg, data, len);

	req = ahash_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_err("ahash request allocation failure\n");
		goto exit;
	}

	crypto_init_wait(&wait);
	ahash_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG,
				   crypto_req_done, &wait);

	ahash_request_set_crypt(req, &sg, out, len);

	ret = crypto_wait_req(crypto_ahash_digest(req), &wait);
	if (ret) {
		pr_err("hashing failed ret=%d\n", ret);
	}
exit:
	ahash_request_free(req);

	return ret;
}

static int hmac_sm3_step_test(struct crypto_ahash *tfm, const u8 *data,
			      unsigned int len, u8 *out)
{
	struct scatterlist sg;
	struct crypto_wait wait;
	struct ahash_request *req;
	int ret = -1;
	unsigned int offset, step_len;
	static u32 flag = 0;

	if (IS_ERR(tfm)) {
		pr_err("failed to load transform for %s: %ld\n", alg,
		       PTR_ERR(tfm));
		return ret;
	}

	if (flag++ == 0)
		pr_info("\ntesting of async %s (%s)\n", alg,
			get_driver_name(crypto_ahash, tfm));

	req = ahash_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_err("ahash request allocation failure\n");
		goto exit;
	}

	crypto_init_wait(&wait);
	ahash_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG,
				   crypto_req_done, &wait);

	ret = crypto_wait_req(crypto_ahash_init(req), &wait);
	if (ret) {
		pr_err("hashing failed ret=%d\n", ret);
		goto exit;
	}

	for (offset = 0; offset < len;) {
		step_len = 0xFF;
		if (offset + step_len > len)
			step_len = len - offset;

		crypto_init_wait(&wait);
		ahash_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG,
					   crypto_req_done, &wait);

		sg_init_table(&sg, 1);
		sg_set_buf(&sg, data + offset, step_len);
		ahash_request_set_crypt(req, &sg, out, step_len);

		ret = crypto_wait_req(crypto_ahash_update(req), &wait);
		if (ret) {
			pr_err("crypto_ahash_update failed ret=%d\n", ret);
			break;
		}
		offset += step_len;
	}

	crypto_init_wait(&wait);
	ahash_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG,
				   crypto_req_done, &wait);

	sg_init_table(&sg, 1);
	sg_set_buf(&sg, data, 0);
	ahash_request_set_crypt(req, &sg, out, 0);
	ret = crypto_wait_req(crypto_ahash_final(req), &wait);
	if (ret) {
		pr_err("crypto_ahash_final failed ret=%d\n", ret);
	}

exit:
	ahash_request_free(req);

	return ret;
}

struct kthread_sm3_data {
	struct crypto_ahash *tfm;
	u8 *key;
	u32 keylen;
};

static int sm3_test_vs_software(struct kthread_sm3_data *thread_data)
{
	u8 *src;
	u8 kdgst[32], sdgst[32];
	u32 len;
	struct sm3_context ctx;
	struct sm3_func func_sm3;
	u64 time = jiffies;
	int ret = -1;
	int err = 0;

	sm3_func_alloc(&func_sm3);

	src = kmalloc(elen + 64, GFP_KERNEL);
	if (src == NULL) {
		pr_err("%s:kmalloc failed!\n", __func__);
		return -1;
	}

	get_random_bytes(src, elen);

	for (len = slen; len <= elen; len++) {
		ret = sm3_dgst_test(thread_data->tfm, src, len, kdgst);
		if (ret) {
			err++;
			break;
		}

		func_sm3.sm3_digest(&ctx, src, len, sdgst);
		if (memcmp(kdgst, sdgst, 32)) {
			pr_err("sm3_dgst_test err!\n");
			print_buf("key", thread_data->key, thread_data->keylen);
			print_buf("src", src, len);
			print_buf("kdgst", kdgst, 32);
			print_buf("sdgst", sdgst, 32);
			err++;
			break;
		}
		ret = sm3_step_test(thread_data->tfm, src, len, kdgst);
		if (ret) {
			err++;
			break;
		}

		if (memcmp(kdgst, sdgst, 32)) {
			pr_err("sm3_step_test err!\n");
			print_buf("key", thread_data->key, thread_data->keylen);
			print_buf("src", src, len);
			print_buf("kdgst", kdgst, 32);
			print_buf("sdgst", sdgst, 32);
			err++;
			break;
		}

		if (time + msecs_to_jiffies(100) < jiffies) {
			time = jiffies;
			schedule();
		}
	}

	kfree(src);

	return err;
}

static int sm3_thread_func(void *data)
{
	struct kthread_sm3_data *thread_data = (struct kthread_sm3_data *)data;
	int ret = 0;
	char name[TASK_COMM_LEN] = { 0 };

	while (1) {
		if (kthread_should_stop())
			break;

		if (ret == 0)
			ret = sm3_test_vs_software(thread_data);
		else
			mdelay(100);
	}

	if (ret == 0) {
		get_task_comm(name, current);
		pr_info("Success to exit  thread %s\n", name);
	}

	return 0;
}

void sm3_test(void)
{
	struct task_struct **kthreads;
	u32 kthread_num = multi ? multi : 1;
	int i;
	struct kthread_sm3_data *thread_data;

	unsigned long start, end;

	kthreads = kmalloc(kthread_num * sizeof(struct task_struct *),
			   GFP_KERNEL | __GFP_ZERO);
	if (kthreads == NULL) {
		pr_err("malloc error\n");
		return;
	}

	thread_data = kmalloc(sizeof(struct kthread_sm3_data), GFP_KERNEL);
	if (thread_data == NULL) {
		kfree(kthreads);
		pr_err("malloc error\n");
		return;
	}

	thread_data->keylen = 0;
	thread_data->key = NULL;

	thread_data->tfm = crypto_alloc_ahash(alg, 0, 0);
	if (IS_ERR(thread_data->tfm)) {
		pr_err("failed to load transform for %s: %ld\n", alg,
		       PTR_ERR(thread_data->tfm));

		kfree(thread_data);
		kfree(kthreads);
		return;
	}

	for (i = 0; i < kthread_num; i++) {
		kthreads[i] = kthread_run(sm3_thread_func, (void *)thread_data,
					  "sm3-%d", i);
	}

	start = jiffies;
	end = start + secs * HZ;

	/* run for secs seconds */
	while (time_before(jiffies, end)) {
		schedule();
	}

	for (i = 0; i < kthread_num; i++) {
		kthread_stop(kthreads[i]);
	}

	mdelay(1000);
	crypto_free_ahash(thread_data->tfm);
	kfree(thread_data->key);
	kfree(thread_data);
	kfree(kthreads);
}

static int hmac_test_vs_software(struct kthread_sm3_data *thread_data)
{
	u8 *src;
	u8 kdgst[32], sdgst[32];
	u32 len;
	struct sm3_context ctx;
	struct sm3_func func_sm3;
	u64 time = jiffies;
	int ret = -1;
	int err = 0;

	sm3_func_alloc(&func_sm3);

	func_sm3.hmac_sm3_setkey(&ctx, thread_data->key, thread_data->keylen);

	src = kmalloc(elen + 64, GFP_KERNEL);
	if (src == NULL) {
		pr_err("%s:kmalloc failed!\n", __func__);
		return -1;
	}

	get_random_bytes(src, elen);
	for (len = slen; len <= elen; len++) {
		ret = hmac_sm3_test(thread_data->tfm, src, len, kdgst);
		if (ret) {
			err++;
			break;
		}

		func_sm3.hmac_sm3_init(&ctx);
		func_sm3.hmac_sm3_finup(&ctx, src, len, sdgst);
		if (memcmp(kdgst, sdgst, 32)) {
			pr_err("hmac_sm3_test err!\n");
			print_buf("key", thread_data->key, thread_data->keylen);
			print_buf("src", src, len);
			print_buf("kdgst", kdgst, 32);
			print_buf("sdgst", sdgst, 32);
			err++;
			break;
		}
		ret = hmac_sm3_step_test(thread_data->tfm, src, len, kdgst);
		if (ret) {
			err++;
			break;
		}

		if (memcmp(kdgst, sdgst, 32)) {
			pr_err("hmac_sm3_step_test err!\n");
			print_buf("key", thread_data->key, thread_data->keylen);
			print_buf("src", src, len);
			print_buf("kdgst", kdgst, 32);
			print_buf("sdgst", sdgst, 32);
			err++;
			break;
		}

		if (time + msecs_to_jiffies(100) < jiffies) {
			time = jiffies;
			schedule();
		}
	}

	kfree(src);

	return err;
}

static int hmac_sm3_thread_func(void *data)
{
	struct kthread_sm3_data *thread_data = (struct kthread_sm3_data *)data;
	int ret = 0;
	char name[TASK_COMM_LEN] = { 0 };

	while (1) {
		if (kthread_should_stop())
			break;
		if (ret == 0)
			ret = hmac_test_vs_software(thread_data);
		else
			mdelay(100);
	}

	if (ret == 0) {
		get_task_comm(name, current);
		pr_info("Success to exit   thread %s\n", name);
	}

	return 0;
}

void hmac_test(void)
{
	struct task_struct **kthreads;
	u32 kthread_num = multi ? multi : 1;
	int i;
	struct kthread_sm3_data *thread_data;
	u32 keylen = 64;
	unsigned long start, end;

	kthreads = kmalloc(kthread_num * sizeof(struct task_struct *),
			   GFP_KERNEL | __GFP_ZERO);
	if (kthreads == NULL) {
		pr_err("malloc error\n");
		return;
	}

	thread_data = kmalloc(sizeof(struct kthread_sm3_data), GFP_KERNEL);
	if (thread_data == NULL) {
		kfree(kthreads);
		pr_err("malloc error\n");
		return;
	}

	thread_data->keylen = keylen;
	thread_data->key = kmalloc(keylen, GFP_KERNEL);
	if (thread_data->key == NULL) {
		kfree(thread_data);
		kfree(kthreads);
		pr_err("malloc error\n");
		return;
	}

	thread_data->tfm = crypto_alloc_ahash(alg, 0, 0);
	if (IS_ERR(thread_data->tfm)) {
		pr_err("failed to load transform for %s: %ld\n", alg,
		       PTR_ERR(thread_data->tfm));
		kfree(thread_data->key);
		kfree(thread_data);
		kfree(kthreads);
		return;
	}

	get_random_bytes(thread_data->key, thread_data->keylen);
	crypto_ahash_setkey(thread_data->tfm, thread_data->key,
			    thread_data->keylen);

	for (i = 0; i < kthread_num; i++) {
		kthreads[i] = kthread_run(hmac_sm3_thread_func,
					  (void *)thread_data, "hmac-%d", i);
	}

	start = jiffies;
	end = start + secs * HZ;

	/* run for secs seconds */
	while (time_before(jiffies, end)) {
		schedule();
	}

	for (i = 0; i < kthread_num; i++) {
		kthread_stop(kthreads[i]);
	}

	mdelay(1000);
	crypto_free_ahash(thread_data->tfm);
	kfree(thread_data->key);
	kfree(thread_data);
	kfree(kthreads);
}
