#include <linux/module.h>
#include <linux/irq.h>
#include <linux/platform_device.h>
#include <linux/dma-mapping.h>
#include <linux/cdev.h>
#include <linux/types.h>
#include <linux/mutex.h>
#include <linux/acpi.h>
#include <crypto/internal/hash.h>
#include <linux/crypto.h>
#include <crypto/hash.h>
#include <crypto/internal/skcipher.h>
#include <crypto/internal/akcipher.h>
#include <linux/ioport.h>
#include <linux/io.h>
#include <linux/delay.h>
#include <linux/kthread.h>

#include "sm4_test.h"
#include "sm4_software.h"
#include "test.h"

#define ENCRYPT 1
#define DECRYPT 0
#define TVMEMSIZE 4
static char *tvmem[TVMEMSIZE];

static inline int do_one_acipher_op(struct skcipher_request *req, int ret)
{
	struct crypto_wait *wait = req->base.data;

	return crypto_wait_req(ret, wait);
}

static int test_acipher_jiffies(struct skcipher_request *req, int enc, int blen,
				int secs)
{
	unsigned long start, end;
	int bcount;
	int ret = -1;

	for (start = jiffies, end = start + secs * HZ, bcount = 0;
	     time_before(jiffies, end); bcount++) {
		if (enc)
			ret = do_one_acipher_op(req,
						crypto_skcipher_encrypt(req));
		else
			ret = do_one_acipher_op(req,
						crypto_skcipher_decrypt(req));

		if (ret)
			return ret;
	}

	pr_cont("%d operations in %d seconds (%llu bytes) %llu Mbps\n", bcount,
		secs, (u64)bcount * blen,
		((u64)bcount * blen * 8) / (secs * 1000000));

	return 0;
}

static int test_acipher_cycles(struct skcipher_request *req, int enc, int blen)
{
	unsigned long cycles = 0;
	int ret = 0;
	int i;

	/* Warm-up run. */
	for (i = 0; i < 4; i++) {
		if (enc)
			ret = do_one_acipher_op(req,
						crypto_skcipher_encrypt(req));
		else
			ret = do_one_acipher_op(req,
						crypto_skcipher_decrypt(req));

		if (ret)
			goto out;
	}

	/* The real thing. */
	for (i = 0; i < 8; i++) {
		cycles_t start, end;

		start = get_cycles();
		if (enc)
			ret = do_one_acipher_op(req,
						crypto_skcipher_encrypt(req));
		else
			ret = do_one_acipher_op(req,
						crypto_skcipher_decrypt(req));
		end = get_cycles();

		if (ret)
			goto out;

		cycles += end - start;
	}

out:
	if (ret == 0)
		pr_cont("1 operation in %lu cycles (%d bytes)\n",
			(cycles + 4) / 8, blen);

	return ret;
}

static void test_skcipher_speed(const char *algo, int enc, unsigned int secs,
				struct cipher_speed_template *template,
				unsigned int tcount, u8 *keysize, bool async)
{
	unsigned int ret, i, j, k, iv_len;
	struct crypto_wait wait;
	const char *key;
	char iv[128];
	struct skcipher_request *req;
	struct crypto_skcipher *tfm;
	const int *b_size;
	const char *e;

	if (enc == ENCRYPT)
		e = "encryption";
	else
		e = "decryption";

	crypto_init_wait(&wait);

	tfm = crypto_alloc_skcipher(algo, 0, async ? 0 : CRYPTO_ALG_ASYNC);

	if (IS_ERR(tfm)) {
		pr_err("failed to load transform for %s: %ld\n", algo,
		       PTR_ERR(tfm));
		return;
	}

	pr_info("\ntesting speed of %s %s (%s) %s\n", async ? "async" : "sync",
		algo, get_driver_name(crypto_skcipher, tfm), e);

	req = skcipher_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_err("tcrypt: skcipher: Failed to allocate request for %s\n",
		       algo);
		goto out;
	}

	skcipher_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG,
				      crypto_req_done, &wait);

	i = 0;
	do {
		b_size = block_sizes;

		do {
			u32 bs = round_up(*b_size,
					  crypto_skcipher_blocksize(tfm));
			struct scatterlist sg[TVMEMSIZE];

			if ((*keysize + bs) > TVMEMSIZE * PAGE_SIZE) {
				pr_err("template (%u) too big for "
				       "tvmem (%lu)\n",
				       *keysize + bs, TVMEMSIZE * PAGE_SIZE);
				goto out_free_req;
			}

			pr_info("test %u (%d bit key, %d byte blocks): ", i,
				*keysize * 8, bs);

			memset(tvmem[0], 0xff, PAGE_SIZE);

			/* set key, plain text and IV */
			key = tvmem[0];
			for (j = 0; j < tcount; j++) {
				if (template[j].klen == *keysize) {
					key = template[j].key;
					break;
				}
			}

			crypto_skcipher_clear_flags(tfm, ~0);

			ret = crypto_skcipher_setkey(tfm, key, *keysize);
			if (ret) {
				pr_err("setkey() failed flags=%x\n",
				       crypto_skcipher_get_flags(tfm));
				goto out_free_req;
			}

			k = *keysize + bs;
			sg_init_table(sg, DIV_ROUND_UP(k, PAGE_SIZE));

			if (k > PAGE_SIZE) {
				sg_set_buf(sg, tvmem[0] + *keysize,
					   PAGE_SIZE - *keysize);
				k -= PAGE_SIZE;
				j = 1;
				while (k > PAGE_SIZE) {
					sg_set_buf(sg + j, tvmem[j], PAGE_SIZE);
					memset(tvmem[j], 0xff, PAGE_SIZE);
					j++;
					k -= PAGE_SIZE;
				}
				sg_set_buf(sg + j, tvmem[j], k);
				memset(tvmem[j], 0xff, k);
			} else {
				sg_set_buf(sg, tvmem[0] + *keysize, bs);
			}

			iv_len = crypto_skcipher_ivsize(tfm);
			if (iv_len)
				memset(&iv, 0xff, iv_len);

			skcipher_request_set_crypt(req, sg, sg, bs, iv);

			if (secs) {
				ret = test_acipher_jiffies(req, enc, bs, secs);
				cond_resched();
			} else {
				ret = test_acipher_cycles(req, enc, bs);
			}

			if (ret) {
				pr_err("%s() failed flags=%x\n", e,
				       crypto_skcipher_get_flags(tfm));
				break;
			}
			b_size++;
			i++;
		} while (*b_size);
		keysize++;
	} while (*keysize);

out_free_req:
	skcipher_request_free(req);
out:
	crypto_free_skcipher(tfm);
}

static void test_cipher_speed(const char *algo, int enc, unsigned int secs,
			      struct cipher_speed_template *template,
			      unsigned int tcount, u8 *keysize)
{
	return test_skcipher_speed(algo, enc, secs, template, tcount, keysize,
				   false);
}

void sm4_speed_test(char *alg, int secs)
{
	int i;

	for (i = 0; i < TVMEMSIZE; i++) {
		tvmem[i] = (void *)__get_free_page(GFP_KERNEL);
		if (!tvmem[i])
			goto err_free_tv;
	}

	test_cipher_speed(alg, ENCRYPT, secs, NULL, 0, speed_template_16);
	test_cipher_speed(alg, DECRYPT, secs, NULL, 0, speed_template_16);

err_free_tv:
	for (i = 0; i < TVMEMSIZE && tvmem[i]; i++)
		free_page((unsigned long)tvmem[i]);
}

static int sm4_ecb_enc_test(const u8 *key, const u8 *data, unsigned int len,
			    u8 *out)
{
	struct crypto_wait wait;
	struct skcipher_request *req;
	struct crypto_skcipher *tfm;
	struct scatterlist ssg, dsg;
	static u32 flag = 0;
	int ret = -1;

	tfm = crypto_alloc_skcipher(alg, 0, 0);
	if (IS_ERR(tfm)) {
		pr_err("failed to load transform for %s: %ld\n", alg,
		       PTR_ERR(tfm));
		return ret;
	}

	if (flag++ == 0)
		pr_info("\ntesting of async %s (%s) enc!\n", alg,
			get_driver_name(crypto_skcipher, tfm));

	req = skcipher_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_info("skcipher: Failed to allocate request for %s\n", alg);
		goto fail;
	}

	crypto_init_wait(&wait);
	skcipher_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG,
				      crypto_req_done, &wait);

	crypto_skcipher_clear_flags(tfm, ~0);
	crypto_skcipher_setkey(tfm, key, 16);

	sg_init_table(&ssg, 1);
	sg_set_buf(&ssg, data, len);
	sg_init_table(&dsg, 1);
	sg_set_buf(&dsg, out, len);
	skcipher_request_set_crypt(req, &ssg, &dsg, len, NULL);

	ret = crypto_wait_req(crypto_skcipher_encrypt(req), &wait);
	if (ret) {
		pr_err("crypto_skcipher_encrypt failed ret=%d\n", ret);
	}

	skcipher_request_free(req);
fail:
	crypto_free_skcipher(tfm);

	return ret;
}

static int sm4_ecb_enc_step_test(const u8 *key, const u8 *data,
				 unsigned int len, u8 *out)
{
	struct crypto_wait wait;
	struct skcipher_request *req;
	struct crypto_skcipher *tfm;
	struct scatterlist ssg, dsg;
	static u32 flag = 0;
	int ret = -1;
	u32 offset, step_len;

	tfm = crypto_alloc_skcipher(alg, 0, 0);
	if (IS_ERR(tfm)) {
		pr_err("failed to load transform for %s: %ld\n", alg,
		       PTR_ERR(tfm));
		return ret;
	}

	if (flag++ == 0)
		pr_info("\ntesting of async %s (%s) enc!\n", alg,
			get_driver_name(crypto_skcipher, tfm));

	req = skcipher_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_info("skcipher: Failed to allocate request for %s\n", alg);
		goto fail;
	}

	crypto_skcipher_clear_flags(tfm, ~0);
	crypto_skcipher_setkey(tfm, key, 16);

	for (offset = 0; offset < len;) {
		step_len = 256;
		if (offset + step_len > len)
			step_len = len - offset;

		crypto_init_wait(&wait);
		skcipher_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG,
					      crypto_req_done, &wait);

		sg_init_table(&ssg, 1);
		sg_set_buf(&ssg, data + offset, step_len);
		sg_init_table(&dsg, 1);
		sg_set_buf(&dsg, out + offset, step_len);
		skcipher_request_set_crypt(req, &ssg, &dsg, step_len, NULL);

		ret = crypto_wait_req(crypto_skcipher_encrypt(req), &wait);
		if (ret) {
			pr_err("crypto_skcipher_encrypt failed ret=%d\n", ret);
			break;
		}
		offset += step_len;
	}

	skcipher_request_free(req);
fail:
	crypto_free_skcipher(tfm);

	return ret;
}

static int sm4_ecb_dec_test(const u8 *key, const u8 *data, unsigned int len,
			    u8 *out)
{
	struct crypto_wait wait;
	struct skcipher_request *req;
	struct crypto_skcipher *tfm;
	struct scatterlist ssg, dsg;
	static u32 flag = 0;
	int ret = -1;

	tfm = crypto_alloc_skcipher(alg, 0, 0);
	if (IS_ERR(tfm)) {
		pr_err("failed to load transform for %s: %ld\n", alg,
		       PTR_ERR(tfm));
		return ret;
	}

	if (flag++ == 0)
		pr_info("\ntesting of async %s (%s) dec!\n", alg,
			get_driver_name(crypto_skcipher, tfm));

	req = skcipher_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_info("skcipher: Failed to allocate request for %s\n", alg);
		goto fail;
	}

	crypto_init_wait(&wait);
	skcipher_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG,
				      crypto_req_done, &wait);

	crypto_skcipher_clear_flags(tfm, ~0);
	crypto_skcipher_setkey(tfm, key, 16);

	sg_init_table(&ssg, 1);
	sg_set_buf(&ssg, data, len);
	sg_init_table(&dsg, 1);
	sg_set_buf(&dsg, out, len);
	skcipher_request_set_crypt(req, &ssg, &dsg, len, NULL);

	ret = crypto_wait_req(crypto_skcipher_decrypt(req), &wait);
	if (ret) {
		pr_err("crypto_skcipher_encrypt failed ret=%d\n", ret);
	}

	skcipher_request_free(req);
fail:
	crypto_free_skcipher(tfm);

	return ret;
}

static int sm4_ecb_dec_step_test(const u8 *key, const u8 *data,
				 unsigned int len, u8 *out)
{
	struct crypto_wait wait;
	struct skcipher_request *req;
	struct crypto_skcipher *tfm;
	struct scatterlist ssg, dsg;
	static u32 flag = 0;
	int ret = -1;
	u32 offset, step_len;

	tfm = crypto_alloc_skcipher(alg, 0, 0);
	if (IS_ERR(tfm)) {
		pr_err("failed to load transform for %s: %ld\n", alg,
		       PTR_ERR(tfm));
		return ret;
	}

	if (flag++ == 0)
		pr_info("\ntesting of async %s (%s) dec!\n", alg,
			get_driver_name(crypto_skcipher, tfm));

	req = skcipher_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_info("skcipher: Failed to allocate request for %s\n", alg);
		goto fail;
	}

	crypto_skcipher_clear_flags(tfm, ~0);
	crypto_skcipher_setkey(tfm, key, 16);

	for (offset = 0; offset < len;) {
		step_len = 256;
		if (offset + step_len > len)
			step_len = len - offset;

		crypto_init_wait(&wait);
		skcipher_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG,
					      crypto_req_done, &wait);
		sg_init_table(&ssg, 1);
		sg_set_buf(&ssg, data + offset, step_len);
		sg_init_table(&dsg, 1);
		sg_set_buf(&dsg, out + offset, step_len);
		skcipher_request_set_crypt(req, &ssg, &dsg, step_len, NULL);

		ret = crypto_wait_req(crypto_skcipher_decrypt(req), &wait);
		if (ret) {
			pr_err("crypto_skcipher_encrypt failed ret=%d\n", ret);
			break;
		}
		offset += step_len;
	}

	skcipher_request_free(req);
fail:
	crypto_free_skcipher(tfm);

	return ret;
}

static int sm4_cbc_enc_test(struct crypto_skcipher *tfm, u8 *iv, const u8 *data,
			    unsigned int len, u8 *out)
{
	struct crypto_wait wait;
	struct skcipher_request *req;
	struct scatterlist ssg, dsg;
	static u32 flag = 0;
	int ret = -1;

	if (flag++ == 0)
		pr_info("\ntesting of async %s (%s) enc!\n", alg,
			get_driver_name(crypto_skcipher, tfm));

	req = skcipher_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_info("skcipher: Failed to allocate request for %s\n", alg);
		goto fail;
	}

	crypto_init_wait(&wait);
	skcipher_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG,
				      crypto_req_done, &wait);

	sg_init_table(&ssg, 1);
	sg_set_buf(&ssg, data, len);
	sg_init_table(&dsg, 1);
	sg_set_buf(&dsg, out, len);
	skcipher_request_set_crypt(req, &ssg, &dsg, len, iv);

	ret = crypto_wait_req(crypto_skcipher_encrypt(req), &wait);
	if (ret) {
		pr_err("crypto_skcipher_encrypt failed ret=%d\n", ret);
	}

	skcipher_request_free(req);
fail:

	return ret;
}

static int sm4_cbc_enc_step_test(struct crypto_skcipher *tfm, u8 *iv,
				 const u8 *data, unsigned int len, u8 *out)
{
	struct crypto_wait wait;
	struct skcipher_request *req;
	struct scatterlist ssg, dsg;
	static u32 flag = 0;
	int ret = -1;
	u32 offset, step_len;

	if (flag++ == 0)
		pr_info("\ntesting of async %s (%s) enc!\n", alg,
			get_driver_name(crypto_skcipher, tfm));

	req = skcipher_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_info("skcipher: Failed to allocate request for %s\n", alg);
		goto fail;
	}

	for (offset = 0; offset < len;) {
		step_len = 256;
		if (offset + step_len > len)
			step_len = len - offset;

		crypto_init_wait(&wait);
		skcipher_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG,
					      crypto_req_done, &wait);

		sg_init_table(&ssg, 1);
		sg_set_buf(&ssg, data + offset, step_len);
		sg_init_table(&dsg, 1);
		sg_set_buf(&dsg, out + offset, step_len);
		skcipher_request_set_crypt(req, &ssg, &dsg, step_len, iv);

		ret = crypto_wait_req(crypto_skcipher_encrypt(req), &wait);
		if (ret) {
			pr_err("crypto_skcipher_encrypt failed ret=%d\n", ret);
			break;
		}
		offset += step_len;
	}

	skcipher_request_free(req);
fail:

	return ret;
}

static int sm4_cbc_dec_test(struct crypto_skcipher *tfm, u8 *iv, const u8 *data,
			    unsigned int len, u8 *out)
{
	struct crypto_wait wait;
	struct skcipher_request *req;
	struct scatterlist ssg, dsg;
	static u32 flag = 0;
	int ret = -1;

	if (flag++ == 0)
		pr_info("\ntesting of async %s (%s) dec!\n", alg,
			get_driver_name(crypto_skcipher, tfm));

	req = skcipher_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_info("skcipher: Failed to allocate request for %s\n", alg);
		goto fail;
	}

	crypto_init_wait(&wait);
	skcipher_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG,
				      crypto_req_done, &wait);

	sg_init_table(&ssg, 1);
	sg_set_buf(&ssg, data, len);
	sg_init_table(&dsg, 1);
	sg_set_buf(&dsg, out, len);
	skcipher_request_set_crypt(req, &ssg, &dsg, len, iv);

	ret = crypto_wait_req(crypto_skcipher_decrypt(req), &wait);
	if (ret) {
		pr_err("crypto_skcipher_encrypt failed ret=%d\n", ret);
	}

	skcipher_request_free(req);
fail:

	return ret;
}

static int sm4_cbc_dec_step_test(struct crypto_skcipher *tfm, u8 *iv,
				 const u8 *data, unsigned int len, u8 *out)
{
	struct crypto_wait wait;
	struct skcipher_request *req;
	struct scatterlist ssg, dsg;
	static u32 flag = 0;
	int ret = -1;
	u32 offset, step_len;

	if (flag++ == 0)
		pr_info("\ntesting of async %s (%s) dec!\n", alg,
			get_driver_name(crypto_skcipher, tfm));

	req = skcipher_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_info("skcipher: Failed to allocate request for %s\n", alg);
		goto fail;
	}

	for (offset = 0; offset < len;) {
		step_len = 256;
		if (offset + step_len > len)
			step_len = len - offset;

		crypto_init_wait(&wait);
		skcipher_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG,
					      crypto_req_done, &wait);

		sg_init_table(&ssg, 1);
		sg_set_buf(&ssg, data + offset, step_len);
		sg_init_table(&dsg, 1);
		sg_set_buf(&dsg, out + offset, step_len);
		skcipher_request_set_crypt(req, &ssg, &dsg, step_len, iv);

		ret = crypto_wait_req(crypto_skcipher_decrypt(req), &wait);
		if (ret) {
			pr_err("crypto_skcipher_encrypt failed ret=%d\n", ret);
			break;
		}
		offset += step_len;
	}

	skcipher_request_free(req);
fail:

	return ret;
}

static int sm4_ctr_enc_test(const u8 *key, u8 *iv, const u8 *data,
			    unsigned int len, u8 *out)
{
	struct crypto_wait wait;
	struct skcipher_request *req;
	struct crypto_skcipher *tfm;
	struct scatterlist ssg, dsg;
	static u32 flag = 0;
	int ret = -1;

	tfm = crypto_alloc_skcipher(alg, 0, 0);
	if (IS_ERR(tfm)) {
		pr_err("failed to load transform for %s: %ld\n", alg,
		       PTR_ERR(tfm));
		return ret;
	}

	if (flag++ == 0)
		pr_info("\ntesting of async %s (%s) enc!\n", alg,
			get_driver_name(crypto_skcipher, tfm));

	req = skcipher_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_info("skcipher: Failed to allocate request for %s\n", alg);
		goto fail;
	}

	crypto_init_wait(&wait);
	skcipher_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG,
				      crypto_req_done, &wait);

	crypto_skcipher_clear_flags(tfm, ~0);
	crypto_skcipher_setkey(tfm, key, 16);

	sg_init_table(&ssg, 1);
	sg_set_buf(&ssg, data, len);
	sg_init_table(&dsg, 1);
	sg_set_buf(&dsg, out, len);
	skcipher_request_set_crypt(req, &ssg, &dsg, len, iv);

	ret = crypto_wait_req(crypto_skcipher_encrypt(req), &wait);
	if (ret) {
		pr_err("crypto_skcipher_encrypt failed ret=%d\n", ret);
	}

	skcipher_request_free(req);
fail:
	crypto_free_skcipher(tfm);

	return ret;
}

static int sm4_ctr_enc_step_test(const u8 *key, u8 *iv, const u8 *data,
				 unsigned int len, u8 *out)
{
	struct crypto_wait wait;
	struct skcipher_request *req;
	struct crypto_skcipher *tfm;
	struct scatterlist ssg, dsg;
	static u32 flag = 0;
	int ret = -1;
	u32 offset, step_len;

	tfm = crypto_alloc_skcipher(alg, 0, 0);
	if (IS_ERR(tfm)) {
		pr_err("failed to load transform for %s: %ld\n", alg,
		       PTR_ERR(tfm));
		return ret;
	}

	if (flag++ == 0)
		pr_info("\ntesting of async %s (%s) enc!\n", alg,
			get_driver_name(crypto_skcipher, tfm));

	req = skcipher_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_info("skcipher: Failed to allocate request for %s\n", alg);
		goto fail;
	}

	crypto_skcipher_clear_flags(tfm, ~0);
	crypto_skcipher_setkey(tfm, key, 16);

	for (offset = 0; offset < len;) {
		step_len = 256;
		if (offset + step_len > len)
			step_len = len - offset;

		crypto_init_wait(&wait);
		skcipher_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG,
					      crypto_req_done, &wait);

		sg_init_table(&ssg, 1);
		sg_set_buf(&ssg, data + offset, step_len);
		sg_init_table(&dsg, 1);
		sg_set_buf(&dsg, out + offset, step_len);
		skcipher_request_set_crypt(req, &ssg, &dsg, step_len, iv);

		ret = crypto_wait_req(crypto_skcipher_encrypt(req), &wait);
		if (ret) {
			pr_err("crypto_skcipher_encrypt failed ret=%d\n", ret);
			break;
		}
		offset += step_len;
	}

	skcipher_request_free(req);
fail:
	crypto_free_skcipher(tfm);

	return ret;
}

static int sm4_ctr_dec_test(const u8 *key, u8 *iv, const u8 *data,
			    unsigned int len, u8 *out)
{
	struct crypto_wait wait;
	struct skcipher_request *req;
	struct crypto_skcipher *tfm;
	struct scatterlist ssg, dsg;
	static u32 flag = 0;
	int ret = -1;

	tfm = crypto_alloc_skcipher(alg, 0, 0);
	if (IS_ERR(tfm)) {
		pr_err("failed to load transform for %s: %ld\n", alg,
		       PTR_ERR(tfm));
		return ret;
	}

	if (flag++ == 0)
		pr_info("\ntesting of async %s (%s) dec!\n", alg,
			get_driver_name(crypto_skcipher, tfm));

	req = skcipher_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_info("skcipher: Failed to allocate request for %s\n", alg);
		goto fail;
	}

	crypto_init_wait(&wait);
	skcipher_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG,
				      crypto_req_done, &wait);

	crypto_skcipher_clear_flags(tfm, ~0);
	crypto_skcipher_setkey(tfm, key, 16);

	sg_init_table(&ssg, 1);
	sg_set_buf(&ssg, data, len);
	sg_init_table(&dsg, 1);
	sg_set_buf(&dsg, out, len);
	skcipher_request_set_crypt(req, &ssg, &dsg, len, iv);

	ret = crypto_wait_req(crypto_skcipher_decrypt(req), &wait);
	if (ret) {
		pr_err("crypto_skcipher_encrypt failed ret=%d\n", ret);
	}

	skcipher_request_free(req);
fail:
	crypto_free_skcipher(tfm);

	return ret;
}

static int sm4_ctr_dec_step_test(const u8 *key, u8 *iv, const u8 *data,
				 unsigned int len, u8 *out)
{
	struct crypto_wait wait;
	struct skcipher_request *req;
	struct crypto_skcipher *tfm;
	struct scatterlist ssg, dsg;
	static u32 flag = 0;
	int ret = -1;
	u32 offset, step_len;

	tfm = crypto_alloc_skcipher(alg, 0, 0);
	if (IS_ERR(tfm)) {
		pr_err("failed to load transform for %s: %ld\n", alg,
		       PTR_ERR(tfm));
		return ret;
	}

	if (flag++ == 0)
		pr_info("\ntesting of async %s (%s) dec!\n", alg,
			get_driver_name(crypto_skcipher, tfm));

	req = skcipher_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_info("skcipher: Failed to allocate request for %s\n", alg);
		goto fail;
	}

	crypto_skcipher_clear_flags(tfm, ~0);
	crypto_skcipher_setkey(tfm, key, 16);

	for (offset = 0; offset < len;) {
		step_len = 256;
		if (offset + step_len > len)
			step_len = len - offset;

		crypto_init_wait(&wait);
		skcipher_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG,
					      crypto_req_done, &wait);

		sg_init_table(&ssg, 1);
		sg_set_buf(&ssg, data + offset, step_len);
		sg_init_table(&dsg, 1);
		sg_set_buf(&dsg, out + offset, step_len);
		skcipher_request_set_crypt(req, &ssg, &dsg, step_len, iv);

		ret = crypto_wait_req(crypto_skcipher_decrypt(req), &wait);
		if (ret) {
			pr_err("crypto_skcipher_encrypt failed ret=%d\n", ret);
			break;
		}
		offset += step_len;
	}

	skcipher_request_free(req);
fail:
	crypto_free_skcipher(tfm);

	return ret;
}

void sm4_ecb_test(void)
{
	u8 *src, *kdst, *sdst;
	u8 key[16];
	u32 len;
	struct sm4_context ctx;
	struct sm4_func func_sm4;
	u64 time = jiffies;
	int ret = -1;
	int err = 0;

	src = kmalloc(elen * 3, GFP_KERNEL);
	if (src == NULL) {
		pr_err("%s:kmalloc failed!\n", __func__);
		return;
	}
	kdst = src + elen;
	sdst = kdst + elen;

	get_random_bytes(src, elen);
	get_random_bytes(key, 16);
	sm4_func_alloc(&func_sm4);

	for (len = slen; len <= elen; len += 16) {
		ret = sm4_ecb_enc_test(key, src, len, kdst);
		if (ret) {
			err++;
			break;
		}

		func_sm4.sm4_setkey_enc(&ctx, key);
		func_sm4.sm4_crypt_ecb_enc(&ctx, len, NULL, src, sdst);
		if (memcmp(kdst, sdst, len)) {
			pr_err("sm4_ecb_enc_test err!\n");
			print_buf("key", key, 16);
			print_buf("src", src, len);
			print_buf("kdst", kdst, len);
			print_buf("sdst", sdst, len);
			err++;
			break;
		}
		ret = sm4_ecb_enc_step_test(key, src, len, kdst);
		if (ret) {
			err++;
			break;
		}
		if (memcmp(kdst, sdst, len)) {
			pr_err("sm4_ecb_enc_step_test err!\n");
			print_buf("key", key, 16);
			print_buf("src", src, len);
			print_buf("kdst", kdst, len);
			print_buf("sdst", sdst, len);
			err++;
			break;
		}

		ret = sm4_ecb_dec_test(key, sdst, len, kdst);
		if (ret) {
			err++;
			break;
		}
		if (memcmp(kdst, src, len)) {
			pr_err("sm4_ecb_dec_test err!\n");
			print_buf("key", key, 16);
			print_buf("src", src, len);
			print_buf("kdst", kdst, len);
			print_buf("sdst", sdst, len);
			err++;
			break;
		}
		ret = sm4_ecb_dec_step_test(key, sdst, len, kdst);
		if (ret) {
			err++;
			break;
		}
		if (memcmp(kdst, src, len)) {
			pr_err("sm4_ecb_dec_step_test err!\n");
			print_buf("key", key, 16);
			print_buf("src", src, len);
			print_buf("kdst", kdst, len);
			print_buf("sdst", sdst, len);
			err++;
			break;
		}

		if (time + msecs_to_jiffies(100) < jiffies) {
			time = jiffies;
			schedule();
		}
	}

	if (err == 0)
		pr_info("Test ecb(sm4) success\r\n");

	kfree(src);
}

struct kthread_sm4_data {
	struct crypto_skcipher *tfm;
	u8 *key;
	u32 keylen;
};

int cbc_sm4_test_vs_software(struct kthread_sm4_data *data)
{
	u8 *src, *kdst, *sdst;
	u8 iv[16], initiv[16];
	u32 len;
	struct sm4_context ctx;
	struct sm4_func func_sm4;
	u64 time = jiffies;
	int ret = -1;
	int err = 0;

	src = kmalloc(elen * 3, GFP_KERNEL);
	if (src == NULL) {
		pr_err("%s:kmalloc failed!\n", __func__);
		return -ENOMEM;
	}
	kdst = src + elen;
	sdst = kdst + elen;

	get_random_bytes(src, elen);
	get_random_bytes(initiv, 16);
	sm4_func_alloc(&func_sm4);

	for (len = slen; len <= elen; len += 16) {
		memcpy(iv, initiv, 16);
		ret = sm4_cbc_enc_test(data->tfm, iv, src, len, kdst);
		if (ret) {
			err++;
			break;
		}

		memcpy(iv, initiv, 16);
		func_sm4.sm4_setkey_enc(&ctx, data->key);
		func_sm4.sm4_crypt_cbc_enc(&ctx, len, iv, src, sdst);
		if (memcmp(kdst, sdst, len)) {
			pr_err("sm4_cbc_enc_test err!\n");
			print_buf("key", data->key, 16);
			print_buf("initiv", initiv, 16);
			print_buf("iv", iv, 16);
			print_buf("src", src, len);
			print_buf("kdst", kdst, len);
			print_buf("sdst", sdst, len);
			err++;
			break;
		}

		memcpy(iv, initiv, 16);
		ret = sm4_cbc_enc_step_test(data->tfm, iv, src, len, kdst);
		if (ret) {
			err++;
			break;
		}

		if (memcmp(kdst, sdst, len)) {
			pr_err("sm4_cbc_enc_step_test err!\n");
			print_buf("key", data->key, 16);
			print_buf("initiv", initiv, 16);
			print_buf("iv", iv, 16);
			print_buf("src", src, len);
			print_buf("kdst", kdst, len);
			print_buf("sdst", sdst, len);
			err++;
			break;
		}

		memcpy(iv, initiv, 16);
		ret = sm4_cbc_dec_test(data->tfm, iv, sdst, len, kdst);
		if (ret) {
			err++;
			break;
		}
		if (memcmp(kdst, src, len)) {
			pr_err("sm4_cbc_dec_test err!\n");
			print_buf("key", data->key, 16);
			print_buf("initiv", initiv, 16);
			print_buf("iv", iv, 16);
			print_buf("src", src, len);
			print_buf("kdst", kdst, len);
			print_buf("sdst", sdst, len);
			err++;
			break;
		}

		memcpy(iv, initiv, 16);
		ret = sm4_cbc_dec_step_test(data->tfm, iv, sdst, len, kdst);
		if (ret) {
			err++;
			break;
		}
		if (memcmp(kdst, src, len)) {
			pr_err("sm4_cbc_dec_step_test err!\n");
			print_buf("key", data->key, 16);
			print_buf("initiv", initiv, 16);
			print_buf("iv", iv, 16);
			print_buf("src", src, len);
			print_buf("kdst", kdst, len);
			print_buf("sdst", sdst, len);
			err++;
			break;
		}

		if (time + msecs_to_jiffies(100) < jiffies) {
			time = jiffies;
			schedule();
		}
	}

	kfree(src);

	return err;
}

static int cbc_sm4_thread_func(void *data)
{
	struct kthread_sm4_data *thread_data = (struct kthread_sm4_data *)data;
	int ret = 0;
	char name[TASK_COMM_LEN] = { 0 };

	while (1) {
		if (kthread_should_stop())
			break;
		if (ret == 0)
			ret = cbc_sm4_test_vs_software(thread_data);
		else
			mdelay(100);
	}

	if (ret == 0) {
		get_task_comm(name, current);
		pr_info("Success to exit   thread %s\n", name);
	}

	return 0;
}

void sm4_cbc_test(void)
{
	struct task_struct **kthreads;
	u32 kthread_num = multi ? multi : 1;
	int i;
	struct kthread_sm4_data *thread_data;
	u32 keylen = 16;
	unsigned long start, end;

	kthreads = kmalloc(kthread_num * sizeof(struct task_struct *),
			   GFP_KERNEL | __GFP_ZERO);
	if (kthreads == NULL) {
		pr_err("malloc error\n");
		return;
	}

	thread_data = kmalloc(sizeof(struct kthread_sm4_data), GFP_KERNEL);
	if (thread_data == NULL) {
		kfree(kthreads);
		pr_err("malloc error\n");
		return;
	}

	thread_data->keylen = keylen;
	thread_data->key = kmalloc(keylen, GFP_KERNEL);
	if (thread_data->key == NULL) {
		kfree(thread_data);
		kfree(kthreads);
		pr_err("malloc error\n");
		return;
	}

	thread_data->tfm = crypto_alloc_skcipher(alg, 0, 0);
	if (IS_ERR(thread_data->tfm)) {
		pr_err("failed to load transform for %s: %ld\n", alg,
		       PTR_ERR(thread_data->tfm));
		kfree(thread_data->key);
		kfree(thread_data);
		kfree(kthreads);
		return;
	}

	get_random_bytes(thread_data->key, thread_data->keylen);
	crypto_skcipher_clear_flags(thread_data->tfm, ~0);
	crypto_skcipher_setkey(thread_data->tfm, thread_data->key, 16);

	for (i = 0; i < kthread_num; i++) {
		kthreads[i] = kthread_run(cbc_sm4_thread_func,
					  (void *)thread_data, "cbcsm4-%d", i);
	}

	start = jiffies;
	end = start + secs * HZ;

	/* run for secs seconds */
	while (time_before(jiffies, end)) {
		schedule();
	}

	for (i = 0; i < kthread_num; i++) {
		kthread_stop(kthreads[i]);
	}

	mdelay(1000);
	crypto_free_skcipher(thread_data->tfm);
	kfree(thread_data->key);
	kfree(thread_data);
	kfree(kthreads);
}

void sm4_ctr_test(void)
{
	u8 *src, *kdst, *sdst;
	u8 key[16], iv[16], initiv[16];
	u32 len;
	struct sm4_context ctx;
	struct sm4_func func_sm4;
	u64 time = jiffies;
	int ret = -1;
	int err = 0;

	src = kmalloc(elen * 3, GFP_KERNEL);
	if (src == NULL) {
		pr_err("%s:kmalloc failed!\n", __func__);
		return;
	}
	kdst = src + elen;
	sdst = kdst + elen;

	get_random_bytes(src, elen);
	get_random_bytes(key, 16);
	get_random_bytes(initiv, 16);
	sm4_func_alloc(&func_sm4);

	for (len = slen; len <= elen; len += 16) {
		memcpy(iv, initiv, 16);
		ret = sm4_ctr_enc_test(key, iv, src, len, kdst);
		if (ret) {
			err++;
			break;
		}
		memcpy(iv, initiv, 16);
		func_sm4.sm4_setkey_enc(&ctx, key);
		func_sm4.sm4_crypt_ctr(&ctx, len, iv, src, sdst);
		if (memcmp(kdst, sdst, len)) {
			pr_err("sm4_ctr_enc_test err!\n");
			print_buf("key", key, 16);
			print_buf("initiv", initiv, 16);
			print_buf("iv", iv, 16);
			print_buf("src", src, len);
			print_buf("kdst", kdst, len);
			print_buf("sdst", sdst, len);
			err++;
			break;
		}

		memcpy(iv, initiv, 16);
		ret = sm4_ctr_enc_step_test(key, iv, src, len, kdst);
		if (ret) {
			err++;
			break;
		}
		if (memcmp(kdst, sdst, len)) {
			pr_err("sm4_ctr_enc_step_test err!\n");
			print_buf("key", key, 16);
			print_buf("initiv", initiv, 16);
			print_buf("iv", iv, 16);
			print_buf("src", src, len);
			print_buf("kdst", kdst, len);
			print_buf("sdst", sdst, len);
			err++;
			break;
		}

		memcpy(iv, initiv, 16);
		ret = sm4_ctr_dec_test(key, iv, sdst, len, kdst);
		if (ret) {
			err++;
			break;
		}
		if (memcmp(kdst, src, len)) {
			pr_err("sm4_ctr_dec_test err!\n");
			print_buf("key", key, 16);
			print_buf("initiv", initiv, 16);
			print_buf("iv", iv, 16);
			print_buf("src", src, len);
			print_buf("kdst", kdst, len);
			print_buf("sdst", sdst, len);
			err++;
			break;
		}

		memcpy(iv, initiv, 16);
		ret = sm4_ctr_dec_step_test(key, iv, sdst, len, kdst);
		if (ret) {
			err++;
			break;
		}
		if (memcmp(kdst, src, len)) {
			pr_err("sm4_ctr_dec_step_test err!\n");
			print_buf("key", key, 16);
			print_buf("initiv", initiv, 16);
			print_buf("iv", iv, 16);
			print_buf("src", src, len);
			print_buf("kdst", kdst, len);
			print_buf("sdst", sdst, len);
			err++;
			break;
		}

		if (time + msecs_to_jiffies(100) < jiffies) {
			time = jiffies;
			schedule();
		}
	}

	if (err == 0)
		pr_info("Test ctr(sm4) success\r\n");

	kfree(src);
}
