/*
 * File Name:test.c - Phytium driver test
 *
 * Copyright (C) 2020 Phytium Technology Co.,Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */

#include <linux/module.h>
#include <linux/irq.h>
#include <linux/platform_device.h>
#include <linux/dma-mapping.h>
#include <linux/cdev.h>
#include <linux/types.h>
#include <linux/mutex.h>
#include <linux/acpi.h>
#include <crypto/internal/hash.h>
#include <linux/crypto.h>
#include <crypto/hash.h>
#include <crypto/internal/skcipher.h>
#include <crypto/internal/akcipher.h>
#include <linux/ioport.h>
#include <linux/io.h>
#include "test.h"
#include "sm4_software.h"
#include "sm3_software.h"

char *alg = NULL;
u32 slen = 256;
u32 elen = 256;
u32 secs = 5;
u32 speed = 0;
u32 multi = 0;

module_param(alg, charp, 0);
MODULE_PARM_DESC(
	alg,
	"Test algo, one of [sm2 sm3 hmac(sm3) ecb(sm4) cbc(sm4) ctr(sm4) trng ]");

module_param(slen, uint, 0);
MODULE_PARM_DESC(slen, "Start of length, default 256.");

module_param(elen, uint, 0);
MODULE_PARM_DESC(elen, "End of length, default 256.");

module_param(secs, uint, 0);
MODULE_PARM_DESC(secs, "Time to test speed, default 5 s.");

module_param(speed, uint, 0);
MODULE_PARM_DESC(speed, "Speed test, default 0.");

module_param(multi, uint, 0);
MODULE_PARM_DESC(multi, "Test  with multi threads in parallel, default 0.");

extern void trng_test(u32 slen, u32 elen);
extern void trng_speed(u32 secs);

extern int sm2_test(void);

void print_buf(u8 *name, u8 *buf, u32 len)
{
	u32 i;

	pr_info("%s[0x%x]:\n", name, len);
	for (i = 0; i < len; i += 16) {
		pr_info("%08x: %02x %02x %02x %02x %02x %02x %02x %02x "
			"%02x %02x %02x %02x %02x %02x %02x %02x\n",
			i, buf[i + 0], buf[i + 1], buf[i + 2], buf[i + 3],
			buf[i + 4], buf[i + 5], buf[i + 6], buf[i + 7],
			buf[i + 8], buf[i + 9], buf[i + 10], buf[i + 11],
			buf[i + 12], buf[i + 13], buf[i + 14], buf[i + 15]);
	}
}

void print_help(void)
{
	pr_info("------BAD PARAM!------\n");
	pr_info("alg: \"sm3\", \"hmac(sm3)\", \"ecb(sm4)\", \"cbc(sm4)\", \"ctr(sm4)\", \"trng\"\n");
	pr_info("slen: test min length\n");
	pr_info("elen: test max length\n");
	pr_info("secs:  Specify duration to test speed, default 5 s\n");
	pr_info("speed: Enabled to test speed, default disabled, only valid for "
		"trng\n");
	pr_info("------BAD PARAM!------\n");
}

static int test_init(void)
{
	u32 len;

	if (slen > elen) {
		len = slen;
		slen = elen;
		elen = len;
	}

	if (alg == NULL) {
		print_help();
		return -1;
	}

	if (memcmp(alg, "sm3", strlen("sm3")) == 0) {
		alg = "sm3-phytium";
		if (speed)
			sm3_speed(secs);
		else
			sm3_test();
		return -1;
	}

	if (memcmp(alg, "hmac(sm3)", strlen("hmac(sm3)")) == 0) {
		alg = "sm3-hmac-phytium";
		if (speed)
			hmac_sm3_speed(secs);
		else
			hmac_test();
		return -1;
	}
	if (memcmp(alg, "ecb(sm4)", strlen("ecb(sm4)")) == 0) {
		alg = "sm4-ecb-phytium";
		slen = (slen + 0xF) & 0xFFFFFFF0;
		elen = (elen + 0xF) & 0xFFFFFFF0;
		if (speed)
			sm4_speed_test("sm4-ecb-phytium", secs);
		else
			sm4_ecb_test();
		return -1;
	}
	if (memcmp(alg, "cbc(sm4)", strlen("cbc(sm4)")) == 0) {
		alg = "sm4-cbc-phytium";
		slen = (slen + 0xF) & 0xFFFFFFF0;
		elen = (elen + 0xF) & 0xFFFFFFF0;
		if (speed)
			sm4_speed_test(alg, secs);
		else
			sm4_cbc_test();
		return -1;
	}
	if (memcmp(alg, "ctr(sm4)", strlen("ctr(sm4)")) == 0) {
		alg = "sm4-ctr-phytium";
		slen = (slen + 0xF) & 0xFFFFFFF0;
		elen = (elen + 0xF) & 0xFFFFFFF0;
		if (speed)
			sm4_speed_test(alg, secs);
		else
			sm4_ctr_test();
		return -1;
	}

	if (memcmp(alg, "trng", strlen("trng")) == 0) {
		alg = "trng-phytium";
		if (speed)
			trng_speed(secs);
		else
			trng_test(slen, elen);

		return -1;
	}

	if (memcmp(alg, "sm2", strlen("sm2")) == 0) {
		alg = "sm2-phytium";
		if (speed)
			sm2_speed(secs);
		else
			sm2_test();

		return -1;
	}

	print_help();

	return -1;
}

static void test_exit(void)
{
}

module_init(test_init);
module_exit(test_exit);
MODULE_DESCRIPTION("PHYTIUM TEST");
MODULE_LICENSE("GPL");
