#include <linux/module.h>
#include <linux/irq.h>
#include <linux/platform_device.h>
#include <linux/dma-mapping.h>
#include <linux/cdev.h>
#include <linux/types.h>
#include <linux/mutex.h>
#include <linux/acpi.h>
#include <crypto/internal/hash.h>
#include <linux/crypto.h>
#include <crypto/hash.h>
#include <crypto/aead.h>
#include <crypto/internal/skcipher.h>
#include <crypto/internal/akcipher.h>
#include <crypto/internal/aead.h>
#include <crypto/internal/rng.h>
#include <linux/ioport.h>
#include <linux/io.h>
#include <linux/err.h>

static void print_buf(u8 *name, u8 *buf, u32 len)
{
	u32 i;

	pr_info("%s[0x%x]:\n", name, len);
	for (i = 0; i < len; i += 16) {
		pr_info("%08x: %02x %02x %02x %02x %02x %02x %02x %02x "
			"%02x %02x %02x %02x %02x %02x %02x %02x\n",
			i, buf[i + 0], buf[i + 1], buf[i + 2], buf[i + 3],
			buf[i + 4], buf[i + 5], buf[i + 6], buf[i + 7],
			buf[i + 8], buf[i + 9], buf[i + 10], buf[i + 11],
			buf[i + 12], buf[i + 13], buf[i + 14], buf[i + 15]);
	}
}

void trng_test(u32 slen, u32 elen)
{
	// generate from sce_trng
	struct crypto_rng *tfm;
	u8 *buf;
	u32 len;
	int err;

	tfm = crypto_alloc_rng("trng-phytium", 0, 0);
	if (IS_ERR(tfm)) {
		pr_err("alg: cprng: Failed to load transform for %s: "
		       "%ld\n",
		       "trng", PTR_ERR(tfm));
		return;
	}

	if (elen < slen) {
		len = slen;
		slen = elen;
		elen = len;
	}

	buf = kmalloc(elen, GFP_KERNEL);
	if (buf == NULL) {
		pr_err("alloc buf err");
		return;
	}

	memset(buf, 0, elen);
	pr_info("get slen %d elen %d", slen, elen);
	for (len = slen; len <= elen; len++) {
		err = crypto_rng_get_bytes(tfm, buf, len);
		if (err) {
			pr_err("generate len %d error", len);
			break;
		}
		print_buf("random data", buf, len);
	}

	kfree(buf);
	crypto_free_rng(tfm);
}

static unsigned long lengths[] = { 16,	 64,   128,  256,  512,
				   1024, 2048, 4096, 8192, 16384 };

static int _trng_speed_test(struct crypto_rng *tfm, char *output, long len,
			    int secs)
{
	unsigned long start, end;
	int count, ret;

	for (start = jiffies, end = start + secs * HZ, count = 0;
	     time_before(jiffies, end); count++) {
		ret = crypto_rng_get_bytes(tfm, output, len);
		if (ret)
			return ret;
	}

	// print_buf("random data", buf, len);
	pr_info("Doing %s for %us on %lu size: %6u opers in %d s %6u opers/sec, "
		"%9lu Mbps",
		"trng-phytium", secs, len, count, secs, count / secs,
		(len * count * 8) / (secs * 1000000));

	return 0;
}

static int _trng_cycles_trng(struct crypto_rng *tfm, char *output, long len)
{
	unsigned long cycles = 0;
	int i;
	cycles_t start, end;

	/*warm up*/
	(void)_trng_speed_test(tfm, output, len, 1);

	for (i = 0; i < sizeof(lengths) / sizeof(lengths[0]); i++) {
		start = get_cycles();
		(void)crypto_rng_get_bytes(tfm, output, len);
		end = get_cycles();
		cycles += end - start;
	}

	pr_info(" %6lu cycles/operation, %4lu cycles/byte", cycles / 8,
		cycles / (8 * len));

	return 0;
}

void trng_speed(u32 secs)
{
	int i, ret;
	unsigned long len;
	struct crypto_rng *tfm;
	u8 *output;

	tfm = crypto_alloc_rng("trng-phytium", 0, 0);
	if (IS_ERR(tfm)) {
		pr_err("alg: cprng: Failed to load transform for %s: "
		       "%ld\n",
		       "trng-phytium", PTR_ERR(tfm));
		return;
	}

	output = kmalloc(16384, GFP_KERNEL);
	if (!output) {
		pr_err("Alloc mem failed.");
		return;
	}

	for (i = 0; i < sizeof(lengths) / sizeof(lengths[0]); i++) {
		len = lengths[i];
		if (secs) {
			// pr_info("Doing %s for %us on %lu size", "trng-phytium", secs,
			// len);
			ret = _trng_speed_test(tfm, output, len, secs);
			cond_resched();
		} else {
			ret = _trng_cycles_trng(tfm, output, len);
		}

		if (ret) {
			pr_err("trng data failed ret=%d\n", ret);
			break;
		}
	}

	crypto_free_rng(tfm);
	kfree(output);
}
