/* SPDX-License-Identifier: GPL-2.0 */
#ifndef __SCE_DRIVER_THREAD_H__
#define __SCE_DRIVER_THREAD_H__

#define SCE_LINK_END_SIZE (0xFFFFFFFFull)

struct sce_link_list_item {
	uint32_t size;
	uint32_t reserved;
	uint64_t address;
};

enum TRMG_MODE {
	TRNG_RNP = 0,
	TRNG_SM4 = 1,
};
enum ALGO_TYPE {
	HASH = 0,
	SCA = 1,
	TRNG = 2,
};
enum HASH_ALGO_TYPE {
	SHA1 = 0,
	SHA224 = 1,
	SHA256 = 2,
	SHA384 = 3,
	SHA512 = 4,
	MD5 = 5,
	SM3 = 8,
};
enum SCA_ALGO_TYPE {
	AES = 0,
	DES = 1,
	TDES = 2,
	GHASH = 3,
	SM4 = 4,
};
enum SCA_ALGO_MODE {
	ECB = 0,
	CTR = 1,
	CBC = 2,
	OFB = 3,
	XTS = 4,
	CBC_MAC = 5,
	CMAC = 6,
	GCM = 7,
	CCM = 8,
};

struct trng_cmd {
	u8 mode;
	u64 data;
	u32 reserved1;
	u32 len;
	u64 tag;
	u32 reserved[10];
} __packed;

struct hash_cmd {
	u8 algo;
	u8 mode;
	u8 dma_mode;
	u8 reserved0;
	u32 in_size;
	u64 in;
	u64 out;
	u64 key;
	u64 tag;
	u32 reserved[6];
} __packed;

struct sca_cmd {
	u8 algo;
	u8 mode;
	u8 op_mode;
	u8 dma_mode;
	u32 in_size;
	u32 aad_len;
	u32 tag_len;
	u64 in;
	u64 out;
	u64 key;
	u64 iv;
	u64 tag;
	u32 reserved[4];
} __packed;

static inline void sce_sync(void)
{
	__asm__ __volatile__("dsb sy" : : : "memory");
}

static inline u32 sce_load_sync(u32 *addr)
{
	uint32_t val;

	__asm__ __volatile__("ldar %w0, %1"
			     : "=r"(val)
			     : "Q"(*addr)
			     : "memory");

	return val;
}

static inline void sce_store_sync(u32 *addr, u32 val)
{
	__asm__ __volatile__("stlr %w1, %0"
			     : "=Q"(*addr)
			     : "r"(val)
			     : "memory");
}

#define REG_WR_64(reg_base, reg_offset, val)                                   \
	(*(volatile uint64_t *)((uint64_t)reg_base + reg_offset) = val)

#define REG_WR_32(reg_base, reg_offset, val)                                   \
	(*(volatile uint32_t *)((uint64_t)reg_base + reg_offset) = val)

#define REG_RD_32(reg_base, reg_offset)                                        \
	(*(volatile uint32_t *)((uint64_t)reg_base + reg_offset))

#define REG_WR_32_SYNC(reg_base, reg_offset, val)                              \
	(sce_store_sync((uint32_t *)((uint64_t)reg_base + reg_offset), \
			val))

#define REG_RD_32_SYNC(reg_base, reg_offset) \
	(sce_load_sync((uint32_t *)((uint64_t)reg_base + reg_offset)))

#define QUEUE_DEPTH (1024)
#define RING_CFG_OFFSET (0)
#define RING_STATUS_OFFSET (0x4)

#define CMD_BADDR0_OFFSET (0x18)
#define CMD_BADDR1_OFFSET (0x1c)
#define RSP_BADDR0_OFFSET (0x20)
#define RSP_BADDR1_OFFSET (0x24)
#define CMD_THRESHOLD_OFFSET (0x28)
#define RSP_THRESHOLD_OFFSET (0x2c)
#define RSP_INT_NUM_OFFSET (0x30)
#define CMD_PENDDING_NUM_OFFSET (0x34)

#define CMD_WRPTR_OFFSET (0x40)
#define RSP_RDPTR_OFFSET (0x44)
#define CMD_RDPTR_OFFSET (0x80)
#define RSP_WRPTR_OFFSET (0x84)

#define TRNG_POOL_SIZE (32)
#define SM4_TRNG_POOL_SIZE (32)
union ring_cfg {
	u32 data;
	struct {
		u32 ring_flush : 1; /* w1s */
		u32 cmd_suspend : 1;
		u32 reserve : 30;
	};
};

union ring_status {
	u32 cmd_ring_busy; /* ro */
};

union cmd_wrptr {
	u32 cmd_wrptr;
};

union cmd_rdptr {
	u32 cmd_rdptr;
};

union rsp_rdptr {
	u32 rsp_rdptr;
};

union rsp_wrptr {
	u32 rsp_wrptr;
};

union cmd_baddr0 {
	u32 cmd_baddr0;
};

union cmd_baddr1 {
	u32 cmd_baddr1;
};

union rsp_baddr0 {
	u32 rsp_baddr0;
};

union rsp_baddr1 {
	u32 rsp_baddr1;
};

union cmd_threshold {
	u32 cmd_timeout_threshold;
};

union rsp_threshold0 {
	u32 rsp_timeout_threshold;
};

union rsp_threshold1 {
	u32 data;
	struct {
		u32 rsp_irq_threshold : 16;
		u32 rsp_count_threshold : 16;
	};
};

struct cmd {
	uint8_t data[64];
};

struct rsp {
	u32 state;
	u32 reserved;
	u64 tag;
} __packed;

int sce_add_thread(unsigned int cpu);
void sce_del_thread(void);
void sce_wake_up_thread(void);

#endif
