/* SPDX-License-Identifier: GPL-2.0 */
#ifndef __SCE_KERNEL_COMMON_H__
#define __SCE_KERNEL_COMMON_H__

#if KERNEL_VERSION(5, 4, 0) <= LINUX_VERSION_CODE
#include <linux/dma-mapping.h>
#endif

struct sce_mutex {
	struct mutex mutex;
};

#define sce_info(fmt, ...) pr_info(pr_fmt(fmt), ##__VA_ARGS__)
#define sce_warning(fmt, ...) pr_warn(pr_fmt(fmt), ##__VA_ARGS__)
#define sce_err(fmt, ...) pr_err(pr_fmt(fmt), ##__VA_ARGS__)

#ifdef SCE_DEBUG
#define sce_debug(fmt, ...) pr_info(pr_fmt(fmt), ##__VA_ARGS__)
#else
#define sce_debug(fmt, ...)
#endif

#define CACHE_LINE_SIZE (64)
#define MAX_LINKS (32)

#define WAIT_POOL_FILLED_US 3
#define WAIT_POOL_CYCLE_INDEX 10000

static inline u64 sce_get_jiffies(void)
{
	return jiffies;
}

static inline void *sce_malloc(size_t size)
{
	return kmalloc(size, GFP_KERNEL);
}

static inline void *sce_malloc_node(size_t size, int node)
{
	return kmalloc_node(size, GFP_KERNEL, node);
}

static inline void *sce_zalloc(size_t size)
{
	return kzalloc(size, GFP_KERNEL);
}

static inline void *sce_zalloc_node(size_t size, int node)
{
	return kzalloc_node(size, GFP_KERNEL, node);
}

static inline void sce_free(void *ptr)
{
	kfree(ptr);
}

static inline void sce_prefetch(void *ptr)
{
	prefetch(ptr);
}

static inline void sce_prefetchw(void *ptr)
{
	prefetchw(ptr);
}

static inline void sce_cache_sync_single(void *dev, u64 ioaddr, size_t size,
					 int dir)
{
	dma_sync_single_for_device(dev, ioaddr, size, dir);
}

static inline void sce_cache_sync_sg(void *dev, void *sg, int nelems, int dir)
{
	dma_sync_sg_for_device(dev, sg, nelems, dir);
}

static inline u64 sce_msecs_to_jiffies(u32 ms)
{
	return msecs_to_jiffies(ms);
}

static inline void sce_msleep(u32 ms)
{
	msleep(ms);
}

static inline void *sce_get_pages(int numa_node, u64 size)
{
	struct page *page;

	page = alloc_pages_node(numa_node, GFP_KERNEL | __GFP_ZERO,
				get_order(size));
	if (!page)
		return NULL;

	return page_address(page);
}

static inline void sce_free_pages(void *buf, u64 size)
{
	free_pages((unsigned long)buf, get_order(size));
}

static inline u64 sce_get_dma_addr(void *dev, void *buf, size_t size,
				   int dma_data_direction)
{
	return dma_map_single(dev, buf, size, dma_data_direction);
}

static inline void sce_free_dma_addr(void *dev, u64 ioaddr, size_t size,
				     int dma_data_direction)
{
	dma_unmap_single(dev, ioaddr, size, dma_data_direction);
}

static inline int sce_dma_map_sg(void *dev, void *sg, int sg_nents,
				 int dma_data_direction)
{
	return dma_map_sg(dev, sg, sg_nents, dma_data_direction);
}

static inline void sce_dma_unmap_sg(void *dev, void *sg, int sg_nents,
				    int dma_data_direction)
{
	return dma_unmap_sg(dev, sg, sg_nents, dma_data_direction);
}

static inline void sce_init_mutex(struct sce_mutex *mutex)
{
	mutex_init(&mutex->mutex);
}

static inline void sce_lock_mutex(struct sce_mutex *mutex)
{
	mutex_lock(&mutex->mutex);
}

static inline void sce_mutex_unlock(struct sce_mutex *mutex)
{
	mutex_unlock(&mutex->mutex);
}

static inline void *sce_kthread_create_on_node(int (*threadfn)(void *data),
					       void *data, int node,
					       const char *name)
{
	void *task;

	task = kthread_create_on_node(threadfn, data, node, name);
	if (IS_ERR(task)) {
		sce_err("%s:%s create failed!\n", __func__, name);
		return NULL;
	}

	return task;
}

static inline void *sce_kthread_create_on_cpu(int (*threadfn)(void *data),
					      void *data, int cpu,
					      const char *name)
{
	void *task;

	task = kthread_create_on_cpu(threadfn, data, cpu, name);
	if (IS_ERR(task)) {
		sce_err("%s:%s create failed!\n", __func__, name);
		return NULL;
	}

	return task;
}

static inline void sce_wake_up_kthread(void *task)
{
	wake_up_process(task);
}

static inline void sce_stop_kthread(void *task)
{
	kthread_stop(task);
}

static inline void sce_print_buf(u8 *name, u8 *buf, u32 len)
{
	u32 i;

	sce_info("%s[0x%x]:\n", name, len);
	for (i = 0; i < len; i += 16) {
		sce_info(
			"\"\\x%02x\\x%02x\\x%02x\\x%02x\\x%02x\\x%02x\\x%02x\\x%02x\\x%"
			"02x\\x%02x\\x%02x\\x%02x\\x%02x\\x%02x\\x%02x\\x%02x\"\n",
			buf[i + 0], buf[i + 1], buf[i + 2], buf[i + 3],
			buf[i + 4], buf[i + 5], buf[i + 6], buf[i + 7],
			buf[i + 8], buf[i + 9], buf[i + 10], buf[i + 11],
			buf[i + 12], buf[i + 13], buf[i + 14], buf[i + 15]);
	}
}

#endif
