// SPDX-License-Identifier: GPL-2.0-or-later
#include <linux/dma-mapping.h>
#include <linux/platform_device.h>
#include <linux/device.h>
#include <linux/cdev.h>
#include <linux/mutex.h>
#include <linux/acpi.h>
#include <linux/delay.h>
#include <linux/kthread.h>
#include <linux/version.h>
#include <asm/cputype.h>
#include "sce_driver_common.h"
#include "sce.h"
#include "sce_driver_mode.h"
#include "sce_kernel_crypto_mode.h"
#include "phytium_scto.h"

static struct sce_mutex proc_lock;
static struct proc_dir_entry *sce_dir;
static struct proc_dir_entry *sce_work_proc;
static u32 sce_work_state = OFF_MODE;
struct sce_module_context *sce;

static ssize_t sce_read_buffer(struct file *filep, char __user *buffer,
				   size_t len, loff_t *offset)
{
	char buf[16] = { 0 };
	u32 buflen;
	char output[3][16] = { "idle", "kernel crypto", "user dirver" };

	sprintf(buf, "%s\n", output[sce_work_state]);
	buflen = strlen(buf);

	if (buflen <= *offset)
		return 0;

	if (*offset + len > buflen)
		len = buflen - *offset;

	if (copy_to_user(buffer, buf + *offset, len)) {
		sce_err("%s:copy to user failed!\n", __func__);
		return -EFAULT;
	}

	*offset += len;

	return len;
}
static ssize_t sce_write_buffer(struct file *filep, const char __user *buffer,
				size_t len, loff_t *offset)
{
	u32 work_state;
	char buf[16] = { 0 };
	int ret = 0;

	if(module_refcount(THIS_MODULE)  > 0)
	{
		sce_err("module:phytium_sce is busy\n");
		return -1;
	}

	if (len >= 15) {
		sce_err("%s:buffer too large!len:0x%lx\n", __func__, len);
		return -EFAULT;
	}

	if (copy_from_user(buf, buffer, len)) {
		sce_err("%s:copy from user failed!\n", __func__);
		return -EFAULT;
	}

	if (kstrtou32(buf, 0, &work_state) < 0) {
		sce_err("Invalid input\r\n");
		return -EINVAL;
	}

	sce_info("%s:work state:%u\n", __func__, work_state);
	sce_lock_mutex(&proc_lock);
	if (work_state == OFF_MODE) {
		if (sce_work_state == DRIVER_MODE)
			sce_driver_unregister();
		if (sce_work_state == KERNEL_MODE)
			sce_kernel_crypto_unregister();
	} else if (work_state == KERNEL_MODE) {
		if (sce_work_state == OFF_MODE) {
			ret = sce_kernel_crypto_register();
			if (ret) {
				sce_err("%s:kernel_register fail!\n", __func__);
				sce_unlock_mutex(&proc_lock);
				return -1;
			}
		} else if (sce_work_state == DRIVER_MODE) {
			sce_driver_unregister();
			ret = sce_kernel_crypto_register();
			if (ret) {
				sce_err("%s:kernel_register fail!\n", __func__);
				sce_unlock_mutex(&proc_lock);
				return -1;
			}
		}
	} else if (work_state == DRIVER_MODE) {
		if (sce_work_state == OFF_MODE) {
			ret = sce_driver_register();
			if (ret) {
				sce_err("%s:driver_register fail!\n", __func__);
				sce_unlock_mutex(&proc_lock);
				return -1;
			}
		} else if (sce_work_state == KERNEL_MODE) {
			sce_kernel_crypto_unregister();
			ret = sce_driver_register();
			if (ret) {
				sce_err("%s:driver_register fail!\n", __func__);
				sce_unlock_mutex(&proc_lock);
				return -1;
			}
		}
	} else {
		sce_err("%s:illegal input work_state!\n", __func__);
		sce_unlock_mutex(&proc_lock);
		return -EINVAL;
	}
	sce_unlock_mutex(&proc_lock);
	sce_work_state = work_state;

	sce_info("%s:modify work state success!", __func__);

	return len;
}
#if KERNEL_VERSION(5, 6, 0) <= LINUX_VERSION_CODE
static const struct proc_ops sce_proc_fops = {
	.proc_read = sce_read_buffer,
	.proc_write = sce_write_buffer,
};
#else
static const struct file_operations sce_proc_fops = {
	.owner = THIS_MODULE,
	.read = sce_read_buffer,
	.write = sce_write_buffer,
};
#endif
static int sce_probe(struct platform_device *sdev)
{
	int ret = 0;
	struct resource *sce_resource = NULL;

	sce_resource = platform_get_resource(sdev, IORESOURCE_MEM, 0);
	sce_info("resource start:0x%016llx end:0x%016llx flag:0x%lx\n",
		sce_resource->start, sce_resource->end, sce_resource->flags);
	
	sce = devm_kzalloc(&sdev->dev, sizeof(struct sce_module_context),
		GFP_KERNEL | GFP_DMA);
	if (!sce)
		return -ENOMEM;
	sce->dev = &sdev->dev;
	sce->base = devm_ioremap_resource(sce->dev, sce_resource);
	if (!sce->base) {
		pr_err("phys_to_virt failed!\n");
		return -1;
	}
	sce->base_ioaddr = sce_resource->start;

	sce_init_mutex(&proc_lock);
	sce_dir = proc_mkdir("sce_dir", NULL);
	if (sce_dir == NULL) {
		sce_err("create proc dir failed!\n");
		return -1;
	}

	sce_work_proc =
		proc_create("sce_work_state", 0600, sce_dir, &sce_proc_fops);
	if (sce_work_proc == NULL) {
		sce_err("create proc entry sce_work_state failed!\n");
		proc_remove(sce_dir);
		return -1;
	}
	sce_info("Loading SCE DRIVER Module version %s success\n", SCE_DRIVER_VERSION);
	ret = scto_device_probe();
	if (ret) {
		sce_err("sm2 register failed!\n");
		proc_remove(sce_work_proc);
		proc_remove(sce_dir);
		return ret;
	}

	return ret;
}

static int sce_remove(struct platform_device *pdev)
{

	if (sce_work_state == KERNEL_MODE)
		sce_kernel_crypto_unregister();
	else if (sce_work_state == DRIVER_MODE)
		sce_driver_unregister();

	proc_remove(sce_work_proc);
	proc_remove(sce_dir);
	scto_device_remove();
	sce_info("Remove SCE DRIVER Module version %s\n", SCE_DRIVER_VERSION);
	return 0;
}

static int sce_suspend(struct device *dev)
{
	return 0;
}

static int sce_resume(struct device *dev)
{
	return 0;
}

/*of*/
static const struct of_device_id crypto_of_id_table[] = { {.compatible = "SCE"}, {}};

MODULE_DEVICE_TABLE(of, crypto_of_id_table);

/*acpi*/
static const struct acpi_device_id crypto_acpi_id_table[] = { { .id = "PHYT0051" }, {} };

MODULE_DEVICE_TABLE(acpi, crypto_acpi_id_table);

/*pm ops*/
static const struct dev_pm_ops sce_pm_ops = {
	.suspend = sce_suspend,
	.resume = sce_resume,
};

static struct platform_driver sce_driver = {
	.probe = sce_probe,
	.remove = sce_remove,
	.driver =
	{
		.name = "phytium-sce",
		.of_match_table = crypto_of_id_table,
		.acpi_match_table = crypto_acpi_id_table,
		.pm = &sce_pm_ops,
	},
};


static int sce_device_probe(void)
{
	int ret = 0;

	ret = platform_driver_register(&sce_driver);
	if (ret) {
		pr_err("Register driver version %s failed, ret %d\r\n", SCE_DRIVER_VERSION, ret);
		return ret;
	}
	return ret;
}

static void sce_device_remove(void)
{
	platform_driver_unregister(&sce_driver);
}
module_init(sce_device_probe);
module_exit(sce_device_remove);
MODULE_DESCRIPTION("PHYTIUM CRYPTO ENGINE");
MODULE_LICENSE("GPL");


