// SPDX-License-Identifier: GPL-2.0-or-later
/*
 * File Name:ccm.c - Phytium SDK for CCM
 *
 * Copyright (C) 2022 Phytium Technology Co.,Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */

#include <asm/memory.h>
#include <linux/delay.h>
#include <crypto/internal/hash.h>
#include <crypto/hash.h>
#include <crypto/internal/skcipher.h>
#include <crypto/skcipher.h>
#include <crypto/internal/aead.h>
#include <crypto/aead.h>
#include <linux/io.h>
#include <linux/crypto.h>
#include <asm/cacheflush.h>
#include <linux/module.h>
#include <linux/irq.h>
#include <linux/platform_device.h>
#include <linux/dma-mapping.h>
#include <linux/device.h>
#include <linux/cdev.h>
#include <linux/types.h>
#include <linux/mutex.h>
#include <linux/acpi.h>
#include <linux/ioport.h>
#include <linux/io.h>
#include <linux/delay.h>
#include <linux/kthread.h>
#include <linux/version.h>

#include "sce_kernel_common.h"
#include "sce_kernel.h"
#include "sce_kernel_reg.h"
#include "sce.h"
#include "ccm.h"
#include "cipher.h"
#define AEAD_INIT_128 0xFFFFFFFFFFFFFFFFull

static char *ccm_alg_name[] = { "ccm(sm4)", "ccm(aes)" };
static u8 alg[] = { SM4, AES };
static u8 block_size[] = { AES_BLOCK_BYTE_LEN, SM4_BLOCK_BYTE_LEN };
static u8 iv_size[] = { AES_IV_BYTE_LEN, SM4_IV_BYTE_LEN };
static u8 key_len[] = { AES_MIN_KEY_BYTE_LEN, SM4_KEY_BYTE_LEN };

static union sca_init_cmd SCA_INIT = {
	.trigger_intr = 0,
	.mode = 2,
	.iv_load = 1,
	.key_len = 0,
	.key_src = 2,
	.algo = 0,
	.reserved = 0,
	.slot_id = 8,
	.opcode = 0x80,
};

static union sca_proc_cmd SCA_PROC = {
	.trigger_intr = 0,
	.dst_src_addr_type = 0,
	.is_last_proc = 1,
	.op_mode = 0,
	.reserved = 0,
	.slot_id = 8,
	.opcode = 0x40,
};

static union sca_finish_cmd SCA_FINISH = {
	.trigger_intr = 0,
	.mac_tlen = 0,
	.reserved = 0,
	.slot_id = 8,
	.opcode = 0x20,
};
static void bit128_xor(uint8_t *a, uint8_t *b, uint8_t *out)
{
	int i;

	for (i = 0; i < 16; i++)
		out[i] = a[i] ^ b[i];
}

static void send_cmd(struct cipher_cmd *cmd)
{
	union sca_init_cmd sca_init_cmd;
	union sca_proc_cmd sca_proc_cmd;
	union sca_finish_cmd sca_finish_cmd;
	union sca_stat_reg stat;
	union sca_csq_reg csq;
	u64 key_paddr, iv_paddr, src_paddr, dst_paddr;
	unsigned long flags;

	key_paddr = cmd->key_ioaddr | 0x300000000000;
	iv_paddr = cmd->iv_ioaddr | 0x300000000000;
	src_paddr = cmd->src_ioaddr | 0x300000000000;
	dst_paddr = cmd->dst_ioaddr | 0x300000000000;
	sca_init_cmd.data = SCA_INIT.data;
	sca_init_cmd.mode = cmd->mode;
	sca_init_cmd.iv_load = cmd->iv_load;
	sca_init_cmd.key_len = cmd->key_len_flag;
	sca_init_cmd.algo = cmd->algo;

	sca_proc_cmd.data = SCA_PROC.data;
	sca_proc_cmd.is_last_proc = cmd->is_last_proc;
	sca_proc_cmd.dst_src_addr_type = cmd->dst_src_addr_type;
	sca_proc_cmd.op_mode = cmd->op_mode;

	sca_finish_cmd.data = SCA_FINISH.data;
	sca_finish_cmd.mac_tlen = cmd->mac_tlen - 1;
	spin_lock_irqsave(&sce_kernel.cipher_lock, flags);

	REG_WR_32(sce_kernel.base, SCA_CQ_FUNC_OFFSET, sca_init_cmd.data);
	do {
		stat.data = REG_RD_32_SYNC(sce_kernel.base, SCA_STAT_OFFSET);
	} while (stat.cq_avail_slots < 4);

	if (cmd->iv_load != 0) {
		REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET,
			  iv_paddr & 0xFFFFFFFF);
		REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET, iv_paddr >> 32);
	}
	REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET, key_paddr & 0xFFFFFFFF);
	REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET, key_paddr >> 32);

	REG_WR_32(sce_kernel.base, SCA_CQ_FUNC_OFFSET, sca_proc_cmd.data);

	do {
		stat.data = REG_RD_32_SYNC(sce_kernel.base, SCA_STAT_OFFSET);
	} while (stat.cq_avail_slots < 5);

	if (cmd->dst_src_addr_type == 0)
		REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET,
			  cmd->src_len - 1);

	REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET, src_paddr & 0xFFFFFFFF);
	REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET, src_paddr >> 32);
	if (cmd->mac_tlen == 0) {
		REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET,
			  dst_paddr & 0xFFFFFFFF);
		REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET, dst_paddr >> 32);
	}

	REG_WR_32(sce_kernel.base, SCA_CQ_FUNC_OFFSET, sca_finish_cmd.data);
	do {
		stat.data = REG_RD_32_SYNC(sce_kernel.base, SCA_STAT_OFFSET);
	} while (stat.cq_avail_slots < 2);

	if (cmd->mac_tlen != 0) {
		REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET,
			  dst_paddr & 0xFFFFFFFF);
		REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET, dst_paddr >> 32);
	}

	do {
		csq.data = REG_RD_32_SYNC(sce_kernel.base, SCA_CSQ_OFFSET);
	} while (csq.opcode_idx != FINISH);

	spin_unlock_irqrestore(&sce_kernel.cipher_lock, flags);
}

static inline int sce_ccm_encrypt(struct aead_request *req)
{
	struct crypto_aead *tfm = __crypto_aead_cast(req->base.tfm);
	struct ccm_alg_context *actx =
		(void *)(((u64)(crypto_aead_ctx(tfm)) + CACHE_LINE_SIZE) &
			 (~(CACHE_LINE_SIZE - 1)));
	struct ccm_session_context *sctx =
		(void *)(((u64)(aead_request_ctx(req)) + CACHE_LINE_SIZE) &
			 (~(CACHE_LINE_SIZE - 1)));
	struct link_list_item *src_link_list = NULL, *dst_link_list = NULL;
	struct scatterlist *sg;
	int src_nents = 0, dst_nents = 0, i, aad_head;
	u32 dma_mode;
	u64 src_ioaddr = 0, dst_ioaddr = 0, iv_ioaddr, temp;
	struct ccm_flag *flag;
	u32 cur_aad_len;
	struct cipher_cmd *cmd;
	int flag1 = 0;
	u8 cmd_id = ENC;
	int authsize = tfm->authsize;

	cmd = &sctx->ccm_cmd;

	if (req->iv[0] < 1 || req->iv[0] > 7) {
		// sce_err("%s:iv[0] value error!\n", __func__);
		return -EINVAL;
	}

	for (i = 0; i < req->iv[0] + 1; i++)
		req->iv[15 - i] = 0x00;

	if (req->iv[0] == 1 && req->cryptlen > 65535) {
		if (cmd_id == ENC)
			// sce_err("%s:bad enc cryptlen error!\n", __func__);
			return -EOVERFLOW;
		return -EBADMSG;

	}
	src_nents = sg_nents(req->src);
	dst_nents = sg_nents(req->dst);
	if (unlikely(!src_nents || !dst_nents)) {
		sce_err("%s:src_nents[%d] dst_nents[%d]! err!\n", __func__,
			src_nents, dst_nents);
		return -1;
	}
	if (unlikely(src_nents + 6 > MAX_LINKS)) {
		src_link_list = sce_malloc(sizeof(struct link_list_item) *
					   (src_nents + 6));
		if (src_link_list == NULL) {
			sce_err("sce_malloc failed!\n");
			return -EMLINK;
		}
	} else {
		src_link_list = sctx->src_list_buf;
	}
	memset(src_link_list, 0, sizeof(*src_link_list));
	if (unlikely(dst_nents + 6 > MAX_LINKS)) {
		dst_link_list = sce_malloc(sizeof(struct link_list_item) *
					   (dst_nents + 6));
		if (dst_link_list == NULL) {
			if (src_link_list != sctx->src_list_buf)
				sce_free(src_link_list);
			sce_err("sce_malloc failed!\n");
			return -EMLINK;
		}
	} else {
		dst_link_list = sctx->dst_list_buf;
	}
	memset(dst_link_list, 0, sizeof(*dst_link_list));
	memset(sctx->iv, 0, actx->iv_size);
	memcpy(sctx->iv, req->iv, actx->iv_size);
	iv_ioaddr = virt_to_phys(sctx->iv);

	if (req->assoclen) {
		if (req->assoclen < (0x10000 - 0x100))
			aad_head = 2;
		else if (req->assoclen <= 0xFFFFFFFF)
			aad_head = 6;
		else
			aad_head = 10;
	} else
		aad_head = 0;

	cur_aad_len = 16 + aad_head + req->assoclen;
	memset(sctx->pad, 0, 16 + aad_head);
	memcpy(sctx->pad, req->iv, 16);
	flag = (struct ccm_flag *)sctx->pad;
	flag->tlen = (authsize - 2) / 2;
	flag->aad = !!req->assoclen;
	flag->reserved = 0;
	for (i = 0; i <= flag->qlen; i++) {
		if (i >= 4)
			sctx->pad[15 - i] = 0;
		else
			sctx->pad[15 - i] = (req->cryptlen >> (8 * i)) & 0xFF;
	}
	switch (aad_head) {
	case 2:
		for (i = 0; i < 2; i++)
			sctx->pad[17 - i] = (req->assoclen >> (8 * i)) & 0xFF;
		break;
	case 6:
		sctx->pad[16] = 0xFF;
		sctx->pad[17] = 0xFE;
		for (i = 0; i < 4; i++)
			sctx->pad[21 - i] = (req->assoclen >> (8 * i)) & 0xFF;
		break;
	case 10:
		sce_err("%s:unsupport assoclen[%d]!\n", __func__,
			req->assoclen);
		return -1;
	default:
		break;
	}
	temp = req->assoclen;
	memset(sctx->cbcmac_result, 0, sizeof(sctx->cbcmac_result));
	memset(sctx->cbcmac_zero2, 0, sizeof(sctx->cbcmac_zero2));
	memset(sctx->cbcmac_zero, 0, sizeof(sctx->cbcmac_zero));
	i = 0;
	flag1 = 0;
	src_ioaddr = virt_to_phys(src_link_list);
	dst_ioaddr = virt_to_phys(sctx->cbcmac_result);

	src_link_list[i].addr = virt_to_phys(sctx->pad);
	src_link_list[i].addr |= 0x300000000000;
	src_link_list[i].size = 16 + aad_head - 1;
	i++;
	for (sg = req->src; sg; sg = sg_next(sg)) {
		if (!sg->length)
			continue;
		if (!temp && (flag1 == 1))
			break;

		if (likely(temp != 0)) {
			src_link_list[i].addr = sg_phys(sg);
			src_link_list[i].addr |= 0x300000000000;
			src_link_list[i].size = temp > sg->length ?
							(sg->length - 1) :
							(temp - 1);
			i++;
		}
		if (unlikely(temp <= sg->length && flag1 == 0)) {
			if (cur_aad_len & 0xf) {
				src_link_list[i].addr =
					virt_to_phys(sctx->cbcmac_zero2);
				src_link_list[i].addr |= 0x300000000000;
				src_link_list[i].size =
					((16 - (cur_aad_len & 0xf)) & 0xf) - 1;
				i++;
			}
			if (temp < sg->length) {
				if (req->cryptlen < sg->length - temp) {
					if (req->cryptlen) {
						src_link_list[i].addr =
							sg_phys(sg) + temp;
						src_link_list[i].addr |=
							0x300000000000;
						src_link_list[i].size =
							req->cryptlen - 1;
						i++;
					}
					break;
				}
				src_link_list[i].addr =
					sg_phys(sg) + temp;
				src_link_list[i].addr |= 0x300000000000;
				src_link_list[i].size =
					sg->length - temp - 1;
				temp = req->cryptlen -
				       (sg->length - temp);
				i++;
			} else {
				temp = req->cryptlen;
			}
			flag1 = 1;
		} else {
			if (temp >= sg->length)
				temp -= sg->length;
			else
				temp = 0;
		}
	}
	if (req->cryptlen & 0xf) {
		src_link_list[i].addr = virt_to_phys(sctx->cbcmac_zero);
		src_link_list[i].addr |= 0x300000000000;
		src_link_list[i].size = 16 - (req->cryptlen & 0xf) - 1;
		i++;
	}
	src_link_list[i].size = AEAD_INIT_128;
	src_link_list[i].addr = 0;
	i++;
	cmd->algo = actx->alg;
	cmd->mode = CBC_MAC;
	cmd->iv_load = 0;
	cmd->key_len_flag = actx->key_len_flag;
	cmd->op_mode = ENC;
	cmd->is_last_proc = 1;
	cmd->dst_src_addr_type = 2;

	cmd->mac_tlen = 16;
	cmd->src_len = 64;
	cmd->key_ioaddr = actx->key_ioaddr;
	cmd->src_ioaddr = src_ioaddr;
	cmd->dst_ioaddr = dst_ioaddr;
	cmd->iv_ioaddr = iv_ioaddr;

	send_cmd(cmd);

	memset(sctx->ecb_result, 0, sizeof(sctx->ecb_result));
	cmd->mode = ECB;
	cmd->dst_src_addr_type = 0;
	cmd->mac_tlen = 0;
	cmd->src_len = 16;

	cmd->src_ioaddr = iv_ioaddr;
	cmd->dst_ioaddr = virt_to_phys(sctx->ecb_result);
	send_cmd(cmd);
	bit128_xor(sctx->ecb_result, sctx->cbcmac_result, sctx->tag);
	flush_dcache_page(virt_to_page(&sctx->tag));

	flag1 = 0;
	i = 0;
	temp = req->assoclen;
	memset(src_link_list, 0, sizeof(*src_link_list));
	dma_mode = 3;
	if (req->cryptlen != 0) {
		for (sg = req->src; sg; sg = sg_next(sg)) {
			if (!sg->length)
				continue;
			if (sg->length <= temp && flag1 == 0) {
				temp -= sg->length;
				continue;
			} else if (flag1 == 0 && sg->length > temp) {
				if (sg->length >= temp + req->cryptlen) {
					src_link_list[i].addr =
						sg_phys(sg) + temp;
					src_link_list[i].addr |= 0x300000000000;
					src_link_list[i].size =
						req->cryptlen - 1;
					i++;
					// dma_mode |= 1;
					break;
				}
			}
			if (flag1 == 0) {
				if (sg->length > temp) {
					src_link_list[i].addr =
						sg_phys(sg) + temp;
					src_link_list[i].addr |= 0x300000000000;
					src_link_list[i].size =
						sg->length - temp - 1;
					temp = req->cryptlen -
					       src_link_list[i].size - 1;
					i++;
					flag1 = 1;
				}
				continue;
			}
			// dma_mode |= 1;
			if (sg->length < temp) {
				src_link_list[i].addr = sg_phys(sg);
				src_link_list[i].addr |= 0x300000000000;
				src_link_list[i].size = sg->length - 1;
				temp -= sg->length;
				i++;
			} else {
				src_link_list[i].addr = sg_phys(sg);
				src_link_list[i].addr |= 0x300000000000;
				src_link_list[i].size = temp - 1;
				i++;
				break;
			}
		}
		src_link_list[i].addr = 0;
		src_link_list[i].size = AEAD_INIT_128;

		flag1 = 0;
		i = 0;
		memset(dst_link_list, 0, sizeof(*dst_link_list));
		temp = req->assoclen;
		for (sg = req->dst; sg; sg = sg_next(sg)) {
			if (!sg->length)
				continue;
			if (sg->length <= temp && flag1 == 0) {
				temp -= sg->length;
				continue;
			} else if (flag1 == 0 && sg->length > temp) {
				if (sg->length >= temp + req->cryptlen) {
					dst_link_list[i].addr =
						sg_phys(sg) + temp;
					dst_link_list[i].addr |= 0x300000000000;
					dst_link_list[i].size =
						req->cryptlen - 1;
					i++;
					break;
				}
			}
			if (flag1 == 0) {
				if (sg->length > temp) {
					dst_link_list[i].addr =
						sg_phys(sg) + temp;
					dst_link_list[i].addr |= 0x300000000000;
					dst_link_list[i].size =
						sg->length - temp - 1;
					temp = req->cryptlen -
					       dst_link_list[i].size - 1;
					i++;
					flag1 = 1;
				}
				continue;
			}
			if (sg->length < temp) {
				dst_link_list[i].addr = sg_phys(sg);
				dst_link_list[i].addr |= 0x300000000000;
				dst_link_list[i].size = sg->length - 1;
				temp -= sg->length;
				i++;
			} else {
				dst_link_list[i].addr = sg_phys(sg);
				dst_link_list[i].addr |= 0x300000000000;
				dst_link_list[i].size = temp - 1;
				i++;
				break;
			}
		}
		dst_link_list[i].addr = 0;
		dst_link_list[i].size = AEAD_INIT_128;
		i++;
		dma_mode = 3;

		src_ioaddr = virt_to_phys(src_link_list);
		dst_ioaddr = virt_to_phys(dst_link_list);

		sctx->iv[15] += 1;
		cmd->mode = CTR;
		cmd->iv_load = 1;
		cmd->dst_src_addr_type = dma_mode;
		cmd->src_len = req->cryptlen;
		cmd->mac_tlen = 0;
		cmd->iv_ioaddr = virt_to_phys(sctx->iv);
		cmd->src_ioaddr = src_ioaddr;
		cmd->dst_ioaddr = dst_ioaddr;
		send_cmd(cmd);
	}
	sg_copy_buffer(req->dst, sg_nents(req->dst), sctx->tag, 16,
		       req->assoclen + req->cryptlen, FALSE);

	return 0;
}

int sce_ccm_decrypt(struct aead_request *req)
{
	struct crypto_aead *tfm = __crypto_aead_cast(req->base.tfm);
	struct ccm_alg_context *actx =
		(void *)(((u64)(crypto_aead_ctx(tfm)) + CACHE_LINE_SIZE) &
			 (~(CACHE_LINE_SIZE - 1)));
	struct ccm_session_context *sctx =
		(void *)(((u64)(aead_request_ctx(req)) + CACHE_LINE_SIZE) &
			 (~(CACHE_LINE_SIZE - 1)));
	struct link_list_item *src_link_list = NULL, *dst_link_list = NULL;
	struct scatterlist *sg;
	int src_nents = 0, dst_nents = 0, i;
	u32 cryptlen, dma_mode;
	u64 src_ioaddr = 0, dst_ioaddr = 0, iv_ioaddr, temp;
	struct cipher_cmd *cmd;
	int flag1 = 0, aad_head;
	int authsize = tfm->authsize;
	struct ccm_flag *flag;
	u32 cur_aad_len;

	u8 cmd_id = DEC;

	cmd = &sctx->ccm_cmd;

	if (req->iv[0] < 1 || req->iv[0] > 7)
		// sce_err("%s:iv[0] value error!\n", __func__);
		return -EINVAL;

	for (i = 0; i < req->iv[0] + 1; i++)
		req->iv[15 - i] = 0x00;

	if (req->iv[0] == 1 && req->cryptlen > 65535) {
		if (cmd_id == ENC)
			// sce_err("%s:bad enc cryptlen error!\n", __func__);
			return -EOVERFLOW;
		return -EBADMSG;
	}

	src_nents = sg_nents(req->src);
	dst_nents = sg_nents(req->dst);

	if (unlikely(!src_nents || !dst_nents)) {
		sce_err("%s:src_nents[%d] dst_nents[%d]! err!\n", __func__,
			src_nents, dst_nents);
		return -1;
	}
	if (unlikely(src_nents + 6 > MAX_LINKS)) {
		src_link_list = sce_malloc(sizeof(struct link_list_item) *
					   (src_nents + 6));
		if (src_link_list == NULL) {
			sce_err("sce_malloc failed!\n");
			return -EMLINK;
		}
	} else {
		src_link_list = sctx->src_list_buf;
	}
	memset(src_link_list, 0, sizeof(*src_link_list));
	if (unlikely(dst_nents + 6 > MAX_LINKS)) {
		dst_link_list = sce_malloc(sizeof(struct link_list_item) *
					   (dst_nents + 6));
		if (dst_link_list == NULL) {
			if (src_link_list != sctx->src_list_buf)
				sce_free(src_link_list);
			sce_err("sce_malloc failed!\n");
			return -EMLINK;
		}
	} else {
		dst_link_list = sctx->dst_list_buf;
	}
	memset(dst_link_list, 0, sizeof(*dst_link_list));

	src_ioaddr = virt_to_phys(src_link_list);
	dst_ioaddr = virt_to_phys(dst_link_list);

	memset(sctx->iv, 0, actx->iv_size);
	memcpy(sctx->iv, req->iv, actx->iv_size);
	iv_ioaddr = virt_to_phys(sctx->iv);
	flag1 = 0;
	i = 0;
	temp = req->assoclen;
	memset(src_link_list, 0, sizeof(*src_link_list));
	dma_mode = 3;

	if (req->cryptlen - authsize) {
		for (sg = req->src; sg; sg = sg_next(sg)) {
			if (!sg->length)
				continue;
			if (sg->length <= temp && flag1 == 0) {
				temp -= sg->length;
				continue;
			} else if (flag1 == 0 && sg->length > temp) {
				if (sg->length >=
				    temp + req->cryptlen - authsize) {
					src_link_list[i].addr =
						sg_phys(sg) + temp;
					src_link_list[i].addr |= 0x300000000000;
					src_link_list[i].size =
						req->cryptlen - authsize - 1;
					temp = 0;
					i++;
					break;
				}
			}
			if (flag1 == 0) {
				if (sg->length > temp) {
					src_link_list[i].addr =
						sg_phys(sg) + temp;
					src_link_list[i].addr |= 0x300000000000;
					src_link_list[i].size =
						sg->length - temp - 1;
					temp = req->cryptlen - authsize -
					       src_link_list[i].size - 1;
					i++;
					flag1 = 1;
				}
				continue;
			}
			if (sg->length < temp) {
				src_link_list[i].addr = sg_phys(sg);
				src_link_list[i].addr |= 0x300000000000;
				src_link_list[i].size = sg->length - 1;
				temp -= sg->length;
				i++;
			} else {
				src_link_list[i].addr = sg_phys(sg);
				src_link_list[i].addr |= 0x300000000000;
				src_link_list[i].size = temp - 1;
				i++;
				break;
			}
		}

		src_link_list[i].addr = 0;
		src_link_list[i].size = AEAD_INIT_128;
		i++;
		flag1 = 0;

		i = 0;
		temp = req->assoclen;
		memset(dst_link_list, 0, sizeof(*dst_link_list));
		for (sg = req->dst; sg; sg = sg_next(sg)) {
			if (!sg->length)
				continue;
			if (sg->length <= temp && flag1 == 0) {
				temp -= sg->length;
				continue;
			} else if (flag1 == 0 && sg->length > temp) {
				if (sg->length >=
				    temp + req->cryptlen - authsize) {
					dst_link_list[i].addr =
						sg_phys(sg) + temp;
					dst_link_list[i].addr |= 0x300000000000;
					dst_link_list[i].size =
						req->cryptlen - authsize - 1;
					temp = 0;
					i++;
					break;
				}
			}
			if (flag1 == 0) {
				if (sg->length > temp) {
					dst_link_list[i].addr =
						sg_phys(sg) + temp;
					dst_link_list[i].addr |= 0x300000000000;
					dst_link_list[i].size =
						sg->length - temp - 1;
					temp = req->cryptlen - authsize -
					       dst_link_list[i].size - 1;
					i++;
					flag1 = 1;
				}
				continue;
			}
			if (sg->length < temp) {
				dst_link_list[i].addr = sg_phys(sg);
				dst_link_list[i].addr |= 0x300000000000;
				dst_link_list[i].size = sg->length - 1;
				temp -= sg->length;
				i++;
			} else {
				dst_link_list[i].addr = sg_phys(sg);
				dst_link_list[i].addr |= 0x300000000000;
				dst_link_list[i].size = temp - 1;
				i++;
				break;
			}
		}
		dst_link_list[i].addr = 0;
		dst_link_list[i].size = AEAD_INIT_128;
		i++;
		dma_mode = 3;
		sctx->iv[15] += 1;
		cmd->algo = actx->alg;
		cmd->mode = CTR;
		cmd->iv_load = 1;
		cmd->key_len_flag = actx->key_len_flag;
		cmd->op_mode = ENC;
		cmd->is_last_proc = 1;
		cmd->dst_src_addr_type = dma_mode;

		cmd->mac_tlen = 0;
		cmd->key_ioaddr = actx->key_ioaddr;
		cmd->src_ioaddr = src_ioaddr;
		cmd->dst_ioaddr = dst_ioaddr;
		cmd->iv_ioaddr = virt_to_phys(sctx->iv);

		send_cmd(cmd);
	}
	if (req->assoclen) {
		if (req->assoclen < (0x10000 - 0x100))
			aad_head = 2;
		else if (req->assoclen <= 0xFFFFFFFF)
			aad_head = 6;
		else
			aad_head = 10;
	} else
		aad_head = 0;


	cur_aad_len = 16 + aad_head + req->assoclen;
	memset(sctx->pad, 0, 16 + aad_head);
	memcpy(sctx->pad, req->iv, 16);
	flag = (struct ccm_flag *)sctx->pad;
	flag->tlen = (authsize - 2) / 2;
	flag->aad = !!req->assoclen;
	flag->reserved = 0;
	for (i = 0; i <= flag->qlen; i++) {
		if (i >= 4) {
			sctx->pad[15 - i] = 0;
		} else {
			sctx->pad[15 - i] =
				((req->cryptlen - authsize) >> (8 * i)) & 0xFF;
		}
	}
	switch (aad_head) {
	case 2:
		for (i = 0; i < 2; i++)
			sctx->pad[17 - i] = (req->assoclen >> (8 * i)) & 0xFF;
		break;
	case 6:
		sctx->pad[16] = 0xFF;
		sctx->pad[17] = 0xFE;
		for (i = 0; i < 4; i++)
			sctx->pad[21 - i] = (req->assoclen >> (8 * i)) & 0xFF;
		break;
	case 10:
		sce_err("%s:unsupport assoclen[%d]!\n", __func__,
			req->assoclen);
		return -1;
	default:
		break;
	}

	cryptlen = req->cryptlen;
	temp = req->assoclen;
	memset(sctx->cbcmac_result, 0, sizeof(sctx->cbcmac_result));

	memset(sctx->cbcmac_zero2, 0, sizeof(sctx->cbcmac_zero2));
	memset(sctx->cbcmac_zero, 0, sizeof(sctx->cbcmac_zero));
	i = 0;
	flag1 = 0;
	src_ioaddr = virt_to_phys(src_link_list);
	dst_ioaddr = virt_to_phys(sctx->cbcmac_result);
	src_link_list[i].addr = virt_to_phys(sctx->pad);
	src_link_list[i].addr |= 0x300000000000;
	src_link_list[i].size = 16 + aad_head - 1;
	i++;
	for (sg = req->src; sg; sg = sg_next(sg)) {
		if (!sg->length)
			continue;
		if (!temp)
			break;
		if (likely(temp)) {
			src_link_list[i].addr = sg_phys(sg);
			src_link_list[i].addr |= 0x300000000000;
			src_link_list[i].size = temp > sg->length ?
							(sg->length - 1) :
							(temp - 1);
			i++;
		}
		if (unlikely(temp <= sg->length)) {
			if (cur_aad_len & 0xf) {
				src_link_list[i].addr =
					virt_to_phys(sctx->cbcmac_zero2);
				src_link_list[i].addr |= 0x300000000000;
				src_link_list[i].size =
					(((16 - (cur_aad_len & 0xf)) & 0xf)) -
					1;
				i++;
			}
			break;
		}
		temp -= sg->length;
	}

	temp = req->assoclen;
	flag1 = 0;
	for (sg = req->dst; sg; sg = sg_next(sg)) {
		if (!sg->length)
			continue;
		if (!temp && (flag1 == 1))
			break;
		if (temp >= sg->length && flag1 == 0) {
			temp -= sg->length;
			continue;
		} else if (flag1 == 0 && sg->length > temp) {
			if (sg->length >= temp + req->cryptlen - authsize) {
				if (req->cryptlen - authsize) {
					src_link_list[i].addr =
						sg_phys(sg) + temp;
					flush_dcache_page(
						phys_to_page(sg_phys(sg)));
					src_link_list[i].addr |= 0x300000000000;
					src_link_list[i].size =
						req->cryptlen - authsize - 1;
					i++;
				}
				break;
			}
			src_link_list[i].addr = sg_phys(sg) + temp;
			src_link_list[i].addr |= 0x300000000000;
			src_link_list[i].size = sg->length - temp - 1;
			temp = req->cryptlen - authsize -
			       src_link_list[i].size - 1;
			i++;
			flag1 = 1;
			continue;
		}

		if (sg->length < temp) {
			src_link_list[i].addr = sg_phys(sg);
			src_link_list[i].addr |= 0x300000000000;
			src_link_list[i].size = sg->length - 1;
			temp -= sg->length;
			i++;
		} else {
			src_link_list[i].addr = sg_phys(sg);
			src_link_list[i].addr |= 0x300000000000;
			src_link_list[i].size = temp - 1;
			i++;
			break;
		}
	}

	if ((req->cryptlen - authsize) & 0xf) {
		src_link_list[i].addr = virt_to_phys(sctx->cbcmac_zero);
		src_link_list[i].addr |= 0x300000000000;
		src_link_list[i].size =
			((16 - ((req->cryptlen - authsize) & 0xf)) & 0xf) - 1;
		i++;
	}

	src_link_list[i].size = AEAD_INIT_128;
	src_link_list[i].addr = 0;
	i++;

	src_ioaddr = virt_to_phys(src_link_list);
	dst_ioaddr = virt_to_phys(sctx->cbcmac_result);
	iv_ioaddr = virt_to_phys(sctx->iv);
	cmd->algo = actx->alg;
	cmd->mode = CBC_MAC;
	cmd->iv_load = 0;
	cmd->key_len_flag = actx->key_len_flag;
	cmd->op_mode = ENC;
	cmd->is_last_proc = 1;
	cmd->dst_src_addr_type = 2;

	cmd->mac_tlen = 16;
	cmd->src_len = req->cryptlen - authsize + req->assoclen;
	cmd->key_ioaddr = actx->key_ioaddr;
	cmd->src_ioaddr = src_ioaddr;
	cmd->dst_ioaddr = dst_ioaddr;
	cmd->iv_ioaddr = iv_ioaddr;

	send_cmd(cmd);

	iv_ioaddr = virt_to_phys(sctx->iv);

	memset(sctx->iv, 0, 16);
	memset(sctx->pad, 0, 16);
	memcpy(sctx->iv, req->iv, 16);
	iv_ioaddr = virt_to_phys(sctx->iv);
	cmd->mode = ECB;
	cmd->dst_src_addr_type = 0;
	cmd->mac_tlen = 0;
	cmd->src_len = 16;
	cmd->iv_load = 0;
	cmd->src_ioaddr = iv_ioaddr;
	cmd->key_ioaddr = actx->key_ioaddr;
	cmd->dst_ioaddr = virt_to_phys(sctx->pad);

	send_cmd(cmd);
	bit128_xor(sctx->pad, sctx->cbcmac_result, sctx->cbcmac_result);

	sg_copy_buffer(req->src, sg_nents(req->src), sctx->tag, 16,
		       req->cryptlen + req->assoclen - authsize, TRUE);
	if (memcmp(sctx->tag, sctx->cbcmac_result, authsize))
		return -EBADMSG;

	return 0;
}

static int sce_ccm_init(struct crypto_aead *tfm)
{
	int i;
	struct ccm_alg_context *actx =
		(void *)(((u64)(crypto_aead_ctx(tfm)) + CACHE_LINE_SIZE) &
			 (~(CACHE_LINE_SIZE - 1)));

	memset(actx, 0, sizeof(struct ccm_alg_context));
	crypto_aead_set_reqsize(tfm, sizeof(struct ccm_session_context) +
					     CACHE_LINE_SIZE);

	for (i = 0; i < ARRAY_SIZE(ccm_alg_name); i++) {
		if (memcmp(tfm->base.__crt_alg->cra_name, ccm_alg_name[i],
			   strlen(ccm_alg_name[i])) == 0) {
			actx->alg = alg[i];
			actx->mode = CCM;
			actx->block_size = block_size[i];
			actx->block_shift = 4;
			actx->iv_size = iv_size[i];
			actx->key_len = key_len[i];
			actx->key_len_flag = 0;
			actx->key_ioaddr = virt_to_phys(actx->key);
		}
	}

	if (actx->block_size == 0)
		return -EINVAL;

	return 0;
}

static int sce_ccm_setkey(struct crypto_aead *tfm, const u8 *key,
			  unsigned int keylen)
{
	struct ccm_alg_context *actx =
		(void *)(((u64)(crypto_aead_ctx(tfm)) + CACHE_LINE_SIZE) &
			 (~(CACHE_LINE_SIZE - 1)));

	if (actx->alg == AES) {
		switch (keylen) {
		case AES_MIN_KEY_BYTE_LEN:
			actx->key_len = AES_MIN_KEY_BYTE_LEN;
			actx->key_len_flag = 0;
			break;
		case AES_MID_KEY_BYTE_LEN:
			actx->key_len = AES_MID_KEY_BYTE_LEN;
			actx->key_len_flag = 1;
			break;
		case AES_MAX_KEY_BYTE_LEN:
			actx->key_len = AES_MAX_KEY_BYTE_LEN;
			actx->key_len_flag = 2;
			break;
		default:
			return -EINVAL;
		}
	} else {
		if (keylen != SM4_KEY_BYTE_LEN)
			return -EINVAL;
		actx->key_len_flag = 0;
		actx->key_len = SM4_KEY_BYTE_LEN;
	}

	if (unlikely(sizeof(actx->key) < keylen))
		return -EINVAL;


	memcpy(actx->key, key, keylen);

	return 0;
}

static int sce_ccm_setauthsize(struct crypto_aead *tfm, unsigned int authsize)
{
	switch (authsize) {
	case 4:
	case 6:
	case 8:
	case 10:
	case 12:
	case 14:
	case 16:
		break;
	default:
		return -EINVAL;
	}

	return 0;
}

static struct aead_alg sce_ccm_alg[] = {
	{
		.setkey = sce_ccm_setkey,
		.setauthsize = sce_ccm_setauthsize,
		.encrypt = sce_ccm_encrypt,
		.decrypt = sce_ccm_decrypt,
		.init = sce_ccm_init,
		.ivsize = CCM_IV_BYTE_LEN,
		.maxauthsize = CCM_DIGEST_BYTE_LEN,
		.chunksize = CCM_BLOCK_BYTE_LEN,
		.base = {
				.cra_flags = CRYPTO_ALG_ASYNC,
				.cra_blocksize = 1,
				.cra_ctxsize = sizeof(struct ccm_alg_context)
					+ CACHE_LINE_SIZE,
				.cra_priority = 300,
				.cra_name = "ccm(sm4)",
				.cra_driver_name = "ccm(sm4)-phytium",
				.cra_module = THIS_MODULE,
			},
	},
	{
		.setkey = sce_ccm_setkey,
		.setauthsize = sce_ccm_setauthsize,
		.encrypt = sce_ccm_encrypt,
		.decrypt = sce_ccm_decrypt,
		.init = sce_ccm_init,
		.ivsize = CCM_IV_BYTE_LEN,
		.maxauthsize = CCM_DIGEST_BYTE_LEN,
		.chunksize = CCM_BLOCK_BYTE_LEN,
		.base = {
				.cra_flags = CRYPTO_ALG_ASYNC,
				.cra_blocksize = 1,
				.cra_ctxsize = sizeof(struct ccm_alg_context)
					+ CACHE_LINE_SIZE,
				.cra_priority = 300,
				.cra_name = "ccm(aes)",
				.cra_driver_name = "ccm(aes)-phytium",
				.cra_module = THIS_MODULE,
			},
	},
};

int sce_ccm_algs_register(void)
{
	int ret;

	ret = crypto_register_aeads(sce_ccm_alg, ARRAY_SIZE(sce_ccm_alg));
	if (ret)
		sce_err("%s:crypto_register_aead failed!\n", __func__);


	return ret;
}

void sce_ccm_algs_unregister(void)
{
	crypto_unregister_aeads(sce_ccm_alg, ARRAY_SIZE(sce_ccm_alg));
}
