// SPDX-License-Identifier: GPL-2.0-or-later
/*
 * File Name:gcm.c - Phytium SDK for GCM
 *
 * Copyright (C) 2022 Phytium Technology Co.,Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */

#include <asm/memory.h>
#include <linux/delay.h>
#include <crypto/internal/hash.h>
#include <crypto/hash.h>
#include <crypto/internal/skcipher.h>
#include <crypto/skcipher.h>
#include <crypto/internal/aead.h>
#include <crypto/aead.h>
#include <linux/io.h>
#include <linux/crypto.h>
#include <asm/cacheflush.h>
#include <linux/module.h>
#include <linux/irq.h>
#include <linux/platform_device.h>
#include <linux/dma-mapping.h>
#include <linux/device.h>
#include <linux/cdev.h>
#include <linux/types.h>
#include <linux/mutex.h>
#include <linux/acpi.h>
#include <linux/ioport.h>
#include <linux/io.h>
#include <linux/delay.h>
#include <linux/kthread.h>
#include <linux/version.h>

#include "sce_kernel_common.h"
#include "sce_kernel.h"
#include "sce_kernel_reg.h"
#include "sce.h"
#include "gcm.h"
#include "cipher.h"

#define AEAD_INIT_128 0xFFFFFFFFFFFFFFFFull

static char *gcm_alg_name[] = { "gcm(sm4)", "gcm(aes)" };
static u8 alg[] = { SM4, AES };
static u8 block_size[] = { AES_BLOCK_BYTE_LEN, SM4_BLOCK_BYTE_LEN };
static u8 iv_size[] = { AES_IV_BYTE_LEN, SM4_IV_BYTE_LEN };
static u8 key_len[] = { AES_MIN_KEY_BYTE_LEN, SM4_KEY_BYTE_LEN };

static union sca_init_cmd SCA_INIT = {
	.trigger_intr = 0,
	.mode = 2,
	.iv_load = 1,
	.key_len = 0,
	.key_src = 2,
	.algo = 0,
	.reserved = 0,
	.slot_id = 8,
	.opcode = 0x80,
};

static union sca_proc_cmd SCA_PROC = {
	.trigger_intr = 0,
	.dst_src_addr_type = 0,
	.is_last_proc = 1,
	.op_mode = 0,
	.reserved = 0,
	.slot_id = 8,
	.opcode = 0x40,
};

static union sca_finish_cmd SCA_FINISH = {
	.trigger_intr = 0,
	.mac_tlen = 0,
	.reserved = 0,
	.slot_id = 8,
	.opcode = 0x20,
};
static void bit128_xor(uint8_t *a, uint8_t *b, uint8_t *out)
{
	int i;

	for (i = 0; i < 16; i++)
		out[i] = a[i] ^ b[i];

}

static void _te_gcm_read_size(uint8_t *dst, size_t ofs, uint64_t size)
{
	size_t i = 0;

	for (i = 0; i < sizeof(size); i++)
		dst[ofs - i] = (size >> (i * 8)) & 0xFF;

}

static void send_cmd(struct cipher_cmd *cmd)
{
	union sca_init_cmd sca_init_cmd;
	union sca_proc_cmd sca_proc_cmd;
	union sca_finish_cmd sca_finish_cmd;
	union sca_stat_reg stat;
	union sca_csq_reg csq;
	u64 key_paddr, iv_paddr, src_paddr, dst_paddr;
	unsigned long flags;

	key_paddr = cmd->key_ioaddr | 0x300000000000;
	iv_paddr = cmd->iv_ioaddr | 0x300000000000;
	src_paddr = cmd->src_ioaddr | 0x300000000000;
	dst_paddr = cmd->dst_ioaddr | 0x300000000000;

	sca_init_cmd.data = SCA_INIT.data;
	sca_init_cmd.mode = cmd->mode;
	sca_init_cmd.iv_load = cmd->iv_load;
	sca_init_cmd.key_len = cmd->key_len_flag;
	sca_init_cmd.algo = cmd->algo;

	sca_proc_cmd.data = SCA_PROC.data;
	sca_proc_cmd.is_last_proc = cmd->is_last_proc;
	sca_proc_cmd.dst_src_addr_type = cmd->dst_src_addr_type;
	sca_proc_cmd.op_mode = cmd->op_mode;

	sca_finish_cmd.data = SCA_FINISH.data;
	sca_finish_cmd.mac_tlen = cmd->mac_tlen - 1;
	spin_lock_irqsave(&sce_kernel.cipher_lock, flags);

	REG_WR_32(sce_kernel.base, SCA_CQ_FUNC_OFFSET, sca_init_cmd.data);
	do {
		stat.data = REG_RD_32_SYNC(sce_kernel.base, SCA_STAT_OFFSET);
	} while (stat.cq_avail_slots < 4);

	if (cmd->iv_load != 0) {
		REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET,
			  iv_paddr & 0xFFFFFFFF);
		REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET, iv_paddr >> 32);
	}
	REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET, key_paddr & 0xFFFFFFFF);
	REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET, key_paddr >> 32);

	REG_WR_32(sce_kernel.base, SCA_CQ_FUNC_OFFSET, sca_proc_cmd.data);

	do {
		stat.data = REG_RD_32_SYNC(sce_kernel.base, SCA_STAT_OFFSET);
	} while (stat.cq_avail_slots < 5);

	if (cmd->dst_src_addr_type == 0)
		REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET,
			  cmd->src_len - 1);

	REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET, src_paddr & 0xFFFFFFFF);
	REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET, src_paddr >> 32);
	if (cmd->mac_tlen == 0) {
		REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET,
			  dst_paddr & 0xFFFFFFFF);
		REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET, dst_paddr >> 32);
	}

	REG_WR_32(sce_kernel.base, SCA_CQ_FUNC_OFFSET, sca_finish_cmd.data);
	do {
		stat.data = REG_RD_32_SYNC(sce_kernel.base, SCA_STAT_OFFSET);
	} while (stat.cq_avail_slots < 2);

	if (cmd->mac_tlen != 0) {
		REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET,
			  dst_paddr & 0xFFFFFFFF);
		REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET, dst_paddr >> 32);
	}

	do {
		csq.data = REG_RD_32_SYNC(sce_kernel.base, SCA_CSQ_OFFSET);
	} while (csq.opcode_idx != FINISH);

	spin_unlock_irqrestore(&sce_kernel.cipher_lock, flags);

}

static inline int sce_gcm_encrypt(struct aead_request *req)
{
	struct crypto_aead *tfm = __crypto_aead_cast(req->base.tfm);
	struct gcm_alg_context *actx =
		(void *)(((u64)(crypto_aead_ctx(tfm)) + CACHE_LINE_SIZE) &
			 (~(CACHE_LINE_SIZE - 1)));
	struct gcm_session_context *sctx =
		(void *)(((u64)(aead_request_ctx(req)) + CACHE_LINE_SIZE) &
			 (~(CACHE_LINE_SIZE - 1)));
	struct link_list_item *src_link_list = NULL, *dst_link_list = NULL;
	struct scatterlist *sg;
	int src_nents = 0, dst_nents = 0, i;
	u32 dma_mode;
	u64 src_ioaddr = 0, dst_ioaddr = 0, temp;
	struct cipher_cmd *cmd;
	int flag1 = 0;
	int authsize = tfm->authsize;
	int j = 0;
	int padding_sz = 0;

	cmd = &sctx->gcm_cmd;
	src_nents = sg_nents(req->src);
	dst_nents = sg_nents(req->dst);
	if (unlikely(!src_nents || !dst_nents)) {
		sce_err("%s:src_nents[%d] dst_nents[%d]! err!\n", __func__,
			src_nents, dst_nents);
		return -EBADMSG;
	}
	if (unlikely(src_nents + 6 > MAX_LINKS)) {
		src_link_list = sce_malloc(sizeof(struct link_list_item) *
					   (src_nents + 6));
		if (src_link_list == NULL) {
			sce_err("sce_malloc failed!\n");
			return -EMLINK;
		}
	} else {
		src_link_list = sctx->src_list_buf;
	}
	memset(src_link_list, 0, sizeof(*src_link_list));
	if (unlikely(dst_nents + 6 > MAX_LINKS)) {
		dst_link_list = sce_malloc(sizeof(struct link_list_item) *
					   (dst_nents + 6));
		if (dst_link_list == NULL) {
			if (src_link_list != sctx->src_list_buf)
				sce_free(src_link_list);
			sce_err("sce_malloc failed!\n");
			return -EMLINK;
		}
	} else {
		dst_link_list = sctx->dst_list_buf;
	}

	memset(sctx->iv, 0, 16);
	if (req->iv)
		memcpy(sctx->iv, req->iv, 12);

	sctx->iv[15] = 1;

	memset(sctx->ghash_zero, 0, 16);
	memset(sctx->ghash_key, 0, 16);

	cmd->algo = actx->alg;
	cmd->mode = ECB;
	cmd->op_mode = 0;
	cmd->iv_load = 0;
	cmd->key_len_flag = actx->key_len_flag;
	cmd->dst_src_addr_type = 0;
	cmd->mac_tlen = 0;
	cmd->src_len = 16;
	cmd->is_last_proc = 1;
	cmd->key_ioaddr = actx->key_ioaddr;
	cmd->src_ioaddr = virt_to_phys(sctx->ghash_zero);
	cmd->dst_ioaddr = virt_to_phys(sctx->ghash_key);

	send_cmd(cmd);

	dsb(sy);
	//ssleep(1);
	//1.do ctr
	flag1 = 0;
	i = 0;
	temp = req->assoclen;
	memset(src_link_list, 0, sizeof(*src_link_list));
	dma_mode = 0;
	if (req->cryptlen != 0) {
		for (sg = req->src; sg; sg = sg_next(sg)) {
			if (!sg->length)
				continue;
			if (sg->length <= temp && flag1 == 0) {
				temp -= sg->length;
				continue;
			} else if (flag1 == 0 && sg->length > temp) {
				if (sg->length >= temp + req->cryptlen) {
					src_link_list[i].addr =
						sg_phys(sg) + temp;
					src_link_list[i].addr |= 0x300000000000;
					src_link_list[i].size =
						req->cryptlen - 1;
					i++;
					//dma_mode |= 1;
					break;
				}
			}
			if (flag1 == 0) {
				if (sg->length > temp) {
					src_link_list[i].addr =
						sg_phys(sg) + temp;
					src_link_list[i].addr |= 0x300000000000;
					src_link_list[i].size =
						sg->length - temp - 1;
					temp = req->cryptlen -
					       src_link_list[i].size - 1;
					i++;
					flag1 = 1;
				}
				continue;
			}
			//dma_mode |= 1;
			if (sg->length < temp) {
				src_link_list[i].addr = sg_phys(sg);
				src_link_list[i].addr |= 0x300000000000;
				src_link_list[i].size = sg->length - 1;
				temp -= sg->length;
				i++;
			} else {
				src_link_list[i].addr = sg_phys(sg);
				src_link_list[i].addr |= 0x300000000000;
				src_link_list[i].size = temp - 1;
				i++;
				break;
			}
		}
		src_link_list[i].addr = 0;
		src_link_list[i].size = AEAD_INIT_128;

		flag1 = 0;
		i = 0;
		memset(dst_link_list, 0, sizeof(*dst_link_list));
		temp = req->assoclen;
		for (sg = req->dst; sg; sg = sg_next(sg)) {
			if (!sg->length)
				continue;
			if (sg->length <= temp && flag1 == 0) {
				temp -= sg->length;
				continue;
			} else if (flag1 == 0 && sg->length > temp) {
				if (sg->length >= temp + req->cryptlen) {
					dst_link_list[i].addr =
						sg_phys(sg) + temp;
					dst_link_list[i].addr |= 0x300000000000;
					dst_link_list[i].size =
						req->cryptlen - 1;
					i++;
					break;
				}
			}
			if (flag1 == 0) {
				if (sg->length > temp) {
					dst_link_list[i].addr =
						sg_phys(sg) + temp;
					dst_link_list[i].addr |= 0x300000000000;
					dst_link_list[i].size =
						sg->length - temp - 1;
					temp = req->cryptlen -
					       dst_link_list[i].size - 1;
					i++;
					flag1 = 1;
				}
				continue;
			}
			//dma_mode |= 1;
			if (sg->length < temp) {
				dst_link_list[i].addr = sg_phys(sg);
				dst_link_list[i].addr |= 0x300000000000;
				dst_link_list[i].size = sg->length - 1;
				temp -= sg->length;
				i++;
			} else {
				dst_link_list[i].addr = sg_phys(sg);
				dst_link_list[i].addr |= 0x300000000000;
				dst_link_list[i].size = temp - 1;
				i++;
				break;
			}
		}
		dst_link_list[i].addr = 0;
		dst_link_list[i].size = AEAD_INIT_128;

		dma_mode = 3;

		src_ioaddr = virt_to_phys(src_link_list);
		dst_ioaddr = virt_to_phys(dst_link_list);

		memset(sctx->ctr_iv, 0, 16);
		memcpy(sctx->ctr_iv, sctx->iv, 16);
		for (j = 16; j > 0; j--)
			if (++sctx->ctr_iv[j - 1])
				break;

		cmd->algo = actx->alg;
		cmd->mode = CTR;
		cmd->iv_load = 1;
		cmd->key_len_flag = actx->key_len_flag;
		cmd->op_mode = ENC;
		cmd->is_last_proc = 1;
		cmd->dst_src_addr_type = dma_mode;

		cmd->mac_tlen = 0;
		cmd->key_ioaddr = actx->key_ioaddr;
		cmd->src_ioaddr = src_ioaddr;
		cmd->dst_ioaddr = dst_ioaddr;
		cmd->iv_ioaddr = virt_to_phys(sctx->ctr_iv);

		send_cmd(cmd);
	}
	dsb(sy);
	flag1 = 0;
	i = 0;
	temp = req->assoclen;
	memset(src_link_list, 0, sizeof(*src_link_list));
	for (sg = req->src; sg; sg = sg_next(sg)) {
		if (!sg->length)
			continue;
		if (!temp)
			break;
		if (temp <= sg->length) {
			src_link_list[i].addr = sg_phys(sg);
			src_link_list[i].addr |= 0x300000000000;
			src_link_list[i].size = temp - 1;
			i++;
			break;
		}
		src_link_list[i].addr = sg_phys(sg);
		src_link_list[i].addr |= 0x300000000000;
		src_link_list[i].size = sg->length - 1;
		i++;
		temp -= sg->length;
	}
	memset(sctx->ghash_zero, 0, 16);
	if (req->assoclen & 0xf) {
		src_link_list[i].addr = virt_to_phys(sctx->ghash_zero);
		src_link_list[i].addr |= 0x300000000000;
		src_link_list[i].size = 16 - (req->assoclen & 0xf) - 1;
		i++;
	}
	flag1 = 0;
	temp = req->assoclen;
	for (sg = req->dst; sg; sg = sg_next(sg)) {
		if (!sg->length)
			continue;
		if (!temp && (flag1 == 1))
			break;
		if (temp >= sg->length && flag1 == 0) {
			temp -= sg->length;
			continue;
		} else if (flag1 == 0 && sg->length > temp) {
			if (sg->length >= temp + req->cryptlen) {
				if (req->cryptlen) {
					src_link_list[i].addr =
						sg_phys(sg) + temp;
					flush_dcache_page(
						phys_to_page(sg_phys(sg)));
					src_link_list[i].addr |= 0x300000000000;
					src_link_list[i].size =
						req->cryptlen - 1;
					i++;
				}
				break;
			}
			src_link_list[i].addr = sg_phys(sg) + temp;
			src_link_list[i].addr |= 0x300000000000;
			src_link_list[i].size = sg->length - temp - 1;
			temp = req->cryptlen - src_link_list[i].size -
			       1;
			i++;
			flag1 = 1;
			continue;

		}

		if (sg->length < temp) {
			src_link_list[i].addr = sg_phys(sg);
			src_link_list[i].addr |= 0x300000000000;
			src_link_list[i].size = sg->length - 1;
			temp -= sg->length;
			i++;
		} else {
			src_link_list[i].addr = sg_phys(sg);
			src_link_list[i].addr |= 0x300000000000;
			src_link_list[i].size = temp - 1;
			i++;
			break;
		}
	}

	padding_sz = (16 - (req->cryptlen & 0xf)) & 0xf;
	memset(sctx->pad, 0, 32);
	_te_gcm_read_size(sctx->pad, padding_sz + 7, req->assoclen * 8);
	_te_gcm_read_size(sctx->pad, padding_sz + 15, req->cryptlen * 8);

	src_link_list[i].addr = virt_to_phys(sctx->pad);
	src_link_list[i].addr |= 0x300000000000;
	src_link_list[i].size = 16 + padding_sz - 1;
	i++;

	src_link_list[i].addr = 0;
	src_link_list[i].size = AEAD_INIT_128;
	cmd->op_mode = ENC;

	cmd->iv_ioaddr = virt_to_phys(sctx->ctr_iv);

	memset(sctx->ghash_result, 0, 16);
	src_ioaddr = virt_to_phys(src_link_list);
	cmd->dst_ioaddr = virt_to_phys(sctx->ghash_result);
	cmd->src_ioaddr = src_ioaddr;
	cmd->iv_load = 0;
	cmd->key_ioaddr = virt_to_phys(sctx->ghash_key);
	cmd->dst_src_addr_type = 2;
	cmd->key_len_flag = 0;
	cmd->mac_tlen = 16;
	cmd->algo = GHASH;
	cmd->src_len = 16;
	cmd->mode = CMAC;
	cmd->is_last_proc = 1;

	send_cmd(cmd);

	memset(sctx->ecb_result, 0, 16);
	cmd->src_ioaddr = virt_to_phys(sctx->iv);
	cmd->dst_ioaddr = virt_to_phys(sctx->ecb_result);

	cmd->algo = actx->alg;
	cmd->key_ioaddr = actx->key_ioaddr;
	cmd->key_len_flag = actx->key_len_flag;
	cmd->mode = ECB;
	cmd->iv_load = 0;
	cmd->dst_src_addr_type = 0;
	cmd->mac_tlen = 0;
	cmd->src_len = 16;
	send_cmd(cmd);

	memset(sctx->tag, 0, 16);
	bit128_xor(sctx->ghash_result, sctx->ecb_result, sctx->tag);
	sg_copy_buffer(req->dst, sg_nents(req->dst), sctx->tag, authsize,
		       req->assoclen + req->cryptlen, FALSE);
	return 0;
}

int sce_gcm_decrypt(struct aead_request *req)
{
	struct crypto_aead *tfm = __crypto_aead_cast(req->base.tfm);
	struct gcm_alg_context *actx =
		(void *)(((u64)(crypto_aead_ctx(tfm)) + CACHE_LINE_SIZE) &
			 (~(CACHE_LINE_SIZE - 1)));
	struct gcm_session_context *sctx =
		(void *)(((u64)(aead_request_ctx(req)) + CACHE_LINE_SIZE) &
			 (~(CACHE_LINE_SIZE - 1)));
	struct link_list_item *src_link_list = NULL, *dst_link_list = NULL;
	struct scatterlist *sg;
	int src_nents = 0, dst_nents = 0, i;
	u32 dma_mode;
	u64 src_ioaddr = 0, dst_ioaddr = 0, temp;
	struct cipher_cmd *cmd;
	int flag1 = 0;
	int authsize = tfm->authsize;
	int j = 0;
	int padding_sz = 0;

	cmd = &sctx->gcm_cmd;

	src_nents = sg_nents(req->src);
	dst_nents = sg_nents(req->dst);

	if (unlikely(!src_nents || !dst_nents)) {
		sce_err("%s:src_nents[%d] dst_nents[%d]! err!\n", __func__,
			src_nents, dst_nents);
		return -EBADMSG;
	}
	if (unlikely(src_nents + 6 > MAX_LINKS)) {
		src_link_list = sce_malloc(sizeof(struct link_list_item) *
					   (src_nents + 6));
		if (src_link_list == NULL) {
			sce_err("sce_malloc failed!\n");
			return -EMLINK;
		}
	} else {
		src_link_list = sctx->src_list_buf;
	}
	memset(src_link_list, 0, sizeof(*src_link_list));
	if (unlikely(dst_nents + 6 > MAX_LINKS)) {
		dst_link_list = sce_malloc(sizeof(struct link_list_item) *
					   (dst_nents + 6));
		if (dst_link_list == NULL) {
			if (src_link_list != sctx->src_list_buf)
				sce_free(src_link_list);
			sce_err("sce_malloc failed!\n");
			return -EMLINK;
		}
	} else {
		dst_link_list = sctx->dst_list_buf;
	}

	memset(sctx->iv, 0, 16);
	if (req->iv)
		memcpy(sctx->iv, req->iv, 12);

	sctx->iv[15] = 1;
	memset(sctx->ghash_zero, 0, 16);
	memset(sctx->ghash_key, 0, 16);

	cmd->algo = actx->alg;
	cmd->mode = ECB;
	cmd->op_mode = 0;
	cmd->iv_load = 0;
	cmd->key_len_flag = actx->key_len_flag;
	cmd->dst_src_addr_type = 0;
	cmd->mac_tlen = 0;
	cmd->src_len = 16;
	cmd->is_last_proc = 1;
	cmd->key_ioaddr = actx->key_ioaddr;
	cmd->src_ioaddr = virt_to_phys(sctx->ghash_zero);
	cmd->dst_ioaddr = virt_to_phys(sctx->ghash_key);

	send_cmd(cmd);

	dsb(sy);

	temp = req->assoclen;
	src_ioaddr = virt_to_phys(src_link_list);
	dst_ioaddr = virt_to_phys(sctx->ghash_result);
	i = 0;
	flag1 = 0;
	memset(sctx->ghash_result, 0, 16);
	for (sg = req->src; sg; sg = sg_next(sg)) {
		if (!sg->length)
			continue;
		if (!temp && (flag1 == 1))
			break;

		if (likely(temp != 0)) {
			src_link_list[i].addr = sg_phys(sg);
			src_link_list[i].addr |= 0x300000000000;
			src_link_list[i].size = temp > sg->length ?
							(sg->length - 1) :
							(temp - 1);
			i++;
		}
		if (unlikely(temp <= sg->length && flag1 == 0)) {
			if (req->assoclen & 0xf) {
				src_link_list[i].addr =
					virt_to_phys(sctx->ghash_zero);
				src_link_list[i].addr |= 0x300000000000;
				src_link_list[i].size =
					((16 - (req->assoclen & 0xf)) & 0xf) -
					1;
				i++;
			}
			if (temp < sg->length) {
				if (req->cryptlen - authsize <
				    sg->length - temp) {
					if (req->cryptlen - authsize) {
						src_link_list[i].addr =
							sg_phys(sg) + temp;
						src_link_list[i].addr |=
							0x300000000000;
						src_link_list[i].size =
							req->cryptlen -
							authsize - 1;
						i++;
					}
					break;
				}
				src_link_list[i].addr =
					sg_phys(sg) + temp;
				src_link_list[i].addr |= 0x300000000000;
				src_link_list[i].size =
					sg->length - temp - 1;
				temp = req->cryptlen - authsize -
				       (sg->length - temp);
				i++;

			} else {
				temp = req->cryptlen - authsize;
			}
			flag1 = 1;
		} else {
			if (temp >= sg->length)
				temp -= sg->length;
			else
				temp = 0;
		}
	}
	padding_sz = (16 - ((req->cryptlen - authsize) & 0xf)) & 0xf;
	memset(sctx->pad, 0, 32);
	_te_gcm_read_size(sctx->pad, padding_sz + 7, req->assoclen * 8);
	_te_gcm_read_size(sctx->pad, padding_sz + 15,
			  (req->cryptlen - authsize) * 8);

	src_link_list[i].addr = virt_to_phys(sctx->pad);
	src_link_list[i].addr |= 0x300000000000;
	src_link_list[i].size = 16 + padding_sz - 1;
	i++;

	src_link_list[i].addr = 0;
	src_link_list[i].size = AEAD_INIT_128;

	cmd->dst_ioaddr = virt_to_phys(sctx->ghash_result);
	cmd->src_ioaddr = src_ioaddr;
	cmd->iv_load = 0;
	cmd->dst_src_addr_type = 2;
	cmd->mac_tlen = 16;
	cmd->algo = GHASH;
	cmd->mode = CMAC;
	cmd->is_last_proc = 1;
	cmd->key_ioaddr = virt_to_phys(sctx->ghash_key);
	cmd->key_len_flag = 0;
	cmd->op_mode = ENC;
	cmd->src_len = 114;

	send_cmd(cmd);
	memset(sctx->ecb_result, 0, 16);
	cmd->src_ioaddr = virt_to_phys(sctx->iv);
	cmd->dst_ioaddr = virt_to_phys(sctx->ecb_result);

	cmd->key_ioaddr = actx->key_ioaddr;
	cmd->key_len_flag = actx->key_len_flag;
	cmd->algo = actx->alg;
	cmd->mode = ECB;
	cmd->iv_load = 0;
	cmd->dst_src_addr_type = 0;
	cmd->mac_tlen = 0;
	cmd->src_len = 16;
	send_cmd(cmd);
	memset(sctx->iv, 0, 16);
	if (req->iv)
		memcpy(sctx->iv, req->iv, 12);

	sctx->iv[15] = 1;
	flag1 = 0;
	i = 0;
	temp = req->assoclen;
	memset(src_link_list, 0, sizeof(*src_link_list));
	dma_mode = 0;
	if (req->cryptlen - authsize) {
		for (sg = req->src; sg; sg = sg_next(sg)) {
			if (!sg->length)
				continue;
			if (sg->length <= temp && flag1 == 0) {
				temp -= sg->length;
				continue;
			} else if (flag1 == 0 && sg->length > temp) {
				if (sg->length >=
				    temp + req->cryptlen - authsize) {
					src_link_list[i].addr =
						sg_phys(sg) + temp;
					src_link_list[i].addr |= 0x300000000000;
					src_link_list[i].size =
						req->cryptlen - authsize - 1;
					i++;
					dma_mode |= 1;
					break;
				}
			}
			if (flag1 == 0) {
				if (sg->length > temp) {
					src_link_list[i].addr =
						sg_phys(sg) + temp;
					src_link_list[i].addr |= 0x300000000000;
					src_link_list[i].size =
						sg->length - temp - 1;
					temp = req->cryptlen - authsize -
					       src_link_list[i].size - 1;
					i++;
					flag1 = 1;
				}
				continue;
			}
			dma_mode |= 1;
			if (sg->length < temp) {
				src_link_list[i].addr = sg_phys(sg);
				src_link_list[i].addr |= 0x300000000000;
				src_link_list[i].size = sg->length - 1;
				temp -= sg->length;
				i++;
			} else {
				src_link_list[i].addr = sg_phys(sg);
				src_link_list[i].addr |= 0x300000000000;
				src_link_list[i].size = temp - 1;
				i++;
				break;
			}
		}
		src_link_list[i].addr = 0;
		src_link_list[i].size = AEAD_INIT_128;

		flag1 = 0;
		i = 0;
		memset(dst_link_list, 0, sizeof(*dst_link_list));
		temp = req->assoclen;
		for (sg = req->dst; sg; sg = sg_next(sg)) {
			if (!sg->length)
				continue;
			if (sg->length <= temp && flag1 == 0) {
				temp -= sg->length;
				continue;
			} else if (flag1 == 0 && sg->length > temp) {
				if (sg->length >=
				    temp + req->cryptlen - authsize) {
					dst_link_list[i].addr =
						sg_phys(sg) + temp;
					dst_link_list[i].addr |= 0x300000000000;
					dst_link_list[i].size =
						req->cryptlen - authsize - 1;
					i++;
					break;
				}
			}
			if (flag1 == 0) {
				if (sg->length > temp) {
					dst_link_list[i].addr =
						sg_phys(sg) + temp;
					dst_link_list[i].addr |= 0x300000000000;
					dst_link_list[i].size =
						sg->length - temp - 1;
					temp = req->cryptlen - authsize -
					       dst_link_list[i].size - 1;
					i++;
					flag1 = 1;
					//continue;
				}
				continue;
			}
			dma_mode |= 1;
			if (sg->length < temp) {
				dst_link_list[i].addr = sg_phys(sg);
				dst_link_list[i].addr |= 0x300000000000;
				dst_link_list[i].size = sg->length - 1;
				temp -= sg->length;
				i++;
			} else {
				dst_link_list[i].addr = sg_phys(sg);
				dst_link_list[i].addr |= 0x300000000000;
				dst_link_list[i].size = temp - 1;
				i++;
				break;
			}
		}
		dst_link_list[i].addr = 0;
		dst_link_list[i].size = AEAD_INIT_128;

		dma_mode = 3;

		src_ioaddr = virt_to_phys(src_link_list);
		dst_ioaddr = virt_to_phys(dst_link_list);

		memset(sctx->ctr_iv, 0, 16);
		memcpy(sctx->ctr_iv, sctx->iv, 16);
		for (j = 16; j > 0; j--)
			if (++sctx->ctr_iv[j - 1])
				break;

		cmd->algo = actx->alg;
		cmd->mode = CTR;
		cmd->iv_load = 1;
		cmd->key_len_flag = actx->key_len_flag;
		cmd->op_mode = ENC;
		cmd->is_last_proc = 1;
		cmd->dst_src_addr_type = dma_mode;

		cmd->mac_tlen = 0;
		cmd->key_ioaddr = actx->key_ioaddr;
		cmd->src_ioaddr = src_ioaddr;
		cmd->dst_ioaddr = dst_ioaddr;
		cmd->iv_ioaddr = virt_to_phys(sctx->ctr_iv);

		send_cmd(cmd);
	}

	memset(sctx->tag, 0, 16);
	bit128_xor(sctx->ghash_result, sctx->ecb_result, sctx->ecb_result);

	sg_copy_buffer(req->src, sg_nents(req->src), sctx->tag, authsize,
		       req->cryptlen + req->assoclen - authsize, TRUE);
	if (memcmp(sctx->tag, sctx->ecb_result, authsize))
		return -EBADMSG;
	return 0;
}

static int sce_gcm_init(struct crypto_aead *tfm)
{
	int i;
	struct gcm_alg_context *actx =
		(void *)(((u64)(crypto_aead_ctx(tfm)) + CACHE_LINE_SIZE) &
			 (~(CACHE_LINE_SIZE - 1)));

	memset(actx, 0, sizeof(struct gcm_alg_context));
	crypto_aead_set_reqsize(tfm, sizeof(struct gcm_session_context) +
					     CACHE_LINE_SIZE);

	for (i = 0; i < ARRAY_SIZE(gcm_alg_name); i++) {
		if (memcmp(tfm->base.__crt_alg->cra_name, gcm_alg_name[i],
			   strlen(gcm_alg_name[i])) == 0) {
			actx->alg = alg[i];
			actx->mode = GCM;
			actx->block_size = block_size[i];
			actx->block_shift = 4;
			actx->iv_size = iv_size[i];
			actx->key_len = key_len[i];
			actx->key_len_flag = 0;
			actx->key_ioaddr = virt_to_phys(actx->key);
		}
	}

	if (actx->block_size == 0)
		return -EINVAL;

	return 0;
}

static int sce_gcm_setkey(struct crypto_aead *tfm, const u8 *key,
			  unsigned int keylen)
{
	struct gcm_alg_context *actx =
		(void *)(((u64)(crypto_aead_ctx(tfm)) + CACHE_LINE_SIZE) &
			 (~(CACHE_LINE_SIZE - 1)));

	if (actx->alg == AES) {
		switch (keylen) {
		case AES_MIN_KEY_BYTE_LEN:
			actx->key_len = AES_MIN_KEY_BYTE_LEN;
			actx->key_len_flag = 0;
			break;
		case AES_MID_KEY_BYTE_LEN:
			actx->key_len = AES_MID_KEY_BYTE_LEN;
			actx->key_len_flag = 1;
			break;
		case AES_MAX_KEY_BYTE_LEN:
			actx->key_len = AES_MAX_KEY_BYTE_LEN;
			actx->key_len_flag = 2;
			break;
		default:
			return -EINVAL;
		}
	} else if (actx->alg == SM4) {
		if (keylen != SM4_KEY_BYTE_LEN)
			return -EINVAL;
		actx->key_len_flag = 0;
		actx->key_len = SM4_KEY_BYTE_LEN;
	} else {
		sce_err("unsupport alg!\n");
		return -EINVAL;
	}

	if (unlikely(sizeof(actx->key) < keylen))
		return -EINVAL;


	memcpy(actx->key, key, keylen);

	return 0;
}

static int sce_gcm_setauthsize(struct crypto_aead *tfm, unsigned int authsize)
{
	switch (authsize) {
	case 4:
	case 8:
	case 12:
	case 13:
	case 14:
	case 15:
	case 16:
		break;
	default:
		return -EINVAL;
	}

	return 0;
}

static struct aead_alg sce_gcm_alg[] = {
	{
		.setkey         = sce_gcm_setkey,
		.setauthsize = sce_gcm_setauthsize,
		.encrypt        = sce_gcm_encrypt,
		.decrypt        = sce_gcm_decrypt,
		.init           = sce_gcm_init,
		.ivsize         = GCM_IV_BYTE_LEN,
		.maxauthsize = GCM_DIGEST_BYTE_LEN,
		.chunksize      = GCM_BLOCK_BYTE_LEN,
		.base = {
			.cra_flags = CRYPTO_ALG_ASYNC,
			.cra_blocksize          = 1,
			.cra_ctxsize
				= sizeof(struct gcm_alg_context)
				+ CACHE_LINE_SIZE,
			.cra_priority           = 300,
			.cra_name               = "gcm(sm4)",
			.cra_driver_name        = "gcm(sm4)-phytium",
			.cra_module     = THIS_MODULE,
		},
	},
	{
		.setkey         = sce_gcm_setkey,
		.setauthsize = sce_gcm_setauthsize,
		.encrypt        = sce_gcm_encrypt,
		.decrypt        = sce_gcm_decrypt,
		.init           = sce_gcm_init,
		.ivsize         = GCM_IV_BYTE_LEN,
		.maxauthsize = GCM_DIGEST_BYTE_LEN,
		.chunksize      = GCM_BLOCK_BYTE_LEN,
		.base = {
			.cra_flags = CRYPTO_ALG_ASYNC,
			.cra_blocksize          = 1,
			.cra_ctxsize
				= sizeof(struct gcm_alg_context)
				+ CACHE_LINE_SIZE,
			.cra_priority           = 300,
			.cra_name               = "gcm(aes)",
			.cra_driver_name        = "gcm(aes)-phytium",
			.cra_module     = THIS_MODULE,
		},
	},
};

int sce_gcm_algs_register(void)
{
	int ret;

	ret = crypto_register_aeads(sce_gcm_alg, ARRAY_SIZE(sce_gcm_alg));
	if (ret)
		sce_err("%s:crypto_register_aead failed!\n", __func__);

	return ret;
}

void sce_gcm_algs_unregister(void)
{
	crypto_unregister_aeads(sce_gcm_alg, ARRAY_SIZE(sce_gcm_alg));
}
