// SPDX-License-Identifier: GPL-2.0-or-later
/*
 * File Name:hash.c - Phytium SDK for HASH
 *
 * Copyright (C) 2022 Phytium Technology Co.,Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */

#include <linux/delay.h>
#include <crypto/internal/hash.h>
#include <crypto/hash.h>
#include <linux/crypto.h>
#include <linux/delay.h>
#include <linux/kthread.h>
#include <linux/version.h>
#include "sce_kernel_common.h"
#include "sce_kernel_reg.h"
#include "sce_kernel.h"
#include "hash.h"

static char *alg_name[] = { "sha1",	    "sha224",	    "sha256",
			    "sha384",	    "sha512",	    "md5",
			    "sm3",	    "hmac(sha1)",   "hmac(sha224)",
			    "hmac(sha256)", "hmac(sha384)", "hmac(sha512)",
			    "hmac(md5)",    "hmac(sm3)" };
static u8 alg[] = { SHA1, SHA224, SHA256, SHA384, SHA512, MD5, SM3,
		    SHA1, SHA224, SHA256, SHA384, SHA512, MD5, SM3 };
static u8 mode[] = { NORMAL, NORMAL, NORMAL, NORMAL, NORMAL, NORMAL, NORMAL,
		     HMAC,   HMAC,   HMAC,   HMAC,   HMAC,   HMAC,   HMAC };
static u8 block_size[] = { SHA1_BLOCK_BYTE_LEN,	  SHA224_BLOCK_BYTE_LEN,
			   SHA256_BLOCK_BYTE_LEN, SHA384_BLOCK_BYTE_LEN,
			   SHA512_BLOCK_BYTE_LEN, MD5_BLOCK_BYTE_LEN,
			   SM3_BLOCK_BYTE_LEN,	  SHA1_BLOCK_BYTE_LEN,
			   SHA224_BLOCK_BYTE_LEN, SHA256_BLOCK_BYTE_LEN,
			   SHA384_BLOCK_BYTE_LEN, SHA512_BLOCK_BYTE_LEN,
			   MD5_BLOCK_BYTE_LEN,	  SM3_BLOCK_BYTE_LEN };
static u8 digest_size[] = { SHA1_DIGEST_BYTE_LEN,   SHA224_DIGEST_BYTE_LEN,
			    SHA256_DIGEST_BYTE_LEN, SHA384_DIGEST_BYTE_LEN,
			    SHA512_DIGEST_BYTE_LEN, MD5_DIGEST_BYTE_LEN,
			    SM3_DIGEST_BYTE_LEN,    SHA1_DIGEST_BYTE_LEN,
			    SHA224_DIGEST_BYTE_LEN, SHA256_DIGEST_BYTE_LEN,
			    SHA384_DIGEST_BYTE_LEN, SHA512_DIGEST_BYTE_LEN,
			    MD5_DIGEST_BYTE_LEN,    SM3_DIGEST_BYTE_LEN };
static u8 key_len[] = { 0,
			0,
			0,
			0,
			0,
			0,
			0,
			SHA1_BLOCK_BYTE_LEN,
			SHA224_BLOCK_BYTE_LEN,
			SHA256_BLOCK_BYTE_LEN,
			SHA384_BLOCK_BYTE_LEN,
			SHA512_BLOCK_BYTE_LEN,
			MD5_BLOCK_BYTE_LEN,
			SM3_BLOCK_BYTE_LEN };
static union hash_init_cmd HASH_INIT = {
	.trigger_intr = 0,
	.key_len = 0,
	.key_src = 2,
	.mode = 0,
	.algo = 8,
	.reserved = 0,
	.slot_id = 8,
	.opcode = 0x80,
};

static union hash_proc_cmd HASH_PROC = {
	.trigger_intr = 0,
	.src_addr_type = 0,
	.reserved0 = 0,
	.padding_mode = 1,
	.is_last_proc = 1,
	.is_zero_proc = 0,
	.reserved1 = 0,
	.slot_id = 8,
	.opcode = 0x40,
};

static union hash_finish_cmd HASH_FINISH = {
	.trigger_intr = 0,
	.mac_tlen = 31,
	.reserved = 0,
	.slot_id = 8,
	.opcode = 0x20,
};

static union hash_warp_cmd HASH_WARPOUT = {
	.trigger_intr = 0,
	.reserved = 0,
	.slot_id = 8,
	.opcode = 0x4,
};

static union hash_warp_cmd HASH_WARPIN = {
	.trigger_intr = 0,
	.reserved = 0,
	.slot_id = 8,
	.opcode = 0x2,
};

static int sce_hash_cra_init(struct crypto_tfm *tfm)
{
	int i;
	struct hash_alg_context *actx =
		(void *)(((u64)(crypto_tfm_ctx(tfm)) + CACHE_LINE_SIZE) &
			 (~(CACHE_LINE_SIZE - 1)));

	memset(actx, 0, sizeof(struct hash_alg_context));
	crypto_ahash_set_reqsize(__crypto_ahash_cast(tfm),
				 sizeof(struct hash_session_context) +
					 CACHE_LINE_SIZE);

	for (i = 0; i < ARRAY_SIZE(alg_name); i++) {
		if (memcmp(tfm->__crt_alg->cra_name, alg_name[i],
			   strlen(alg_name[i])) == 0) {
			actx->alg = alg[i];
			actx->mode = mode[i];
			actx->block_size = block_size[i];
			actx->digest_size = digest_size[i];
			actx->key_len = key_len[i];
			actx->key_ioaddr = virt_to_phys(actx->key);
			actx->key_ioaddr |= 0x300000000000;
		}
	}

	if (actx->block_size == 0)
		return -EINVAL;

	return 0;
}

static int sce_hash_init(struct ahash_request *req)
{
	struct hash_session_context *sctx =
		(struct hash_session_context *)(((u64)(ahash_request_ctx(req)) +
						 CACHE_LINE_SIZE) &
						(~(CACHE_LINE_SIZE - 1)));

	sctx->total = 0;
	sctx->buffer_data_len = 0;
	sctx->hash_flag = INIT;
	sctx->state_ioaddr = virt_to_phys(sctx->state);
	sctx->state_ioaddr |= 0x300000000000;
	sctx->buffer_ioaddr = virt_to_phys(sctx->buffer);
	sctx->buffer_ioaddr |= 0x300000000000;
	sctx->list_buf_ioaddr = virt_to_phys(sctx->list_buf);
	sctx->list_buf_ioaddr |= 0x300000000000;
	return 0;
}

static int sce_hash_update(struct ahash_request *req)
{
	struct hash_session_context *sctx =
		(struct hash_session_context *)(((u64)(ahash_request_ctx(req)) +
						 CACHE_LINE_SIZE) &
						(~(CACHE_LINE_SIZE - 1)));
	struct hash_alg_context *actx =
		(void *)(((u64)(crypto_tfm_ctx(req->base.tfm)) +
			  CACHE_LINE_SIZE) &
			 (~(CACHE_LINE_SIZE - 1)));
	struct scatterlist *sg;
	int src_nents;
	u32 last, process_len, temp;
	u64 src_addr = 0;
	struct link_list_item *link_list = NULL;
	int i;
	u32 link_mode = 0;
	u64 link_ioaddr;
	union hash_init_cmd hash_init_cmd;
	union hash_proc_cmd hash_proc_cmd;
	union hash_stat_reg stat = { 0 };
	union hash_csq_reg csq;
	unsigned long flags;

	if (unlikely(!req->nbytes))
		return 0;

	if (likely(sctx->buffer_data_len + req->nbytes < HASH_BUFFER_SIZE)) {
		sg_copy_buffer(req->src, sg_nents(req->src),
			       sctx->buffer + sctx->buffer_data_len,
			       req->nbytes, 0, true);
		sctx->buffer_data_len += req->nbytes;
		sctx->total += req->nbytes;
		return 0;
	}

	src_nents = sg_nents(req->src);
	if (unlikely(src_nents > MAX_LINKS)) {
		link_list = sce_malloc(sizeof(struct link_list_item) *
				       (src_nents + 2));
		if (link_list == NULL) {
			sce_debug("sce_malloc failed!\n");
			return -EMLINK;
		}
		link_ioaddr = virt_to_phys(link_list);
		link_ioaddr |= 0x300000000000;
	} else {
		link_list = sctx->list_buf;
		link_ioaddr = sctx->list_buf_ioaddr;
		link_ioaddr |= 0x300000000000;
	}

	sctx->total += req->nbytes;
	last = sctx->total & (actx->block_size - 1);
	process_len = req->nbytes - last;

	if (unlikely((!!sctx->buffer_data_len + src_nents) <= 1)) {
		src_addr = sg_phys(req->src);
		src_addr |= 0x300000000000;
		link_mode = 0;
	} else {
		src_addr = link_ioaddr;
		src_addr |= 0x300000000000;
		link_mode = 1;
		i = 0;
		temp = process_len;
		if (likely(sctx->buffer_data_len)) {
			link_list[i].size = sctx->buffer_data_len - 1;
			link_list[i].addr = sctx->buffer_ioaddr;
			link_list[i].addr |= 0x300000000000;
			i++;
		}

		for (sg = req->src; sg; sg = sg_next(sg)) {
			if (!sg->length)
				continue;
			if (!temp)
				break;
			link_list[i].size = (temp > sg->length) ? sg->length :
								  temp;
			link_list[i].addr = sg_phys(sg);
			link_list[i].addr |= 0x300000000000;
			temp -= link_list[i].size;
			link_list[i].size -= 1;
			i++;
		}
		link_list[i].size = 0xFFFFFFFFFFFFFFFFull;
		link_list[i].addr = 0;
	}

	hash_proc_cmd.data = HASH_PROC.data;
	hash_proc_cmd.src_addr_type = link_mode;
	hash_proc_cmd.is_last_proc = 0;

	spin_lock_irqsave(&sce_kernel.hash_lock, flags);

	if (likely(sctx->hash_flag == INIT)) {
		hash_init_cmd.data = HASH_INIT.data;
		hash_init_cmd.mode = actx->mode;
		hash_init_cmd.algo = actx->alg;
		REG_WR_32(sce_kernel.base, HASH_CQ_FUNC_OFFSET,
			  hash_init_cmd.data);
		REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
			  (iv_paddr + actx->alg * 128u) & 0xFFFFFFFF);
		REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
			  (iv_paddr + actx->alg * 128u) >> 32);
		if (unlikely(actx->mode)) {
			REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
				  actx->key_ioaddr & 0xFFFFFFFF);
			REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
				  actx->key_ioaddr >> 32);
		}
		sctx->hash_flag = WARPOUT;
	} else if (sctx->hash_flag == WARPOUT) {
		REG_WR_32(sce_kernel.base, HASH_CQ_FUNC_OFFSET,
			  HASH_WARPIN.data);
		REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
			  sctx->state_ioaddr & 0xFFFFFFFF);
		REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
			  sctx->state_ioaddr >> 32);
	} else {
		spin_unlock_irqrestore(&sce_kernel.hash_lock, flags);
		if (link_list != sctx->list_buf)
			sce_free(link_list);
		sce_err("hash_flag[%lld] err!\n", sctx->hash_flag);
		return -EINVAL;
	}

	REG_WR_32(sce_kernel.base, HASH_CQ_FUNC_OFFSET, hash_proc_cmd.data);
	if (likely(link_mode == 0)) {
		REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
			  process_len - 1);
	}
	REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET, src_addr & 0xFFFFFFFF);
	do {
		stat.data = REG_RD_32_SYNC(sce_kernel.base, HASH_STAT_OFFSET);
	} while (stat.cq_avail_slots < 4);
	REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET, src_addr >> 32);

	REG_WR_32(sce_kernel.base, HASH_CQ_FUNC_OFFSET, HASH_WARPOUT.data);
	REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
		  sctx->state_ioaddr & 0xFFFFFFFF);
	REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
		  sctx->state_ioaddr >> 32);
	do {
		csq.data = REG_RD_32_SYNC(sce_kernel.base, HASH_CSQ_OFFSET);
	} while (csq.opcode_idx != WARPOUT);

	spin_unlock_irqrestore(&sce_kernel.hash_lock, flags);

	if (last) {
		sg_copy_buffer(req->src, src_nents, sctx->buffer, last,
			       process_len, true);
		sctx->buffer_data_len = last;
	} else {
		sctx->buffer_data_len = 0;
	}

	if (unlikely(link_list != sctx->list_buf))
		sce_free(link_list);

	return 0;
}

static int sce_hash_final(struct ahash_request *req)
{
	struct hash_session_context *sctx =
		(struct hash_session_context *)(((u64)(ahash_request_ctx(req)) +
						 CACHE_LINE_SIZE) &
						(~(CACHE_LINE_SIZE - 1)));
	struct hash_alg_context *actx =
		(void *)(((u64)(crypto_tfm_ctx(req->base.tfm)) +
			  CACHE_LINE_SIZE) &
			 (~(CACHE_LINE_SIZE - 1)));
	union hash_init_cmd hash_init_cmd;
	union hash_proc_cmd hash_proc_cmd;
	union hash_finish_cmd hash_finish_cmd;
	union hash_stat_reg stat = { 0 };
	union hash_csq_reg csq;
	unsigned long flags;

	hash_proc_cmd.data = HASH_PROC.data;
	if (unlikely(sctx->buffer_data_len == 0))
		hash_proc_cmd.is_zero_proc = 1;

	hash_finish_cmd.data = HASH_FINISH.data;
	hash_finish_cmd.mac_tlen = actx->digest_size - 1;

	spin_lock_irqsave(&sce_kernel.hash_lock, flags);

	if (likely(sctx->hash_flag == INIT)) {
		hash_init_cmd.data = HASH_INIT.data;
		hash_init_cmd.mode = actx->mode;
		hash_init_cmd.algo = actx->alg;
		REG_WR_32(sce_kernel.base, HASH_CQ_FUNC_OFFSET,
			  hash_init_cmd.data);
		REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
			  (iv_paddr + actx->alg * 128u) & 0xFFFFFFFF);
		REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
			  (iv_paddr + actx->alg * 128u) >> 32);
		if (unlikely(actx->mode)) {
			REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
				  actx->key_ioaddr & 0xFFFFFFFF);
			REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
				  actx->key_ioaddr >> 32);
		}
	} else if (sctx->hash_flag == WARPOUT) {
		REG_WR_32(sce_kernel.base, HASH_CQ_FUNC_OFFSET,
			  HASH_WARPIN.data);
		REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
			  sctx->state_ioaddr & 0xFFFFFFFF);
		REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
			  sctx->state_ioaddr >> 32);
	} else {
		spin_unlock_irqrestore(&sce_kernel.hash_lock, flags);
		sce_err("hash_flag[%lld] err!\n", sctx->hash_flag);
		return -EINVAL;
	}
	sctx->hash_flag = FINISH;

	REG_WR_32(sce_kernel.base, HASH_CQ_FUNC_OFFSET, hash_proc_cmd.data);
	if (likely(sctx->buffer_data_len)) {
		REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
			  sctx->buffer_data_len - 1);
		REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
			  sctx->buffer_ioaddr & 0xFFFFFFFF);
		do {
			stat.data = REG_RD_32_SYNC(sce_kernel.base,
						   HASH_STAT_OFFSET);
		} while (stat.cq_avail_slots < 2);
		REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
			  sctx->buffer_ioaddr >> 32);
	}

	REG_WR_32(sce_kernel.base, HASH_CQ_FUNC_OFFSET, hash_finish_cmd.data);
	do {
		stat.data = REG_RD_32_SYNC(sce_kernel.base, HASH_STAT_OFFSET);
	} while (stat.cq_avail_slots < 6);
	if (unlikely(actx->mode)) {
		REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
			  (iv_paddr + actx->alg * 128u) & 0xFFFFFFFF);
		REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
			  (iv_paddr + actx->alg * 128u) >> 32);
		REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
			  actx->key_ioaddr & 0xFFFFFFFF);
		REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
			  actx->key_ioaddr >> 32);
	}
	REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
		  sctx->state_ioaddr & 0xFFFFFFFF);
	REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
		  sctx->state_ioaddr >> 32);
	do {
		csq.data = REG_RD_32_SYNC(sce_kernel.base, HASH_CSQ_OFFSET);
	} while (csq.opcode_idx != FINISH);

	spin_unlock_irqrestore(&sce_kernel.hash_lock, flags);

	memcpy(req->result, sctx->state, actx->digest_size);

	return 0;
}

static int sce_hash_finup(struct ahash_request *req)
{
	struct hash_session_context *sctx =
		(struct hash_session_context *)(((u64)(ahash_request_ctx(req)) +
						 CACHE_LINE_SIZE) &
						(~(CACHE_LINE_SIZE - 1)));
	struct hash_alg_context *actx =
		(void *)(((u64)(crypto_tfm_ctx(req->base.tfm)) +
			  CACHE_LINE_SIZE) &
			 (~(CACHE_LINE_SIZE - 1)));
	struct scatterlist *sg;
	int src_nents;
	u32 process_len, temp;
	u64 src_addr;
	struct link_list_item *link_list = NULL;
	int i;
	u32 link_mode = 0;
	u64 link_ioaddr;
	union hash_init_cmd hash_init_cmd;
	union hash_proc_cmd hash_proc_cmd;
	union hash_finish_cmd hash_finish_cmd;
	union hash_stat_reg stat = { 0 };
	union hash_csq_reg csq;
	unsigned long flags;

	src_nents = sg_nents(req->src);
	if (unlikely(src_nents > MAX_LINKS)) {
		link_list = sce_malloc(sizeof(struct link_list_item) *
				       (src_nents + 2));
		if (link_list == NULL) {
			sce_debug("sce_malloc failed!\n");
			return -EMLINK;
		}
		link_ioaddr = virt_to_phys(link_list);
		link_ioaddr |= 0x300000000000;
	} else {
		link_list = sctx->list_buf;
		link_ioaddr = sctx->list_buf_ioaddr;
		link_ioaddr |= 0x300000000000;
	}

	process_len = req->nbytes;

	if (unlikely((!!sctx->buffer_data_len + src_nents) <= 1)) {
		if (process_len) {
			src_addr = sg_phys(req->src);
			src_addr |= 0x300000000000;
		} else {
			src_addr = sctx->buffer_ioaddr;
			src_addr |= 0x300000000000;
		}
		link_mode = 0;
	} else {
		src_addr = link_ioaddr;
		link_mode = 1;
		i = 0;
		temp = process_len;
		if (likely(sctx->buffer_data_len)) {
			link_list[i].size = sctx->buffer_data_len - 1;
			link_list[i].addr = sctx->buffer_ioaddr;
			link_list[i].addr |= 0x300000000000;
			i++;
		}

		for (sg = req->src; sg; sg = sg_next(sg)) {
			if (!sg->length)
				continue;
			if (!temp)
				break;
			link_list[i].size = (temp > sg->length) ? sg->length :
								  temp;
			link_list[i].addr = sg_phys(sg);
			link_list[i].addr |= 0x300000000000;
			temp -= link_list[i].size;
			link_list[i].size -= 1;
			i++;
		}
		link_list[i].size = 0xFFFFFFFFFFFFFFFFull;
		link_list[i].addr = 0;
	}

	hash_proc_cmd.data = HASH_PROC.data;
	hash_proc_cmd.src_addr_type = link_mode;
	if ((sctx->buffer_data_len + process_len) == 0)
		hash_proc_cmd.is_zero_proc = 1;

	hash_finish_cmd.data = HASH_FINISH.data;
	hash_finish_cmd.mac_tlen = actx->digest_size - 1;

	spin_lock_irqsave(&sce_kernel.hash_lock, flags);

	if (likely(sctx->hash_flag == INIT)) {
		hash_init_cmd.data = HASH_INIT.data;
		hash_init_cmd.mode = actx->mode;
		hash_init_cmd.algo = actx->alg;
		REG_WR_32(sce_kernel.base, HASH_CQ_FUNC_OFFSET,
			  hash_init_cmd.data);
		REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
			  (iv_paddr + actx->alg * 128u) & 0xFFFFFFFF);
		REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
			  (iv_paddr + actx->alg * 128u) >> 32);
		if (unlikely(actx->mode)) {
			REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
				  actx->key_ioaddr & 0xFFFFFFFF);
			REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
				  actx->key_ioaddr >> 32);
		}
	} else if (sctx->hash_flag == WARPOUT) {
		REG_WR_32(sce_kernel.base, HASH_CQ_FUNC_OFFSET,
			  HASH_WARPIN.data);
		REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
			  sctx->state_ioaddr & 0xFFFFFFFF);
		REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
			  sctx->state_ioaddr >> 32);
	} else {
		spin_unlock_irqrestore(&sce_kernel.hash_lock, flags);
		if (link_list != sctx->list_buf)
			sce_free(link_list);
		sce_err("hash_flag[%lld] err!\n", sctx->hash_flag);
		return -EINVAL;
	}
	sctx->hash_flag = FINISH;

	REG_WR_32(sce_kernel.base, HASH_CQ_FUNC_OFFSET, hash_proc_cmd.data);
	if (likely(sctx->buffer_data_len + process_len)) {
		if (likely(link_mode == 0)) {
			REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
				  sctx->buffer_data_len + process_len - 1);
		}
		REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
			  src_addr & 0xFFFFFFFF);
		do {
			stat.data = REG_RD_32_SYNC(sce_kernel.base,
						   HASH_STAT_OFFSET);
		} while (stat.cq_avail_slots < 2);
		REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET, src_addr >> 32);
	}

	REG_WR_32(sce_kernel.base, HASH_CQ_FUNC_OFFSET, hash_finish_cmd.data);
	do {
		stat.data = REG_RD_32_SYNC(sce_kernel.base, HASH_STAT_OFFSET);
	} while (stat.cq_avail_slots < 6);
	if (unlikely(actx->mode)) {
		REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
			  (iv_paddr + actx->alg * 128u) & 0xFFFFFFFF);
		REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
			  (iv_paddr + actx->alg * 128u) >> 32);
		REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
			  actx->key_ioaddr & 0xFFFFFFFF);
		REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
			  actx->key_ioaddr >> 32);
	}
	REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
		  sctx->state_ioaddr & 0xFFFFFFFF);
	REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
		  sctx->state_ioaddr >> 32);
	do {
		csq.data = REG_RD_32_SYNC(sce_kernel.base, HASH_CSQ_OFFSET);
	} while (csq.opcode_idx != FINISH);

	spin_unlock_irqrestore(&sce_kernel.hash_lock, flags);

	memcpy(req->result, sctx->state, actx->digest_size);

	if (unlikely(link_list != sctx->list_buf))
		sce_free(link_list);

	return 0;
}

static int sce_hash_digest(struct ahash_request *req)
{
	struct hash_session_context *sctx =
		(struct hash_session_context *)(((u64)(ahash_request_ctx(req)) +
						 CACHE_LINE_SIZE) &
						(~(CACHE_LINE_SIZE - 1)));
	struct hash_alg_context *actx =
		(void *)(((u64)(crypto_tfm_ctx(req->base.tfm)) +
			  CACHE_LINE_SIZE) &
			 (~(CACHE_LINE_SIZE - 1)));
	struct scatterlist *sg;
	int src_nents;
	u32 process_len, temp;
	u64 src_addr = 0;
	struct link_list_item *link_list = NULL;
	int i;
	u32 link_mode = 0;
	u64 link_ioaddr;
	union hash_init_cmd hash_init_cmd;
	union hash_proc_cmd hash_proc_cmd;
	union hash_finish_cmd hash_finish_cmd;
	union hash_stat_reg stat = { 0 };
	union hash_csq_reg csq;
	unsigned long flags;

	src_nents = sg_nents(req->src);
	if (unlikely(src_nents > MAX_LINKS)) {
		link_list = sce_malloc(sizeof(struct link_list_item) *
				       (src_nents + 2));
		if (link_list == NULL) {
			sce_debug("sce_malloc failed!\n");
			return -EMLINK;
		}
	} else {
		link_list = sctx->list_buf;
	}
	link_ioaddr = virt_to_phys(link_list);
	link_ioaddr |= 0x300000000000;
	sctx->state_ioaddr = virt_to_phys(sctx->state);
	sctx->state_ioaddr |= 0x300000000000;
	process_len = req->nbytes;

	if (unlikely(src_nents <= 1)) {
		if (process_len) {
			src_addr = sg_phys(req->src);
			src_addr |= 0x300000000000;
		}
		link_mode = 0;
	} else {
		src_addr = link_ioaddr;
		link_mode = 1;
		i = 0;
		temp = process_len;
		for (sg = req->src; sg; sg = sg_next(sg)) {
			if (!sg->length)
				continue;
			if (!temp)
				break;
			link_list[i].size = (temp > sg->length) ? sg->length :
								  temp;
			link_list[i].addr = sg_phys(sg);
			link_list[i].addr |= 0x300000000000;
			temp -= link_list[i].size;
			link_list[i].size -= 1;
			i++;
		}
		link_list[i].size = 0xFFFFFFFFFFFFFFFFull;
		link_list[i].addr = 0;
	}

	hash_init_cmd.data = HASH_INIT.data;
	hash_init_cmd.mode = actx->mode;
	hash_init_cmd.algo = actx->alg;

	hash_proc_cmd.data = HASH_PROC.data;
	hash_proc_cmd.src_addr_type = link_mode;
	if (process_len == 0)
		hash_proc_cmd.is_zero_proc = 1;

	hash_finish_cmd.data = HASH_FINISH.data;
	hash_finish_cmd.mac_tlen = actx->digest_size - 1;

	spin_lock_irqsave(&sce_kernel.hash_lock, flags);

	REG_WR_32(sce_kernel.base, HASH_CQ_FUNC_OFFSET, hash_init_cmd.data);
	REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
		  (iv_paddr + actx->alg * 128u) & 0xFFFFFFFF);
	REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
		  (iv_paddr + actx->alg * 128u) >> 32);
	if (unlikely(actx->mode)) {
		REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
			  actx->key_ioaddr & 0xFFFFFFFF);
		REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
			  actx->key_ioaddr >> 32);
	}

	REG_WR_32(sce_kernel.base, HASH_CQ_FUNC_OFFSET, hash_proc_cmd.data);
	if (likely(process_len)) {
		if (likely(link_mode == 0)) {
			REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
				  process_len - 1);
		}
		REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
			  src_addr & 0xFFFFFFFF);
		do {
			stat.data =
				REG_RD_32(sce_kernel.base, HASH_STAT_OFFSET);
		} while (stat.cq_avail_slots < 1);
		REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET, src_addr >> 32);
	}

	REG_WR_32(sce_kernel.base, HASH_CQ_FUNC_OFFSET, hash_finish_cmd.data);
	do {
		stat.data = REG_RD_32(sce_kernel.base, HASH_STAT_OFFSET);
	} while (stat.cq_avail_slots < 6);
	if (unlikely(actx->mode)) {
		REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
			  (iv_paddr + actx->alg * 128u) & 0xFFFFFFFF);
		REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
			  (iv_paddr + actx->alg * 128u) >> 32);
		REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
			  actx->key_ioaddr & 0xFFFFFFFF);
		REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
			  actx->key_ioaddr >> 32);
	}
	REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
		  sctx->state_ioaddr & 0xFFFFFFFF);
	REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
		  sctx->state_ioaddr >> 32);
	do {
		csq.data = REG_RD_32_SYNC(sce_kernel.base, HASH_CSQ_OFFSET);
	} while (csq.opcode_idx != FINISH);

	spin_unlock_irqrestore(&sce_kernel.hash_lock, flags);

	memcpy(req->result, sctx->state, actx->digest_size);

	if (unlikely(link_list != sctx->list_buf))
		sce_free(link_list);

	return 0;
}

static int sce_hash_export(struct ahash_request *req, void *out)
{
	struct hash_session_context *sctx =
		(struct hash_session_context *)(((u64)(ahash_request_ctx(req)) +
						 CACHE_LINE_SIZE) &
						(~(CACHE_LINE_SIZE - 1)));
	struct hash_alg_context *actx =
		(void *)(((u64)(crypto_tfm_ctx(req->base.tfm)) +
			  CACHE_LINE_SIZE) &
			 (~(CACHE_LINE_SIZE - 1)));
	union hash_init_cmd hash_init_cmd;
	union hash_proc_cmd hash_proc_cmd;
	union hash_stat_reg stat = { 0 };
	union hash_csq_reg csq;
	u32 last, process_len;
	unsigned long flags;

	if (out == NULL) {
		sce_err("Invalid params(null)");
		return -EINVAL;
	}

	if (likely(sctx->buffer_data_len > 256)) {
		last = sctx->buffer_data_len & (actx->block_size - 1);
		process_len = sctx->buffer_data_len - last;
		sctx->buffer_data_len = last;

		hash_proc_cmd.data = HASH_PROC.data;
		hash_proc_cmd.is_last_proc = 0;

		spin_lock_irqsave(&sce_kernel.hash_lock, flags);

		if (likely(sctx->hash_flag == INIT)) {
			hash_init_cmd.data = HASH_INIT.data;
			hash_init_cmd.mode = actx->mode;
			hash_init_cmd.algo = actx->alg;
			REG_WR_32(sce_kernel.base, HASH_CQ_FUNC_OFFSET,
				  hash_init_cmd.data);
			REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
				  (iv_paddr + actx->alg * 128u) & 0xFFFFFFFF);
			REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
				  (iv_paddr + actx->alg * 128u) >> 32);
			if (unlikely(actx->mode)) {
				REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
					  actx->key_ioaddr & 0xFFFFFFFF);
				REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
					  actx->key_ioaddr >> 32);
			}
			sctx->hash_flag = WARPOUT;
		} else if (sctx->hash_flag == WARPOUT) {
			REG_WR_32(sce_kernel.base, HASH_CQ_FUNC_OFFSET,
				  HASH_WARPIN.data);
			REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
				  sctx->state_ioaddr & 0xFFFFFFFF);
			REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
				  sctx->state_ioaddr >> 32);
		} else {
			spin_unlock_irqrestore(&sce_kernel.hash_lock, flags);
			sce_err("hash_flag[%lld] err!\n", sctx->hash_flag);
			return -EINVAL;
		}

		REG_WR_32(sce_kernel.base, HASH_CQ_FUNC_OFFSET,
			  hash_proc_cmd.data);
		REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
			  process_len - 1);
		REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
			  sctx->buffer_ioaddr & 0xFFFFFFFF);
		do {
			stat.data = REG_RD_32_SYNC(sce_kernel.base,
						   HASH_STAT_OFFSET);
		} while (stat.cq_avail_slots < 4);
		REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
			  sctx->buffer_ioaddr >> 32);

		REG_WR_32(sce_kernel.base, HASH_CQ_FUNC_OFFSET,
			  HASH_WARPOUT.data);
		REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
			  sctx->state_ioaddr & 0xFFFFFFFF);
		REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
			  sctx->state_ioaddr >> 32);
		do {
			csq.data = REG_RD_32_SYNC(sce_kernel.base,
						  HASH_CSQ_OFFSET);
		} while (csq.opcode_idx != WARPOUT);

		spin_unlock_irqrestore(&sce_kernel.hash_lock, flags);
		memcpy(sctx->buffer, sctx->buffer + process_len, last);
	}

	memcpy(out, sctx->buffer, 256);
	memcpy(out + 256, sctx->state, 128);
	memcpy(out + 384, sctx->data, 24);

	return 0;
}

static int sce_hash_import(struct ahash_request *req, const void *in)
{
	struct hash_session_context *sctx =
		(struct hash_session_context *)(((u64)(ahash_request_ctx(req)) +
						 CACHE_LINE_SIZE) &
						(~(CACHE_LINE_SIZE - 1)));

	if (in == NULL) {
		sce_err("Invalid params(null)");
		return -EINVAL;
	}

	memcpy(sctx->buffer, in, 256);
	memcpy(sctx->state, in + 256, 128);
	memcpy(sctx->data, in + 384, 24);
	sctx->state_ioaddr = virt_to_phys(sctx->state);
	sctx->state_ioaddr |= 0x300000000000;
	sctx->buffer_ioaddr = virt_to_phys(sctx->buffer);
	sctx->buffer_ioaddr |= 0x300000000000;
	sctx->list_buf_ioaddr = virt_to_phys(sctx->list_buf);
	sctx->list_buf_ioaddr |= 0x300000000000;
	return 0;
}

static int sce_hash_setkey(struct crypto_ahash *tfm, const u8 *key,
			   unsigned int keylen)
{
	struct hash_alg_context *actx =
		(void *)(((u64)(crypto_tfm_ctx(crypto_ahash_tfm(tfm))) +
			  CACHE_LINE_SIZE) &
			 (~(CACHE_LINE_SIZE - 1)));
	union hash_init_cmd hash_init_cmd;
	union hash_finish_cmd hash_finish_cmd;
	union hash_stat_reg stat;
	union hash_csq_reg csq;
	u64 key_ioaddr;
	u8 *tempkey;
	unsigned long flags;

	if (unlikely(key == NULL)) {
		sce_err("Invalid key(null)");
		return -EINVAL;
	}

	if (likely(actx->key_len)) {
		if (unlikely(keylen > actx->block_size)) {
			tempkey = sce_malloc(keylen);
			if (tempkey == NULL)
				return -ENOMEM;
			memcpy(tempkey, key, keylen);
			key_ioaddr = virt_to_phys(tempkey);
			key_ioaddr |= 0x300000000000;
			hash_init_cmd.data = HASH_INIT.data;
			hash_init_cmd.algo = actx->alg;

			hash_finish_cmd.data = HASH_FINISH.data;
			hash_finish_cmd.mac_tlen = actx->digest_size - 1;

			spin_lock_irqsave(&sce_kernel.hash_lock, flags);

			REG_WR_32(sce_kernel.base, HASH_CQ_FUNC_OFFSET,
				  hash_init_cmd.data);
			REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
				  (iv_paddr + actx->alg * 128u) & 0xFFFFFFFF);
			REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
				  (iv_paddr + actx->alg * 128u) >> 32);

			REG_WR_32(sce_kernel.base, HASH_CQ_FUNC_OFFSET,
				  HASH_PROC.data);
			REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
				  keylen - 1);
			REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
				  key_ioaddr & 0xFFFFFFFF);
			REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
				  key_ioaddr >> 32);

			do {
				stat.data = REG_RD_32_SYNC(sce_kernel.base,
							   HASH_STAT_OFFSET);
			} while (stat.cq_avail_slots < 3);
			REG_WR_32(sce_kernel.base, HASH_CQ_FUNC_OFFSET,
				  hash_finish_cmd.data);
			REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
				  actx->key_ioaddr & 0xFFFFFFFF);
			REG_WR_32(sce_kernel.base, HASH_CQ_PARA_OFFSET,
				  actx->key_ioaddr >> 32);
			do {
				csq.data = REG_RD_32_SYNC(sce_kernel.base,
							  HASH_CSQ_OFFSET);
			} while (csq.opcode_idx != FINISH);

			spin_unlock_irqrestore(&sce_kernel.hash_lock, flags);

			memset(actx->key + actx->digest_size, 0,
			       actx->key_len - actx->digest_size);
			sce_free(tempkey);
		} else {
			memcpy(actx->key, key, keylen);
			memset(actx->key + keylen, 0, actx->key_len - keylen);
		}
	}

	return 0;
}

#define AHASH_ALG(SETKEY, DIGESTSIZE, BLOCKSIZE, NAME, DRIVER_NAME)      \
	{                                                                \
		.init = sce_hash_init, .update = sce_hash_update,        \
		.final = sce_hash_final, .finup = sce_hash_finup,        \
		.digest = sce_hash_digest, .export = sce_hash_export,    \
		.import = sce_hash_import, .setkey = SETKEY,             \
		.halg.digestsize = DIGESTSIZE, .halg.statesize = 512,    \
		.halg.base = {                                           \
			.cra_flags = CRYPTO_ALG_ASYNC,                   \
			.cra_blocksize = BLOCKSIZE,                      \
			.cra_ctxsize = sizeof(struct hash_alg_context) + \
				       CACHE_LINE_SIZE,                  \
			.cra_priority = 300,                             \
			.cra_name = NAME,                                \
			.cra_driver_name = DRIVER_NAME,                  \
			.cra_module = THIS_MODULE,                       \
			.cra_init = sce_hash_cra_init,                   \
		}                                                        \
	}

static struct ahash_alg sce_hash_alg[] = {
	AHASH_ALG(NULL, SM3_DIGEST_BYTE_LEN, SM3_BLOCK_BYTE_LEN, "sm3",
		  "sm3-phytium"),
	AHASH_ALG(sce_hash_setkey, SM3_DIGEST_BYTE_LEN, SM3_BLOCK_BYTE_LEN,
		  "hmac(sm3)", "hmac(sm3)-phytium"),

	AHASH_ALG(NULL, SHA1_DIGEST_BYTE_LEN, SHA1_BLOCK_BYTE_LEN, "sha1",
		  "sha1-phytium"),
	AHASH_ALG(sce_hash_setkey, SHA1_DIGEST_BYTE_LEN, SHA1_BLOCK_BYTE_LEN,
		  "hmac(sha1)", "hmac(sha1)-phytium"),

	AHASH_ALG(NULL, SHA224_DIGEST_BYTE_LEN, SHA224_BLOCK_BYTE_LEN, "sha224",
		  "sha224-phytium"),
	AHASH_ALG(sce_hash_setkey, SHA224_DIGEST_BYTE_LEN,
		  SHA224_BLOCK_BYTE_LEN, "hmac(sha224)",
		  "hmac(sha224)-phytium"),

	AHASH_ALG(NULL, SHA256_DIGEST_BYTE_LEN, SHA256_BLOCK_BYTE_LEN, "sha256",
		  "sha256-phytium"),
	AHASH_ALG(sce_hash_setkey, SHA256_DIGEST_BYTE_LEN,
		  SHA256_BLOCK_BYTE_LEN, "hmac(sha256)",
		  "hmac(sha256)-phytium"),

	AHASH_ALG(NULL, SHA384_DIGEST_BYTE_LEN, SHA384_BLOCK_BYTE_LEN, "sha384",
		  "sha384-phytium"),
	AHASH_ALG(sce_hash_setkey, SHA384_DIGEST_BYTE_LEN,
		  SHA384_BLOCK_BYTE_LEN, "hmac(sha384)",
		  "hmac(sha384)-phytium"),

	AHASH_ALG(NULL, SHA512_DIGEST_BYTE_LEN, SHA512_BLOCK_BYTE_LEN, "sha512",
		  "sha512-phytium"),
	AHASH_ALG(sce_hash_setkey, SHA512_DIGEST_BYTE_LEN,
		  SHA512_BLOCK_BYTE_LEN, "hmac(sha512)",
		  "hmac(sha512)-phytium"),

	AHASH_ALG(NULL, MD5_DIGEST_BYTE_LEN, MD5_BLOCK_BYTE_LEN, "md5",
		  "md5-phytium"),
	AHASH_ALG(sce_hash_setkey, MD5_DIGEST_BYTE_LEN, MD5_BLOCK_BYTE_LEN,
		  "hmac(md5)", "hmac(md5)-phytium"),
};

int sce_hash_algs_register(void)
{
	int ret;

	ret = crypto_register_ahashes(sce_hash_alg, ARRAY_SIZE(sce_hash_alg));
	if (ret) {
		sce_err("%s:crypto_register_ahashes failed! ret:%d\n", __func__,
			ret);
	}

	return ret;
}

void sce_hash_algs_unregister(void)
{
	crypto_unregister_ahashes(sce_hash_alg, ARRAY_SIZE(sce_hash_alg));
}
