// SPDX-License-Identifier: GPL-2.0-or-later
/*
 * File Name:sce_kernel.c - Phytium driver for SDK
 *
 * Copyright (C) 2022 Phytium Technology Co.,Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */

#include <linux/module.h>
#include <linux/mutex.h>
#include <crypto/internal/hash.h>
#include <linux/crypto.h>
#include <crypto/hash.h>
#include <crypto/internal/skcipher.h>
#include <crypto/internal/akcipher.h>
#include <linux/delay.h>
#include <linux/kthread.h>
#include <linux/version.h>
#include "sce_kernel_reg.h"
#include "sce_kernel_common.h"
#include "sce_kernel.h"
#include "sce.h"
#include "hash.h"
#include "cipher.h"
#include "trng.h"
#include "mac.h"
#include "ccm.h"
#include "gcm.h"

struct sce_kernel sce_kernel;
u64 iv_paddr;
static u8 *iv_vaddr;

static u8 sha1_iv_data[20] = {
	0x67, 0x45, 0x23, 0x01, 0xef, 0xcd, 0xab, 0x89, 0x98, 0xba,
	0xdc, 0xfe, 0x10, 0x32, 0x54, 0x76, 0xc3, 0xd2, 0xe1, 0xf0,
};

static u8 sha224_iv_data[32] = {
	0xc1, 0x05, 0x9e, 0xd8, 0x36, 0x7c, 0xd5, 0x07, 0x30, 0x70, 0xdd,
	0x17, 0xf7, 0x0e, 0x59, 0x39, 0xff, 0xc0, 0x0b, 0x31, 0x68, 0x58,
	0x15, 0x11, 0x64, 0xf9, 0x8f, 0xa7, 0xbe, 0xfa, 0x4f, 0xa4,
};

static u8 sha256_iv_data[32] = {
	0x6a, 0x09, 0xe6, 0x67, 0xbb, 0x67, 0xae, 0x85, 0x3c, 0x6e, 0xf3,
	0x72, 0xa5, 0x4f, 0xf5, 0x3a, 0x51, 0x0e, 0x52, 0x7f, 0x9b, 0x05,
	0x68, 0x8c, 0x1f, 0x83, 0xd9, 0xab, 0x5b, 0xe0, 0xcd, 0x19,
};

static u8 sha384_iv_data[64] = {
	0xcb, 0xbb, 0x9d, 0x5d, 0xc1, 0x05, 0x9e, 0xd8, 0x62, 0x9a, 0x29,
	0x2a, 0x36, 0x7c, 0xd5, 0x07, 0x91, 0x59, 0x01, 0x5a, 0x30, 0x70,
	0xdd, 0x17, 0x15, 0x2f, 0xec, 0xd8, 0xf7, 0x0e, 0x59, 0x39, 0x67,
	0x33, 0x26, 0x67, 0xff, 0xc0, 0x0b, 0x31, 0x8e, 0xb4, 0x4a, 0x87,
	0x68, 0x58, 0x15, 0x11, 0xdb, 0x0c, 0x2e, 0x0d, 0x64, 0xf9, 0x8f,
	0xa7, 0x47, 0xb5, 0x48, 0x1d, 0xbe, 0xfa, 0x4f, 0xa4,
};

static u8 sha512_iv_data[64] = {
	0x6a, 0x09, 0xe6, 0x67, 0xf3, 0xbc, 0xc9, 0x08, 0xbb, 0x67, 0xae,
	0x85, 0x84, 0xca, 0xa7, 0x3b, 0x3c, 0x6e, 0xf3, 0x72, 0xfe, 0x94,
	0xf8, 0x2b, 0xa5, 0x4f, 0xf5, 0x3a, 0x5f, 0x1d, 0x36, 0xf1, 0x51,
	0x0e, 0x52, 0x7f, 0xad, 0xe6, 0x82, 0xd1, 0x9b, 0x05, 0x68, 0x8c,
	0x2b, 0x3e, 0x6c, 0x1f, 0x1f, 0x83, 0xd9, 0xab, 0xfb, 0x41, 0xbd,
	0x6b, 0x5b, 0xe0, 0xcd, 0x19, 0x13, 0x7e, 0x21, 0x79,
};

static u8 md5_iv_data[16] = {
	0x01, 0x23, 0x45, 0x67, 0x89, 0xab, 0xcd, 0xef,
	0xfe, 0xdc, 0xba, 0x98, 0x76, 0x54, 0x32, 0x10,
};

static u8 sm3_iv_data[32] = { 0x73, 0x80, 0x16, 0x6f, 0x49, 0x14, 0xb2, 0xb9,
			      0x17, 0x24, 0x42, 0xd7, 0xda, 0x8a, 0x06, 0x00,
			      0xa9, 0x6f, 0x30, 0xbc, 0x16, 0x31, 0x38, 0xaa,
			      0xe3, 0x8d, 0xee, 0x4d, 0xb0, 0xfb, 0x0e, 0x4e };

static int sce_init(u32 *base)
{
	REG_WR_32(base, HASH_CTRL_OFFSET, 0x49);
	REG_WR_32(base, HASH_INTR_STAT0_OFFSET, 0x3F);
	REG_WR_32(base, HASH_INTR_STAT1_OFFSET, 0xFFFFFFFF);
	REG_WR_32(base, HASH_INTR_STAT2_OFFSET, 0xFFFFFFFF);

	REG_WR_32(base, SCA_CTRL_OFFSET, 0x49);
	REG_WR_32(base, SCA_INTR_STAT0_OFFSET, 0x3F);
	REG_WR_32(base, SCA_INTR_STAT1_OFFSET, 0xFFFFFFFF);
	REG_WR_32(base, SCA_INTR_STAT2_OFFSET, 0xFFFFFFFF);

	REG_WR_32(base, RN_POOL_INTR_MSK_OFFSET, 0x1);

	REG_RD_32(base, HASH_CSQ_OFFSET);
	REG_RD_32(base, HASH_CSQ_OFFSET);
	REG_RD_32(base, HASH_CSQ_OFFSET);
	REG_RD_32(base, HASH_CSQ_OFFSET);

	REG_RD_32(base, SCA_CSQ_OFFSET);
	REG_RD_32(base, SCA_CSQ_OFFSET);
	REG_RD_32(base, SCA_CSQ_OFFSET);
	REG_RD_32(base, SCA_CSQ_OFFSET);

	REG_WR_32(sce_kernel.base, 0x6000, 0x10100F0);
	REG_WR_32(sce_kernel.base, 0x6008, 7);
	REG_WR_32(sce_kernel.base, 0x6004, 0);
	REG_WR_32_SYNC(sce_kernel.base, 0x6000, 0x10100F1);

	iv_vaddr = sce_malloc(0x800);
	if (iv_vaddr) {
		iv_paddr = virt_to_phys(iv_vaddr);
		iv_paddr |= 0x300000000000;
		memcpy(iv_vaddr + SHA1 * 128, sha1_iv_data, 20);
		memcpy(iv_vaddr + SHA224 * 128, sha224_iv_data, 32);
		memcpy(iv_vaddr + SHA256 * 128, sha256_iv_data, 32);
		memcpy(iv_vaddr + SHA384 * 128, sha384_iv_data, 64);
		memcpy(iv_vaddr + SHA512 * 128, sha512_iv_data, 64);
		memcpy(iv_vaddr + MD5 * 128, md5_iv_data, 16);
		memcpy(iv_vaddr + SM3 * 128, sm3_iv_data, 32);
	} else {
		return -1;
	}

	return 0;
}

int sce_kernel_crypto_register(void)
{
	int ret = 0;
	
	sce_kernel.base = sce->base;
	if (IS_ERR(sce_kernel.base)) {
		sce_err("ioremap_nocache fail!\n");
		return -1;
	}

	spin_lock_init(&sce_kernel.trng_sm4_lock);
	spin_lock_init(&sce_kernel.trng_lock);
	spin_lock_init(&sce_kernel.hash_lock);
	spin_lock_init(&sce_kernel.cipher_lock);

	ret = sce_init(sce_kernel.base);
	if (ret)
		goto err0;

	ret = sce_hash_algs_register();
	if (ret)
		goto err1;

	ret = sce_cipher_algs_register();
	if (ret)
		goto err2;

	ret = sce_trng_alg_register();
	if (ret)
		goto err3;
#if 0
	ret = sce_mac_algs_register();
	if (ret)
		goto err4;

	ret = sce_ccm_algs_register();
	if (ret)
		goto err5;

	ret = sce_gcm_algs_register();
	if (ret)
		goto err6;
#endif
	return 0;
#if 0
err6:
	sce_ccm_algs_unregister();
err5:
	sce_mac_algs_unregister();
err4:
#endif
	sce_trng_alg_unregister();
err3:
	sce_cipher_algs_unregister();
err2:
	sce_hash_algs_unregister();
err1:
	sce_free(iv_vaddr);
err0:
	//iounmap(sce_kernel.base);

	return ret;
}

void sce_kernel_crypto_unregister(void)
{
#if 0
	sce_gcm_algs_unregister();
	sce_ccm_algs_unregister();
	sce_mac_algs_unregister();
#endif
	sce_trng_alg_unregister();
	sce_cipher_algs_unregister();
	sce_hash_algs_unregister();
	sce_free(iv_vaddr);
}
