// SPDX-License-Identifier: GPL-2.0-or-later
/*
 * File Name:mac.c - Phytium SDK for MAC
 *
 * Copyright (C) 2022 Phytium Technology Co.,Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */

#include <linux/delay.h>
#include <crypto/internal/hash.h>
#include <crypto/hash.h>
#include <linux/crypto.h>
#include <linux/delay.h>
#include <linux/kthread.h>
#include <linux/version.h>
#include "sce_kernel_common.h"
#include "sce_kernel_reg.h"
#include "sce_kernel.h"
#include "mac.h"
#include "cipher.h"

static int register_alg[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };

static char *alg_name[] = { "cmac(aes)",      "cbcmac(aes)",	 "cmac(sm4)",
			    "cbcmac(sm4)",    "cmac(des)",	 "cbcmac(des)",
			    "cmac(des3_ede)", "cbcmac(des3_ede)" };
static u8 alg[] = { AES, AES, SM4, SM4, DES, DES, TDES, TDES };
static u8 mode[] = {
	CMAC, CBC_MAC, CMAC, CBC_MAC, CMAC, CBC_MAC, CMAC, CBC_MAC
};
static u8 block_size[] = { AES_BLOCK_BYTE_LEN,	AES_BLOCK_BYTE_LEN,
			   SM4_BLOCK_BYTE_LEN,	SM4_BLOCK_BYTE_LEN,
			   DES_BLOCK_BYTE_LEN,	DES_BLOCK_BYTE_LEN,
			   TDES_BLOCK_BYTE_LEN, TDES_BLOCK_BYTE_LEN };
static u8 digest_size[] = { CMAC_DIGEST_BYTE_LEN, CBC_MAC_DIGEST_BYTE_LEN,
			    CMAC_DIGEST_BYTE_LEN, CBC_MAC_DIGEST_BYTE_LEN,
			    DES_BLOCK_BYTE_LEN,	  DES_BLOCK_BYTE_LEN,
			    TDES_BLOCK_BYTE_LEN,  TDES_BLOCK_BYTE_LEN };
static u8 key_len[] = {
	AES_MIN_KEY_BYTE_LEN, AES_MIN_KEY_BYTE_LEN, SM4_KEY_BYTE_LEN,
	SM4_KEY_BYTE_LEN,     DES_KEY_BYTE_LEN,	    DES_KEY_BYTE_LEN,
	TDES_KEY_BYTE_LEN,    TDES_KEY_BYTE_LEN,
};

static union sca_init_cmd MAC_INIT = {
	.trigger_intr = 0,
	.mode = 2,
	.iv_load = 0,
	.key_len = 0,
	.key_src = 2,
	.algo = 0,
	.reserved = 0,
	.slot_id = 7,
	.opcode = 0x80,
};

static union sca_proc_cmd MAC_PROC = {
	.trigger_intr = 0,
	.dst_src_addr_type = 0,
	.is_last_proc = 1,
	.op_mode = 0,
	.reserved = 0,
	.slot_id = 7,
	.opcode = 0x40,
};

static union sca_finish_cmd MAC_FINISH = {
	.trigger_intr = 0,
	.mac_tlen = 15,
	.reserved = 0,
	.slot_id = 7,
	.opcode = 0x20,
};

static union sca_warp_cmd MAC_WARPOUT = {
	.trigger_intr = 0,
	.reserved = 0,
	.slot_id = 7,
	.opcode = 0x4,
};

static union sca_warp_cmd MAC_WARPIN = {
	.trigger_intr = 0,
	.reserved = 0,
	.slot_id = 7,
	.opcode = 0x2,
};

static int sce_mac_cra_init(struct crypto_tfm *tfm)
{
	int i;
	struct mac_alg_context *actx =
		(void *)(((u64)(crypto_tfm_ctx(tfm)) + CACHE_LINE_SIZE) &
			 (~(CACHE_LINE_SIZE - 1)));

	memset(actx, 0, sizeof(struct mac_alg_context));
	crypto_ahash_set_reqsize(__crypto_ahash_cast(tfm),
				 sizeof(struct mac_session_context) +
					 CACHE_LINE_SIZE);

	for (i = 0; i < ARRAY_SIZE(alg_name); i++) {
		if (memcmp(tfm->__crt_alg->cra_name, alg_name[i],
			   strlen(alg_name[i])) == 0) {
			actx->alg = alg[i];
			actx->mode = mode[i];
			actx->block_size = block_size[i];
			actx->digest_size = digest_size[i];
			actx->key_len = key_len[i];
			actx->key_ioaddr = virt_to_phys(actx->key);
			actx->key_ioaddr |= 0x300000000000;
		}
	}

	if (actx->block_size == 0)
		return -EINVAL;

	return 0;
}

static int sce_mac_generic(struct ahash_request *req)
{
	u8 alg[64] = { 0 };
	struct crypto_wait wait_tmp;
	struct mac_alg_context *actx =
		(void *)(((u64)(crypto_tfm_ctx(req->base.tfm)) +
			  CACHE_LINE_SIZE) &
			 (~(CACHE_LINE_SIZE - 1)));
	struct ahash_request *req_tmp;
	struct crypto_ahash *tfm_tmp;
	int ret = 0;

	if (actx->mode == CBC_MAC) {
		if (actx->alg == AES)
			memcpy(alg, "cbcmac(aes-generic)",
			       strlen("cbcmac(aes-generic)"));
		else if (actx->alg == SM4)
			memcpy(alg, "cbcmac(sm4-generic)",
			       strlen("cbcmac(sm4-generic)"));
		else if (actx->alg == DES)
			memcpy(alg, "cbcmac(des-generic)",
			       strlen("cbcmac(des-generic)"));
		else
			memcpy(alg, "cbcmac(des3_ede-generic)",
			       strlen("cbcmac(des3_ede-generic)"));
	} else {
		if (actx->alg == AES)
			memcpy(alg, "cmac(aes-generic)",
			       strlen("cmac(aes-generic)"));
		else if (actx->alg == SM4)
			memcpy(alg, "cmac(sm4-generic)",
			       strlen("cmac(sm4-generic)"));
		else if (actx->alg == DES)
			memcpy(alg, "cmac(des-generic)",
			       strlen("cmac(des-generic)"));
		else
			memcpy(alg, "cmac(des3_ede-generic)",
			       strlen("cmac(des3_ede-generic)"));
	}

	tfm_tmp = crypto_alloc_ahash(alg, 0, 0);
	if (IS_ERR(tfm_tmp)) {
		pr_err("failed to load transform for %s: %ld\n",
		       "hash", PTR_ERR(tfm_tmp));
		return -1;
	}

	req_tmp = ahash_request_alloc(tfm_tmp, GFP_KERNEL);
	if (!req_tmp) {
		pr_err("ahash request allocation failure\n");
		crypto_free_ahash(tfm_tmp);
		return -1;
	}

	crypto_init_wait(&wait_tmp);
	ahash_request_set_callback(req_tmp, CRYPTO_TFM_REQ_MAY_BACKLOG,
				   crypto_req_done, &wait_tmp);

	ret = crypto_ahash_setkey(tfm_tmp, actx->key, actx->key_len);
	if (ret) {
		pr_err("hash setkey failed ret=%d\n", ret);
		ahash_request_free(req_tmp);
		crypto_free_ahash(tfm_tmp);
		return -1;
	}

	ahash_request_set_crypt(req_tmp, req->src, req->result, req->nbytes);

	ret = crypto_wait_req(crypto_ahash_digest(req_tmp), &wait_tmp);
	if (ret) {
		pr_err("hash init failed ret=%d\n", ret);
		ahash_request_free(req_tmp);
		crypto_free_ahash(tfm_tmp);
		return -1;
	}

	ahash_request_free(req_tmp);
	crypto_free_ahash(tfm_tmp);
	return 0;
}

static int sce_mac_init(struct ahash_request *req)
{
	struct mac_session_context *sctx =
		(struct mac_session_context *)(((u64)(ahash_request_ctx(req)) +
						CACHE_LINE_SIZE) &
					       (~(CACHE_LINE_SIZE - 1)));

	memset(sctx, 0, sizeof(struct mac_session_context));
	sctx->total = 0;
	sctx->buffer_data_len = 0;
	sctx->mac_flag = INIT;
	sctx->is_first_req = 1;
	sctx->state_ioaddr = virt_to_phys(sctx->state);
	memset(sctx->buffer, 0, 16);

	sctx->state_ioaddr |= 0x300000000000;
	sctx->buffer_ioaddr = virt_to_phys(sctx->buffer);
	sctx->buffer_ioaddr |= 0x300000000000;
	sctx->list_buf_ioaddr = virt_to_phys(sctx->list_buf);
	sctx->list_buf_ioaddr |= 0x300000000000;
	return 0;
}

static int sce_mac_update(struct ahash_request *req)
{
	struct mac_session_context *sctx =
		(struct mac_session_context *)(((u64)(ahash_request_ctx(req)) +
						CACHE_LINE_SIZE) &
					       (~(CACHE_LINE_SIZE - 1)));
	struct mac_alg_context *actx =
		(void *)(((u64)(crypto_tfm_ctx(req->base.tfm)) +
			  CACHE_LINE_SIZE) &
			 (~(CACHE_LINE_SIZE - 1)));
	struct scatterlist *sg;
	int src_nents;
	u32 last, src_len = 0, temp, process_len;
	u64 src_addr = 0;
	struct link_list_item *link_list = NULL;
	int i;
	int skip = 0;
	u32 link_mode = 0;
	u64 link_ioaddr;
	union sca_init_cmd mac_init_cmd;
	union sca_proc_cmd mac_proc_cmd;
	union sca_stat_reg stat = { 0 };
	union sca_csq_reg csq;
	unsigned long flags;

	if (unlikely(!req->nbytes))
		return 0;

	if (likely(sctx->buffer_data_len + req->nbytes <= 16)) {
		temp = req->nbytes;
		for (sg = req->src; sg; sg = sg_next(sg)) {
			if (!sg->length)
				continue;
			if (!temp)
				break;
			if (sg->length <= temp) {
				sg_copy_buffer(req->src, sg_nents(req->src),
					       sctx->buffer +
						       sctx->buffer_data_len,
					       sg->length, skip, true);
				temp -= sg->length;
				skip += sg->length;
				sctx->buffer_data_len += sg->length;
			} else {
				sg_copy_buffer(req->src, sg_nents(req->src),
					       sctx->buffer +
						       sctx->buffer_data_len,
					       temp, skip, true);
				sctx->buffer_data_len += temp;
				temp = 0;
			}
		}
		sctx->total += req->nbytes;
		return 0;
	}
	sctx->total += req->nbytes;
	last = sctx->total & (actx->block_size - 1);
	if (!last)
		last += actx->block_size;
	process_len = req->nbytes - last;
	src_nents = sg_nents(req->src);
	if (unlikely(src_nents > MAX_LINKS)) {
		link_list = sce_malloc(sizeof(struct link_list_item) *
				       (src_nents + 3));
		if (link_list == NULL) {
			sce_debug("sce_malloc failed!\n");
			return -EMLINK;
		}
		link_ioaddr = virt_to_phys(link_list);
		link_ioaddr |= 0x300000000000;
	} else {
		link_list = sctx->list_buf;
		link_ioaddr = sctx->list_buf_ioaddr;
		link_ioaddr |= 0x300000000000;
	}
	src_addr = link_ioaddr;

	i = 0;
	link_mode = 2;
	temp = process_len;

	if (likely(sctx->buffer_data_len)) {
		link_list[i].size = sctx->buffer_data_len - 1;
		link_list[i].addr = sctx->buffer_ioaddr;
		link_list[i].addr |= 0x300000000000;
		i++;
	}
	for (sg = req->src; sg; sg = sg_next(sg)) {
		if (!sg->length)
			continue;
		if (!temp)
			break;
		link_list[i].size = (temp > sg->length) ? sg->length : temp;
		link_list[i].addr = sg_phys(sg);
		link_list[i].addr |= 0x300000000000;
		temp -= link_list[i].size;
		link_list[i].size -= 1;
		i++;
	}
	link_list[i].size = 0xFFFFFFFFFFFFFFFFull;
	link_list[i].addr = 0;
	mac_proc_cmd.data = MAC_PROC.data;
	mac_proc_cmd.dst_src_addr_type = link_mode;
	mac_proc_cmd.is_last_proc = 0;

	src_len = process_len;

	spin_lock_irqsave(&sce_kernel.cipher_lock, flags);
	stat.data = REG_RD_32_SYNC(sce_kernel.base, SCA_STAT_OFFSET);

	if (likely(sctx->mac_flag == INIT)) {
		mac_init_cmd.data = MAC_INIT.data;
		mac_init_cmd.mode = actx->mode;
		mac_init_cmd.algo = actx->alg;
		mac_init_cmd.iv_load = 0;
		mac_init_cmd.key_len = actx->key_len_flag;

		REG_WR_32(sce_kernel.base, SCA_CQ_FUNC_OFFSET,
			  mac_init_cmd.data);
		if (unlikely(actx->mode)) {
			REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET,
				  actx->key_ioaddr & 0xFFFFFFFF);
			REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET,
				  actx->key_ioaddr >> 32);
		}
		sctx->mac_flag = WARPOUT;
	} else if (sctx->mac_flag == WARPOUT) {
		REG_WR_32(sce_kernel.base, SCA_CQ_FUNC_OFFSET, MAC_WARPIN.data);
		REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET,
			  sctx->state_ioaddr & 0xFFFFFFFF);
		REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET,
			  sctx->state_ioaddr >> 32);
	} else {
		spin_unlock_irqrestore(&sce_kernel.cipher_lock, flags);
		if (link_list != sctx->list_buf)
			sce_free(link_list);
		sce_err("mac_flag[%lld] err!\n", sctx->mac_flag);
		return -EINVAL;
	}
	if (i == 0 && actx->mode != CMAC)
		goto out;

	if (sctx->total) {
		REG_WR_32(sce_kernel.base, SCA_CQ_FUNC_OFFSET,
			  mac_proc_cmd.data);
		if (unlikely((link_mode & 0x2) == 0)) {
			REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET,
				  src_len - 1);
		}
		REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET,
			  src_addr & 0xFFFFFFFF);
		do {
			stat.data = REG_RD_32(sce_kernel.base, SCA_STAT_OFFSET);
		} while (stat.cq_avail_slots < 3);
		REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET, src_addr >> 32);
	}
out:
	stat.data = REG_RD_32_SYNC(sce_kernel.base, SCA_STAT_OFFSET);

	REG_WR_32(sce_kernel.base, SCA_CQ_FUNC_OFFSET, MAC_WARPOUT.data);
	stat.data = REG_RD_32_SYNC(sce_kernel.base, SCA_STAT_OFFSET);

	REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET,
		  sctx->state_ioaddr & 0xFFFFFFFF);
	REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET,
		  sctx->state_ioaddr >> 32);
	do {
		csq.data = REG_RD_32_SYNC(sce_kernel.base, SCA_CSQ_OFFSET);
	} while (csq.opcode_idx != WARPOUT);
	sctx->mac_flag = WARPOUT;
	spin_unlock_irqrestore(&sce_kernel.cipher_lock, flags);

	if (last) {
		sg_copy_buffer(req->src, src_nents, sctx->buffer, last,
			       process_len, true);
		sctx->buffer_data_len = last;
	} else {
		sctx->buffer_data_len = 0;
	}

	if (unlikely(link_list != sctx->list_buf))
		sce_free(link_list);

	return 0;
}

static int sce_mac_final(struct ahash_request *req)
{
	struct mac_session_context *sctx =
		(struct mac_session_context *)(((u64)(ahash_request_ctx(req)) +
						CACHE_LINE_SIZE) &
					       (~(CACHE_LINE_SIZE - 1)));
	struct mac_alg_context *actx =
		(void *)(((u64)(crypto_tfm_ctx(req->base.tfm)) +
			  CACHE_LINE_SIZE) &
			 (~(CACHE_LINE_SIZE - 1)));
	union sca_init_cmd mac_init_cmd;
	union sca_proc_cmd mac_proc_cmd;
	union sca_finish_cmd mac_finish_cmd;
	union sca_stat_reg stat = { 0 };
	union sca_csq_reg csq;
	struct link_list_item *link_list = NULL;
	int i, src_len = 0;
	u32 link_mode = 0;
	u64 link_ioaddr;
	u64 src_addr;
	u8 *cbcmac_last;
	unsigned long flags;

	if (!sctx->total)
		return sce_mac_generic(req);

	cbcmac_last = sce_malloc(16);
	if (cbcmac_last == NULL) {
		sce_debug("sce_malloc failed!\n");
		return -EMLINK;
	}
	memset(cbcmac_last, 0, 16);
	mac_proc_cmd.data = MAC_PROC.data;

	mac_finish_cmd.data = MAC_FINISH.data;
	mac_finish_cmd.mac_tlen = actx->digest_size - 1;

	spin_lock_irqsave(&sce_kernel.cipher_lock, flags);

	if (likely(sctx->mac_flag == INIT)) {
		mac_init_cmd.data = MAC_INIT.data;
		mac_init_cmd.mode = actx->mode;
		mac_init_cmd.algo = actx->alg;
		mac_init_cmd.iv_load = 0;
		mac_init_cmd.key_len = actx->key_len_flag;

		REG_WR_32(sce_kernel.base, SCA_CQ_FUNC_OFFSET,
			  mac_init_cmd.data);
		if (unlikely(actx->mode)) {
			REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET,
				  actx->key_ioaddr & 0xFFFFFFFF);
			REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET,
				  actx->key_ioaddr >> 32);
		}
	} else if (sctx->mac_flag == WARPOUT) {
		sctx->state_ioaddr = virt_to_phys(sctx->state);
		sctx->state_ioaddr |= 0x300000000000;
		REG_WR_32(sce_kernel.base, SCA_CQ_FUNC_OFFSET, MAC_WARPIN.data);
		REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET,
			  sctx->state_ioaddr & 0xFFFFFFFF);
		REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET,
			  sctx->state_ioaddr >> 32);
	} else {
		spin_unlock_irqrestore(&sce_kernel.cipher_lock, flags);
		sce_err("mac_flag[%lld] err!\n", sctx->mac_flag);
		sce_free(cbcmac_last);
		return -EINVAL;
	}
	sctx->mac_flag = FINISH;
	i = 0;
	link_list = sctx->list_buf;
	link_ioaddr = sctx->list_buf_ioaddr;
	link_ioaddr |= 0x300000000000;
	link_list[i].size = sctx->buffer_data_len - 1;
	link_list[i].addr = sctx->buffer_ioaddr;
	link_list[i].addr |= 0x300000000000;
	src_len += sctx->buffer_data_len;
	i++;
	if (sctx->total & (actx->block_size - 1) && actx->mode == CBC_MAC) {
		link_list[i].size = actx->block_size -
				    (sctx->total & (actx->block_size - 1)) - 1;
		src_len += actx->block_size -
			   (sctx->total & (actx->block_size - 1));
		link_list[i].addr = virt_to_phys(cbcmac_last);
		link_list[i].addr |= 0x300000000000;
		i++;
	}
	link_list[i].size = 0xFFFFFFFFFFFFFFFFull;
	link_list[i].addr = 0;
	i++;
	link_mode = 2;

	src_addr = link_ioaddr;

	mac_proc_cmd.data = MAC_PROC.data;
	mac_proc_cmd.dst_src_addr_type = link_mode;
	mac_proc_cmd.is_last_proc = 1;

	REG_WR_32(sce_kernel.base, SCA_CQ_FUNC_OFFSET, mac_proc_cmd.data);
	if (likely(link_mode == 0))
		REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET, src_len - 1);

	REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET, src_addr & 0xFFFFFFFF);
	do {
		stat.data = REG_RD_32_SYNC(sce_kernel.base, SCA_STAT_OFFSET);
	} while (stat.cq_avail_slots < 2);
	REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET, src_addr >> 32);
	REG_WR_32(sce_kernel.base, SCA_CQ_FUNC_OFFSET, mac_finish_cmd.data);
	do {
		stat.data = REG_RD_32_SYNC(sce_kernel.base, SCA_STAT_OFFSET);
	} while (stat.cq_avail_slots < 6);

	REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET,
		  sctx->state_ioaddr & 0xFFFFFFFF);
	REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET,
		  sctx->state_ioaddr >> 32);

	do {
		csq.data = REG_RD_32_SYNC(sce_kernel.base, SCA_CSQ_OFFSET);
	} while (csq.opcode_idx != FINISH);

	spin_unlock_irqrestore(&sce_kernel.cipher_lock, flags);

	memcpy(req->result, sctx->state, actx->digest_size);

	if (cbcmac_last)
		sce_free(cbcmac_last);

	return 0;
}

static int sce_mac_finup(struct ahash_request *req)
{
	struct mac_session_context *sctx =
		(struct mac_session_context *)(((u64)(ahash_request_ctx(req)) +
						CACHE_LINE_SIZE) &
					       (~(CACHE_LINE_SIZE - 1)));
	struct mac_alg_context *actx =
		(void *)(((u64)(crypto_tfm_ctx(req->base.tfm)) +
			  CACHE_LINE_SIZE) &
			 (~(CACHE_LINE_SIZE - 1)));
	struct scatterlist *sg;
	int src_nents;
	u64 src_addr;
	int src_len = 0;
	struct link_list_item *link_list = NULL;
	int i;
	u32 link_mode = 0;
	u64 link_ioaddr;
	union sca_init_cmd mac_init_cmd;
	union sca_proc_cmd mac_proc_cmd;
	union sca_finish_cmd mac_finish_cmd;
	union sca_stat_reg stat = { 0 };
	union sca_csq_reg csq;
	u8 *cbcmac_last;
	unsigned long flags;

	sctx->total += req->nbytes;
	if (!sctx->total)
		return sce_mac_generic(req);
	cbcmac_last = sce_malloc(16);
	if (cbcmac_last == NULL) {
		sce_debug("sce_malloc failed!\n");
		return -EMLINK;
	}
	memset(cbcmac_last, 0, 16);

	src_nents = sg_nents(req->src);
	if (unlikely(src_nents > MAX_LINKS)) {
		link_list = sce_malloc(sizeof(struct link_list_item) *
				       (src_nents + 3));
		if (link_list == NULL) {
			sce_debug("sce_malloc failed!\n");
			sce_free(cbcmac_last);
			return -EMLINK;
		}
		link_ioaddr = virt_to_phys(link_list);
		link_ioaddr |= 0x300000000000;
	} else {
		link_list = sctx->list_buf;
		link_ioaddr = sctx->list_buf_ioaddr;
		link_ioaddr |= 0x300000000000;
	}

	src_addr = link_ioaddr;
	link_mode = 2;
	i = 0;
	if (likely(sctx->buffer_data_len)) {
		link_list[i].size = sctx->buffer_data_len - 1;
		src_len += sctx->buffer_data_len;
		link_list[i].addr = sctx->buffer_ioaddr;
		link_list[i].addr |= 0x300000000000;
		i++;
	}

	for (sg = req->src; sg; sg = sg_next(sg)) {
		if (!sg->length)
			continue;
		link_list[i].size = sg->length - 1;
		link_list[i].addr = sg_phys(sg);
		src_len += sg->length;
		link_list[i].addr |= 0x300000000000;
		i++;
	}

	if (sctx->total & (actx->block_size - 1) && actx->mode == CBC_MAC) {
		link_list[i].size = actx->block_size -
				    (sctx->total & (actx->block_size - 1)) - 1;
		link_list[i].addr = virt_to_phys(cbcmac_last);
		link_list[i].addr |= 0x300000000000;
		src_len += actx->block_size -
			   (sctx->total & (actx->block_size - 1));
		i++;
	}
	link_list[i].size = 0xFFFFFFFFFFFFFFFFull;
	link_list[i].addr = 0;
	mac_proc_cmd.data = MAC_PROC.data;
	mac_proc_cmd.dst_src_addr_type = link_mode;
	mac_proc_cmd.is_last_proc = 1;
	mac_finish_cmd.data = MAC_FINISH.data;
	mac_finish_cmd.mac_tlen = actx->digest_size - 1;

	spin_lock_irqsave(&sce_kernel.cipher_lock, flags);

	if (likely(sctx->mac_flag == INIT)) {
		mac_init_cmd.data = MAC_INIT.data;
		mac_init_cmd.mode = actx->mode;
		mac_init_cmd.algo = actx->alg;
		mac_init_cmd.key_len = actx->key_len_flag;
		mac_init_cmd.iv_load = 0;
		REG_WR_32(sce_kernel.base, SCA_CQ_FUNC_OFFSET,
			  mac_init_cmd.data);
		if (unlikely(actx->mode)) {
			REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET,
				  actx->key_ioaddr & 0xFFFFFFFF);
			REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET,
				  actx->key_ioaddr >> 32);
		}
	} else if (sctx->mac_flag == WARPOUT) {
		REG_WR_32(sce_kernel.base, SCA_CQ_FUNC_OFFSET, MAC_WARPIN.data);
		REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET,
			  sctx->state_ioaddr & 0xFFFFFFFF);
		REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET,
			  sctx->state_ioaddr >> 32);
	} else {
		spin_unlock_irqrestore(&sce_kernel.cipher_lock, flags);
		if (link_list != sctx->list_buf)
			sce_free(link_list);
		sce_err("mac_flag[%lld] err!\n", sctx->mac_flag);
		sce_free(cbcmac_last);
		return -EINVAL;
	}
	sctx->mac_flag = FINISH;

	REG_WR_32(sce_kernel.base, SCA_CQ_FUNC_OFFSET, mac_proc_cmd.data);
	if (likely(link_mode == 0))
		REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET, src_len - 1);

	REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET, src_addr & 0xFFFFFFFF);
	do {
		stat.data = REG_RD_32_SYNC(sce_kernel.base, SCA_STAT_OFFSET);
	} while (stat.cq_avail_slots < 2);
	REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET, src_addr >> 32);

	REG_WR_32(sce_kernel.base, SCA_CQ_FUNC_OFFSET, mac_finish_cmd.data);
	do {
		stat.data = REG_RD_32_SYNC(sce_kernel.base, SCA_STAT_OFFSET);
	} while (stat.cq_avail_slots < 6);
	REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET,
		  sctx->state_ioaddr & 0xFFFFFFFF);
	REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET,
		  sctx->state_ioaddr >> 32);
	do {
		csq.data = REG_RD_32_SYNC(sce_kernel.base, SCA_CSQ_OFFSET);
	} while (csq.opcode_idx != FINISH);

	spin_unlock_irqrestore(&sce_kernel.cipher_lock, flags);

	memcpy(req->result, sctx->state, actx->digest_size);

	if (unlikely(link_list != sctx->list_buf))
		sce_free(link_list);

	if (cbcmac_last)
		sce_free(cbcmac_last);
	return 0;
}

static int sce_mac_digest(struct ahash_request *req)
{
	struct mac_session_context *sctx =
		(struct mac_session_context *)(((u64)(ahash_request_ctx(req)) +
						CACHE_LINE_SIZE) &
					       (~(CACHE_LINE_SIZE - 1)));
	struct mac_alg_context *actx =
		(void *)(((u64)(crypto_tfm_ctx(req->base.tfm)) +
			  CACHE_LINE_SIZE) &
			 (~(CACHE_LINE_SIZE - 1)));
	struct scatterlist *sg;
	int src_nents;
	u32 process_len;
	u64 src_addr = 0;
	u64 src_len = 0;
	struct link_list_item *link_list = NULL;
	int i = 0;
	u32 link_mode = 0;
	u64 link_ioaddr;
	union sca_init_cmd mac_init_cmd;
	union sca_proc_cmd mac_proc_cmd;
	union sca_finish_cmd mac_finish_cmd;
	union sca_stat_reg stat = { 0 };
	union sca_csq_reg csq;
	u8 *cbcmac_last;
	unsigned long flags;

	if (!req->nbytes)
		return sce_mac_generic(req);

	cbcmac_last = sce_malloc(16);
	if (cbcmac_last == NULL) {
		sce_debug("sce_malloc failed!\n");
		return -EMLINK;
	}
	src_nents = sg_nents(req->src);
	if (unlikely(src_nents > MAX_LINKS)) {
		link_list = sce_malloc(sizeof(struct link_list_item) *
				       (src_nents + 2));
		if (link_list == NULL) {
			sce_debug("sce_malloc failed!\n");
			sce_free(cbcmac_last);
			return -EMLINK;
		}
	} else {
		link_list = sctx->list_buf;
	}
	memset(cbcmac_last, 0, 16);
	link_ioaddr = virt_to_phys(link_list);
	link_ioaddr |= 0x300000000000;
	src_addr = link_ioaddr;
	sctx->state_ioaddr = virt_to_phys(sctx->state);
	sctx->state_ioaddr |= 0x300000000000;
	process_len = req->nbytes;
	sctx->total = req->nbytes;
	sctx->is_first_req = 1;

	link_mode = 2;
	for (sg = req->src; sg; sg = sg_next(sg)) {
		if (!sg->length)
			continue;
		if (sctx->is_first_req) {
			sctx->size = sg->length;
			sctx->addr = sg_phys(sg);
			sctx->addr |= 0x300000000000;
			sctx->is_first_req = 0;
			continue;
		} else {
			link_list[i].size = sctx->size - 1;
			link_list[i].addr = sctx->addr;
			link_list[i].addr |= 0x300000000000;
			src_len += sctx->size;
			sctx->size = sg->length;
			sctx->addr = sg_phys(sg);
			sctx->addr |= 0x300000000000;
			i++;
		}
	}

	if (sctx->is_first_req && actx->mode != CMAC)
		goto out;

	link_list[i].size = sctx->size - 1;
	link_list[i].addr = sctx->addr;
	link_list[i].addr |= 0x300000000000;
	src_len += sctx->size;
	i++;

	if (sctx->total & (actx->block_size - 1) && actx->mode == CBC_MAC) {
		link_list[i].size = actx->block_size -
				    (sctx->total & (actx->block_size - 1)) - 1;
		link_list[i].addr = virt_to_phys(cbcmac_last);
		link_list[i].addr |= 0x300000000000;
		i++;
		src_len += actx->block_size -
			   (sctx->total & (actx->block_size - 1));
	}
	link_list[i].size = 0xFFFFFFFFFFFFFFFFull;
	link_list[i].addr = 0;
	mac_init_cmd.data = MAC_INIT.data;
	mac_init_cmd.mode = actx->mode;
	mac_init_cmd.algo = actx->alg;
	mac_init_cmd.iv_load = 0;
	mac_init_cmd.key_len = actx->key_len_flag;

	mac_proc_cmd.data = MAC_PROC.data;
	mac_proc_cmd.dst_src_addr_type = link_mode;
	mac_proc_cmd.is_last_proc = 1;

	mac_finish_cmd.data = MAC_FINISH.data;
	mac_finish_cmd.mac_tlen = actx->digest_size - 1;
	spin_lock_irqsave(&sce_kernel.cipher_lock, flags);

	REG_WR_32(sce_kernel.base, SCA_CQ_FUNC_OFFSET, mac_init_cmd.data);
	if (unlikely(actx->mode)) {
		REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET,
			  actx->key_ioaddr & 0xFFFFFFFF);
		REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET,
			  actx->key_ioaddr >> 32);
	}
	REG_WR_32(sce_kernel.base, SCA_CQ_FUNC_OFFSET, mac_proc_cmd.data);
	if (unlikely((link_mode & 0x2) == 0))
		REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET, src_len - 1);

	REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET, src_addr & 0xFFFFFFFF);
	do {
		stat.data = REG_RD_32(sce_kernel.base, SCA_STAT_OFFSET);
	} while (stat.cq_avail_slots < 1);
	REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET, src_addr >> 32);

	REG_WR_32(sce_kernel.base, SCA_CQ_FUNC_OFFSET, mac_finish_cmd.data);
	do {
		stat.data = REG_RD_32(sce_kernel.base, SCA_STAT_OFFSET);
	} while (stat.cq_avail_slots < 6);
	REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET,
		  sctx->state_ioaddr & 0xFFFFFFFF);
	REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET,
		  sctx->state_ioaddr >> 32);
	do {
		csq.data = REG_RD_32_SYNC(sce_kernel.base, SCA_CSQ_OFFSET);
	} while (csq.opcode_idx != FINISH);

	spin_unlock_irqrestore(&sce_kernel.cipher_lock, flags);

	memcpy(req->result, sctx->state, actx->digest_size);
out:
	if (cbcmac_last)
		sce_free(cbcmac_last);
	if (unlikely(link_list != sctx->list_buf))
		sce_free(link_list);

	return 0;
}

static int sce_mac_export(struct ahash_request *req, void *out)
{
	struct mac_session_context *sctx =
		(struct mac_session_context *)(((u64)(ahash_request_ctx(req)) +
						CACHE_LINE_SIZE) &
					       (~(CACHE_LINE_SIZE - 1)));
	union sca_csq_reg csq;
	unsigned long flags;

	if (out == NULL) {
		sce_err("Invalid params(null)");
		return -EINVAL;
	}

	if (sctx->mac_flag != INIT && sctx->mac_flag != WARPOUT) {
		spin_lock_irqsave(&sce_kernel.cipher_lock, flags);
		REG_WR_32(sce_kernel.base, SCA_CQ_FUNC_OFFSET,
			  MAC_WARPOUT.data);
		REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET,
			  sctx->state_ioaddr & 0xFFFFFFFF);
		REG_WR_32(sce_kernel.base, SCA_CQ_PARA_OFFSET,
			  sctx->state_ioaddr >> 32);
		do {
			csq.data =
				REG_RD_32_SYNC(sce_kernel.base, SCA_CSQ_OFFSET);
		} while (csq.opcode_idx != WARPOUT);

		spin_unlock_irqrestore(&sce_kernel.cipher_lock, flags);
		sctx->mac_flag = WARPOUT;
	}

	memcpy(out, sctx->list_data, 24);
	memcpy(out + 256, sctx->state, 128);
	memcpy(out + 384, sctx->data, 24);
	memcpy(out + 408, sctx->buffer, 16);

	return 0;
}

static int sce_mac_import(struct ahash_request *req, const void *in)
{
	struct mac_session_context *sctx =
		(struct mac_session_context *)(((u64)(ahash_request_ctx(req)) +
						CACHE_LINE_SIZE) &
					       (~(CACHE_LINE_SIZE - 1)));

	if (sctx->mac_flag == INIT)
		return -EINVAL;
	if (in == NULL) {
		sce_err("Invalid params(null)");
		return -EINVAL;
	}

	memcpy(sctx->list_data, in, 24);
	memcpy(sctx->state, in + 256, 128);
	memcpy(sctx->data, in + 384, 24);
	memcpy(sctx->buffer, in + 408, 16);
	sctx->state_ioaddr = virt_to_phys(sctx->state);
	sctx->state_ioaddr |= 0x300000000000;
	sctx->buffer_ioaddr = virt_to_phys(sctx->buffer);
	sctx->buffer_ioaddr |= 0x300000000000;
	sctx->list_buf_ioaddr = virt_to_phys(sctx->list_buf);
	sctx->list_buf_ioaddr |= 0x300000000000;
	return 0;
}

static int sce_mac_setkey(struct crypto_ahash *tfm, const u8 *key,
			  unsigned int keylen)
{
	struct mac_alg_context *actx =
		(void *)(((u64)(crypto_tfm_ctx(crypto_ahash_tfm(tfm))) +
			  CACHE_LINE_SIZE) &
			 (~(CACHE_LINE_SIZE - 1)));

	if (unlikely(key == NULL)) {
		sce_err("Invalid key(null)");
		return -EINVAL;
	}

	if (actx->alg == AES) {
		switch (keylen) {
		case AES_MIN_KEY_BYTE_LEN:
			actx->key_len = AES_MIN_KEY_BYTE_LEN;
			actx->key_len_flag = 0;
			break;
		case AES_MID_KEY_BYTE_LEN:
			actx->key_len = AES_MID_KEY_BYTE_LEN;
			actx->key_len_flag = 1;
			break;
		case AES_MAX_KEY_BYTE_LEN:
			actx->key_len = AES_MAX_KEY_BYTE_LEN;
			actx->key_len_flag = 2;
			break;
		default:
			return -EINVAL;
		}
	} else {
		if (unlikely(keylen != actx->key_len))
			return -EINVAL;
		actx->key_len_flag = 0;
	}
	if (unlikely(sizeof(actx->key) < keylen))
		return -EINVAL;

	memset(actx->key, 0, keylen);
	memcpy(actx->key, key, keylen);
	return 0;
}

static struct ahash_alg sce_mac_alg[] = {
	{
		.init           = sce_mac_init,
		.update         = sce_mac_update,
		.final          = sce_mac_final,
		.finup          = sce_mac_finup,
		.digest         = sce_mac_digest,
		.export         = sce_mac_export,
		.import         = sce_mac_import,
		.setkey         = sce_mac_setkey,
		.halg = {
			.digestsize = MAC_BLOCK_BYTE_LEN,
			.statesize      = MAC_BUFFER_SIZE+CACHE_LINE_SIZE+24,
			.base   = {
				.cra_name               = "cbcmac(aes)",
				.cra_driver_name        = "cbcmac-aes-phytium",
				.cra_priority           = 300,
				.cra_flags              = CRYPTO_ALG_ASYNC,
				.cra_blocksize          = 1,
				.cra_module     = THIS_MODULE,
				.cra_ctxsize = sizeof(struct mac_alg_context)
					+ CACHE_LINE_SIZE,
				.cra_init = sce_mac_cra_init,
			}
		}
	},
	{
		.init           = sce_mac_init,
		.update         = sce_mac_update,
		.final          = sce_mac_final,
		.finup          = sce_mac_finup,
		.digest         = sce_mac_digest,
		.export         = sce_mac_export,
		.import         = sce_mac_import,
		.setkey         = sce_mac_setkey,
		.halg = {
			.digestsize = MAC_BLOCK_BYTE_LEN,
			.statesize      = MAC_BUFFER_SIZE+CACHE_LINE_SIZE+24,
			.base   = {
				.cra_name               = "cbcmac(sm4)",
				.cra_driver_name        = "cbcmac-sm4-phytium",
				.cra_priority           = 300,
				.cra_flags              = CRYPTO_ALG_ASYNC,
				.cra_blocksize          = 1,
				.cra_module     = THIS_MODULE,
				.cra_ctxsize = sizeof(struct mac_alg_context)
					+ CACHE_LINE_SIZE,
				.cra_init = sce_mac_cra_init,
			}
		}
	},

	{
		.init           = sce_mac_init,
		.update         = sce_mac_update,
		.final          = sce_mac_final,
		.finup          = sce_mac_finup,
		.digest         = sce_mac_digest,
		.export         = sce_mac_export,
		.import         = sce_mac_import,
		.setkey         = sce_mac_setkey,
		.halg = {
			.digestsize = MAC_BLOCK_BYTE_LEN,
			.statesize      = MAC_BUFFER_SIZE+CACHE_LINE_SIZE+24,
			.base   = {
				.cra_name               = "cmac(aes)",
				.cra_driver_name        = "cmac-aes-phytium",
				.cra_priority           = 300,
				.cra_flags              = CRYPTO_ALG_ASYNC,
				.cra_blocksize          = MAC_BLOCK_BYTE_LEN,
				.cra_module     = THIS_MODULE,
				.cra_ctxsize = sizeof(struct mac_alg_context)
					+ CACHE_LINE_SIZE,
				.cra_init = sce_mac_cra_init,
			}
		}
	},
	{
		.init           = sce_mac_init,
		.update         = sce_mac_update,
		.final          = sce_mac_final,
		.finup          = sce_mac_finup,
		.digest         = sce_mac_digest,
		.export         = sce_mac_export,
		.import         = sce_mac_import,
		.setkey         = sce_mac_setkey,
		.halg = {
			.digestsize = MAC_BLOCK_BYTE_LEN,
			.statesize      = MAC_BUFFER_SIZE+CACHE_LINE_SIZE+24,
			.base   = {
				.cra_name               = "cmac(sm4)",
				.cra_driver_name        = "cmac-sm4-phytium",
				.cra_priority           = 300,
				.cra_flags              = CRYPTO_ALG_ASYNC,
				.cra_blocksize          = MAC_BLOCK_BYTE_LEN,
				.cra_module     = THIS_MODULE,
				.cra_ctxsize = sizeof(struct mac_alg_context)
					+ CACHE_LINE_SIZE,
				.cra_init = sce_mac_cra_init,
			}
		}
	},
	{
		.init           = sce_mac_init,
		.update         = sce_mac_update,
		.final          = sce_mac_final,
		.finup          = sce_mac_finup,
		.digest         = sce_mac_digest,
		.export         = sce_mac_export,
		.import         = sce_mac_import,
		.setkey         = sce_mac_setkey,
		.halg = {
			.digestsize = TDES_BLOCK_BYTE_LEN,
			.statesize      = MAC_BUFFER_SIZE+CACHE_LINE_SIZE+24,
			.base   = {
				.cra_name               = "cmac(des3_ede)",
				.cra_driver_name
					= "cmac-des3_ede-phytium",
				.cra_priority           = 300,
				.cra_flags              = CRYPTO_ALG_ASYNC,
				.cra_blocksize          = TDES_BLOCK_BYTE_LEN,
				.cra_module     = THIS_MODULE,
				.cra_ctxsize = sizeof(struct mac_alg_context)
					+ CACHE_LINE_SIZE,
				.cra_init = sce_mac_cra_init,
			}
		}
	},
	{
		.init           = sce_mac_init,
		.update         = sce_mac_update,
		.final          = sce_mac_final,
		.finup          = sce_mac_finup,
		.digest         = sce_mac_digest,
		.export         = sce_mac_export,
		.import         = sce_mac_import,
		.setkey         = sce_mac_setkey,
		.halg = {
			.digestsize = DES_BLOCK_BYTE_LEN,
			.statesize      = MAC_BUFFER_SIZE+CACHE_LINE_SIZE+24,
			.base   = {
				.cra_name               = "cmac(des)",
				.cra_driver_name        = "cmac-des-phytium",
				.cra_priority           = 300,
				.cra_flags              = CRYPTO_ALG_ASYNC,
				.cra_blocksize          = DES_BLOCK_BYTE_LEN,
				.cra_module     = THIS_MODULE,
				.cra_ctxsize = sizeof(struct mac_alg_context)
					+ CACHE_LINE_SIZE,
				.cra_init = sce_mac_cra_init,
			}
		}
	},
	{
		.init           = sce_mac_init,
		.update         = sce_mac_update,
		.final          = sce_mac_final,
		.finup          = sce_mac_finup,
		.digest         = sce_mac_digest,
		.export         = sce_mac_export,
		.import         = sce_mac_import,
		.setkey         = sce_mac_setkey,
		.halg = {
			.digestsize = TDES_BLOCK_BYTE_LEN,
			.statesize      = MAC_BUFFER_SIZE+CACHE_LINE_SIZE+24,
			.base   = {
				.cra_name        = "cbcmac(des3_ede)",
				.cra_driver_name =
					"cbcmac-des3_ede-phytium",
				.cra_priority    = 300,
				.cra_flags       = CRYPTO_ALG_ASYNC,
				.cra_blocksize   = 1,
				.cra_module     = THIS_MODULE,
				.cra_ctxsize = sizeof(struct mac_alg_context)
					+ CACHE_LINE_SIZE,
				.cra_init = sce_mac_cra_init,
			}
		}
	},
	{
		.init           = sce_mac_init,
		.update         = sce_mac_update,
		.final          = sce_mac_final,
		.finup          = sce_mac_finup,
		.digest         = sce_mac_digest,
		.export         = sce_mac_export,
		.import         = sce_mac_import,
		.setkey         = sce_mac_setkey,
		.halg = {
			.digestsize = DES_BLOCK_BYTE_LEN,
			.statesize      = MAC_BUFFER_SIZE+CACHE_LINE_SIZE+24,
			.base   = {
				.cra_name               = "cbcmac(des)",
				.cra_driver_name
					= "cbcmac-des-phytium",
				.cra_priority           = 300,
				.cra_flags              = CRYPTO_ALG_ASYNC,
				.cra_blocksize          = 1,
				.cra_module     = THIS_MODULE,
				.cra_ctxsize = sizeof(struct mac_alg_context)
					+ CACHE_LINE_SIZE,
				.cra_init = sce_mac_cra_init,
			}
		}
	},

};

int sce_mac_algs_register(void)
{
	int ret = 0, i = 0, k = 0;
	u8 alg[64] = { 0 };
	struct crypto_ahash *tfm_sce;

	for (i = 0; i < 8; i++) {
		memset(alg, 0, 64);
		switch (i) {
		case 0:
			memcpy(alg, "cbcmac(aes-generic)",
			strlen("cbcmac(aes-generic)"));
			break;
		case 1:
			memcpy(alg, "cbcmac(sm4-generic)",
			strlen("cbcmac(sm4-generic)"));
			break;
		case 2:
			memcpy(alg, "cmac(aes-generic)",
			strlen("cmac(aes-generic)"));
			break;
		case 3:
			memcpy(alg, "cmac(sm4-generic)",
			strlen("cmac(sm4-generic)"));
			break;
		case 4:
			memcpy(alg, "cmac(des3_ede-generic)",
			strlen("cmac(des3_ede-generic)"));
			break;
		case 5:
			memcpy(alg, "cmac(des-generic)",
			strlen("cmac(des-generic)"));
			break;
		case 6:
			memcpy(alg, "cbcmac(des3_ede-generic)",
			strlen("cbcmac(des3_ede-generic)"));
			break;
		case 7:
			memcpy(alg, "cbcmac(des-generic)",
			strlen("cbcmac(des-generic)"));
			break;
		default:
			sce_err("mac algs num err!\n");
			break;
		}

		tfm_sce = crypto_alloc_ahash(alg, 0, 0);
		if (IS_ERR(tfm_sce)) {
			sce_err("failed to load transform for %s: %ld\n", alg,
				PTR_ERR(tfm_sce));
		} else {
			crypto_free_ahash(tfm_sce);
			register_alg[i] = 1;
		}
	}

	for (i = 0; i < 8; i++) {
		if (register_alg[i] == 1) {
			ret = crypto_register_ahash(&sce_mac_alg[i]);
			if (ret) {
				for (k = 0; k < i; k++) {
					if (register_alg[k] == 1)
						crypto_unregister_ahash(
							&sce_mac_alg[i]);
				}
				sce_err("%s:register cbcmac failed!\n",
					__func__);
				return -1;
			}
		}
	}
	return 0;
}

void sce_mac_algs_unregister(void)
{
	int i = 0;

	for (i = 0; i < 8; i++) {
		if (register_alg[i] == 1)
			crypto_unregister_ahash(&sce_mac_alg[i]);
	}
}
