// SPDX-License-Identifier: GPL-2.0-or-later
/*
 * File Name:phytium_scto.c - Phytium driver for SDK
 *
 * Copyright (C) 2020 Phytium Technology Co.,Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */

#include <linux/module.h>
#include <linux/irq.h>
#include <linux/platform_device.h>
#include <linux/dma-mapping.h>
#include <linux/cdev.h>
#include <linux/types.h>
#include <linux/mutex.h>
#include <linux/acpi.h>
#include <crypto/internal/hash.h>
#include <linux/crypto.h>
#include <crypto/hash.h>
#include <crypto/internal/skcipher.h>
#include <crypto/internal/akcipher.h>
#include <linux/ioport.h>
#include <linux/io.h>
#include "sce.h"
#include "smx_common.h"
#include "phytium_scto.h"
#include "sm2_phytium.h"

struct scto_dev scto;
static struct resource *scto_resource;

int scto_open(struct inode *inode, struct file *filep)
{
	return 0;
}

int scto_release(struct inode *inode, struct file *filep)
{
	return 0;
}

#ifdef SM2_TEST
static long sm2_sign(u8 *priv_key, u8 *e, u8 *sign, u32 *sign_len)
{
	struct crypto_wait cwait;
	struct crypto_akcipher *tfm;
	struct akcipher_request *req;
	struct scatterlist e_sg, sign_sg;
	int ret;

	tfm = crypto_alloc_akcipher("sm2", 0, 0);
	if (IS_ERR(tfm)) {
		pr_err("crypto_alloc_akcipher failed!\n");
		return -1;
	}

	ret = -ENOMEM;
	req = akcipher_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_err("akcipher_request_alloc failed!\n");
		goto error_free_tfm;
	}

	ret = crypto_akcipher_set_priv_key(tfm, priv_key, 32);
	if (ret) {
		pr_err("crypto_akcipher_set_priv_key failed!\n");
		goto error_free_req;
	}

	sg_init_one(&e_sg, e, 32);
	sg_init_one(&sign_sg, sign, *sign_len);
	akcipher_request_set_crypt(req, &e_sg, &sign_sg, 32, *sign_len);
	crypto_init_wait(&cwait);
	akcipher_request_set_callback(
		req, CRYPTO_TFM_REQ_MAY_BACKLOG | CRYPTO_TFM_REQ_MAY_SLEEP,
		crypto_req_done, &cwait);

	ret = crypto_wait_req(crypto_akcipher_sign(req), &cwait);
	if (ret)
		goto error_free_req;

	*sign_len = req->dst_len;

error_free_req:
	akcipher_request_free(req);
error_free_tfm:
	crypto_free_akcipher(tfm);

	return ret;
}

static long sm2_verify(u8 *pub_key, u8 *sign_e, u32 len)
{
	struct crypto_wait cwait;
	struct crypto_akcipher *tfm;
	struct akcipher_request *req;
	struct scatterlist sign_e_sg, null_sg;
	u8 null[64];
	int ret;

	tfm = crypto_alloc_akcipher("sm2", 0, 0);
	if (IS_ERR(tfm)) {
		pr_err("crypto_alloc_akcipher failed!\n");
		return -1;
	}

	ret = -ENOMEM;
	req = akcipher_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_err("akcipher_request_alloc failed!\n");
		goto error_free_tfm;
	}

	ret = crypto_akcipher_set_pub_key(tfm, pub_key, 65);
	if (ret) {
		pr_err("crypto_akcipher_set_pub_key failed!\n");
		goto error_free_req;
	}

	sg_init_one(&sign_e_sg, sign_e, len);
	sg_init_one(&null_sg, null, 32);
	akcipher_request_set_crypt(req, &sign_e_sg, &null_sg, len - 32, 32);
	crypto_init_wait(&cwait);
	akcipher_request_set_callback(
		req, CRYPTO_TFM_REQ_MAY_BACKLOG | CRYPTO_TFM_REQ_MAY_SLEEP,
		crypto_req_done, &cwait);

	ret = crypto_wait_req(crypto_akcipher_verify(req), &cwait);
	if (ret)
		goto error_free_req;

error_free_req:
	akcipher_request_free(req);
error_free_tfm:
	crypto_free_akcipher(tfm);

	return ret;
}

static long sm2_enc(u8 *pub_key, u8 *src, u8 *dst, u32 src_len, u32 *dst_len)
{
	struct crypto_wait cwait;
	struct crypto_akcipher *tfm;
	struct akcipher_request *req;
	struct scatterlist src_sg, dst_sg;
	int ret;

	tfm = crypto_alloc_akcipher("sm2", 0, 0);
	if (IS_ERR(tfm)) {
		pr_err("crypto_alloc_akcipher failed!\n");
		return -1;
	}

	ret = -ENOMEM;
	req = akcipher_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_err("akcipher_request_alloc failed!\n");
		goto error_free_tfm;
	}

	ret = crypto_akcipher_set_pub_key(tfm, pub_key, 65);
	if (ret) {
		pr_err("crypto_akcipher_set_pub_key failed!\n");
		goto error_free_req;
	}

	sg_init_one(&src_sg, src, src_len);
	sg_init_one(&dst_sg, dst, *dst_len);
	akcipher_request_set_crypt(req, &src_sg, &dst_sg, src_len, *dst_len);
	crypto_init_wait(&cwait);
	akcipher_request_set_callback(
		req, CRYPTO_TFM_REQ_MAY_BACKLOG | CRYPTO_TFM_REQ_MAY_SLEEP,
		crypto_req_done, &cwait);

	ret = crypto_wait_req(crypto_akcipher_encrypt(req), &cwait);
	if (ret)
		goto error_free_req;

	*dst_len = req->dst_len;

error_free_req:
	akcipher_request_free(req);
error_free_tfm:
	crypto_free_akcipher(tfm);

	return ret;
}

static long sm2_dec(u8 *priv_key, u8 *src, u8 *dst, u32 src_len, u32 *dst_len)
{
	struct crypto_wait cwait;
	struct crypto_akcipher *tfm;
	struct akcipher_request *req;
	struct scatterlist src_sg, dst_sg;
	int ret;

	tfm = crypto_alloc_akcipher("sm2", 0, 0);
	if (IS_ERR(tfm)) {
		pr_err("crypto_alloc_akcipher failed!\n");
		return -1;
	}

	ret = -ENOMEM;
	req = akcipher_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_err("akcipher_request_alloc failed!\n");
		goto error_free_tfm;
	}

	ret = crypto_akcipher_set_priv_key(tfm, priv_key, 32);
	if (ret) {
		pr_err("crypto_akcipher_set_priv_key failed!\n");
		goto error_free_req;
	}

	sg_init_one(&src_sg, src, src_len);
	sg_init_one(&dst_sg, dst, *dst_len);
	akcipher_request_set_crypt(req, &src_sg, &dst_sg, src_len, *dst_len);
	crypto_init_wait(&cwait);
	akcipher_request_set_callback(
		req, CRYPTO_TFM_REQ_MAY_BACKLOG | CRYPTO_TFM_REQ_MAY_SLEEP,
		crypto_req_done, &cwait);

	ret = crypto_wait_req(crypto_akcipher_decrypt(req), &cwait);
	if (ret)
		goto error_free_req;

	*dst_len = req->dst_len;

error_free_req:
	akcipher_request_free(req);
error_free_tfm:
	crypto_free_akcipher(tfm);

	return ret;
}

struct sm2_info {
	u32 type;
	u32 src_len;
	u32 dst_len;
	u8 *src;
	u8 *dst;
	u8 *key;
};
long scto_sm2(unsigned int cmd, unsigned long arg)
{
	struct sm2_info info;
	long ret;
	u8 *src, *dst, *key;

	if (unlikely(cmd != SCTO_SM2)) {
		pr_err("cmd[0x%x] not support!\n", cmd);
		return -1;
	}

	ret = copy_from_user(&info, (void *)arg, sizeof(info));
	if (ret) {
		pr_err("copy_from_user[arg:0x%lx] failed!\n", arg);
		return ret;
	}

	switch (info.type) {
	case ALG_SM2_SIGN:
		if (info.src_len != 32) {
			pr_err("ALG_SM2_SIGN srclen[0x%x] err!\n",
			       info.src_len);
			return -1;
		}
		src = kmalloc(info.src_len + info.dst_len + 256, GFP_KERNEL);
		if (src == NULL)
			return -ENOMEM;


		ret = copy_from_user(src, info.src, info.src_len);
		if (ret) {
			pr_err("copy_from_user[src:%p] failed!\n",
			       info.src);
			kfree(src);
			return ret;
		}
		dst = src + info.src_len + 64;
		key = dst + info.dst_len + 64;
		ret = copy_from_user(key, info.key, 32);
		if (ret) {
			pr_err("copy_from_user[key:%p] failed!\n",
			       info.key);
			kfree(src);
			return ret;
		}

		ret = sm2_sign(key, src, dst, &info.dst_len);
		if (ret) {
			pr_err("sm2_sign[ret:0x%lx] failed!\n", ret);
			kfree(src);
			return ret;
		}

		ret = copy_to_user(info.dst, dst, info.dst_len);
		if (ret) {
			pr_err("copy_to_user[dst:%p] failed!\n",
			       info.dst);
			kfree(src);
			return ret;
		}
		ret = copy_to_user((void *)arg, &info, sizeof(info));
		if (ret) {
			pr_err("copy_to_user[arg:0x%lx] failed!\n",
			       arg);
			kfree(src);
			return ret;
		}
		kfree(src);
		return ret;
	case ALG_SM2_VERIFY:
		src = kmalloc(info.src_len + 256, GFP_KERNEL);
		if (src == NULL)
			return -ENOMEM;

		ret = copy_from_user(src, info.src, info.src_len);
		if (ret) {
			pr_err("copy_from_user[src:%p] failed!\n",
			       info.src);
			kfree(src);
			return ret;
		}
		key = src + info.src_len + 64;
		ret = copy_from_user(key, info.key, 65);
		if (ret) {
			pr_err("copy_from_user[key:%p] failed!\n",
			       info.key);
			kfree(src);
			return ret;
		}

		ret = sm2_verify(key, src, info.src_len);
		if (ret) {
			pr_err("sm2_verify[ret:0x%lx] failed!\n", ret);
			kfree(src);
			return ret;
		}

		kfree(src);
		return ret;
	case ALG_SM2_ENC:
		src = kmalloc(info.src_len + info.dst_len + 256, GFP_KERNEL);
		if (src == NULL)
			return -ENOMEM;

		ret = copy_from_user(src, info.src, info.src_len);
		if (ret) {
			pr_err("copy_from_user[src:%p] failed!\n",
			       info.src);
			kfree(src);
			return ret;
		}
		dst = src + info.src_len + 64;
		key = dst + info.dst_len + 64;
		ret = copy_from_user(key, info.key, 65);
		if (ret) {
			pr_err("copy_from_user[key:%p] failed!\n",
			       info.key);
			kfree(src);
			return ret;
		}

		ret = sm2_enc(key, src, dst, info.src_len, &info.dst_len);
		if (ret) {
			pr_err("sm2_enc[ret:0x%lx] failed!\n", ret);
			kfree(src);
			return ret;
		}

		ret = copy_to_user(info.dst, dst, info.dst_len);
		if (ret) {
			pr_err("copy_to_user[dst:%p] failed!\n",
			       info.dst);
			kfree(src);
			return ret;
		}
		ret = copy_to_user((void *)arg, &info, sizeof(info));
		if (ret) {
			pr_err("copy_to_user[arg:0x%lx] failed!\n",
			       arg);
			kfree(src);
			return ret;
		}

		kfree(src);
		return ret;
	case ALG_SM2_DEC:
		src = kmalloc(info.src_len + info.dst_len + 256, GFP_KERNEL);
		if (src == NULL)
			return -ENOMEM;

		ret = copy_from_user(src, info.src, info.src_len);
		if (ret) {
			pr_err("copy_from_user[src:%p] failed!\n",
			       info.src);
			kfree(src);
			return ret;
		}
		dst = src + info.src_len + 64;
		key = dst + info.dst_len + 64;
		ret = copy_from_user(key, info.key, 32);
		if (ret) {
			pr_err("copy_from_user[key:%p] failed!\n",
			       info.key);
			kfree(src);
			return ret;
		}

		ret = sm2_dec(key, src, dst, info.src_len, &info.dst_len);
		if (ret) {
			pr_err("sm2_dec[ret:0x%lx] failed!\n", ret);
			kfree(src);
			return ret;
		}

		ret = copy_to_user(info.dst, dst, info.dst_len);
		if (ret) {
			pr_err("copy_to_user[dst:%p] failed!\n",
			       info.dst);
			kfree(src);
			return ret;
		}
		ret = copy_to_user((void *)arg, &info, sizeof(info));
		if (ret) {
			pr_err("copy_to_user[arg:0x%lx] failed!\n",
			       arg);
			kfree(src);
			return ret;
		}

		kfree(src);
		return ret;
	default:
		return -1;
	}
}

#endif
long scto_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
	if (unlikely(cmd != SCTO_SM2))
		return -1;

#ifdef SM2_TEST
	return scto_sm2(cmd, arg);
#else
	return -1;
#endif
}

static const struct file_operations scto_fops = {
	.owner = THIS_MODULE,
	.open = scto_open,
	.release = scto_release,
	.unlocked_ioctl = scto_ioctl,

};

static void phytium_algs_unregister(void)
{
	sm2_phytium_algs_unregister();
}

static int phytium_algs_register(void)
{
	int ret;

	ret = sm2_phytium_algs_register();
	if (ret)
		return -1;

	return 0;
}

static struct cdev scto_pcdev;
static struct class *scto_pclass;
static dev_t scto_dev;

static int scto_cdev_register(void)
{
	int ret;

	ret = alloc_chrdev_region(&scto_dev, 0, 1, "dsce_pke");
	if (ret < 0) {
		pr_err("alloc_chrdev_region fail\n");
		return -1;
	}

	cdev_init(&scto_pcdev, &scto_fops);

	ret = cdev_add(&scto_pcdev, scto_dev, 1);
	if (ret) {
		pr_err("cdev_add error!\n");
		goto err_0;
	}

	scto_pclass = class_create(THIS_MODULE, "dsce_pke");
	if (IS_ERR(scto_pclass)) {
		pr_err("class_create fail!\n");
		goto err_1;
	}

	if (IS_ERR(device_create(scto_pclass, NULL, scto_dev, NULL, "dsce_pke"))) {
		pr_err("device_create fail!\n\n");
		goto err_2;
	}

	return 0;
err_2:
	class_destroy(scto_pclass);
err_1:
	cdev_del(&scto_pcdev);
err_0:
	unregister_chrdev_region(scto_dev, 1);

	return -1;
}

static void scto_cdev_unregister(void)
{
	device_destroy(scto_pclass, scto_dev);
	class_destroy(scto_pclass);
	cdev_del(&scto_pcdev);
	unregister_chrdev_region(scto_dev, 1);
}

static int scto_probe(struct platform_device *pdev)
{
	struct device *dev = &pdev->dev;
	int err = 0;
	int i;

	acpi_dma_configure(dev, DEV_DMA_COHERENT);
	dma_set_mask_and_coherent(dev, DMA_BIT_MASK(48));

	scto_resource = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	scto.regs = devm_ioremap_resource(&pdev->dev, scto_resource);
	if (IS_ERR(scto.regs))
		return PTR_ERR(scto.regs);

	scto.pke_reg[0] = (struct pke_reg *)(scto.regs + 0x2000);
	scto.pke_reg[1] = (struct pke_reg *)(scto.regs + 0x6000);
	scto.pke_reg[2] = (struct pke_reg *)(scto.regs + 0xa000);
	scto.pke_reg[3] = (struct pke_reg *)(scto.regs + 0xe000);
	scto.pke_reg[0]->cfg = 0;
	scto.pke_reg[1]->cfg = 0;
	scto.pke_reg[2]->cfg = 0;
	scto.pke_reg[3]->cfg = 0;
	scto.trng_reg = sce->base + 0x6000;
	scto.trng_reg->cr = 0x10100F0;
	scto.trng_reg->sr = 7;
	scto.trng_reg->rtcr = 0;
	scto.trng_reg->cr = 0x10100F1;

	mutex_init(&scto.pke_lock);

	for (i = 0; i < NR_CPUS; i++) {
		spin_lock_init(&scto.sm2_queue[i].queue_lock);
		sqlist_init(&scto.sm2_queue[i].list);
	}

#if SCTO_KERNEL_MODE
	err = phytium_algs_register();
	if (err) {
		pr_err("phytium_algs_register fail!\n");
		return -1;
	}
#endif
	err = scto_cdev_register();
	if (err) {
		pr_err("scto_cdev_register fail!\n");
		phytium_algs_unregister();
		return -1;
	}

	platform_set_drvdata(pdev, &scto);

	return 0;
}

static int scto_remove(struct platform_device *pdev)
{
#if SCTO_KERNEL_MODE
	phytium_algs_unregister();
#endif
	scto_cdev_unregister();

	platform_set_drvdata(pdev, NULL);

	return 0;
}

static const struct of_device_id scto_dma_of_match[] = {
	{ .compatible = "PKE" },
	{}
};
MODULE_DEVICE_TABLE(of, scto_dma_of_match);

#ifdef CONFIG_ACPI
static const struct acpi_device_id phytium_scto_acpi_ids[] = {
	{ .id = "PHYT0052" },
	{}
};

MODULE_DEVICE_TABLE(acpi, phytium_scto_acpi_ids);
#else
#define phytium_scto_acpi_ids NULL
#endif

static int scto_suspend(struct device *dev)
{
	pr_info("ENTER %s...\n", __func__);
	mutex_lock(&scto.pke_lock);

	return 0;
}

static int scto_resume(struct device *dev)
{
	pr_info("ENTER %s...\n", __func__);
	scto.pke_reg[0]->cfg = 0;
	scto.pke_reg[1]->cfg = 0;
	scto.pke_reg[2]->cfg = 0;
	scto.pke_reg[3]->cfg = 0;
	scto.trng_reg->cr = 0x10100F0;
	scto.trng_reg->sr = 7;
	scto.trng_reg->rtcr = 0;
	scto.trng_reg->cr = 0x10100F1;
	mutex_unlock(&scto.pke_lock);

	return 0;
}

static const struct dev_pm_ops scto_pm_ops = {
	.suspend = scto_suspend,
	.resume = scto_resume,
};

static struct platform_driver
	scto_driver = { .probe = scto_probe,
			.remove = scto_remove,
			.driver = {
				.name = "PHYTSCTO",
				.of_match_table = scto_dma_of_match,
				.acpi_match_table = phytium_scto_acpi_ids,
				.pm = &scto_pm_ops,
			} };

int scto_device_probe(void)
{
	int ret;

	ret = platform_driver_register(&scto_driver);
	return ret;
}

void scto_device_remove(void)
{
	platform_driver_unregister(&scto_driver);
}
