// SPDX-License-Identifier: GPL-2.0-or-later
#include <linux/module.h>
#include <linux/irq.h>
#include <linux/dma-mapping.h>
#include <linux/mutex.h>
#include <linux/acpi.h>
#include <linux/delay.h>
#include <linux/kthread.h>
#include "sce_driver_common.h"
#include "sce_driver_resource.h"
#include "sce_driver.h"

static struct cdev sce_cdev;
static struct class *sce_class;
static dev_t sce_dev;

int sce_open(struct inode *inode, struct file *filep)
{
	return 0;
}

int sce_release(struct inode *inode, struct file *filep)
{
	return 0;
}

long sce_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
	union sce_ioctl_info info;
	long ret;
	struct buf_desc *buf_desc;

	switch (cmd) {
	case SCE_QUEUE_ALLOC:
		ret = copy_from_user(&info, (void *)arg, sizeof(info));
		if (ret) {
			sce_err("%s:copy_from_user[arg:0x%lx] failed!\n",
				__func__, arg);
			return ret;
		}

		ret = sce_request_queue(info.queue_type, &info.tag);
		if (ret)
			return ret;

		ret = copy_to_user((void *)arg, &info, sizeof(info));
		if (ret) {
			sce_err("%s:copy_to_user[arg:0x%lx] failed!\n",
				__func__, arg);
			return ret;
		}
		return 0;
	case SCE_QUEUE_FREE:
		ret = copy_from_user(&info, (void *)arg, sizeof(info));
		if (ret) {
			sce_err("%s:copy_from_user[arg:0x%lx] failed!\n",
				__func__, arg);
			return ret;
		}

		ret = sce_release_queue(info.tag);
		if (ret)
			return ret;
		return 0;
	case SCE_MEM_ALLOC:
		ret = copy_from_user(&info, (void *)arg, sizeof(info));
		if (ret) {
			sce_err("%s:copy_from_user[arg:0x%lx] failed!\n",
				__func__, arg);
			return ret;
		}

		ret = sce_alloc_phys_mem(info.insize, &info.tag);
		if (ret)
			return ret;

		buf_desc = phys_to_virt(info.tag);
		info.paddr = virt_to_phys(buf_desc->vaddr);
		info.ioaddr = buf_desc->ioaddr;
		info.outsize = buf_desc->size;

		ret = copy_to_user((void *)arg, &info, sizeof(info));
		if (ret) {
			sce_err("%s:copy_to_user[arg:0x%lx] failed!\n",
				__func__, arg);
			return ret;
		}

		return 0;
	case SCE_MEM_FREE:
		ret = copy_from_user(&info, (void *)arg, sizeof(info));
		if (ret) {
			sce_err("%s:copy_from_user[arg:0x%lx] failed!\n",
				__func__, arg);
			return ret;
		}

		ret = sce_free_phys_mem(info.tag);
		if (ret)
			return ret;

		return 0;
	default:
		sce_err("%s:cmd[0x%x] err!\n", __func__, cmd);
		return -1;
	}

	return 0;
}

int sce_mmap(struct file *filep, struct vm_area_struct *vma)
{
	struct buf_desc *buf_desc;
	u64 tag, pfn;
	int ret = -1;

	tag = vma->vm_pgoff;

	buf_desc = phys_to_virt(tag);
	if (!virt_addr_valid(buf_desc)) {
		sce_err("%s:tag[0x%llx] err!\n", __func__, tag);
		return -1;
	}

	if (buf_desc->magic != SCE_MAGIC_NUMBER) {
		sce_err("%s:tag[0x%llx] invalid!\n", __func__, tag);
		return -1;
	}

	sce_debug("magic:0x%llx, vmstart:0x%lx, vmend:0x%lx\n", buf_desc->magic,
		  vma->vm_start, vma->vm_end);

	sce_lock_mutex(&sce_driver.lock);
	if (buf_desc->magic == SCE_MAGIC_NUMBER) {
		if (buf_desc->isio) {
			pfn = buf_desc->ioaddr >> PAGE_SHIFT;
			vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
		} else {
			pfn = virt_to_phys(buf_desc->vaddr) >> PAGE_SHIFT;
		}

		vma->vm_flags |= VM_LOCKED | VM_DONTCOPY | VM_DONTEXPAND |
				 VM_DONTDUMP;

		sce_debug("%s:pfn[0x%llx]\n", __func__, pfn);
		ret = remap_pfn_range(vma, vma->vm_start, pfn,
				      vma->vm_end - vma->vm_start,
				      vma->vm_page_prot);
	}
	sce_unlock_mutex(&sce_driver.lock);

	return ret;
}

static const struct file_operations sce_fops = {
	.owner = THIS_MODULE,
	.open = sce_open,
	.release = sce_release,
	.mmap = sce_mmap,
	.unlocked_ioctl = sce_ioctl,

};

int sce_register_cdev(void)
{
	int ret;

	ret = alloc_chrdev_region(&sce_dev, 0, 1, "dsce");
	if (ret < 0) {
		sce_err("alloc_chrdev_region fail\n");
		return -1;
	}

	cdev_init(&sce_cdev, &sce_fops);

	ret = cdev_add(&sce_cdev, sce_dev, 1);
	if (ret) {
		sce_err("cdev_add error!\n");
		goto err_0;
	}

	sce_class = class_create(THIS_MODULE, "dsce");
	if (IS_ERR(sce_class)) {
		sce_err("class_create fail!\n");
		goto err_1;
	}

	if (IS_ERR(device_create(sce_class, NULL, sce_dev, NULL, "dsce"))) {
		sce_err("device_create fail!\n\n");
		goto err_2;
	}

	return 0;
err_2:
	class_destroy(sce_class);
err_1:
	cdev_del(&sce_cdev);
err_0:
	unregister_chrdev_region(sce_dev, 1);

	return -1;
}

void sce_unregister_cdev(void)
{
	device_destroy(sce_class, sce_dev);
	class_destroy(sce_class);
	cdev_del(&sce_cdev);
	unregister_chrdev_region(sce_dev, 1);
}
