// SPDX-License-Identifier: GPL-2.0-or-later
/*
 * File Name:sce_driver.c - Phytium driver for SDK
 *
 * Copyright (C) 2022 Phytium Technology Co.,Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */

#include <linux/module.h>
#include <linux/irq.h>
#include <linux/dma-mapping.h>
#include <linux/mutex.h>
#include <linux/acpi.h>
#include <linux/delay.h>
#include <linux/kthread.h>
#include <linux/io.h>
#include <linux/version.h>
#include "sce.h"
#include "sce_driver_reg.h"
#include "sce_driver_common.h"
#include "sce_driver_resource.h"
#include "sce_driver_char.h"
#include "sce_driver_thread.h"
#include "sce_driver.h"

//#define SCE_ALG_TYPE_NUM 3

static unsigned int cpu = NR_CPUS;
module_param(cpu, uint, 0);

MODULE_PARM_DESC(cpu, "Specify cpu to bind by sce kthread.");

struct sce_msg sce_driver;
struct sce_queue_desc queue_desc[SCE_ALG_TYPE_NUM] = { 0 };

int sce_driver_register(void)
{
	int i;

	sce_driver.base = sce->base;
	if (IS_ERR(sce_driver.base)) {
		sce_err("ioremap_nocache fail!\n");
		return -1;
	}

	sce_driver.hash_cmd_space = 8;
	sce_driver.hash_rsp_space = 4;
	sce_driver.sca_cmd_space = 8;
	sce_driver.sca_rsp_space = 4;
	sce_init_mutex(&sce_driver.lock);
	sqlist_init(&sce_driver.list);

	for (i = 0; i < SCE_ALG_TYPE_NUM; i++) {
		queue_desc[i].reg = (struct sce_queue_reg *)sce_get_pages(
			NUMA_NO_NODE, 0x100000);
		if (queue_desc[i].reg == NULL) {
			sce_err("%s:sce_get_pages out of memory!\n", __func__);
			goto out;
		}
	}

	if (sce_register_cdev())
		goto out;

	if (sce_add_thread(cpu)) {
		sce_unregister_cdev();
		goto out;
	}

	return 0;

out:
	for (i = 0; i < SCE_ALG_TYPE_NUM; i++) {
		if (queue_desc[i].reg != NULL)
			sce_free_pages(queue_desc[i].reg, 0x100000);
	}

	return -1;
}

void sce_driver_unregister(void)
{
	struct sqlist *list;
	struct buf_desc *buf_desc;
	int i;

	sce_del_thread();

	for (i = 0; i < SCE_ALG_TYPE_NUM; i++) {
		if (queue_desc[i].reg != NULL)
			sce_free_pages(queue_desc[i].reg, 0x100000);
	}

	sce_unregister_cdev();
	do {
		list = sqlist_get_head(&sce_driver.list);
		if (list) {
			buf_desc = container_of(list, struct buf_desc, node);
			if (buf_desc->magic == SCE_MAGIC_NUMBER) {
				buf_desc->magic = 0;
				sce_free_pages(buf_desc->vaddr, buf_desc->size);
				sce_free(buf_desc);
			}
		}
	} while (list);
}
