// SPDX-License-Identifier: GPL-2.0-or-later
#include <linux/dma-mapping.h>
#include <linux/mutex.h>
#include <linux/acpi.h>
#include <linux/delay.h>
#include <linux/kthread.h>
#include "sce_driver_common.h"
#include "sce_driver_resource.h"
#include "sce_driver.h"

int sce_request_queue(int alg_type, u64 *sce_queue_desc)
{
	int queue = 0, mask, i;
	struct buf_desc *buf_desc;

	if (sce_queue_desc == NULL) {
		sce_err("Invalid index sce_queue_desc");
		return -EINVAL;
	}
	// TRNG
	if (alg_type >= SCE_ALG_TYPE_NUM) {
		sce_err("%s:alg_type[%d] err!MAX alg_type[1]\n", __func__,
			alg_type);
		return -1;
	}

	buf_desc = sce_malloc_node(sizeof(*buf_desc), NUMA_NO_NODE);
	if (buf_desc == NULL) {
		sce_err("%s:sce_malloc_node out of memory!\n", __func__);
		return -1;
	}

	sce_lock_mutex(&sce_driver.lock);
	mask = sce_driver.queue_mask[alg_type];
	//if no queue free
	if (mask == 0xF) {
		sce_warning("%s:alg_type[%d] out of queue!\n", __func__,
			    alg_type);
		sce_free(buf_desc);
		sce_unlock_mutex(&sce_driver.lock);
		return -1;
	}

	for (i = 0; i < 4; i++, queue++) {
		if (queue >= 4)
			queue = 0;
		if ((mask & (1 << queue)) == 0)
			break;
	}

	sce_driver.queue_mask[alg_type] |= 1 << queue;
	sce_unlock_mutex(&sce_driver.lock);

	buf_desc->vaddr =
		(void *)((u64)queue_desc[alg_type].reg + queue * 0x40000);
	buf_desc->magic = SCE_MAGIC_NUMBER;
	buf_desc->size = 0x40000;
	buf_desc->isio = 0;
	buf_desc->ioaddr = virt_to_phys(buf_desc->vaddr);
	buf_desc->type = alg_type;

	*sce_queue_desc = virt_to_phys(buf_desc);

	return 0;
}

int sce_release_queue(u64 sce_queue_desc)
{
	int queue = -1;
	struct buf_desc *buf_desc;
	int type;

	buf_desc = phys_to_virt(sce_queue_desc);
	if (!virt_addr_valid(buf_desc)) {
		sce_err("%s:sce_queue_desc[0x%llx] err!\n", __func__,
			sce_queue_desc);
		return -1;
	}

	sce_lock_mutex(&sce_driver.lock);

	if (buf_desc->magic != SCE_MAGIC_NUMBER) {
		sce_err("%s:sce_queue_desc[0x%llx] invalid!\n", __func__,
			sce_queue_desc);
		sce_unlock_mutex(&sce_driver.lock);
		return -1;
	}

	type = buf_desc->type;

	if (type >= SCE_ALG_TYPE_NUM) {
		sce_err("%s:type[%d] err!MAX type[1]\n", __func__, type);
		sce_unlock_mutex(&sce_driver.lock);
		return -1;
	}

	if ((u64)buf_desc->vaddr >= (u64)queue_desc[type].reg)
		queue = ((u64)buf_desc->vaddr - (u64)queue_desc[type].reg) /
			0x40000;
	if ((queue < 8) && (queue >= 0))
		sce_driver.queue_mask[type] &= ~(1 << queue);

	buf_desc->magic = 0;
	sce_free(buf_desc);

	sce_unlock_mutex(&sce_driver.lock);

	return 0;
}

int sce_alloc_phys_mem(u64 size, u64 *sce_mem_desc)
{
	struct buf_desc *buf_desc;

	if (sce_mem_desc == NULL) {
		sce_err("Invalid index sce_mem_desc");
		return -EINVAL;
	}

	if (size & (PAGE_SIZE - 1)) {
		size -= size & (PAGE_SIZE - 1);
		size += PAGE_SIZE;
	}

	buf_desc = sce_malloc_node(sizeof(struct buf_desc), NUMA_NO_NODE);
	if (buf_desc == NULL) {
		sce_err("%s:sce_malloc_node out of memory!\n", __func__);
		return -1;
	}

	buf_desc->vaddr = sce_get_pages(NUMA_NO_NODE, size);
	if (buf_desc->vaddr == NULL) {
		sce_err("%s:sce_get_pages out of memory!\n", __func__);
		sce_free(buf_desc);
		return -1;
	}

	buf_desc->magic = SCE_MAGIC_NUMBER;
	buf_desc->ioaddr = virt_to_phys(buf_desc->vaddr);
	buf_desc->size = size;
	buf_desc->isio = 0;

	sce_lock_mutex(&sce_driver.lock);
	sqlist_insert_tail(&sce_driver.list, &buf_desc->node);
	sce_unlock_mutex(&sce_driver.lock);

	*sce_mem_desc = virt_to_phys(buf_desc);

	return 0;
}

int sce_free_phys_mem(u64 sce_mem_desc)
{
	struct buf_desc *buf_desc;

	buf_desc = phys_to_virt(sce_mem_desc);
	if (!virt_addr_valid(buf_desc)) {
		sce_err("%s:sce_mem_desc[0x%llx] err!\n", __func__,
			sce_mem_desc);
		return -1;
	}

	if (buf_desc->magic != SCE_MAGIC_NUMBER) {
		sce_err("%s:sce_mem_desc[0x%llx] invalid!\n", __func__,
			sce_mem_desc);
		return -1;
	}

	sce_lock_mutex(&sce_driver.lock);
	if (buf_desc->magic == SCE_MAGIC_NUMBER) {
		buf_desc->magic = 0;
		sqlist_remove(&buf_desc->node);
		sce_free_pages(buf_desc->vaddr, buf_desc->size);
		sce_free(buf_desc);
	}
	sce_unlock_mutex(&sce_driver.lock);

	return 0;
}
