// SPDX-License-Identifier: GPL-2.0-or-later
#include <linux/delay.h>
#include <crypto/hash.h>
#include <linux/crypto.h>
#include <linux/delay.h>
#include <linux/kthread.h>
#include <crypto/internal/rng.h>
#include <linux/version.h>

#include "sce_kernel_common.h"
#include "sce_kernel_reg.h"
#include "sce_kernel.h"
#include "trng.h"

#define TRNG_POOL_SIZE 32
#define TRNG_SM4_SIZE 32

struct trng_context {
	struct sce_mutex *lock;
};

static int trng_init(struct crypto_tfm *tfm)
{
	return 0;
}

static int trng_sm4_init(struct crypto_tfm *tfm)
{
	return 0;
}

static void trng_cleanup(struct crypto_tfm *tfm)
{
}

static int trng_reset(struct crypto_rng *tfm, const u8 *seed, unsigned int slen)
{
	return 0;
}

static inline int trng_read(u8 *buf, size_t len)
{
	size_t i = 0, step;
	u32 val;

	for (i = 0; i < len; i += step) {
		step = (len - i) > 4 ? 4 : (len - i);
		val = REG_RD_32_SYNC(sce_kernel.base,
				     RN_POOL_DATA_WORD_OFFSET(i >> 2));
		memcpy((buf + i), &val, step);
	}

	return i;
}

static int trng_sm4_read(void *buf, size_t len)
{
	size_t i = 0, step;
	u32 val;

	for (i = 0; i < len; i += step) {
		step = (len - i) > 4 ? 4 : (len - i);
		val = REG_RD_32_SYNC(sce_kernel.base, TRNG_DR_OFFSET);
		memcpy(buf + i, &val, step);
	}

	return i;
}

static int trng_sm4_generate(struct crypto_rng *tfm, const u8 *src,
			     unsigned int slen, u8 *rdata, unsigned int dlen)
{
	size_t offset = 0;
	size_t step_len = 0;
	union trng_sr_reg initr = { 0 };
	int count;
	unsigned long flags;

	if (!rdata) {
		pr_err("Wrong rdata value.");
		return -1;
	}

	spin_lock_irqsave(&sce_kernel.trng_sm4_lock, flags);
	while (offset < dlen) {
		step_len = (dlen - offset) > TRNG_SM4_SIZE ? TRNG_SM4_SIZE :
							     dlen - offset;

		count = 0;
		//init
		do {
			initr.data = REG_RD_32(sce_kernel.base, TRNG_SR_OFFSET);
			if (initr.drdy) {
				trng_sm4_read(rdata + offset, step_len);
				offset += step_len;
				REG_WR_32_SYNC(sce_kernel.base, 0x6008, 0x7);
				break;
			}

			udelay(WAIT_POOL_FILLED_US);
			count++;
			if (count > WAIT_POOL_CYCLE_INDEX) {
				spin_unlock_irqrestore(
					&sce_kernel.trng_sm4_lock, flags);
				pr_err("Wait timeout for trng-sm4");
				return -1;
			}

		} while (1);
	}
	spin_unlock_irqrestore(&sce_kernel.trng_sm4_lock, flags);

	return 0;
}

static int trng_generate(struct crypto_rng *tfm, const u8 *src,
			 unsigned int slen, u8 *rdata, unsigned int dlen)
{
	size_t offset = 0;
	size_t step_len = 0;
	union rnpool_ctrl_reg ctrl = { 0 };
	union rnpool_intr_reg intr = { 0 };
	int count;
	unsigned long flags;

	if (!rdata) {
		pr_err("Wrong rdata value.");
		return -1;
	}

	spin_lock_irqsave(&sce_kernel.trng_lock, flags);
	while (offset < dlen) {
		step_len = (dlen - offset) > TRNG_POOL_SIZE ? TRNG_POOL_SIZE :
							      dlen - offset;

		// set_ctl
		ctrl.trng_clock_enable = 1;
		ctrl.rn_pool_fill_req = 1;
		REG_WR_32_SYNC(sce_kernel.base, RN_POOL_CTL_OFFSET, ctrl.data);
		count = 0;
		do {
			// get int state
			intr.data = REG_RD_32_SYNC(sce_kernel.base,
						   RN_POOL_INTR_OFFSET);
			if (intr.intr_fill_done) {
				trng_read(rdata + offset, step_len);
				offset += step_len;

				// eoi
				REG_WR_32_SYNC(sce_kernel.base,
					       RN_POOL_INTR_OFFSET, intr.data);
				break;
			}
			udelay(WAIT_POOL_FILLED_US);
			count++;
			if (count > WAIT_POOL_CYCLE_INDEX) {
				spin_unlock_irqrestore(&sce_kernel.trng_lock,
					flags);
				pr_err("Wait timeout for trng");
				return -1;
			}
		} while (1);
	}
	spin_unlock_irqrestore(&sce_kernel.trng_lock, flags);

	return 0;
}

static struct rng_alg sce_trng_alg = { .generate = trng_generate,
				       .seed = trng_reset,
				       .seedsize = 0,
				       .base = {
					       .cra_name = "trng",
					       .cra_driver_name =
						       "trng-phytium",
					       .cra_priority = 300,
					       .cra_ctxsize = sizeof(
						       struct trng_context),
					       .cra_module = THIS_MODULE,
					       .cra_init = trng_init,
					       .cra_exit = trng_cleanup,
				       } };

static struct rng_alg sce_trng_sm4_alg = { .generate = trng_sm4_generate,
					   .seed = trng_reset,
					   .seedsize = 0,
					   .base = {
						   .cra_name = "trng-sm4",
						   .cra_driver_name =
							   "trng-sm4-phytium",
						   .cra_priority = 300,
						   .cra_ctxsize = sizeof(
							   struct trng_context),
						   .cra_module = THIS_MODULE,
						   .cra_init = trng_sm4_init,
						   .cra_exit = trng_cleanup,
					   } };

int sce_trng_alg_register(void)
{
	int ret = 0;

	ret = crypto_register_rng(&sce_trng_alg);
	if (ret) {
		sce_err("%s:crypto_register_rng failed!ret:%d\n", __func__,
			ret);
	}

	ret = crypto_register_rng(&sce_trng_sm4_alg);
	if (ret) {
		crypto_unregister_rng(&sce_trng_alg);
		sce_err("%s:crypto_register_sm4_rng failed!ret:%d\n", __func__,
			ret);
	}

	return ret;
}

void sce_trng_alg_unregister(void)
{
	crypto_unregister_rng(&sce_trng_alg);
	crypto_unregister_rng(&sce_trng_sm4_alg);
}
