#include <linux/module.h>
#include <linux/irq.h>
#include <linux/platform_device.h>
#include <linux/dma-mapping.h>
#include <linux/cdev.h>
#include <linux/types.h>
#include <linux/mutex.h>
#include <linux/acpi.h>
#include <crypto/internal/hash.h>
#include <linux/crypto.h>
#include <crypto/hash.h>
#include <crypto/aead.h>
#include <crypto/internal/skcipher.h>
#include <crypto/internal/akcipher.h>
#include <crypto/internal/aead.h>
#include <linux/ioport.h>
#include <linux/io.h>

#include "common.h"

#define XBUFSIZE 8
static int testmgr_alloc_buf(char *buf[XBUFSIZE])
{
	int i;

	for (i = 0; i < XBUFSIZE; i++) {
		buf[i] = (void *)__get_free_page(GFP_KERNEL);
		if (!buf[i])
			goto err_free_buf;
	}

	return 0;

err_free_buf:
	while (i-- > 0)
		free_page((unsigned long)buf[i]);

	return -ENOMEM;
}

static void testmgr_free_buf(char *buf[XBUFSIZE])
{
	int i;

	for (i = 0; i < XBUFSIZE; i++)
		free_page((unsigned long)buf[i]);
}
static void hexdump(unsigned char *buf, unsigned int len)
{
	print_hex_dump(KERN_CONT, "", DUMP_PREFIX_OFFSET,
			16, 1,
			buf, len, false);
}
int akcipher_test(char *alg, const struct sce_akcipher_testvec *vecs)
{
	struct crypto_wait wait;
	struct akcipher_request *req;
	struct crypto_akcipher *tfm;
	static u32 flag = 0;
	char *xbuf[XBUFSIZE];
	void *outbuf_enc = NULL;
	void *outbuf_dec = NULL;
	unsigned int out_len_max, out_len = 0;
	int err = -ENOMEM;
	struct scatterlist src, dst, src_tab[3];
	const char *m, *c;
	unsigned int m_size, c_size;
	const char *op;
	u8 *key;

	tfm = crypto_alloc_akcipher(alg, 0, 0);
	if (IS_ERR(tfm)) {
		pr_err("failed to load transform for %s: %ld\n", alg, PTR_ERR(tfm));
		return -1;
	}

	if (flag++ == 0)
		pr_info("\nsingle testing of async %s (%s)!\n", alg, get_driver_name(crypto_akcipher, tfm));

	if (testmgr_alloc_buf(xbuf))
			goto free_tfm;

	req = akcipher_request_alloc(tfm, GFP_KERNEL);
	if (!req)
		goto free_xbuf;

	crypto_init_wait(&wait);

	key = kmalloc(vecs->key_len + sizeof(u32) * 2 + vecs->param_len,
			  GFP_KERNEL);
	if (!key)
		goto free_req;
	memcpy(key, vecs->key, vecs->key_len);
	if (vecs->public_key_vec)
		err = crypto_akcipher_set_pub_key(tfm, key, vecs->key_len);
	else
		err = crypto_akcipher_set_priv_key(tfm, key, vecs->key_len);
	if (err)
		goto free_key;

	/*
	 * First run test which do not require a private key, such as
	 * encrypt or verify.
	 */
	err = -ENOMEM;
	out_len_max = crypto_akcipher_maxsize(tfm);
	outbuf_enc = kzalloc(out_len_max, GFP_KERNEL);
	if (!outbuf_enc)
		goto free_key;
	
	if (!vecs->siggen_sigver_test) {
		m = vecs->m;
		m_size = vecs->m_size;
		c = vecs->c;
		c_size = vecs->c_size;
		op = "encrypt";
	} else {
		/* Swap args so we could keep plaintext (digest)
		 * in vecs->m, and cooked signature in vecs->c.
		 */
		m = vecs->c; /* signature */
		m_size = vecs->c_size;
		c = vecs->m; /* digest */
		c_size = vecs->m_size;
		op = "verify";
	}
	if (vecs->public_key_vec){
		err = -E2BIG;
		if (WARN_ON(m_size > PAGE_SIZE))
			goto free_all;
		memcpy(xbuf[0], m, m_size);

		sg_init_table(src_tab, 3);
		sg_set_buf(&src_tab[0], xbuf[0], 8);
		sg_set_buf(&src_tab[1], xbuf[0] + 8, m_size - 8);
		if (vecs->siggen_sigver_test) {
			if (WARN_ON(c_size > PAGE_SIZE))
				goto free_all;
			memcpy(xbuf[1], c, c_size);
			sg_set_buf(&src_tab[2], xbuf[1], c_size);
			akcipher_request_set_crypt(req, src_tab, NULL, m_size, c_size);
		} else {
			sg_init_one(&dst, outbuf_enc, out_len_max);
			akcipher_request_set_crypt(req, src_tab, &dst, m_size,
					   out_len_max);
		}
		akcipher_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG,
					  crypto_req_done, &wait);

		err = crypto_wait_req(vecs->siggen_sigver_test ?
				  /* Run asymmetric signature verification */
				  crypto_akcipher_verify(req) :
				  /* Run asymmetric encrypt */
				  crypto_akcipher_encrypt(req), &wait);
		if (err) {
			pr_err("alg: akcipher: %s test failed. err %d\n", op, err);
			goto free_all;
		}
		if (!vecs->siggen_sigver_test && c) {
			if (req->dst_len != c_size) {
				pr_err("alg: akcipher: %s test failed. Invalid output len!\n",op);
				err = -EINVAL;
				goto free_all;
			}
			/* verify that encrypted message is equal to expected */
			if (memcmp(c, outbuf_enc, c_size) != 0) {
				pr_err("alg: akcipher: %s test failed. Invalid output\n",
				   op);
				hexdump(outbuf_enc, c_size);
				err = -EINVAL;
				goto free_all;
			}
		}
		pr_info("alg: akcipher: %s test success!\n", op);
		err = 0;
        	goto free_all;
	}
	/*
	 * Don't invoke (decrypt or sign) test which require a private key
	 * for vectors with only a public key.
	 */
	outbuf_dec = kzalloc(out_len_max, GFP_KERNEL);
	if (!outbuf_dec) {
		err = -ENOMEM;
		goto free_all;
	}

	if (!vecs->siggen_sigver_test && !c) {
		c = outbuf_enc;
		c_size = req->dst_len;
	}

	err = -E2BIG;
	op = vecs->siggen_sigver_test ? "sign" : "decrypt";

	if (WARN_ON(c_size > PAGE_SIZE))
		goto free_all;
	memcpy(xbuf[0], c, c_size);
	
	sg_init_one(&src, xbuf[0], c_size);
	sg_init_one(&dst, outbuf_dec, out_len_max);
	crypto_init_wait(&wait);
	akcipher_request_set_crypt(req, &src, &dst, c_size, out_len_max);
	akcipher_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG,
						  crypto_req_done, &wait);
	err = crypto_wait_req(vecs->siggen_sigver_test ?
				  /* Run asymmetric signature generation */
				  crypto_akcipher_sign(req) :
				  /* Run asymmetric decrypt */
				  crypto_akcipher_decrypt(req), &wait);
	if (err) {
		pr_err("alg: akcipher: %s test failed. err %d\n", op, err);
		goto free_all;
	}
	out_len = req->dst_len;
	if (out_len < m_size) {
		pr_err("alg: akcipher: %s test failed. Invalid output len %u\n",
			   op, out_len);
		err = -EINVAL;
		goto free_all;
	}
	/* verify that decrypted message is equal to the original msg */
	if(!vecs->siggen_sigver_test){
	if (memchr_inv(outbuf_dec, 0, out_len - m_size) ||
		memcmp(m, outbuf_dec + out_len - m_size, m_size)) {
		pr_err("alg: akcipher: %s test failed. Invalid output\n", op);
		hexdump(outbuf_dec, out_len);
		err = -EINVAL;
	}
	}
	pr_info("alg: akcipher: %s test success!\n", op);
free_all:
	kfree(outbuf_dec);
	kfree(outbuf_enc);
free_key:
	kfree(key);
free_req:
	akcipher_request_free(req);
free_xbuf:
	testmgr_free_buf(xbuf);
free_tfm:
	 crypto_free_akcipher(tfm);
	 return err;
}

#ifdef TESTX86
void get_random_bytes(void *buf, int nbytes)
{
    u64 jf = jiffies;
    u8 *pjf = (u8 *)&jf;
    u8 *out = (u8 *)buf;
    u32 i;
    for (i = 0; i < nbytes; i++) {
        out[i] = pjf[i & 0x7] + i;
    }
}

#endif

void print_buf(u8 *name, u8 *buf, u32 len)
{
    u32 i;

    pr_info("%s[0x%x]:\n", name, len);
    for (i = 0; i < len; i += 16) {
        pr_info("\"\\x%02x\\x%02x\\x%02x\\x%02x\\x%02x\\x%02x\\x%02x\\x%02x\\x%"
               "02x\\x%02x\\x%02x\\x%02x\\x%02x\\x%02x\\x%02x\\x%02x\"\n",
               buf[i + 0], buf[i + 1], buf[i + 2], buf[i + 3], buf[i + 4],
               buf[i + 5], buf[i + 6], buf[i + 7], buf[i + 8], buf[i + 9],
               buf[i + 10], buf[i + 11], buf[i + 12], buf[i + 13], buf[i + 14],
               buf[i + 15]);
    }
}

void ahash_dgst_test(char *alg, const u8 *key, unsigned int klen,
                     const u8 *data, unsigned int len, u8 *out)
{
    struct scatterlist *sg = NULL;
    struct crypto_wait wait;
    struct ahash_request *req;
    struct crypto_ahash *tfm;
    int ret, i;
    static u32 flag = 0;

    tfm = crypto_alloc_ahash(alg, 0, 0);
    if (IS_ERR(tfm)) {
        pr_err("failed to load transform for %s: %ld\n", alg, PTR_ERR(tfm));
        return;
    }

    if (flag++ == 0)
        pr_info("\ndgst testing of async %s (%s)\n", alg,
                get_driver_name(crypto_ahash, tfm));

    req = ahash_request_alloc(tfm, GFP_KERNEL);
    if (!req) {
        pr_err("ahash request allocation failure\n");
        goto fail;
    }

    crypto_init_wait(&wait);
    ahash_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG, crypto_req_done,
                               &wait);

    if (key && klen){
        ret = crypto_ahash_setkey(tfm, key, klen);
        if (ret){
            pr_err("setkey failed, ret = %d!\n", ret);
            goto fail1;
        }
    }

    if (len) {
        sg = kmalloc(sizeof(struct scatterlist) * len, GFP_KERNEL);
        if (sg == NULL) {
            pr_err("%s:kmalloc sg failed!\n", __func__);
            goto fail1;
        }
        sg_init_table(sg, len);
        for (i = 0; i < len; i++) {
            sg_set_buf(sg + i, data + i, 1);
        }
    }
    ahash_request_set_crypt(req, sg, out, len);

    ret = crypto_wait_req(crypto_ahash_digest(req), &wait);
    if (ret) {
        pr_err("hashing failed ret=%d\n", ret);
    }

    if (sg)
        kfree(sg);
fail1:
    ahash_request_free(req);
fail:
    crypto_free_ahash(tfm);
}

void ahash_finup_test(char *alg, const u8 *key, unsigned int klen,
                      const u8 *data, unsigned int len, u8 *out)
{
    struct scatterlist *sg = NULL;
    struct crypto_wait wait;
    struct ahash_request *req;
    struct crypto_ahash *tfm;
    int ret, i;
    static u32 flag = 0;

    tfm = crypto_alloc_ahash(alg, 0, 0);
    if (IS_ERR(tfm)) {
        pr_err("failed to load transform for %s: %ld\n", alg, PTR_ERR(tfm));
        return;
    }

    if (flag++ == 0)
        pr_info("\nfinup testing of async %s (%s)\n", alg,
                get_driver_name(crypto_ahash, tfm));

    req = ahash_request_alloc(tfm, GFP_KERNEL);
    if (!req) {
        pr_err("ahash request allocation failure\n");
        goto fail;
    }

    crypto_init_wait(&wait);
    ahash_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG, crypto_req_done,
                               &wait);

    if (key && klen){
        ret = crypto_ahash_setkey(tfm, key, klen);
        if (ret){
            pr_err("setkey failed, ret = %d!\n", ret);
            goto fail1;
        }
    }

    if (len) {
        sg = kmalloc(sizeof(struct scatterlist) * len, GFP_KERNEL);
        if (sg == NULL) {
            pr_err("%s:kmalloc sg failed!\n", __func__);
            goto fail1;
        }
        sg_init_table(sg, len);
        for (i = 0; i < len; i++) {
            sg_set_buf(sg + i, data + i, 1);
        }
    }
    ahash_request_set_crypt(req, sg, out, len);

    ret = crypto_wait_req(crypto_ahash_init(req), &wait);
    if (ret) {
        pr_err("hash init failed ret=%d\n", ret);
    }

    ret = crypto_wait_req(crypto_ahash_finup(req), &wait);
    if (ret) {
        pr_err("hash finup failed ret=%d\n", ret);
    }

    if (sg)
        kfree(sg);
fail1:
    ahash_request_free(req);
fail:
    crypto_free_ahash(tfm);
}

void ahash_step_test(char *alg, const u8 *key, unsigned int klen,
                     const u8 *data, unsigned int len, u8 *out)
{
    struct scatterlist sg;
    struct crypto_wait wait;
    struct ahash_request *req;
    struct crypto_ahash *tfm;
    int ret;
    unsigned int offset, step_len;
    static u32 flag = 0;

    tfm = crypto_alloc_ahash(alg, 0, 0);
    if (IS_ERR(tfm)) {
        pr_err("failed to load transform for %s: %ld\n", alg, PTR_ERR(tfm));
        return;
    }

    if (flag++ == 0)
        pr_info("\nstep testing of async %s (%s)\n", alg,
                get_driver_name(crypto_ahash, tfm));

    req = ahash_request_alloc(tfm, GFP_KERNEL);
    if (!req) {
        pr_err("ahash request allocation failure\n");
        goto fail;
    }

    if (key && klen){
        ret = crypto_ahash_setkey(tfm, key, klen);
        if (ret){
            pr_err("setkey failed, ret = %d!\n", ret);
            goto fail1;
        }
    }

    crypto_init_wait(&wait);
    ahash_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG, crypto_req_done,
                               &wait);

    ret = crypto_wait_req(crypto_ahash_init(req), &wait);
    if (ret) {
        pr_err("hashing failed ret=%d\n", ret);
        goto fail1;
    }

    for (offset = 0; offset < len;) {
        step_len = 0xF;
        if (offset + step_len > len)
            step_len = len - offset;

        sg_init_table(&sg, 1);
        sg_set_buf(&sg, data + offset, step_len);
        ahash_request_set_crypt(req, &sg, out, step_len);

        ret = crypto_wait_req(crypto_ahash_update(req), &wait);
        if (ret) {
            pr_err("crypto_ahash_update failed ret=%d\n", ret);
            break;
        }
        offset += step_len;
    }

    ret = crypto_wait_req(crypto_ahash_final(req), &wait);
    if (ret) {
        pr_err("crypto_ahash_final failed ret=%d\n", ret);
    }

fail1:
    ahash_request_free(req);
fail:
    crypto_free_ahash(tfm);
}

void skcipher_test(char *alg, int enc, const u8 *key, int klen, u8 *iv,
                   const u8 *data, unsigned int len, u8 *out)
{
    struct crypto_wait wait;
    struct skcipher_request *req;
    struct crypto_skcipher *tfm;
    struct scatterlist *ssg = NULL, *dsg = NULL;
    static u32 flag = 0;
    int ret, i;

    tfm = crypto_alloc_skcipher(alg, 0, 0);
    if (IS_ERR(tfm)) {
        pr_err("failed to load transform for %s: %ld\n", alg, PTR_ERR(tfm));
        return;
    }

    if (flag++ == 0)
        pr_info("\nsingle testing of async %s (%s)!\n", alg,
                get_driver_name(crypto_skcipher, tfm));

    req = skcipher_request_alloc(tfm, GFP_KERNEL);
    if (!req) {
        pr_info("skcipher: Failed to allocate request for %s\n", alg);
        goto fail;
    }

    crypto_init_wait(&wait);
    skcipher_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG,
                                  crypto_req_done, &wait);

    crypto_skcipher_clear_flags(tfm, ~0);
    ret = crypto_skcipher_setkey(tfm, key, klen);
    if (ret){
        pr_err("setkey failed, ret = %d!\n", ret);
        goto fail1;
    }


    ssg = kmalloc(sizeof(struct scatterlist) * len, GFP_KERNEL);
    if (ssg == NULL) {
        pr_err("%s:kmalloc ssg failed!\n", __func__);
        goto fail1;
    }
    sg_init_table(ssg, len);
    for (i = 0; i < len; i++) {
        sg_set_buf(ssg + i, data + i, 1);
    }

    dsg = kmalloc(sizeof(struct scatterlist) * len, GFP_KERNEL);
    if (dsg == NULL) {
        pr_err("%s:kmalloc dsg failed!\n", __func__);
        goto fail2;
    }
    sg_init_table(dsg, len);
    for (i = 0; i < len; i++) {
        sg_set_buf(dsg + i, out + i, 1);
    }

    skcipher_request_set_crypt(req, ssg, dsg, len, iv);

    if (enc)
        ret = crypto_wait_req(crypto_skcipher_encrypt(req), &wait);
    else
        ret = crypto_wait_req(crypto_skcipher_decrypt(req), &wait);
    if (ret) {
        pr_err("failed ret=%d\n", ret);
    }

    if (dsg)
        kfree(dsg);
fail2:
    if (ssg)
        kfree(ssg);
fail1:
    skcipher_request_free(req);
fail:
    crypto_free_skcipher(tfm);
}

void skcipher_step_test(char *alg, int enc, const u8 *key, int klen, u8 *iv,
                        const u8 *data, unsigned int len, u8 *out)
{
    struct crypto_wait wait;
    struct skcipher_request *req;
    struct crypto_skcipher *tfm;
    struct scatterlist ssg, dsg;
    static u32 flag = 0;
    int ret;
    u32 offset, step_len;

    tfm = crypto_alloc_skcipher(alg, 0, 0);
    if (IS_ERR(tfm)) {
        pr_err("failed to load transform for %s: %ld\n", alg, PTR_ERR(tfm));
        return;
    }

    if (flag++ == 0)
        pr_info("\nstep testing of async %s (%s)!\n", alg,
                get_driver_name(crypto_skcipher, tfm));

    req = skcipher_request_alloc(tfm, GFP_KERNEL);
    if (!req) {
        pr_info("skcipher: Failed to allocate request for %s\n", alg);
        goto fail;
    }

    crypto_skcipher_clear_flags(tfm, ~0);
    ret = crypto_skcipher_setkey(tfm, key, klen);
    if (ret){
        pr_err("setkey failed, ret = %d!\n", ret);
        goto fail1;
    }

    for (offset = 0; offset < len;) {
        step_len = 64;
        if (offset + step_len > len)
            step_len = len - offset;

        crypto_init_wait(&wait);
        skcipher_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG,
                                      crypto_req_done, &wait);

        sg_init_table(&ssg, 1);
        sg_set_buf(&ssg, data + offset, step_len);
        sg_init_table(&dsg, 1);
        sg_set_buf(&dsg, out + offset, step_len);
        skcipher_request_set_crypt(req, &ssg, &dsg, step_len, iv);

        if (enc)
            ret = crypto_wait_req(crypto_skcipher_encrypt(req), &wait);
        else
            ret = crypto_wait_req(crypto_skcipher_decrypt(req), &wait);
        if (ret) {
            pr_err("failed ret=%d\n", ret);
            break;
        }
        offset += step_len;
    }
fail1:
    skcipher_request_free(req);
fail:
    crypto_free_skcipher(tfm);
}

void aead_test(char *alg, int enc, const u8 *key, int klen, u8 *iv,
               const u8 *data, unsigned int len, unsigned int aad_len, u8 *out, unsigned int out_len)
{
    struct crypto_wait wait;
    struct aead_request *req;
    struct crypto_aead *tfm;
    struct scatterlist ssg , dsg;
    static u32 flag = 0;
    int ret;
    u32 authsize, src;

    tfm = crypto_alloc_aead(alg, 0, 0);
    if (IS_ERR(tfm)) {
        pr_err("failed to load transform for %s: %ld\n", alg, PTR_ERR(tfm));
        return;
    }

    if (flag++ == 0)
        pr_info("\nsingle testing of async %s (%s)!\n", alg,
                get_driver_name(crypto_aead, tfm));

    req = aead_request_alloc(tfm, GFP_KERNEL);
    if (!req) {
        pr_info("aead: Failed to allocate request for %s\n", alg);
        goto fail;
    }

    crypto_init_wait(&wait);
    aead_request_set_callback(req, CRYPTO_TFM_REQ_MAY_BACKLOG, crypto_req_done,
                              &wait);

    crypto_aead_clear_flags(tfm, ~0);
    ret = crypto_aead_setkey(tfm, key, klen);
    if (ret){
        pr_err("setkey failed, ret = %d!\n", ret);
        goto fail1;
    }
    if (enc) {
	authsize = out_len - len;
	src = len;
    } else {
	authsize = len - out_len;
	src = len;
    }
    crypto_aead_setauthsize(tfm, authsize);

    sg_init_one(&ssg, data, len + aad_len);
    sg_init_one(&dsg, out, out_len + aad_len);
    aead_request_set_crypt(req, &ssg, &dsg, src, iv);
    aead_request_set_ad(req, aad_len);

    if (enc)
        ret = crypto_wait_req(crypto_aead_encrypt(req), &wait);
    else
        ret = crypto_wait_req(crypto_aead_decrypt(req), &wait);
    if (ret) {
        pr_err("failed ret=%d\n", ret);
    }
fail1:
    aead_request_free(req);
fail:
    crypto_free_aead(tfm);
}

void sce_get_alg_name(char *alg_name, char *alg, char *mode)
{
    if (!alg || !mode) {
        return;
    }

    memcpy(alg_name, mode, strlen(mode));
    alg_name[strlen(mode)] = '(';
    memcpy(alg_name + strlen(mode) + 1, alg, strlen(alg));
    alg_name[strlen(mode) + strlen(alg) + 1] = ')';

    pr_info("------test alg_name %s!------\n", alg_name);
}
