// SPDX-License-Identifier: GPL-2.0-only
/*
 * File Name:phytium_scto.c - Phytium driver for SDK
 *
 * Copyright (C) 2020 Phytium Technology Co.,Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */

#include <linux/module.h>
#include <linux/irq.h>
#include <linux/platform_device.h>
#include <linux/dma-mapping.h>
#include <linux/cdev.h>
#include <linux/types.h>
#include <linux/mutex.h>
#include <linux/acpi.h>
#include <crypto/internal/hash.h>
#include <linux/crypto.h>
#include <crypto/hash.h>
#include <crypto/internal/skcipher.h>
#include <crypto/internal/akcipher.h>
#include <linux/ioport.h>
#include <linux/io.h>
#include <linux/version.h>
#include <linux/delay.h>

#include <asm/cputype.h>
#include "smx_common.h"
#include "sm4_phytium.h"
#include "sm3_phytium.h"
#include "phytium_scto.h"
#include "version.h"
#include "trng_phytium.h"

struct phytium_scto_dev scto;
static struct resource *scto_resource;
static u64 *common_info_vaddr;

static int trng;
module_param(trng, int, 0);

#define DESC_VADDR_COUNT \
	((SCTO_DESC_NUM + 1) * sizeof(phytium_scto_context) / 0x400000 + 1)

/**
 * Alloc 0x400000 bytes buffer per desc_vaddr item.
 * Each desc_vaddr item contains 0x400000/sizeof(phytium_scto_context) phytium_scto_context items.
 * So, max DESC_VADDR_COUNT items for desc_vaddr. Use desc_vaddr to save phytium_scto_context
 */
static phytium_scto_context *desc_vaddr[DESC_VADDR_COUNT] = { NULL };

static dma_addr_t paddr[SCTO_DESC_NUM];
static void *vaddr[SCTO_DESC_NUM] = { NULL };

/**
 * Alloc 0x400000 bytes buffer for per mepool item.
 * Each mepool item contains 0x400000/sizeof(mepool_t) mepool_t items.
 * So, max MEPOOL_COUNT items for mepool. Use mepool to save mepool_t
 *
 */
#define MEPOOL_COUNT ((SCTO_DESC_NUM + 1) * sizeof(mepool_t) / 0x400000 + 1)
static mepool_t *mepool[MEPOOL_COUNT] = { NULL };

static inline void set_sm3_data(SM3_CTX *sm3_ctx)
{
	scto.hash_reg->hash_in[0] = sm3_ctx->A;
	scto.hash_reg->hash_in[1] = sm3_ctx->B;
	scto.hash_reg->hash_in[2] = sm3_ctx->C;
	scto.hash_reg->hash_in[3] = sm3_ctx->D;
	scto.hash_reg->hash_in[4] = sm3_ctx->E;
	scto.hash_reg->hash_in[5] = sm3_ctx->F;
	scto.hash_reg->hash_in[6] = sm3_ctx->G;
	scto.hash_reg->hash_in[7] = sm3_ctx->H;
}

static inline void process_sm3(long in, long out, uint32_t byteLen)
{
	int count = 0;

	//src addr
	scto.dma_reg->saddr0 = (in >> 2) & 0xFFFFFFFF;
	scto.dma_reg->saddr1 = (in >> 34) & 0x0FFF;

	//dst addr
	scto.dma_reg->daddr0 = (out >> 2) & 0xFFFFFFFF;
	scto.dma_reg->daddr1 = (out >> 34) & 0x0FFF;

	//data word length
	scto.dma_reg->len = (byteLen >> 2);

	//clear flag
	scto.smx_reg->sr_2 = 2;

	//store cfg
	scto.smx_reg->cmd = 2;

	dsb(sy);

	//start
	scto.smx_reg->cr = 1;

	do {
		dsb(sy);
		if (unlikely(count++ > 1024))
			schedule();
		if (unlikely(count > 65536)) {
			pr_err("sm3 timeout!\n");
			break;
		}
	} while (!(scto.smx_reg->sr_2 & 2));
}

static inline void set_sm4_key(uint32_t *key)
{
	scto.ske_reg->key[0] = key[0];
	scto.ske_reg->key[1] = key[1];
	scto.ske_reg->key[2] = key[2];
	scto.ske_reg->key[3] = key[3];
}

static inline void set_sm4_iv(uint32_t *iv)
{
	scto.ske_reg->iv[0] = iv[0];
	scto.ske_reg->iv[1] = iv[1];
	scto.ske_reg->iv[2] = iv[2];
	scto.ske_reg->iv[3] = iv[3];
}

static inline int init_sm4(sm4_mode_e mode, int crypto, uint32_t *key,
			   uint32_t *iv)
{
	uint32_t scto_iv[4] = { 0 };
	uint32_t cfg = 0;

	SCTO_CHECK_AND_RET(key == NULL, -EINVAL, "Invalid sm4 key(null)");

	if (likely(mode != SM4_MODE_ECB)) {
		SCTO_CHECK_AND_RET(iv == NULL, -EINVAL, "Invalid sm4 iv(null)");

#ifdef DMA_LITTLE_ENDIAN
		memcpy(scto_iv, iv, 16);
#else
		smx_reverse_word(iv, scto_iv, SM4_BLOCK_WORD_LEN);
#endif

		set_sm4_iv(scto_iv);
		cfg |= (1 << SKE_UPDATE_IV_OFFSET);
	}

	cfg |= (mode << SKE_MODE_OFFSET) | (crypto << SKE_CRYPTO_OFFSET);
	scto.ske_reg->cfg = cfg;

	set_sm4_key(key);

	return 0;
}

static void process_sm4(long in, long out, uint32_t byteLen)
{
	int count = 0;

	//src addr
	scto.dma_reg->saddr0 = (in >> 2) & 0xFFFFFFFF;
	scto.dma_reg->saddr1 = (in >> 34) & 0x0FFF;

	//dst addr
	scto.dma_reg->daddr0 = (out >> 2) & 0xFFFFFFFF;
	scto.dma_reg->daddr1 = (out >> 34) & 0x0FFF;

	//data word length
	scto.dma_reg->len = (byteLen >> 2);

	//clear flag
	scto.smx_reg->sr_2 = 1;

	//store cfg
	scto.smx_reg->cmd = 1;

	dsb(sy);
	//start
	scto.smx_reg->cr = 1;

	do {
		dsb(sy);
		if (unlikely(count++ > 1024))
			schedule();
		if (unlikely(count > 65536)) {
			pr_err("sm4 timeout!\n");
			break;
		}
	} while (!(scto.smx_reg->sr_2 & 1));
}

static u32 wait_trng_ready(void)
{
	u32 count = 0;

	//wait till done
	while (scto.trng_reg->fifo_sr & 0x1000000) {
		if (unlikely(count++ > 1024))
			schedule();

		if ((scto.trng_reg->sr & 1) || (count > 65536)) {
			scto.trng_reg->cr = 0x1E;
			scto.trng_reg->sr = 7; //clear flag
			scto.trng_reg->rtcr = 0;
			dsb(sy);
			scto.trng_reg->cr = 0x1F;
			dsb(sy);

			pr_warn("trng timeout!\n");
			break;
		}
	}

	// counts of random in FIFO
	return ((scto.trng_reg->fifo_sr & 0xFC0000) >> 16);
}

static int get_rand(char *a, size_t byteLen)
{
	u32 rng_data, rng_data_len;
	size_t count = 0;

	rng_data_len = wait_trng_ready();
	while (rng_data_len) {
		rng_data = scto.trng_reg->dr;
		if (byteLen >= 4) {
			if (unlikely(copy_to_user(a, &rng_data, 4)))
				return -1;
			a += 4;
			byteLen -= 4;
			rng_data_len -= 4;
			count += 4;
		} else {
			if (unlikely(copy_to_user(a, &rng_data, byteLen)))
				return -1;
			count += byteLen;
			break;
		}
	}

	return count;
}

static int phytium_scto_open(struct inode *inode, struct file *filep)
{
	return 0;
}

int phytium_scto_release(struct inode *inode, struct file *filep)
{
	return 0;
}

ssize_t phytium_scto_read(struct file *filep, char __user *buf, size_t len,
			  loff_t *offset)
{
	int count;

	SCTO_CHECK_AND_RET(unlikely(0 == len || NULL == buf || NULL == offset),
			   0, "Invalid params for read.");

	spin_lock(&scto.rng_lock);
	count = get_rand(buf, len);
	spin_unlock(&scto.rng_lock);

	return count;
}

static long driver_info_ioctl(struct file *fp, unsigned int cmd,
			      unsigned long arg)
{
	int ret;
	struct phy_crypto_kernel_info info = { 0 };

	ret = copy_from_user(&info, (void *)arg, sizeof(info));
	if (ret) {
		pr_err("%s copy_from_user failed", __func__);
		return ret;
	}

	strncpy(info.version, VERSION, 32);
	info.version_number = 0x010603;
	if (info.magic_number != PHY_CRYPTO_DRIVER_MAGIC) {
		pr_err("Invalid lib, Magic number %#llx, expected %#lx!\r\n",
		       info.magic_number, PHY_CRYPTO_DRIVER_MAGIC);
		info.magic_number = PHY_CRYPTO_DRIVER_MAGIC;
	}

	if (info.reserved != (SCTO_DESC_NUM << 8) + PER_DESC_DMA_BUF_SIZE) {
		pr_err("Macro dismath, expected SCTO_DESC_NUM %d, PER_DESC_DMA_BUF_SIZE %d!",
		       SCTO_DESC_NUM, PER_DESC_DMA_BUF_SIZE);
		return -1;
	}

	info.response = 1;

	/* default big endian, should reverse it */
#ifdef DMA_LITTLE_ENDIAN
	info.endian = scto.smx_reg->endian & 0x3;
#else
	info.endian = 0x3;
#endif
	info.reserved = 1;

	ret = copy_to_user((void *)arg, &info, sizeof(info));

	return ret;
}

long _algs_sm34_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
	phytium_scto_context *ctx;
	int ret;
	unsigned int desc_id, buf_len;
	unsigned long flags;

	volatile int *user_count = (volatile int *)((long)common_info_vaddr +
						    (SCTO_DESC_NUM * 8) + 128);

	desc_id = arg >> 32;
	buf_len = arg & 0xFFFFFFFF;

	SCTO_CHECK_AND_RET((desc_id > SCTO_DESC_NUM), -1, "Invalid desc_id");
	SCTO_CHECK_AND_RET((buf_len > PER_DESC_DMA_BUF_SIZE), -1,
			   "Invalid buf_len");

	atomic_inc((atomic_t*)user_count);

	ctx = &desc_vaddr[desc_id / (0x400000 / sizeof(phytium_scto_context))]
			 [desc_id &
			  ((0x400000 / sizeof(phytium_scto_context)) - 1)];

	spin_lock_irqsave(&scto.scto_lock, flags);

	if (cmd == SCTO_SM3) {
		set_sm3_data(&ctx->psm3_ctx.sm3_ctx);
		process_sm3(common_info_vaddr[desc_id],
			    common_info_vaddr[desc_id], buf_len);
	} else if (cmd == SCTO_SM4) {
		ret = init_sm4(ctx->psm4_ctx.mode, ctx->psm4_ctx.cryptomode,
			       ctx->psm4_ctx.scto_key,
			       (u32 *)ctx->psm4_ctx.evp_cipher_ctx.iv);
		if (ret)
			return ret;

		process_sm4(common_info_vaddr[desc_id],
			    common_info_vaddr[desc_id], buf_len);
	}

	spin_unlock_irqrestore(&scto.scto_lock, flags);

	atomic_dec((atomic_t*)user_count);

	return 0;
}

long phytium_scto_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
	switch (cmd) {
#ifdef SCTO_ENABLED_SM2
	case SCTO_SM2:
		return scto_sm2(cmd, arg);
#endif

	case PHY_CRYPTO_DRIVER_INFO:
		return driver_info_ioctl(file, cmd, arg);
	case SCTO_SM3:
	case SCTO_SM4:
		return _algs_sm34_ioctl(file, cmd, arg);
	default:
		return -1;
	}
}

int phytium_scto_mmap(struct file *filep, struct vm_area_struct *vma)
{
	unsigned long mmap_flag = vma->vm_pgoff;
	unsigned long pfn = vma->vm_pgoff;
	u32 index;

	index = (mmap_flag >> 2) & 0xfff;
	mmap_flag &= ~0x3ffc;

	switch (mmap_flag) {
	case 0:
		pfn = scto_resource->start >> PAGE_SHIFT;
		vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
		break;
	case 1:
		pfn = virt_to_phys(common_info_vaddr) >> PAGE_SHIFT;
		break;
	case 2:
		SCTO_CHECK_AND_RET((index >= DESC_VADDR_COUNT), -EINVAL,
				   "Invalid index for mmap");
		pfn = virt_to_phys(desc_vaddr[index]) >> PAGE_SHIFT;
		break;
	case 3:
		SCTO_CHECK_AND_RET((index >= MEPOOL_COUNT), -EINVAL,
				   "Invalid index for mmap");
		pfn = virt_to_phys(mepool[index]) >> PAGE_SHIFT;
		break;
	default:
		break;
	}

	return remap_pfn_range(vma, vma->vm_start, pfn,
			       vma->vm_end - vma->vm_start, vma->vm_page_prot);
}

static const struct file_operations phytium_scto_fops = {
	.owner = THIS_MODULE,
	.open = phytium_scto_open,
	.release = phytium_scto_release,
	.read = phytium_scto_read,
	.mmap = phytium_scto_mmap,
	.unlocked_ioctl = phytium_scto_ioctl,

};

#if SCTO_KERNEL_MODE

static void phytium_scto_unregister_algs(void)
{
	phytium_scto_unregister_sm3();
	phytium_scto_unregister_sm4();

	if (trng)
		scto_trng_alg_unregister();
}

static int phytium_scto_register_algs(struct phytium_scto_dev *scto)
{
	int ret;

	ret = phytium_scto_register_sm3();
	if (ret)
		return -1;

	ret = phytium_scto_register_sm4();
	if (ret) {
		phytium_scto_unregister_sm3();
		return -1;
	}

	if (trng) {
		ret = scto_trng_alg_register();
		if (ret) {
			phytium_scto_unregister_sm3();
			phytium_scto_unregister_sm4();
		}
	}

	return ret;
}

#endif

static struct cdev scto_pcdev;
static struct class *scto_pclass;
static dev_t scto_dev;

static int phytium_scto_cdev_register(void)
{
	int ret;

	ret = alloc_chrdev_region(&scto_dev, 0, 1, "dscto");
	if (ret < 0) {
		pr_err("alloc chrdev region fail\n");
		return -1;
	}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 4, 0)
	scto_pclass = class_create("dscto");
#else
	scto_pclass = class_create(THIS_MODULE, "dscto");
#endif

	if (IS_ERR(scto_pclass)) {
		pr_err("class create fail!\n");
		goto err_0;
	}

	cdev_init(&scto_pcdev, &phytium_scto_fops);

	ret = cdev_add(&scto_pcdev, scto_dev, 1);
	if (ret) {
		pr_err("cdev add error!\n");
		goto err_1;
	}

	if (IS_ERR(device_create(scto_pclass, NULL, MKDEV(MAJOR(scto_dev), 0),
				 NULL, "dscto"))) {
		pr_err("device create fail!\n\n");
		goto err_2;
	}

	return 0;
err_2:
	cdev_del(&scto_pcdev);
err_1:
	class_destroy(scto_pclass);
err_0:
	unregister_chrdev_region(scto_dev, 1);

	return -1;
}

static void phytium_scto_cdev_unregister(void)
{
	device_destroy(scto_pclass, scto_dev);
	cdev_del(&scto_pcdev);
	class_destroy(scto_pclass);
	unregister_chrdev_region(scto_dev, 1);
}

static int phytium_scto_probe(struct platform_device *pdev)
{
	struct device *dev = &pdev->dev;
	int err = 0;
	int i;

#if SCTO_KERNEL_MODE
	pr_info("SCTO kernel mode ON\r\n");
	acpi_dma_configure(dev, DEV_DMA_NON_COHERENT);
#else
	pr_info("SCTO kernel mode OFF\r\n");
	acpi_dma_configure(dev, DEV_DMA_COHERENT);
#endif

#ifdef DMA_IO_BUFFER
	acpi_dma_configure(dev, DEV_DMA_COHERENT);
	dma_set_mask(dev, DMA_BIT_MASK(32));
	dma_set_coherent_mask(dev, DMA_BIT_MASK(32));
	pr_info("SCTO use DMA_IO_BUFFER\r\n");
#else
	dma_set_mask_and_coherent(dev, DMA_BIT_MASK(48));
#endif

	scto_resource = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	scto.regs = devm_ioremap_resource(&pdev->dev, scto_resource);
	if (IS_ERR(scto.regs))
		return -1;

	scto.pke_reg[0] = (pke_reg_t *)(scto.regs + PKE_BASE_ADDR);
	scto.hash_reg = (hash_reg_t *)(scto.regs + HASH_BASE_ADDR);
	scto.ske_reg = (ske_reg_t *)(scto.regs + SKE_BASE_ADDR);
	scto.dma_reg = (dma_reg_t *)(scto.regs + DMA_BASE_ADDR);
	scto.trng_reg = (trng_reg_t *)(scto.regs + TRNG_BASE_ADDR);
	scto.smx_reg = (smx_reg_t *)(scto.regs);

	scto.hash_reg->hash_cfg = 0;
	scto.hash_reg->hash_ctrl = 0x10000;
	scto.ske_reg->cfg = 0;
	scto.ske_reg->ctrl = 0x10000;
	scto.dma_reg->cfg |= 0x7;
	scto.dma_reg->cfg_aw = 0x01000270;
	scto.dma_reg->cfg_ar = 0x010002b0;

	scto.smx_reg->cfg = 0;
	scto.trng_reg->cr = 0x1E;
	scto.trng_reg->sr = 7;
	scto.trng_reg->rtcr = 0;
	scto.trng_reg->cr = 0x1F;
	scto.pke_reg[0]->cfg = 0;
	scto.suspending = 0;

#ifdef DMA_LITTLE_ENDIAN
	scto.smx_reg->endian = 0;
	pr_info("Endian value is %u ===========\r\n", scto.smx_reg->endian);

	{
		u32 *dma_addr = scto.regs + 0x34;
		*dma_addr = 0; /* little endian */
	}
#endif

	atomic_set(&scto.wait_count, 0);
	spin_lock_init(&scto.scto_lock);
	spin_lock_init(&scto.rng_lock);
	spin_lock_init(&scto.pke_lock);

	for (i = 0; i < NR_CPUS; i++) {
		spin_lock_init(&scto.sm2_queue[i].queue_lock);
		sqlist_init(&scto.sm2_queue[i].list);
	}

	scto.dev = dev;

	common_info_vaddr = (void *)__get_free_pages(
		GFP_KERNEL | __GFP_ZERO, get_order(SCTO_DESC_NUM * 8 + 1024));
	if (common_info_vaddr == NULL) {
		pr_err("get free pages fail!\n");
		return -1;
	}

	for (i = 0;
	     i <
	     (SCTO_DESC_NUM + 1) * sizeof(phytium_scto_context) / 0x400000 + 1;
	     i++) {
		desc_vaddr[i] = (void *)__get_free_pages(
			GFP_KERNEL | __GFP_ZERO, get_order(0x400000));
		if (desc_vaddr[i] == NULL) {
			pr_err("get free pages fail!\n");
			goto err_desc;
		}
	}

	for (i = 0; i < (SCTO_DESC_NUM + 1) * sizeof(mepool_t) / 0x400000 + 1;
	     i++) {
		mepool[i] = (void *)__get_free_pages(GFP_KERNEL | __GFP_ZERO,
						     get_order(0x400000));
		if (mepool[i] == NULL) {
			pr_err("get free pages fail!\n");
			goto err_mepool;
		}
	}

	for (i = 0; i < SCTO_DESC_NUM; i++) {
		vaddr[i] = dma_alloc_coherent(dev, PER_DESC_DMA_BUF_SIZE,
					      &paddr[i], GFP_KERNEL);
		if (vaddr[i] == NULL) {
			pr_err("dma alloc coherent fail!\n");
			goto err_dma;
		}
	}

#if SCTO_KERNEL_MODE
	err = phytium_scto_register_algs(&scto);
	if (err) {
		pr_err("algs register fail!\n");
		goto err_reg;
	}
#endif

#ifdef SCTO_ENABLED_SM2
	err = phytium_scto_register_sm2();
	if (err) {
		pr_err("Register sm2 failed\r\n");
		goto err_sm2;
	}
#endif

	err = phytium_scto_cdev_register();
	if (err) {
		pr_err("cdev register fail!\n");
		goto err_dev;
	}

	for (i = 0; i < SCTO_DESC_NUM; i++) {
		common_info_vaddr[i + 1] = paddr[i];

		mepool[i / (0x400000 / sizeof(mepool_t))]
		      [i & ((0x400000 / sizeof(mepool_t)) - 1)]
			      .counter = i + 1;
	}

	platform_set_drvdata(pdev, &scto);

	pr_info("scto ver:%s probe success!\n", VERSION);

	return 0;

err_dev:

#ifdef SCTO_ENABLED_SM2
	phytium_scto_unregister_sm2();

err_sm2:
#endif

#if SCTO_KERNEL_MODE
	phytium_scto_unregister_algs();

err_reg:
#endif

	for (i = 0; i < SCTO_DESC_NUM; i++) {
		if (vaddr[i])
			dma_free_coherent(scto.dev, PER_DESC_DMA_BUF_SIZE,
					  vaddr[i], paddr[i]);
	}

err_dma:
	for (i = 0; i < (SCTO_DESC_NUM + 1) * sizeof(mepool_t) / 0x400000 + 1;
	     i++) {
		if (mepool[i])
			free_pages((unsigned long)mepool[i],
				   get_order(0x400000));
	}
err_mepool:
	for (i = 0;
	     i <
	     (SCTO_DESC_NUM + 1) * sizeof(phytium_scto_context) / 0x400000 + 1;
	     i++) {
		if (desc_vaddr[i])
			free_pages((unsigned long)desc_vaddr[i],
				   get_order(0x400000));
	}

err_desc:
	if (common_info_vaddr)
		free_pages((unsigned long)common_info_vaddr,
			   get_order(SCTO_DESC_NUM * 8 + 1024));

	return -1;
}

static int phytium_scto_remove(struct platform_device *pdev)
{
	int i;

#if SCTO_KERNEL_MODE
	phytium_scto_unregister_algs();
#endif

#ifdef SCTO_ENABLED_SM2
	phytium_scto_unregister_sm2();
#endif

	phytium_scto_cdev_unregister();

	if (common_info_vaddr)
		free_pages((unsigned long)common_info_vaddr,
			   get_order(SCTO_DESC_NUM * 8 + 1024));

	for (i = 0; i < DESC_VADDR_COUNT; i++) {
		if (desc_vaddr[i])
			free_pages((unsigned long)desc_vaddr[i],
				   get_order(0x400000));
	}

	for (i = 0; i < MEPOOL_COUNT; i++) {
		if (mepool[i])
			free_pages((unsigned long)mepool[i],
				   get_order(0x400000));
	}

	for (i = 0; i < SCTO_DESC_NUM; i++) {
		if (vaddr[i])
			dma_free_coherent(scto.dev, PER_DESC_DMA_BUF_SIZE,
					  vaddr[i], paddr[i]);
	}

	platform_set_drvdata(pdev, NULL);

	pr_info("scto removed!\n");

	return 0;
}

static const struct of_device_id phytium_scto_dma_of_match[] = {
#if SCTO_REGISTER_SELF
	{ .compatible = "PHYTSCTO" },
#endif
	{}
};
MODULE_DEVICE_TABLE(of, phytium_scto_dma_of_match);

#ifdef CONFIG_ACPI
static const struct acpi_device_id phytium_scto_acpi_ids[] = {
#if SCTO_REGISTER_SELF
	{ .id = "PHYTSCTO" },
#else
	{ .id = "PHYT0012" },
#endif
	{}
};

MODULE_DEVICE_TABLE(acpi, phytium_scto_acpi_ids);
#else
#define phytium_scto_acpi_ids NULL
#endif

static int phytium_scto_suspend(struct device *dev)
{
	pr_info("ENTER suspend\n");

	scto.suspending = 1;
	spin_lock(&scto.pke_lock);
	spin_lock(&scto.scto_lock);
	spin_lock(&scto.rng_lock);

	return 0;
}

static int phytium_scto_resume(struct device *dev)
{
	pr_info("ENTER resume\n");

	scto.hash_reg->hash_cfg = 0;
	scto.hash_reg->hash_ctrl = 0x10000;
	scto.ske_reg->cfg = 0;
	scto.ske_reg->ctrl = 0x10000;
	scto.dma_reg->cfg |= 0x7;
	scto.dma_reg->cfg_aw = 0x01000270;
	scto.dma_reg->cfg_ar = 0x010002b0;
	scto.smx_reg->cfg = 0;
	scto.trng_reg->cr = 0x1E;
	scto.trng_reg->sr = 7;
	scto.trng_reg->rtcr = 1;
	scto.trng_reg->cr = 0x1F;
	scto.pke_reg[0]->cfg = 0;

	scto.suspending = 0;

#ifdef SCTO_ENABLED_SM2
	wake_up_process(sm2_task);
#endif

	spin_unlock(&scto.rng_lock);
	spin_unlock(&scto.scto_lock);
	spin_unlock(&scto.pke_lock);

	return 0;
}

static const struct dev_pm_ops phytium_scto_pm_ops = {
	.suspend = phytium_scto_suspend,
	.resume = phytium_scto_resume,
};

static struct platform_driver
	phytium_scto_driver = { .probe = phytium_scto_probe,
				.remove = phytium_scto_remove,
				.driver = {
					.name = "PHYTSCTO",
					.of_match_table =
						phytium_scto_dma_of_match,
					.acpi_match_table =
						phytium_scto_acpi_ids,
					.pm = &phytium_scto_pm_ops,
				} };

#if SCTO_REGISTER_SELF
static struct resource phytium_scto_resource = {
	.start = 0x28220000,
	.end = 0x2822ffff,
	.flags = IORESOURCE_MEM,
};

static void phytium_scto_dev_release(struct device *dev)
{
}

static struct platform_device phytium_scto_device = {
	.name = "PHYTSCTO",
	.id = -1,
	.num_resources = 1,
	.resource = &phytium_scto_resource,
	.dev = {
		.release = phytium_scto_dev_release,
	},
};
#endif

static int phytium_scto_device_probe(void)
{
	int ret;

#ifdef D2000
	int cpu_type = read_cpuid_part_number();

	if (cpu_type == 0x661 || cpu_type == 0x662 || cpu_type == 0x663) {
#endif

#if SCTO_REGISTER_SELF
		ret = platform_device_register(&phytium_scto_device);
		if (ret)
			return ret;
#endif
		ret = platform_driver_register(&phytium_scto_driver);

#if SCTO_REGISTER_SELF
		if (ret)
			platform_device_unregister(&phytium_scto_device);
#endif

#ifdef D2000
	} else {
		pr_err("cpu not support d2000_scto!\n");
		ret = -ENODEV;
	}
#endif

	return ret;
}

static void phytium_scto_device_remove(void)
{
#ifdef D2000
	int cpu_type = read_cpuid_part_number();

	if (cpu_type == 0x661 || cpu_type == 0x662 || cpu_type == 0x663) {
#endif
		platform_driver_unregister(&phytium_scto_driver);
#if SCTO_REGISTER_SELF
		platform_device_unregister(&phytium_scto_device);
#endif

#ifdef D2000
	} else
		return;
#endif
}

module_init(phytium_scto_device_probe);
module_exit(phytium_scto_device_remove);
MODULE_DESCRIPTION("PHYTIUM SCTO");
MODULE_LICENSE("GPL");
