// SPDX-License-Identifier: GPL-2.0-only
/*
 * File Name:sm2_phytium.c - Phytium SDK for SM2
 *
 * Copyright (C) 2020 Phytium Technology Co.,Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 */
#include <linux/scatterlist.h>
#include <asm/memory.h>
#include <linux/delay.h>
#include <crypto/internal/akcipher.h>
#include <crypto/akcipher.h>
#include <linux/io.h>
#include <linux/crypto.h>
#include <linux/kthread.h>
#include <asm/cacheflush.h>
#include <linux/version.h>
#include "smx_common.h"
#include "sm2_phytium.h"
#include "sm3_phytium.h"
#include "phytium_scto.h"

#ifdef SCTO_ENABLED_SM2

#include "sm3_phytium.h"

static volatile int running;
static u32 sm2_rng_index;
static struct sm3_context sm2_rng_ctx;
struct task_struct *sm2_task;

static sm2p256v1_params_t sm2p256v1_params = {
	.p = { 0xFFFFFFFF, 0xFFFFFFFF, 0x00000000, 0xFFFFFFFF, 0xFFFFFFFF,
	       0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFE },
	.p_h = { 0x00000003, 0x00000002, 0xFFFFFFFF, 0x00000002, 0x00000001,
		 0x00000001, 0x00000002, 0x00000004 },
	.a = { 0xFFFFFFFC, 0xFFFFFFFF, 0x00000000, 0xFFFFFFFF, 0xFFFFFFFF,
	       0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFE },
	.b = { 0x4D940E93, 0xDDBCBD41, 0x15AB8F92, 0xF39789F5, 0xCF6509A7,
	       0x4D5A9E4B, 0x9D9F5E34, 0x28E9FA9E },
	.Gx = { 0x334C74C7, 0x715A4589, 0xF2660BE1, 0x8FE30BBF, 0x6A39C994,
		0x5F990446, 0x1F198119, 0x32C4AE2C },
	.Gy = { 0x2139F0A0, 0x02DF32E5, 0xC62A4740, 0xD0A9877C, 0x6B692153,
		0x59BDCEE3, 0xF4F6779C, 0xBC3736A2 },
	.n = { 0x39D54123, 0x53BBF409, 0x21C6052B, 0x7203DF6B, 0xFFFFFFFF,
	       0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFE },
	.n_h = { 0x7C114F20, 0x901192AF, 0xDE6FA2FA, 0x3464504A, 0x3AFFE0D4,
		 0x620FC84C, 0xA22B3D3B, 0x1EB5E412 },
	.n_1 = { 0x39D54122, 0x53BBF409, 0x21C6052B, 0x7203DF6B, 0xFFFFFFFF,
		 0xFFFFFFFF, 0xFFFFFFFF, 0xFFFFFFFE },
	.Gx_2_128 = { 0xD13A42ED, 0xEAE3D9A9, 0x484E1B38, 0x2B2308F6,
		      0x88C21F3A, 0x3DB7B248, 0x74D55DA9, 0xB692E5B5 },
	.Gy_2_128 = { 0xE295E5AB, 0xD186469D, 0x73438E6D, 0xDB61AC17,
		      0x544926F9, 0x5A924F85, 0x0F3FB613, 0xA175051B },
};

static inline int get_valid_words(u32 *a, int max_words)
{
	int i;

	for (i = max_words; i > 0; i--) {
		if (a[i - 1])
			return i;
	}

	return 0;
}

static inline int bignum_cmp(u32 *a, int aWordLen, u32 *b, int bWordLen)
{
	int i;

	aWordLen = get_valid_words(a, aWordLen);
	bWordLen = get_valid_words(b, bWordLen);

	if (aWordLen > bWordLen)
		return 1;

	if (aWordLen < bWordLen)
		return -1;

	for (i = (aWordLen - 1); i >= 0; i--) {
		if (a[i] > b[i])
			return 1;

		if (a[i] < b[i])
			return -1;
	}

	return 0;
}

static inline int bignum_check_zero_u32(u32 *a, int aWordLen)
{
	int i;

	for (i = 0; i < aWordLen; i++) {
		if (a[i])
			return 0;
	}

	return 1;
}

static inline int bignum_check_zero_u8(void *a, int aByteLen)
{
	int i;
	u64 *b = (u64 *)a;
	u8 *c = (u8 *)a;
	int dwordlen = aByteLen >> 3;

	for (i = 0; i < dwordlen; i++) {
		if (b[i])
			return 0;
	}

	for (i = (dwordlen << 3); i < aByteLen; i++) {
		if (c[i])
			return 0;
	}

	return 1;
}

static inline void bignum_add_one(u32 *a, int wordLen)
{
	int i;

	for (i = 0; i < wordLen; i++) {
		a[i] += 1;
		if (a[i] > 0)
			break;
	}
}

/**
 *
 * in[0x20]:
 * "\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f\x10"
 * "\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f\x20"
 * out[0x20]:
 * "\x20\x1f\x1e\x1d\x1c\x1b\x1a\x19\x18\x17\x16\x15\x14\x13\x12\x11"
 * "\x10\x0f\x0e\x0d\x0c\x0b\x0a\x09\x08\x07\x06\x05\x04\x03\x02\x01"
 */
static inline void bignum_convert_byte_order_u256(const void *in, void *out)
{
	u32 i, tmp;
	const u32 *input = in;
	u32 *output = out;

	for (i = 0; i < 4; i++) {
		tmp = input[i];
		output[i] = swap32(input[0x7 - i]);
		output[0x7 - i] = swap32(tmp);
	}
}

static inline volatile u32 *pke_a(volatile pke_reg_t *pke_reg, unsigned long a,
				  int step)
{
	volatile u32 *addr;

	addr = (volatile u32 *)((unsigned long)pke_reg + 0x0400 + a * step);

	return addr;
}

static inline volatile u32 *pke_b(volatile pke_reg_t *pke_reg, unsigned long a,
				  int step)
{
	volatile u32 *addr;

	addr = (volatile u32 *)((unsigned long)pke_reg + 0x1000 + a * step);

	return addr;
}

static inline void pke_load_operand(volatile u32 *baseaddr, u32 *data,
				    int wordLen)
{
	int i;

	for (i = 0; i < wordLen; i++)
		baseaddr[i] = data[i];
}

static inline void pke_read_operand(volatile u32 *baseaddr, u32 *data,
				    int wordLen)
{
	int i;

	for (i = 0; i < wordLen; i++)
		data[i] = baseaddr[i];
}

static inline void pke_set_microcode(volatile pke_reg_t *pke_reg, u32 addr)
{
	pke_reg->mc_ptr = addr;
}

static inline void pke_clear_interrupt(volatile pke_reg_t *pke_reg)
{
	pke_reg->stat = 0x01;
}

static inline void pke_start(volatile pke_reg_t *pke_reg)
{
	pke_reg->ctrl = PKE_START_CALC;
}

static inline int pke_check_rt_code(volatile pke_reg_t *pke_reg)
{
	int ret;

	ret = pke_reg->rt_code & 0x07;
	if (unlikely(ret))
		pr_err("pke rt code[0x%x] error!\n", ret);

	return ret;
}

static inline void pke_load_pre_calc_mont_H(volatile pke_reg_t *pke_reg, u32 *H,
					    u32 wordLen)
{
	pke_load_operand(pke_b(pke_reg, 0, SM2_MEM_STEP), H, wordLen);
}

static u32 trng_check_ready(void)
{
	static u32 count;
	u32 rng_byte = 0;

	if (scto.trng_reg->fifo_sr & 0x1000000) {
		count++;
		if ((scto.trng_reg->sr & 1) || (count > 65536)) {
			scto.trng_reg->cr = 0x1E;

			scto.trng_reg->sr = 7; //clear flag

			scto.trng_reg->rtcr = 0;

			dsb(sy);

			scto.trng_reg->cr = 0x1F;

			dsb(sy);

			count = 0;
			pr_err("trng timeout!\n");
		}
	} else {
		rng_byte = (scto.trng_reg->fifo_sr & 0xFC0000) >> 16;
		if (rng_byte)
			count = 0;
	}

	//return counts of random in FIFO
	return rng_byte;
}

static u32 get_rand(u8 *a, u32 byteLen)
{
	u32 rng_data, rng_data_len, rng_byte = 0;

	if (byteLen) {
		rng_data_len = trng_check_ready();
		while (rng_data_len) {
			rng_data = scto.trng_reg->dr;
			if (byteLen >= 4) {
				memcpy(a, &rng_data, 4);
				rng_byte += 4;
				a += 4;
				byteLen -= 4;
				rng_data_len -= 4;
			} else {
				memcpy(a, &rng_data, byteLen);
				rng_byte += byteLen;
				break;
			}
		}
	}

	return rng_byte;
}

static void crypto_rng_update(u8 sm2_rng_pool[64])
{
	u32 len, rng_byte;
	u8 *addr;

	addr = sm2_rng_pool + sm2_rng_index;
	len = 64 - sm2_rng_index;

	spin_lock(&scto.rng_lock);
	rng_byte = get_rand(addr, len);
	spin_unlock(&scto.rng_lock);

	sm2_rng_index = (sm2_rng_index + rng_byte) & 0x3F;
}

static int sm2_point_mul_shamir(u32 *k1, u32 *P1x, u32 *P1y, u32 *k2, u32 *P2x,
				u32 *P2y, u32 *Qx, u32 *Qy,
				volatile pke_reg_t *pke_reg, int work_step)
{
	int i;

	switch (work_step) {
	case 0:
		break;
	case 1:
		goto STEP_1;
	case 2:
		goto STEP_2;
	default:
		pr_err("%s:work step[0x%x] error!\n", __func__, work_step);
		return -1;
	}

	pke_load_operand(pke_a(pke_reg, 0, SM2_MEM_STEP), sm2p256v1_params.p,
			 SM2_WORD_LEN); //p
	pke_set_microcode(pke_reg, MICROCODE_MGMR_PRE_N0);
	pke_clear_interrupt(pke_reg);
	pke_start(pke_reg);
	return 1;

STEP_1:
	if ((pke_reg->stat & 1) == 0)
		return 1;

	if (pke_check_rt_code(pke_reg))
		return -1;

	pke_load_operand(pke_b(pke_reg, 0, SM2_MEM_STEP), sm2p256v1_params.p_h,
			 SM2_WORD_LEN); //p_h
	pke_load_operand(pke_b(pke_reg, 1, SM2_MEM_STEP), P1x,
			 SM2_WORD_LEN); //P1x
	pke_load_operand(pke_b(pke_reg, 2, SM2_MEM_STEP), P1y,
			 SM2_WORD_LEN); //P1y                           //P1z=1
	*(pke_a(pke_reg, 3, SM2_MEM_STEP)) = 1;
	for (i = 1; i < SM2_WORD_LEN; i++)
		*(pke_a(pke_reg, 3, SM2_MEM_STEP) + i) = 0;
	pke_load_operand(pke_b(pke_reg, 5, SM2_MEM_STEP), P2x,
			 SM2_WORD_LEN); //P2x
	pke_load_operand(pke_b(pke_reg, 6, SM2_MEM_STEP), P2y,
			 SM2_WORD_LEN); //P2y
	pke_load_operand(pke_b(pke_reg, 4, SM2_MEM_STEP), sm2p256v1_params.a,
			 SM2_WORD_LEN); //a
	pke_load_operand(pke_a(pke_reg, 4, SM2_MEM_STEP), k1,
			 SM2_WORD_LEN); //k1
	pke_load_operand(pke_a(pke_reg, 5, SM2_MEM_STEP), k2,
			 SM2_WORD_LEN); //k2
	pke_set_microcode(pke_reg, MICROCODE_PMULF);
	pke_clear_interrupt(pke_reg);
	pke_start(pke_reg);
	return 2;

STEP_2:
	if ((pke_reg->stat & 1) == 0)
		return 2;

	if (pke_check_rt_code(pke_reg))
		return -1;

	pke_read_operand(pke_a(pke_reg, 1, SM2_MEM_STEP), Qx, SM2_WORD_LEN);
	if (Qy != NULL)
		pke_read_operand(pke_a(pke_reg, 2, SM2_MEM_STEP), Qy,
				 SM2_WORD_LEN);

	return 0;
}

static int sm2_point_mul_base(u32 *k, u32 *x, u32 *y,
			      volatile pke_reg_t *pke_reg, int work_step)
{
	u32 k1[SM2_WORD_LEN], k2[SM2_WORD_LEN];
	int i;

	//k1: high half part
	for (i = 0; i < SM2_WORD_LEN / 2; i++)
		k1[i] = k[i + SM2_WORD_LEN / 2];
	for (i = SM2_WORD_LEN / 2; i < SM2_WORD_LEN; i++)
		k1[i] = 0;

	//k2: low half part
	for (i = 0; i < SM2_WORD_LEN / 2; i++)
		k2[i] = k[i];
	for (i = SM2_WORD_LEN / 2; i < SM2_WORD_LEN; i++)
		k2[i] = 0;

	return sm2_point_mul_shamir(k1, sm2p256v1_params.Gx_2_128,
				    sm2p256v1_params.Gy_2_128, k2,
				    sm2p256v1_params.Gx, sm2p256v1_params.Gy, x,
				    y, pke_reg, work_step);
}

int sm2_point_mul(u32 *k, u32 *Px, u32 *Py, u32 *Qx, u32 *Qy,
		  volatile pke_reg_t *pke_reg, int work_step)
{
	int i;

	switch (work_step) {
	case 0:
		break;
	case 1:
		goto STEP_1;
	case 2:
		goto STEP_2;
	default:
		pr_err("%s:work_step[0x%x] error!\n", __func__, work_step);
		return -1;
	}

	pke_load_operand(pke_a(pke_reg, 0, SM2_MEM_STEP), sm2p256v1_params.p,
			 SM2_WORD_LEN); //p
	pke_set_microcode(pke_reg, MICROCODE_MGMR_PRE_N0);
	pke_clear_interrupt(pke_reg);
	pke_start(pke_reg);
	return 1;

STEP_1:
	if ((pke_reg->stat & 1) == 0)
		return 1;

	if (pke_check_rt_code(pke_reg))
		return -1;

	pke_load_operand(pke_b(pke_reg, 0, SM2_MEM_STEP), sm2p256v1_params.p_h,
			 SM2_WORD_LEN); //p_h
	pke_load_operand(pke_b(pke_reg, 1, SM2_MEM_STEP), Px,
			 SM2_WORD_LEN); //Px
	pke_load_operand(pke_b(pke_reg, 2, SM2_MEM_STEP), Py,
			 SM2_WORD_LEN); //Py                            //Pz=1
	*((uint32_t *)pke_a(pke_reg, 3, SM2_MEM_STEP)) = 1;
	for (i = 1; i < SM2_WORD_LEN; i++)
		*(pke_a(pke_reg, 3, SM2_MEM_STEP) + i) = 0;
	pke_load_operand(pke_b(pke_reg, 4, SM2_MEM_STEP), sm2p256v1_params.a,
			 SM2_WORD_LEN); //a
	pke_load_operand(pke_a(pke_reg, 4, SM2_MEM_STEP), k, SM2_WORD_LEN); //k
	pke_set_microcode(pke_reg, MICROCODE_PMUL);
	pke_clear_interrupt(pke_reg);
	pke_start(pke_reg);
	return 2;

STEP_2:
	if ((pke_reg->stat & 1) == 0)
		return 2;

	if (pke_check_rt_code(pke_reg))
		return -1;

	pke_read_operand(pke_a(pke_reg, 1, SM2_MEM_STEP), Qx, SM2_WORD_LEN);
	if (Qy != NULL)
		pke_read_operand(pke_a(pke_reg, 2, SM2_MEM_STEP), Qy,
				 SM2_WORD_LEN);

	return 0;
}

static int sm2_modadd(u32 *modulus, u32 *a, u32 *b, u32 *out, u32 wordLen,
		      volatile pke_reg_t *pke_reg, int work_step)
{
	switch (work_step) {
	case 0:
		break;
	case 1:
		goto STEP_1;
	default:
		pr_err("%s:work_step[0x%x] error!\n", __func__, work_step);
		return -1;
	}

	pke_load_operand(pke_a(pke_reg, 0, SM2_MEM_STEP), modulus,
			 wordLen); //A0 modulus
	pke_load_operand(pke_a(pke_reg, 1, SM2_MEM_STEP), a, wordLen); //A1 a
	pke_load_operand(pke_b(pke_reg, 1, SM2_MEM_STEP), b, wordLen); //B1 b
	pke_set_microcode(pke_reg, MICROCODE_MODADD);
	pke_clear_interrupt(pke_reg);
	pke_start(pke_reg);
	return 1;

STEP_1:
	if ((pke_reg->stat & 1) == 0)
		return 1;

	if (pke_check_rt_code(pke_reg))
		return -1;

	pke_read_operand(pke_a(pke_reg, 1, SM2_MEM_STEP), out,
			 wordLen); //A1 result

	return 0;
}

static int sm2_modsub(u32 *modulus, u32 *a, u32 *b, u32 *out, u32 wordLen,
		      volatile pke_reg_t *pke_reg, int work_step)
{
	switch (work_step) {
	case 0:
		break;
	case 1:
		goto STEP_1;
	default:
		pr_err("%s:work_step[0x%x] error!\n", __func__, work_step);
		return -1;
	}

	pke_load_operand(pke_a(pke_reg, 0, SM2_MEM_STEP), modulus,
			 wordLen); //A0 modulus
	pke_load_operand(pke_a(pke_reg, 1, SM2_MEM_STEP), a, wordLen); //A1 a
	pke_load_operand(pke_b(pke_reg, 1, SM2_MEM_STEP), b, wordLen); //B1 b
	pke_set_microcode(pke_reg, MICROCODE_MODSUB);
	pke_clear_interrupt(pke_reg);
	pke_start(pke_reg);
	return 1;

STEP_1:
	if ((pke_reg->stat & 1) == 0)
		return 1;

	if (pke_check_rt_code(pke_reg))
		return -1;

	pke_read_operand(pke_a(pke_reg, 1, SM2_MEM_STEP), out,
			 wordLen); //A1 result

	return 0;
}

static int sm2_modinv(u32 *modulus, u32 *a, u32 *ainv, int modWordLen,
		      int aWordLen, volatile pke_reg_t *pke_reg, int work_step)
{
	int i;

	switch (work_step) {
	case 0:
		break;
	case 1:
		goto STEP_1;
	default:
		pr_err("%s:work_step[0x%x] error!\n", __func__, work_step);
		return -1;
	}

	pke_load_operand(pke_a(pke_reg, 0, SM2_MEM_STEP), modulus,
			 modWordLen); //A0 modulus
	pke_load_operand(pke_b(pke_reg, 1, SM2_MEM_STEP), a, aWordLen); //B1 a
	if (modWordLen > aWordLen) {
		for (i = aWordLen; i < modWordLen; i++)
			*(pke_b(pke_reg, 1, SM2_MEM_STEP) + i) = 0;
	}
	pke_set_microcode(pke_reg, MICROCODE_MODINV);
	pke_clear_interrupt(pke_reg);
	pke_start(pke_reg);
	return 1;

STEP_1:
	if ((pke_reg->stat & 1) == 0)
		return 1;

	if (pke_check_rt_code(pke_reg))
		return -1;

	pke_read_operand(pke_a(pke_reg, 1, SM2_MEM_STEP), ainv,
			 modWordLen); //A1 ainv

	return 0;
}

static int sm2_modmul_internal(u32 *modulus, u32 *a, u32 *b, u32 *out,
			       int wordLen, volatile pke_reg_t *pke_reg,
			       int work_step)
{
	switch (work_step) {
	case 0:
		break;
	case 1:
		goto STEP_1;
	case 2:
		goto STEP_2;
	default:
		pr_err("%s:work_step[0x%x] error!\n", __func__, work_step);
		return -1;
	}

	pke_load_pre_calc_mont_H(pke_reg, sm2p256v1_params.n_h, SM2_WORD_LEN);
	pke_load_operand(pke_a(pke_reg, 0, SM2_MEM_STEP), modulus,
			 wordLen); //A0 modulus
	pke_set_microcode(pke_reg, MICROCODE_MGMR_PRE_N0);
	pke_clear_interrupt(pke_reg);
	pke_start(pke_reg);
	return 1;

STEP_1:
	if ((pke_reg->stat & 1) == 0)
		return 1;

	if (pke_check_rt_code(pke_reg))
		return -1;

	pke_load_operand(pke_a(pke_reg, 1, SM2_MEM_STEP), a, wordLen); //A1 a
	pke_load_operand(pke_b(pke_reg, 1, SM2_MEM_STEP), b, wordLen); //B1 b
	pke_set_microcode(pke_reg, MICROCODE_MODMUL);
	pke_clear_interrupt(pke_reg);
	pke_start(pke_reg);
	return 2;

STEP_2:
	if ((pke_reg->stat & 1) == 0)
		return 2;

	if (pke_check_rt_code(pke_reg))
		return -1;

	pke_read_operand(pke_a(pke_reg, 1, SM2_MEM_STEP), out,
			 wordLen); //A0 out

	return 0;
}

int sm2_point_verify(u32 *Px, u32 *Py, volatile pke_reg_t *pke_reg,
		     int work_step)
{
	switch (work_step) {
	case 0:
		break;
	case 1:
		goto STEP_1;
	case 2:
		goto STEP_2;
	default:
		pr_err("%s:work_step[0x%x] error!\n", __func__, work_step);
		return -1;
	}

	pke_load_operand(pke_a(pke_reg, 0, SM2_MEM_STEP), sm2p256v1_params.p,
			 SM2_WORD_LEN); //p
	pke_set_microcode(pke_reg, MICROCODE_MGMR_PRE_N0);
	pke_clear_interrupt(pke_reg);
	pke_start(pke_reg);
	return 1;

STEP_1:
	if ((pke_reg->stat & 1) == 0)
		return 1;

	if (pke_check_rt_code(pke_reg))
		return -1;

	pke_load_operand(pke_b(pke_reg, 0, SM2_MEM_STEP), sm2p256v1_params.p_h,
			 SM2_WORD_LEN); //p_h
	pke_load_operand(pke_a(pke_reg, 1, SM2_MEM_STEP), Px,
			 SM2_WORD_LEN); //Px
	pke_load_operand(pke_a(pke_reg, 2, SM2_MEM_STEP), Py,
			 SM2_WORD_LEN); //Py
	pke_load_operand(pke_b(pke_reg, 4, SM2_MEM_STEP), sm2p256v1_params.a,
			 SM2_WORD_LEN); //a
	pke_load_operand(pke_a(pke_reg, 4, SM2_MEM_STEP), sm2p256v1_params.b,
			 SM2_WORD_LEN); //b
	pke_set_microcode(pke_reg, MICROCODE_PVER);
	pke_clear_interrupt(pke_reg);
	pke_start(pke_reg);
	return 2;

STEP_2:
	if ((pke_reg->stat & 1) == 0)
		return 2;

	if (pke_check_rt_code(pke_reg))
		return -1;

	return 0;
}

static inline void sm2_kdf_counter_add_one(u8 a[4])
{
	int i;

	for (i = 3; i >= 0; i--) {
		a[i] += 1;
		if (a[i] > 0)
			break;
	}
}

static void sm2_kdf(u8 *x2, u8 *y2, u8 *k, int kLen)
{
	u8 digest[SM3_DIGEST_BYTE_LEN];
	int i, t;
	u8 counter[4] = { 0x00, 0x00, 0x00, 0x01 }; // count = 1;
	struct sm3_context basectx, ctx;

	_software_sm3_init(&basectx);
	_software_sm3_update(&basectx, x2, SM2_BYTE_LEN);
	_software_sm3_update(&basectx, y2, SM2_BYTE_LEN);

	t = kLen >> 5; // t = kByteLen/32;
	for (i = 0; i < t; i++) {
		memcpy(&ctx, &basectx, sizeof(ctx));
		_software_sm3_update(&ctx, counter, 4);
		_software_sm3_final(&ctx, k + (i << 5));
		sm2_kdf_counter_add_one(counter);
	}

	kLen = kLen & 0x1F;
	if (kLen) {
		memcpy(&ctx, &basectx, sizeof(ctx));
		_software_sm3_update(&ctx, counter, 4);
		_software_sm3_final(&ctx, digest);
		memcpy(k + (t << 5), digest, kLen);
	}
}

static inline void xor
	(const void *A, const void *B, void *C, int byteLen) {
		int i;
		int dwordlen = byteLen >> 3;
		u64 *dA = (u64 *)A, *dB = (u64 *)B, *dC = (u64 *)C;
		u8 *bA = (u8 *)A, *bB = (u8 *)B, *bC = (u8 *)C;

		for (i = 0; i < dwordlen; i++)
			dC[i] = dA[i] ^ dB[i];

		for (i = (dwordlen << 3); i < byteLen; i++)
			bC[i] = bA[i] ^ bB[i];
	}

	static void sm2_rng(u32 k[8], u8 sm2_rng_pool[64])
{
	crypto_rng_update(sm2_rng_pool);
READ_RNG_K:
	_software_sm3_update(&sm2_rng_ctx, sm2_rng_pool, 64);
	memcpy(k, sm2_rng_ctx.state, SM2_BYTE_LEN);
	memcpy(sm2_rng_pool, k, SM2_BYTE_LEN);

	if (bignum_check_zero_u32(k, SM2_WORD_LEN))
		goto READ_RNG_K;
	if (bignum_cmp(k, SM2_WORD_LEN, sm2p256v1_params.n, SM2_WORD_LEN) >= 0)
		goto READ_RNG_K;
}

static int __sm2_sign(u8 *r_s, volatile pke_reg_t *pke_reg,
		      phytium_scto_sm2_work_list_t *work_list,
		      u8 sm2_rng_pool[64])
{
	int ret;

	switch (work_list->work_step >> 8) {
	case 0:
		break;
	case 1:
		goto STEP_1;
	case 2:
		goto STEP_2;
	case 3:
		goto STEP_3;
	case 4:
		goto STEP_4;
	case 5:
		goto STEP_5;
	case 6:
		goto STEP_6;
	case 7:
		goto STEP_7;
	default:
		pr_err("%s:work_step[0x%x] error!\n", __func__,
		       work_list->work_step);
		return -1;
	}

READ_RNG_K:
	sm2_rng(work_list->k, sm2_rng_pool);

STEP_1:
	ret = sm2_point_mul_base(work_list->k, work_list->x1, NULL, pke_reg,
				 work_list->work_step & 0xFF);

	work_list->work_step = ret | (1 << 8);
	if (ret) {
		if (unlikely(ret < 0))
			pr_err("%s:sm2_point_mul_base fail!ret:%d\n", __func__,
			       ret);
		return ret;
	}

STEP_2:
	ret = sm2_modadd(sm2p256v1_params.n, work_list->e, work_list->x1,
			 work_list->r, SM2_WORD_LEN, pke_reg,
			 work_list->work_step & 0xFF);

	work_list->work_step = ret | (2 << 8);
	if (ret) {
		if (unlikely(ret < 0))
			pr_err("%s:sm2_modadd fail!ret:%d\n", __func__, ret);
		return ret;
	}

	//make sure r is not zero
	if (bignum_check_zero_u32(work_list->r, SM2_WORD_LEN))
		goto READ_RNG_K;

	//tmp1 = r + k mod n
STEP_3:
	ret = sm2_modadd(sm2p256v1_params.n, work_list->r, work_list->k,
			 work_list->tmp1, SM2_WORD_LEN, pke_reg,
			 work_list->work_step & 0xFF);

	work_list->work_step = ret | (3 << 8);
	if (ret) {
		if (unlikely(ret < 0))
			pr_err("%s:sm2_modadd fail!ret:%d\n", __func__, ret);
		return ret;
	}

	//make sure r+k is not n
	if (bignum_check_zero_u32(work_list->tmp1, SM2_WORD_LEN))
		goto READ_RNG_K;

STEP_4:
	ret = sm2_modmul_internal(sm2p256v1_params.n, work_list->r,
				  work_list->ctx->d, work_list->tmp1,
				  SM2_WORD_LEN, pke_reg,
				  work_list->work_step & 0xFF);

	work_list->work_step = ret | (4 << 8);
	if (ret) {
		if (unlikely(ret < 0))
			pr_err("%s:sm2_modmul_internal fail!ret:%d\n", __func__,
			       ret);
		return ret;
	}

STEP_5:
	ret = sm2_modsub(sm2p256v1_params.n, work_list->k, work_list->tmp1,
			 work_list->tmp1, SM2_WORD_LEN, pke_reg,
			 work_list->work_step & 0xFF);

	work_list->work_step = ret | (5 << 8);
	if (ret) {
		if (unlikely(ret < 0))
			pr_err("%s:sm2_modsub fail!ret:%d\n", __func__, ret);
		return ret;
	}

	memcpy(work_list->tmp2, work_list->ctx->d, SM2_BYTE_LEN);
	bignum_add_one(work_list->tmp2, SM2_WORD_LEN);

STEP_6:
	ret = sm2_modinv(sm2p256v1_params.n, work_list->tmp2, work_list->tmp2,
			 SM2_WORD_LEN, SM2_WORD_LEN, pke_reg,
			 work_list->work_step & 0xFF);

	work_list->work_step = ret | (6 << 8);
	if (ret) {
		if (unlikely(ret < 0))
			pr_err("%s:sm2_modinv fail!ret:%d\n", __func__, ret);
		return ret;
	}

STEP_7:
	ret = sm2_modmul_internal(sm2p256v1_params.n, work_list->tmp1,
				  work_list->tmp2, work_list->s, SM2_WORD_LEN,
				  pke_reg, work_list->work_step & 0xFF);

	work_list->work_step = ret | (7 << 8);
	if (ret) {
		if (unlikely(ret < 0))
			pr_err("%s:sm2_modmul_internal fail!ret:%d\n", __func__,
			       ret);
		return ret;
	}

	//make sure s is not zero
	if (bignum_check_zero_u32(work_list->s, SM2_WORD_LEN))
		goto READ_RNG_K;

	bignum_convert_byte_order_u256(work_list->r, r_s);
	bignum_convert_byte_order_u256(work_list->s, r_s + SM2_BYTE_LEN);

	return 0;
}

static int __sm2_verify(volatile pke_reg_t *pke_reg,
			phytium_scto_sm2_work_list_t *work_list)
{
	int ret;

	switch (work_list->work_step >> 8) {
	case 0:
		break;
	case 1:
		goto STEP_1;
	case 2:
		goto STEP_2;
	case 3:
		goto STEP_3;
	default:
		pr_err("%s:work_step[0x%x] error!\n", __func__,
		       work_list->work_step);
		return -1;
	}

STEP_1:
	ret = sm2_modadd(sm2p256v1_params.n, work_list->r, work_list->s,
			 work_list->t, SM2_WORD_LEN, pke_reg,
			 work_list->work_step & 0xFF);

	work_list->work_step = ret | (1 << 8);
	if (ret) {
		if (unlikely(ret < 0))
			pr_err("%s:sm2_modadd fail!ret:%d\n", __func__, ret);
		return ret;
	}
	if (bignum_check_zero_u32(work_list->t, SM2_WORD_LEN)) {
		pr_err("%s:t == zero, error!\n", __func__);
		return -1;
	}

STEP_2:
	ret = sm2_point_mul_shamir(work_list->s, sm2p256v1_params.Gx,
				   sm2p256v1_params.Gy, work_list->t,
				   work_list->ctx->x, work_list->ctx->y,
				   work_list->x1, NULL, pke_reg,
				   work_list->work_step & 0xFF);

	work_list->work_step = ret | (2 << 8);
	if (ret) {
		if (unlikely(ret < 0))
			pr_err("%s:sm2_point_mul_shamir fail!ret:%d\n",
			       __func__, ret);
		return ret;
	}

STEP_3:
	ret = sm2_modadd(sm2p256v1_params.n, work_list->e, work_list->x1,
			 work_list->R, SM2_WORD_LEN, pke_reg,
			 work_list->work_step & 0xFF);

	work_list->work_step = ret | (3 << 8);
	if (ret) {
		if (unlikely(ret < 0))
			pr_err("%s:sm2 modadd fail!ret:%d\n", __func__, ret);
		return ret;
	}

	//cmp R == r
	if (bignum_cmp(work_list->R, SM2_WORD_LEN, work_list->r,
		       SM2_WORD_LEN)) {
		pr_err("%s:R != r, error!\n", __func__);
		return -1;
	}

	return 0;
}

static int __sm2_encrypt(u8 *M, u32 MLen, u8 *out, u32 *outlen,
			 volatile pke_reg_t *pke_reg,
			 phytium_scto_sm2_work_list_t *work_list,
			 u8 sm2_rng_pool[64])
{
	u8 *C1 = out, *C2 = out + 65, *C3 = out + 65 + MLen;
	int ret;
	struct sm3_context ctx;

	switch (work_list->work_step >> 8) {
	case 0:
		break;
	case 1:
		goto STEP_1;
	case 2:
		goto STEP_2;
	default:
		pr_err("%s:work_step[0x%x] error!\n", __func__,
		       work_list->work_step);
		return -1;
	}

READ_RNG_K:
	sm2_rng(work_list->k, sm2_rng_pool);

STEP_1:
	ret = sm2_point_mul_base(work_list->k, work_list->x1, work_list->y1,
				 pke_reg, work_list->work_step & 0xFF);

	work_list->work_step = ret | (1 << 8);
	if (ret) {
		if (unlikely(ret < 0))
			pr_err("%s:sm2_point_mul_base fail!ret:%d\n", __func__,
			       ret);
		return ret;
	}

	C1[0] = POINT_NOT_COMPRESSED;
	bignum_convert_byte_order_u256(work_list->x1, &C1[1]);
	bignum_convert_byte_order_u256(work_list->y1, &C1[1 + SM2_BYTE_LEN]);

STEP_2:
	ret = sm2_point_mul(work_list->k, work_list->ctx->x, work_list->ctx->y,
			    work_list->x1, work_list->y1, pke_reg,
			    work_list->work_step & 0xFF);

	work_list->work_step = ret | (2 << 8);
	if (ret) {
		if (unlikely(ret < 0))
			pr_err("%s:sm2_point_mul fail!ret:%d\n", __func__, ret);
		return ret;
	}

	bignum_convert_byte_order_u256(work_list->x1, work_list->x1);
	bignum_convert_byte_order_u256(work_list->y1, work_list->y1);
	sm2_kdf((u8 *)work_list->x1, (u8 *)work_list->y1, C2, MLen);
	if (bignum_check_zero_u8(C2, MLen))
		goto READ_RNG_K;

	xor(C2, M, C2, MLen);

	_software_sm3_init(&ctx);
	_software_sm3_update(&ctx, (u8 *)work_list->x1, SM2_BYTE_LEN);
	_software_sm3_update(&ctx, M, MLen);
	_software_sm3_update(&ctx, (u8 *)work_list->y1, SM2_BYTE_LEN);
	_software_sm3_final(&ctx, C3);

	*outlen = MLen + 1 + 3 * SM2_BYTE_LEN;

	return 0;
}

static int __sm2_decrypt(const u8 *C, int CLen, u8 *M, u32 *MLen,
			 volatile pke_reg_t *pke_reg,
			 phytium_scto_sm2_work_list_t *work_list)
{
	int temLen;
	u8 digest[SM2_BYTE_LEN];
	const u8 *C2, *C3;
	int ret;
	struct sm3_context ctx;

	switch (work_list->work_step >> 8) {
	case 0:
		break;
	case 1:
		goto STEP_1;
	case 2:
		goto STEP_2;
	default:
		pr_err("%s:work_step[0x%x] error!\n", __func__,
		       work_list->work_step);
		return -1;
	}

	bignum_convert_byte_order_u256(C + 1, work_list->x1);
	bignum_convert_byte_order_u256(C + 1 + SM2_BYTE_LEN, work_list->y1);
STEP_1:
	ret = sm2_point_verify(work_list->x1, work_list->y1, pke_reg,
			       work_list->work_step & 0xFF);

	work_list->work_step = ret | (1 << 8);
	if (ret) {
		if (unlikely(ret < 0))
			pr_err("%s:sm2_point_verify fail!ret:%d\n", __func__,
			       ret);
		return ret;
	}
STEP_2:
	ret = sm2_point_mul(work_list->ctx->d, work_list->x1, work_list->y1,
			    work_list->x1, work_list->y1, pke_reg,
			    work_list->work_step & 0xFF);

	work_list->work_step = ret | (2 << 8);
	if (ret) {
		if (unlikely(ret < 0))
			pr_err("%s:sm2_point_mul fail!ret:%d\n", __func__, ret);
		return ret;
	}
	bignum_convert_byte_order_u256(work_list->x1, work_list->x1);
	bignum_convert_byte_order_u256(work_list->y1, work_list->y1);

	temLen = CLen - (1 + 3 * SM2_BYTE_LEN);
	sm2_kdf((u8 *)work_list->x1, (u8 *)work_list->y1, M, temLen);
	if (bignum_check_zero_u8(M, temLen)) {
		pr_err("%s:t == zero, error!\n", __func__);
		return -1;
	}

	C2 = C + (1 + 2 * SM2_BYTE_LEN);
	xor(M, C2, M, temLen);

	C3 = C2 + temLen;
	_software_sm3_init(&ctx);
	_software_sm3_update(&ctx, (u8 *)work_list->x1, SM2_BYTE_LEN);
	_software_sm3_update(&ctx, M, temLen);
	_software_sm3_update(&ctx, (u8 *)work_list->y1, SM2_BYTE_LEN);
	_software_sm3_final(&ctx, digest);
	if (memcmp(C3, digest, SM2_BYTE_LEN))
		return -1;

	*MLen = temLen;

	return 0;
}

static void sm2signature_encode(u8 *input, u8 *sig, u32 *sig_len)
{
	u32 offset;
	u32 rlen, slen;
	u8 r[32], s[32];

	for (offset = 0; offset < 31; offset++) {
		if (input[offset])
			break;
	}

	rlen = 32 - offset;
	memcpy(r, input + 32 - rlen, rlen);

	for (offset = 32; offset < 63; offset++) {
		if (input[offset])
			break;
	}

	slen = 64 - offset;
	memcpy(s, input + 64 - slen, slen);

	*sig_len = 6 + rlen + slen;
	sig[0] = 0x30; //Tag 0, ASN1_SEQ(bit 0-4), Value = TLV + TLV + ...(bit5)
	sig[1] = 4 + rlen + slen; //Len 0,
	sig[2] = 0x02; //Tag 1, ASN1_INT(bit 0-4)
	sig[3] = rlen; //Len 1,

	offset = 4;
	if (r[0] & 0x80) { //Highest bit, r is unsigned Integer
		*sig_len += 1;
		sig[1]++;
		sig[3]++;
		sig[4] = 0x00;
		offset++;
	}

	memcpy(sig + offset, r, rlen);
	offset += rlen;

	sig[offset++] = 0x02; //Tag 2, ASN1_INT(bit 0-4)
	sig[offset++] = slen; //Len 2,

	if (s[0] & 0x80) { //Highest bit, s is unsigned Integer
		*sig_len += 1;
		sig[1]++;
		sig[offset - 1]++;
		sig[offset++] = 0x00;
	}

	memcpy(sig + offset, s, slen);
}

static int sm2signature_decode(u8 *sig, u32 sig_len, u8 *out)
{
	u32 offset = 4, len;
	u32 rlen, slen;
	u8 r[32] = { 0 }, s[32] = { 0 };

	if (unlikely((sig[0] != 0x30) ||
		     (sig[2] != 0x02))) { //check Tag 0 and Tag 1
		pr_err("%s:((sig[0] != 0x30) || (sig[2] != 0x02)), error!\n",
		       __func__);
		return -1;
	}

	len = sig[1] + 2;
	if (unlikely(len != sig_len)) { //check Len 0
		pr_err("%s:len != sig_len, error!\n", __func__);
		return -1;
	}

	rlen = sig[3]; //check Len 1
	if (rlen == 0x21) {
		offset++;
		rlen = 0x20;
	} else if (rlen > 0x21) {
		pr_err("%s:rlen > 0x21, error!\n", __func__);
		return -1;
	}

	memcpy(r + 32 - rlen, sig + offset, rlen);
	offset += rlen;

	if (unlikely(sig[offset++] != 0x02)) { //check Tag 2
		pr_err("%s:Tag 2 != 0x02, error!\n", __func__);
		return -1;
	}

	slen = sig[offset]; //check Len 2
	if (slen == 0x21) {
		offset++;
		slen = 0x20;
	} else if (slen > 0x21) {
		pr_err("%s:slen > 0x21, error!\n", __func__);
		return -1;
	}
	offset++;

	if (offset + slen == sig_len) //check Value Length
		memcpy(s + 32 - slen, sig + offset, slen);
	else {
		pr_err("%s:offset + slen != sig_len, error!\n", __func__);
		return -1;
	}

	memcpy(out, r, 32);
	memcpy(out + 32, s, 32);

	return 0;
}

static void sm2encrypt_encode(u8 *input, u32 inlen, u8 *output, u32 *outlen)
{
	u8 *c2;
	u8 x[32], y[32], hash[32];
	u8 xhead[4], yhead[4], hashhead[4], c2head[8], head[8];
	u32 offset, length, c2headlen, xheadlen, yheadlen, hashheadlen, headlen,
		c2len, i;
	u32 xlen, ylen;

	c2 = input + 65;
	memcpy(x, input + 1, 32);
	memcpy(y, input + 33, 32);
	memcpy(hash, input + inlen - 32, 32);

	c2len = inlen - 97;
	c2headlen = 2;
	if (c2len > 127)
		c2headlen = 3;
	if (c2len > 255)
		c2headlen = 4;
	if (c2len > 65535)
		c2headlen = 5;
	if (c2len > 16777215)
		c2headlen = 6;

	c2head[0] = 0x04; // Tag, ASN1_OTS
	switch (c2headlen) {
	case 2:
		c2head[1] = c2len; //Len(bit7 = 0)
		break;
	case 3:
		c2head[1] = 0x81; //Len(bit7 = 1, bit0-6 = Len length)
		c2head[2] = c2len;
		break;
	case 4:
		c2head[1] = 0x82;
		c2head[2] = c2len >> 8;
		c2head[3] = c2len;
		break;
	case 5:
		c2head[1] = 0x83;
		c2head[2] = c2len >> 16;
		c2head[3] = c2len >> 8;
		c2head[4] = c2len;
		break;
	case 6:
		c2head[1] = 0x84;
		c2head[2] = c2len >> 24;
		c2head[3] = c2len >> 16;
		c2head[4] = c2len >> 8;
		c2head[5] = c2len;
		break;
	default:
		break;
	}

	for (i = 0; i < 31; i++) {
		if (x[i])
			break;
	}
	xlen = 32 - i;
	xheadlen = 2;
	if (x[i] & 0x80) //Highest bit, x is unsigned Integer
		xheadlen = 3;
	xhead[0] = 0x2; //Tag, ASN1_INT
	switch (xheadlen) {
	case 2:
		xhead[1] = xlen;
		break;
	case 3:
		xhead[1] = xlen + 1;
		xhead[2] = 0x00;
		break;
	default:
		break;
	}

	for (i = 0; i < 31; i++) {
		if (y[i])
			break;
	}
	ylen = 32 - i;
	yheadlen = 2;
	if (y[i] & 0x80) //Highest bit, y is unsigned Integer
		yheadlen = 3;
	yhead[0] = 0x2; //Tag, ASN1_INT
	switch (yheadlen) {
	case 2:
		yhead[1] = ylen;
		break;
	case 3:
		yhead[1] = ylen + 1;
		yhead[2] = 0x00;
		break;
	default:
		break;
	}

	hashheadlen = 2;
	hashhead[0] = 0x04; // Tag, ASN1_OTS
	hashhead[1] = 0x20;

	length = c2headlen + c2len;
	length += xheadlen + xlen;
	length += yheadlen + ylen;
	length += hashheadlen + 32;

	headlen = 2;
	if (length > 127)
		headlen = 3;
	if (length > 255)
		headlen = 4;
	if (length > 65535)
		headlen = 5;
	if (length > 16777215)
		headlen = 6;

	head[0] = 0x30; //Tag, ASN1_SEQ, Value = TLV + TLV + ...
	switch (headlen) {
	case 2:
		head[1] = length;
		break;
	case 3:
		head[1] = 0x81;
		head[2] = length;
		break;
	case 4:
		head[1] = 0x82;
		head[2] = length >> 8;
		head[3] = length;
		break;
	case 5:
		head[1] = 0x83;
		head[2] = length >> 16;
		head[3] = length >> 8;
		head[4] = length;
		break;
	case 6:
		head[1] = 0x84;
		head[2] = length >> 24;
		head[3] = length >> 16;
		head[4] = length >> 8;
		head[5] = length;
		break;
	default:
		break;
	}

	offset = 0;
	memcpy(output + offset, head, headlen);
	offset += headlen;
	memcpy(output + offset, xhead, xheadlen);
	offset += xheadlen;
	memcpy(output + offset, x + 32 - xlen, xlen);
	offset += xlen;
	memcpy(output + offset, yhead, yheadlen);
	offset += yheadlen;
	memcpy(output + offset, y + 32 - ylen, ylen);
	offset += ylen;
	memcpy(output + offset, hashhead, hashheadlen);
	offset += hashheadlen;
	memcpy(output + offset, hash, 32);
	offset += 32;
	memcpy(output + offset, c2head, c2headlen);
	offset += c2headlen;
	memmove(output + offset, c2, c2len);
	offset += c2len;

	*outlen = offset;
}

static int sm2decrypt_decode(u8 *input, u32 inlen, u8 *output, u32 *outlen)
{
	u8 x[32] = { 0 }, y[32] = { 0 }, hash[32];
	u32 offset, length, c2headlen, xheadlen, yheadlen, hashheadlen, headlen,
		c2len, i;
	u32 xlen, ylen;

	if (unlikely(inlen < 45)) { //head:10 hash:32 x:1 y:1 c2:1
		pr_err("%s:inlen[%u], error!\n", __func__, inlen);
		return -1;
	}

	if (unlikely(input[0] != 0x30)) { //check Tag
		pr_err("%s:input[0] != 0x30, error!\n", __func__);
		return -1;
	}
	if (input[1] & 0x80) {
		headlen = 2 + (input[1] & 0x7f);
		if (unlikely(headlen > 6)) {
			pr_err("%s:headlen > 6, error!\n", __func__);
			return -1;
		}

		length = 0;
		for (i = 2; i < headlen; i++) {
			length <<= 8;
			length += input[i];
		}
	} else {
		headlen = 2;
		length = input[1];
	}

	if (unlikely(headlen + length != inlen)) {
		pr_err("%s:(headlen + length != inlen), error!\n", __func__);
		return -1;
	}

	offset = headlen;
	if (unlikely(input[offset] != 0x02)) {
		pr_err("%s:(input[offset] != 0x02), error!\n", __func__);
		return -1;
	}

	xlen = input[offset + 1];
	xheadlen = 2;
	if (xlen == 0x21) {
		xheadlen = 3;
		xlen = 0x20;
	} else if (xlen > 0x21) {
		pr_err("%s:(xlen > 0x21), error!\n", __func__);
		return -1;
	}

	offset += xheadlen;
	memcpy(x + 32 - xlen, input + offset, xlen);
	offset += xlen;

	if (unlikely(input[offset] != 0x02)) {
		pr_err("%s:(input[offset] != 0x02), error!\n", __func__);
		return -1;
	}
	ylen = input[offset + 1];
	yheadlen = 2;
	if (ylen == 0x21) {
		yheadlen = 3;
		ylen = 0x20;
	}

	if (unlikely(offset + yheadlen + ylen + 2 > inlen)) {
		pr_err("%s:(offset + yheadlen + ylen + 2 > inlen), error!\n",
		       __func__);
		return -1;
	}

	offset += yheadlen;
	memcpy(y + 32 - ylen, input + offset, ylen);
	offset += ylen;

	if (unlikely((input[offset] != 0x04) || (input[offset + 1] != 0x20))) {
		pr_err("%s:((input[offset] != 0x04) || (input[offset + 1] != 0x20)), error!\n",
		       __func__);
		return -1;
	}
	hashheadlen = 2;
	offset += hashheadlen;

	if (unlikely(offset + 35 > inlen)) {
		pr_err("%s:(offset + 35 > inlen), error!\n", __func__);
		return -1;
	}

	memcpy(hash, input + offset, 32);
	offset += 32;

	if (unlikely(input[offset] != 0x04)) {
		pr_err("%s:(input[offset] != 0x04), error!\n", __func__);
		return -1;
	}
	if (input[offset + 1] & 0x80) {
		c2headlen = 2 + (input[offset + 1] & 0x7f);
		if (unlikely(c2headlen > 6)) {
			pr_err("%s:(c2headlen > 6), error!\n", __func__);
			return -1;
		}

		if (unlikely(offset + c2headlen > inlen)) {
			pr_err("%s:(offset + c2headlen > inlen), error!\n",
			       __func__);
			return -1;
		}

		c2len = 0;
		for (i = 2; i < c2headlen; i++) {
			c2len <<= 8;
			c2len += input[offset + i];
		}
	} else {
		c2headlen = 2;
		c2len = input[offset + 1];
	}

	offset += c2headlen;
	if (unlikely(offset + c2len != inlen)) {
		pr_err("%s:(offset + c2len != inlen), error!\n", __func__);
		return -1;
	}

	output[0] = 0x04;
	memcpy(output + 1, x, 32);
	memcpy(output + 33, y, 32);
	memmove(output + 65, input + offset, c2len);
	memcpy(output + 65 + c2len, hash, 32);
	*outlen = c2len + 97;

	return 0;
}

static int sm2_sign(volatile pke_reg_t *pke_reg,
		    phytium_scto_sm2_work_list_t *work_list,
		    u8 sm2_rng_pool[64])
{
	struct akcipher_request *req = work_list->req;
	u8 r_s[64], dercode[72];
	u32 sign_len;
	int ret;

	if (unlikely(work_list->work_step == 0)) {
		/*
		 e check and convert
		*/
		if (unlikely(req->src_len != 32)) {
			pr_err("%s:src_len[0x%x], error!\n", __func__,
			       req->src_len);
			return -1;
		}
		if (unlikely(req->dst_len < 72)) {
			pr_err("%s:dst_len[0x%x], error!\n", __func__,
			       req->dst_len);
			return -1;
		}
		sg_copy_buffer(req->src, sg_nents(req->src), r_s, 32, 0, true);
		bignum_convert_byte_order_u256(r_s, work_list->e);
	}

	ret = __sm2_sign(r_s, pke_reg, work_list, sm2_rng_pool);
	if (ret) {
		if (unlikely(ret < 0))
			pr_err("%s:__sm2_sign failed!ret:%d\n", __func__, ret);
		return ret;
	}

	sm2signature_encode(r_s, dercode, &sign_len);
	sg_copy_buffer(req->dst, sg_nents(req->dst), dercode, sign_len, 0,
		       false);
	req->dst_len = sign_len;

	return ret;
}

static int sm2_verify(volatile pke_reg_t *pke_reg,
		      phytium_scto_sm2_work_list_t *work_list)
{
	struct akcipher_request *req = work_list->req;
	u8 r_s[64], code[128];
	int ret;

	if (unlikely(work_list->work_step == 0)) {
		/*
		 e check and convert
		*/
		if (unlikely(req->src_len > 72)) {
			pr_err("%s:src_len[0x%x], error!\n", __func__,
			       req->src_len);
			return -1;
		}
		if (unlikely(req->dst_len != 32)) {
			pr_err("%s:dst_len[0x%x], error!\n", __func__,
			       req->dst_len);
			return -1;
		}
		sg_copy_buffer(req->src, sg_nents(req->src), code,
			       req->src_len + req->dst_len, 0, true);
		bignum_convert_byte_order_u256(code + req->src_len,
					       work_list->e);
		ret = sm2signature_decode(code, req->src_len, r_s);
		if (unlikely(ret)) {
			pr_err("%s:sm2signature_decode failed!ret:%d\n",
			       __func__, ret);
			return ret;
		}
		bignum_convert_byte_order_u256(r_s, work_list->r);
		if (unlikely(bignum_check_zero_u32(work_list->r,
						   SM2_WORD_LEN))) {
			pr_err("%s:r == 0, error!\n", __func__);
			return -1;
		}
		if (unlikely(bignum_cmp(work_list->r, SM2_WORD_LEN,
					sm2p256v1_params.n,
					SM2_WORD_LEN) >= 0)) {
			pr_err("%s:r > n - 1, error!\n", __func__);
			return -1;
		}

		bignum_convert_byte_order_u256(r_s + SM2_BYTE_LEN,
					       work_list->s);
		if (unlikely(bignum_check_zero_u32(work_list->s,
						   SM2_WORD_LEN))) {
			pr_err("%s:s == 0, error!\n", __func__);
			return -1;
		}

		if (unlikely(bignum_cmp(work_list->s, SM2_WORD_LEN,
					sm2p256v1_params.n,
					SM2_WORD_LEN) >= 0)) {
			pr_err("%s:s > n - 1, error!\n", __func__);
			return -1;
		}
	}

	ret = __sm2_verify(pke_reg, work_list);
	if (ret) {
		if (unlikely(ret < 0))
			pr_err("%s:__sm2_verify failed!ret:%d\n", __func__,
			       ret);
		return ret;
	}

	return ret;
}

static int sm2_encrypt(volatile pke_reg_t *pke_reg,
		       phytium_scto_sm2_work_list_t *work_list,
		       u8 sm2_rng_pool[64])
{
	struct akcipher_request *req = work_list->req;
	u8 *M, *C;
	int ret;
	u32 outlen;

	if (unlikely(work_list->work_step == 0)) {
		/*
		 e check and convert
		*/
		M = kmalloc(req->src_len + 512, GFP_KERNEL);
		if (unlikely(M == NULL)) {
			pr_err("%s:kmalloc failed!\n", __func__);
			return -ENOMEM;
		}
		C = kmalloc(req->src_len + 512, GFP_KERNEL);
		if (unlikely(C == NULL)) {
			pr_err("%s:kmalloc failed!\n", __func__);
			kfree(M);
			return -ENOMEM;
		}
		sg_copy_buffer(req->src, sg_nents(req->src), M, req->src_len, 0,
			       true);
		work_list->M = M;
		work_list->C = C;
	}

	ret = __sm2_encrypt(work_list->M, req->src_len, work_list->C, &outlen,
			    pke_reg, work_list, sm2_rng_pool);
	if (ret) {
		if (unlikely(ret < 0)) {
			pr_err("%s:__sm2_encrypt failed!ret:%d\n", __func__,
			       ret);
			kfree(work_list->M);
			kfree(work_list->C);
		}
		return ret;
	}

	sm2encrypt_encode(work_list->C, outlen, work_list->M, &outlen);
	if (unlikely(outlen > req->dst_len)) {
		pr_err("%s:outlen[0x%x] > dst_len[0x%x], fail!\n", __func__,
		       outlen, req->dst_len);
		kfree(work_list->M);
		kfree(work_list->C);
		return -1;
	}

	sg_copy_buffer(req->dst, sg_nents(req->dst), work_list->M, outlen, 0,
		       false);
	req->dst_len = outlen;
	kfree(work_list->M);
	kfree(work_list->C);

	return ret;
}

static int sm2_decrypt(volatile pke_reg_t *pke_reg,
		       phytium_scto_sm2_work_list_t *work_list)
{
	struct akcipher_request *req = work_list->req;
	u8 *M, *C;
	int ret;
	u32 outlen;

	if (unlikely(work_list->work_step == 0)) {
		/*
		 e check and convert
		*/
		M = kmalloc(req->src_len + 512, GFP_KERNEL);
		if (unlikely(M == NULL)) {
			pr_err("%s:kmalloc failed!\n", __func__);
			return -ENOMEM;
		}

		C = kmalloc(req->src_len + 512, GFP_KERNEL);
		if (unlikely(C == NULL)) {
			pr_err("%s:kmalloc failed!\n", __func__);
			kfree(M);
			return -ENOMEM;
		}
		sg_copy_buffer(req->src, sg_nents(req->src), M, req->src_len, 0,
			       true);
		ret = sm2decrypt_decode(M, req->src_len, C,
					&work_list->decode_len);
		if (unlikely(ret)) {
			pr_err("%s:sm2decrypt_decode failed!ret:%d\n", __func__,
			       ret);
			kfree(C);
			kfree(M);
			return ret;
		}
		work_list->M = M;
		work_list->C = C;
	}

	ret = __sm2_decrypt(work_list->C, work_list->decode_len, work_list->M,
			    &outlen, pke_reg, work_list);
	if (ret) {
		if (unlikely(ret < 0)) {
			pr_err("%s:__sm2_decrypt failed!ret:%d\n", __func__,
			       ret);
			kfree(work_list->M);
			kfree(work_list->C);
		}
		return ret;
	}

	if (unlikely(outlen > req->dst_len)) {
		pr_err("%s:outlen[0x%x] > dst_len[0x%x], fail!\n", __func__,
		       outlen, req->dst_len);
		kfree(work_list->M);
		kfree(work_list->C);
		return -1;
	}

	sg_copy_buffer(req->dst, sg_nents(req->dst), work_list->M, outlen, 0,
		       false);
	req->dst_len = outlen;
	kfree(work_list->M);
	kfree(work_list->C);

	return ret;
}

static int sm2_genkey(volatile pke_reg_t *pke_reg,
		      phytium_scto_sm2_work_list_t *work_list,
		      u8 sm2_rng_pool[64])
{
	int ret = -1;

	switch (work_list->work_step >> 8) {
	case 0:
		break;
	case 1:
		goto STEP_1;
	case 2:
		goto STEP_2;
	case 3: // pubkey
		bignum_convert_byte_order_u256(work_list->prikey, work_list->k);
		work_list->work_step = 0;
		if (bignum_check_zero_u32(work_list->k, SM2_WORD_LEN)) {
			pr_err("Invalid prikey\r\n");
			return -1;
		}

		if (bignum_cmp(work_list->k, SM2_WORD_LEN, sm2p256v1_params.n,
			       SM2_WORD_LEN) >= 0) {
			pr_err("Invalid prikey\r\n");
			return -1;
		}

		goto STEP_1;
	default:
		pr_err("%s:work_step[0x%x] error!\n", __func__,
		       work_list->work_step);
		return -1;
	}

	sm2_rng(work_list->k, sm2_rng_pool);

STEP_1:
	ret = sm2_point_mul_base(work_list->k, work_list->x1, work_list->y1,
				 pke_reg, work_list->work_step & 0xFF);

	// step 1
	work_list->work_step = ret | (1 << 8);
	if (ret) {
		if (unlikely(ret < 0))
			pr_err("%s:sm2_point_mul_base fail!ret:%d\n", __func__,
			       ret);
		return ret;
	}

STEP_2:
	// step 2
	ret = sm2_point_mul_base(work_list->k, work_list->x1, work_list->y1,
				 pke_reg, work_list->work_step & 0xFF);

	work_list->work_step = ret | (2 << 8);
	if (ret) {
		if (unlikely(ret < 0))
			pr_err("%s:sm2_point_mul fail!ret:%d\n", __func__, ret);
		return ret;
	}

	work_list->pubkey[0] = POINT_NOT_COMPRESSED;

	bignum_convert_byte_order_u256(work_list->k, work_list->prikey);
	bignum_convert_byte_order_u256(work_list->x1, work_list->pubkey + 1);
	bignum_convert_byte_order_u256(work_list->y1,
				       work_list->pubkey + 1 + SM2_BYTE_LEN);

	return ret;
}

static void kdf(u8 *z, u32 zlen, u8 *k, int kLen)
{
	u8 digest[32];
	int i, t;
	u8 counter[4] = { 0x00, 0x00, 0x00, 0x01 }; // count = 1;
	struct sm3_context basectx, ctx;

	_software_sm3_init(&basectx);
	_software_sm3_update(&basectx, z, zlen);

	t = kLen >> 5; // t = kByteLen/32;
	for (i = 0; i < t; i++) {
		memcpy(&ctx, &basectx, sizeof(ctx));
		_software_sm3_update(&ctx, counter, 4);
		_software_sm3_final(&ctx, k + (i << 5));
		sm2_kdf_counter_add_one(counter);
	}

	kLen = kLen & 0x1F;
	if (kLen) {
		memcpy(&ctx, &basectx, sizeof(ctx));
		_software_sm3_update(&ctx, counter, 4);
		_software_sm3_final(&ctx, digest);
		memcpy(k + (t << 5), digest, kLen);
	}
}

static int _sm2_exchangekey_hw(volatile pke_reg_t *pke_reg, int role, u8 *dA,
			       u8 *PB, u8 *rA, uint8_t *RA, u8 *RB, u8 *ZA,
			       u8 *ZB, u32 kByteLen, u8 *KA, u8 *S1, u8 *SA)
{
	u32 x1[SM2_WORD_LEN], t[SM2_WORD_LEN], temp[SM2_WORD_LEN];
	u32 x2[SM2_WORD_LEN], y2[SM2_WORD_LEN], x3[SM2_WORD_LEN],
		y3[SM2_WORD_LEN];
	u32 xu[SM2_WORD_LEN], yu[SM2_WORD_LEN];
	u8 z[256];
	u8 dgst[32];
	struct sm3_context ctx[1];
	int work_step;

	//	print_buf("dA", dA, 32);
	//	print_buf("PB", PB, 64);
	//	print_buf("rA", rA, 32);
	//	print_buf("RA", RA, 64);
	//	print_buf("RB", RB, 64);
	//	print_buf("ZA", ZA, 32);
	//	print_buf("ZB", ZB, 32);

	//	if (role == 0) {//用户A为发起者
	//记w=⌈(⌈log2(n)⌉/2)⌉ −1, n为G的阶，这里w = 127
	//A1：用随机数发生器产生随机数rA ∈ [1, n-1]；
	//A2：计算椭圆曲线点RA = [rA]G=(x1;y1)；
	//A3：将RA发送给用户B；
	//A4：从RA中取出域元素x1，按本文本第1部分4.2.7给出的方法将x1的数据类型转换为整数，计
	//算x¯1 = 2^w + (x1&(2^w − 1))；
	bignum_convert_byte_order_u256(RA, x1);
	memset((u8 *)x1 + SM2_BYTE_LEN / 2, 0, SM2_BYTE_LEN / 2);
	x1[SM2_WORD_LEN / 2 - 1] |= 0x80000000;
	//		print_buf("x1", (u8 *)x1, 32);
	//A5：计算tA = (dA +x¯1 · rA)mod n；
	//t = x1*rA mod n
	bignum_convert_byte_order_u256(rA, temp);
	work_step = 0;
	do {
		work_step = sm2_modmul_internal(sm2p256v1_params.n, x1, temp, t,
						SM2_WORD_LEN, pke_reg,
						work_step);
	} while (work_step > 0);
	//t = (dA +x¯1 · rA) mod n
	bignum_convert_byte_order_u256(dA, temp);
	work_step = 0;
	do {
		work_step = sm2_modadd(sm2p256v1_params.n, temp, t, t,
				       SM2_WORD_LEN, pke_reg, work_step);
	} while (work_step > 0);
	//		print_buf("t", (u8 *)t, 32);
	//A6：验证RB是否满足椭圆曲线方程，若不满足则协商失败；否则从RB中取出域元素x2，按本文本
	//第1部分4.2.7给出的方法将x2 的数据类型转换为整数，计算x¯2 = 2w + (x2&(2w − 1))；
	bignum_convert_byte_order_u256(RB, x1);
	memset((u8 *)x1 + SM2_BYTE_LEN / 2, 0, SM2_BYTE_LEN / 2);
	x1[SM2_WORD_LEN / 2 - 1] |= 0x80000000;
	//		print_buf("x2", (u8 *)x1, 32);
	//A7：计算椭圆曲线点U = [h ·tA](PB + [x¯2]RB) = (xU;yU) ，若U是无穷远点，则A协商失败；否则按
	//本文本第1部分4.2.5和4.2.4给出的方法将xU、 yU 的数据类型转换为比特串
	//temp = tA * [x¯2]
	work_step = 0;
	do {
		work_step = sm2_modmul_internal(sm2p256v1_params.n, t, x1, temp,
						SM2_WORD_LEN, pke_reg,
						work_step);
	} while (work_step > 0);

	//[tA]PB + tA * [x¯2] *RB
	bignum_convert_byte_order_u256(PB, x2);
	bignum_convert_byte_order_u256(PB + 32, y2);
	bignum_convert_byte_order_u256(RB, x3);
	bignum_convert_byte_order_u256(RB + 32, y3);
	work_step = 0;
	do {
		work_step = sm2_point_mul_shamir(t, x2, y2, temp, x3, y3, xu,
						 yu, pke_reg, work_step);
	} while (work_step > 0);
	//		print_buf("xu", (u8 *)xu, 32);
	//		print_buf("yu", (u8 *)yu, 32);

	bignum_convert_byte_order_u256(xu, xu);
	bignum_convert_byte_order_u256(yu, yu);
	memcpy(z, xu, 32);
	memcpy(z + 32, yu, 32);
	if (role) {
		memcpy(z + 64, ZB, 32);
		memcpy(z + 96, ZA, 32);
	} else {
		memcpy(z + 64, ZA, 32);
		memcpy(z + 96, ZB, 32);
	}

	kdf(z, 128, KA, kByteLen);

	if (role) {
		_software_sm3_init(ctx);
		_software_sm3_update(ctx, (u8 *)xu, 32);
		_software_sm3_update(ctx, ZB, 32);
		_software_sm3_update(ctx, ZA, 32);
		_software_sm3_update(ctx, RB, 64);
		_software_sm3_update(ctx, RA, 64);
		_software_sm3_final(ctx, dgst);

		_software_sm3_init(ctx);
		_software_sm3_update(ctx, "\x03", 1);
		_software_sm3_update(ctx, (u8 *)yu, 32);
		_software_sm3_update(ctx, dgst, 32);
		_software_sm3_final(ctx, S1);

		_software_sm3_init(ctx);
		_software_sm3_update(ctx, "\x02", 1);
		_software_sm3_update(ctx, (u8 *)yu, 32);
		_software_sm3_update(ctx, dgst, 32);
		_software_sm3_final(ctx, SA);
	} else {
		_software_sm3_init(ctx);
		_software_sm3_update(ctx, (u8 *)xu, 32);
		_software_sm3_update(ctx, ZA, 32);
		_software_sm3_update(ctx, ZB, 32);
		_software_sm3_update(ctx, RA, 64);
		_software_sm3_update(ctx, RB, 64);
		_software_sm3_final(ctx, dgst);

		_software_sm3_init(ctx);
		_software_sm3_update(ctx, "\x02", 1);
		_software_sm3_update(ctx, (u8 *)yu, 32);
		_software_sm3_update(ctx, dgst, 32);
		_software_sm3_final(ctx, S1);

		_software_sm3_init(ctx);
		_software_sm3_update(ctx, "\x03", 1);
		_software_sm3_update(ctx, (u8 *)yu, 32);
		_software_sm3_update(ctx, dgst, 32);
		_software_sm3_final(ctx, SA);
	}

	return 0;
}

static int sm2_exchangekey(volatile pke_reg_t *pke_reg,
			   phytium_scto_sm2_work_list_t *work_list)
{
	u8 *data = &work_list->M[1];
	u8 *result = work_list->C;

	return _sm2_exchangekey_hw(pke_reg, (int)work_list->M[0], data,
				   data + 192, data + 96, data + 128,
				   data + 256, data + 320, data + 352,
				   work_list->decode_len, result,
				   result + work_list->decode_len,
				   result + work_list->decode_len + 32);
}

static int phytium_scto_sm2_work_thread(void *arg)
{
	int i, engine_index, queue_index, work_num, ret;
	phytium_scto_sm2_work_list_t *work_list[4] = { NULL };
	u64 time = jiffies;
	sqlist_t *node;
	u8 sm2_rng_pool[64] = { 0 };
	unsigned long flags;

	queue_index = 0;
	_software_sm3_init(&sm2_rng_ctx);
	while (1) {
		if (scto.suspending) {
			schedule();
			continue;
		}

		spin_lock(&scto.pke_lock);
		work_num = 0;

		for (engine_index = 0; engine_index < 1; engine_index++) {
			if (work_list[engine_index] == NULL) {
				for (i = 0; i < NR_CPUS; i++) {
					spin_lock_irqsave(
						&scto.sm2_queue[queue_index]
							 .queue_lock,
						flags);
					node = sqlist_get_head(
						&scto.sm2_queue[queue_index]
							 .list);
					spin_unlock_irqrestore(
						&scto.sm2_queue[queue_index]
							 .queue_lock,
						flags);

					queue_index++;
					if (queue_index >= NR_CPUS)
						queue_index = 0;

					if (node) {
						work_list[engine_index] = container_of(
							node,
							phytium_scto_sm2_work_list_t,
							node);
						break;
					}
				}
			}

			if (work_list[engine_index] == NULL)
				continue;

			work_num++;
			switch (work_list[engine_index]->work_type) {
			case SM2_SIGN:
				ret = sm2_sign(scto.pke_reg[engine_index],
					       work_list[engine_index],
					       sm2_rng_pool);
				if (ret <= 0) {
					akcipher_request_complete(
						work_list[engine_index]->req,
						ret);

					kfree(work_list[engine_index]);
					work_list[engine_index] = NULL;
				}
				break;
			case SM2_VERIFY:
				ret = sm2_verify(scto.pke_reg[engine_index],
						 work_list[engine_index]);
				if (ret <= 0) {
					akcipher_request_complete(
						work_list[engine_index]->req,
						ret);

					kfree(work_list[engine_index]);
					work_list[engine_index] = NULL;
				}
				break;
			case SM2_ENCRYPT:
				ret = sm2_encrypt(scto.pke_reg[engine_index],
						  work_list[engine_index],
						  sm2_rng_pool);
				if (ret <= 0) {
					akcipher_request_complete(
						work_list[engine_index]->req,
						ret);

					kfree(work_list[engine_index]);
					work_list[engine_index] = NULL;
				}
				break;
			case SM2_DECRYPT:
				ret = sm2_decrypt(scto.pke_reg[engine_index],
						  work_list[engine_index]);
				if (ret <= 0) {
					akcipher_request_complete(
						work_list[engine_index]->req,
						ret);

					kfree(work_list[engine_index]);
					work_list[engine_index] = NULL;
				}
				break;
			case SM2_KEY:
				/* reuse req?? */
				ret = sm2_genkey(scto.pke_reg[engine_index],
						 work_list[engine_index],
						 sm2_rng_pool);
				if (ret <= 0) {
					akcipher_request_complete(
						work_list[engine_index]->req,
						ret);
					work_list[engine_index] = NULL;
				}
				break;
			case SM2_EXCHANGEKEY: {
				ret = sm2_exchangekey(
					scto.pke_reg[engine_index],
					work_list[engine_index]);
				if (ret <= 0) {
					akcipher_request_complete(
						work_list[engine_index]->req,
						ret);
					work_list[engine_index] = NULL;
				}
			} break;
			default:
				pr_err("work_type[0x%x] error!\n",
				       work_list[engine_index]->work_type);
				akcipher_request_complete(
					work_list[engine_index]->req, -1);

				kfree(work_list[engine_index]);
				work_list[engine_index] = NULL;
				break;
			}
		}

		crypto_rng_update(sm2_rng_pool);
		spin_unlock(&scto.pke_lock);

		if (work_num == 0) {
			if (running == 0)
				break;
			time = jiffies;
			schedule_timeout_interruptible(HZ);
		}

		if (time + msecs_to_jiffies(100) < jiffies) {
			time = jiffies;
			schedule();
		}
	}

	running = 1;

	return 0;
}

static int phytium_scto_sm2_insert_queue(struct akcipher_request *req,
					 int work_type)
{
	struct crypto_akcipher *tfm = crypto_akcipher_reqtfm(req);
	phytium_scto_sm2_context_t *ctx = akcipher_tfm_ctx(tfm);
	int cpuid;
	unsigned long flags;
	struct phytium_scto_dev *dev = ctx->dev;
	phytium_scto_sm2_work_list_t *work_list =
		kmalloc(sizeof(phytium_scto_sm2_work_list_t), GFP_KERNEL);

	if (unlikely(work_list == NULL)) {
		pr_err("%s:kmalloc failed!\n", __func__);
		return -ENOMEM;
	}

	if (unlikely((req->base.complete == NULL) || (dev == NULL))) {
		pr_err("%s:complete[%p], dev[%p], error!\n", __func__,
		       req->base.complete, dev);
		kfree(work_list);
		return -EINVAL;
	}

	if ((work_type == SM2_SIGN) || (work_type == SM2_DECRYPT)) {
		if (unlikely(!ctx->private_key_valid)) {
			pr_err("%s:private_key invalid!\n", __func__);
			kfree(work_list);
			return -EINVAL;
		}
	}

	if ((work_type == SM2_VERIFY) || (work_type == SM2_ENCRYPT)) {
		if (unlikely(!ctx->public_key_valid)) {
			pr_err("%s:public_key invalid!\n", __func__);
			kfree(work_list);
			return -EINVAL;
		}
	}

	work_list->work_type = work_type;
	work_list->work_step = 0;
	work_list->req = req;
	work_list->ctx = ctx;

	cpuid = smp_processor_id();
	if (unlikely(cpuid >= NR_CPUS))
		cpuid = 0;

	spin_lock_irqsave(&dev->sm2_queue[cpuid].queue_lock, flags);
	sqlist_insert_tail(&dev->sm2_queue[cpuid].list, &work_list->node);
	spin_unlock_irqrestore(&dev->sm2_queue[cpuid].queue_lock, flags);

	wake_up_process(sm2_task);

	return -EINPROGRESS;
}

static int phytium_scto_sm2_sign(struct akcipher_request *req)
{
	return phytium_scto_sm2_insert_queue(req, SM2_SIGN);
}

static int phytium_scto_sm2_verify(struct akcipher_request *req)
{
	return phytium_scto_sm2_insert_queue(req, SM2_VERIFY);
}

static int phytium_scto_sm2_encrypt(struct akcipher_request *req)
{
	return phytium_scto_sm2_insert_queue(req, SM2_ENCRYPT);
}

static int phytium_scto_sm2_decrypt(struct akcipher_request *req)
{
	return phytium_scto_sm2_insert_queue(req, SM2_DECRYPT);
}

static int phytium_scto_sm2_set_pub_key(struct crypto_akcipher *tfm,
					const void *key, unsigned int keylen)
{
	phytium_scto_sm2_context_t *ctx = akcipher_tfm_ctx(tfm);
	u32 x[SM2_WORD_LEN], y[SM2_WORD_LEN];

	if (unlikely(key == NULL)) {
		pr_err("pubkey NULL!\n");
		return -1;
	}

	if (unlikely(((u8 *)key)[0] != POINT_NOT_COMPRESSED)) {
		pr_err("pubkey only support POINT_NOT_COMPRESSED(0x%x)!\n",
		       POINT_NOT_COMPRESSED);
		return -1;
	}
	if (unlikely(keylen != 1 + 2 * SM2_BYTE_LEN)) {
		pr_err("pubkey keylen:%u no support!\n", keylen);
		return -1;
	}

	bignum_convert_byte_order_u256(key + 1, x);
	bignum_convert_byte_order_u256(key + 1 + SM2_BYTE_LEN, y);

	memcpy((u8 *)ctx->x, (u8 *)x, SM2_BYTE_LEN);
	memcpy((u8 *)ctx->y, (u8 *)y, SM2_BYTE_LEN);
	ctx->public_key_valid = 1;

	return 0;
}

static int phytium_scto_sm2_set_priv_key(struct crypto_akcipher *tfm,
					 const void *key, unsigned int keylen)
{
	phytium_scto_sm2_context_t *ctx = akcipher_tfm_ctx(tfm);
	u32 d[SM2_WORD_LEN];

	if (unlikely(key == NULL)) {
		pr_err("prikey NULL!\n");
		return -1;
	}

	if (unlikely(keylen != SM2_BYTE_LEN)) {
		pr_err("prikey keylen:%u no support!\n", keylen);
		return -1;
	}

	bignum_convert_byte_order_u256(key, d);

	memcpy((u8 *)ctx->d, (u8 *)d, SM2_BYTE_LEN);
	ctx->private_key_valid = 1;

	return 0;
}

static unsigned int phytium_scto_sm2_max_size(struct crypto_akcipher *tfm)
{
	/* Unlimited max size */
	return PAGE_SIZE;
}

static int phytium_scto_sm2_init(struct crypto_akcipher *tfm)
{
	phytium_scto_sm2_context_t *ctx = akcipher_tfm_ctx(tfm);

	if (unlikely(ctx == NULL)) {
		pr_err("%s:ctx NULL!\n", __func__);
		return -1;
	}

	memset(ctx, 0, sizeof(phytium_scto_sm2_context_t));
	ctx->dev = &scto;

	return 0;
}

static void phytium_scto_sm2_exit(struct crypto_akcipher *tfm)
{
	phytium_scto_sm2_context_t *ctx = akcipher_tfm_ctx(tfm);

	if (unlikely(ctx == NULL)) {
		pr_err("%s:ctx NULL!\n", __func__);
		return;
	}

	memset(ctx, 0, sizeof(phytium_scto_sm2_context_t));
}

static struct akcipher_alg sm2_phytium_alg = {
	.sign = phytium_scto_sm2_sign,
	.verify = phytium_scto_sm2_verify,
	.encrypt = phytium_scto_sm2_encrypt,
	.decrypt = phytium_scto_sm2_decrypt,
	.set_pub_key = phytium_scto_sm2_set_pub_key,
	.set_priv_key = phytium_scto_sm2_set_priv_key,
	.max_size = phytium_scto_sm2_max_size,
	.init = phytium_scto_sm2_init,
	.exit = phytium_scto_sm2_exit,
	.base = {
		.cra_name = "sm2",
		.cra_driver_name = "sm2-phytium",
		.cra_priority = 300,
		.cra_module = THIS_MODULE,
		.cra_ctxsize = sizeof(phytium_scto_sm2_context_t),
	},
};

int phytium_scto_register_sm2(void)
{
	int ret;

	running = 1;
	sm2_task = kthread_run(phytium_scto_sm2_work_thread, NULL,
			       "scto_sm2_kthread");
	if (IS_ERR(sm2_task)) {
		pr_err("%s:kthread_run failed!\n", __func__);
		running = 0;
		return -1;
	}

	ret = crypto_register_akcipher(&sm2_phytium_alg);
	if (ret) {
		pr_err("%s:crypto_register_akcipher failed!\n", __func__);
		running = 0;
		while (1) {
			msleep(10);
			if (running)
				break;
		}
		return ret;
	}

	return 0;
}

void phytium_scto_unregister_sm2(void)
{
	crypto_unregister_akcipher(&sm2_phytium_alg);
	running = 0;
	while (1) {
		msleep(10);
		if (running)
			break;
	}
}

static int
phytium_scto_sm2_insert_queue2(struct akcipher_request *req,
			       phytium_scto_sm2_work_list_t *work_list,
			       int work_type)
{
	//phytium_scto_sm2_work_list_t *work_list = NULL;
	struct crypto_akcipher *tfm = crypto_akcipher_reqtfm(req);
	phytium_scto_sm2_context_t *ctx = akcipher_tfm_ctx(tfm);
	struct phytium_scto_dev *dev = ctx->dev;
	int cpuid;
	unsigned long flags;

	if (req == NULL) {
		pr_err("%s: invalid null req!\n", __func__);
		return -EINVAL;
	}

	if ((work_type != SM2_KEY) && (work_type != SM2_EXCHANGEKEY) &&
	    (work_type != SM2_PUBKEY)) {
		pr_err("%s: invalid work_type!\n", __func__);

		return -EINVAL;
	}

	if (unlikely((req->base.complete == NULL) || (dev == NULL))) {
		pr_err("%s:complete[%p], dev[%p], error!\n", __func__,
		       req->base.complete, dev);
		return -EINVAL;
	}

	work_list->work_type = work_type;
	work_list->work_step = 0;
	work_list->req = req;
	work_list->ctx = ctx;

	if (work_type == SM2_PUBKEY) {
		work_type = SM2_KEY;
		work_list->work_type = work_type;
		work_list->work_step = 3 << 8;
	}

	cpuid = smp_processor_id();
	if (unlikely(cpuid >= NR_CPUS))
		cpuid = 0;

	spin_lock_irqsave(&dev->sm2_queue[cpuid].queue_lock, flags);
	sqlist_insert_tail(&dev->sm2_queue[cpuid].list, &work_list->node);
	spin_unlock_irqrestore(&dev->sm2_queue[cpuid].queue_lock, flags);

	wake_up_process(sm2_task);

	return -EINPROGRESS;
}

static int phytium_scto_sm2_genkey(struct akcipher_request *req,
				   phytium_scto_sm2_work_list_t *work_list)
{
	return phytium_scto_sm2_insert_queue2(req, work_list, SM2_KEY);
}

static long _sm2_genkey(u8 privkey[32], u8 pubkey[65])
{
	struct crypto_wait cwait;
	struct crypto_akcipher *tfm;
	struct akcipher_request *req;
	struct scatterlist sign_e_sg, null_sg;
	u8 null[64];
	int ret;
	phytium_scto_sm2_work_list_t *work_list;

	work_list = kmalloc(sizeof(phytium_scto_sm2_work_list_t), GFP_KERNEL);
	if (unlikely(work_list == NULL)) {
		pr_err("%s:kmalloc failed!\n", __func__);
		return -ENOMEM;
	}
	memset(work_list, 0, sizeof(phytium_scto_sm2_work_list_t));
	work_list->prikey = privkey;
	work_list->pubkey = pubkey;

	pubkey[0] = POINT_NOT_COMPRESSED;

	tfm = crypto_alloc_akcipher("sm2", 0, 0);
	if (IS_ERR(tfm)) {
		pr_err("crypto_alloc_akcipher failed!\n");
		kfree(work_list);
		return -1;
	}

	ret = -ENOMEM;
	req = akcipher_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_err("akcipher_request_alloc failed!\n");
		goto error_free_tfm;
	}

	ret = crypto_akcipher_set_pub_key(tfm, pubkey, 65);
	if (ret) {
		pr_err("crypto_akcipher_set_pub_key failed!\n");
		goto error_free_req;
	}

	sg_init_one(&sign_e_sg, null, 64);
	sg_init_one(&null_sg, null, 32);
	akcipher_request_set_crypt(req, &sign_e_sg, &null_sg, 0, 32);
	crypto_init_wait(&cwait);
	akcipher_request_set_callback(
		req, CRYPTO_TFM_REQ_MAY_BACKLOG | CRYPTO_TFM_REQ_MAY_SLEEP,
		crypto_req_done, &cwait);

	ret = crypto_wait_req(phytium_scto_sm2_genkey(req, work_list), &cwait);
	if (ret)
		goto error_free_req;

error_free_req:
	akcipher_request_free(req);
error_free_tfm:
	crypto_free_akcipher(tfm);

	kfree(work_list);

	return ret;
}

static long _sm2_pubkey(u8 privkey[32], u8 pubkey[65])
{
	struct crypto_wait cwait;
	struct crypto_akcipher *tfm;
	struct akcipher_request *req;
	struct scatterlist sign_e_sg, null_sg;
	u8 null[64];
	int ret;
	phytium_scto_sm2_work_list_t *work_list;

	work_list = kmalloc(sizeof(phytium_scto_sm2_work_list_t), GFP_KERNEL);
	if (unlikely(work_list == NULL)) {
		pr_err("%s:kmalloc failed!\n", __func__);
		return -ENOMEM;
	}
	memset(work_list, 0, sizeof(phytium_scto_sm2_work_list_t));
	work_list->prikey = privkey;
	work_list->pubkey = pubkey;

	pubkey[0] = POINT_NOT_COMPRESSED;

	tfm = crypto_alloc_akcipher("sm2", 0, 0);
	if (IS_ERR(tfm)) {
		pr_err("crypto_alloc_akcipher failed!\n");
		kfree(work_list);
		return -1;
	}

	ret = -ENOMEM;
	req = akcipher_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_err("akcipher_request_alloc failed!\n");
		goto error_free_tfm;
	}

	ret = crypto_akcipher_set_pub_key(tfm, pubkey, 65);
	if (ret) {
		pr_err("crypto_akcipher_set_pub_key failed!\n");
		goto error_free_req;
	}

	sg_init_one(&sign_e_sg, null, 64);
	sg_init_one(&null_sg, null, 32);
	akcipher_request_set_crypt(req, &sign_e_sg, &null_sg, 0, 32);
	crypto_init_wait(&cwait);
	akcipher_request_set_callback(
		req, CRYPTO_TFM_REQ_MAY_BACKLOG | CRYPTO_TFM_REQ_MAY_SLEEP,
		crypto_req_done, &cwait);

	ret = crypto_wait_req(phytium_scto_sm2_insert_queue2(req, work_list,
							     SM2_PUBKEY),
			      &cwait);
	if (ret)
		goto error_free_req;

error_free_req:
	akcipher_request_free(req);
error_free_tfm:
	crypto_free_akcipher(tfm);

	kfree(work_list);

	return ret;
}

static int phytium_scto_sm2_exchangekey(struct akcipher_request *req,
					phytium_scto_sm2_work_list_t *work_list)
{
	return phytium_scto_sm2_insert_queue2(req, work_list, SM2_EXCHANGEKEY);
}

static long _sm2_exchangekey(u8 *key_z, u8 *dst, uint32_t keylen)
{
	struct crypto_wait cwait;
	struct crypto_akcipher *tfm;
	struct akcipher_request *req;
	struct scatterlist sign_e_sg, null_sg;
	u8 null[64];
	int ret;
	phytium_scto_sm2_work_list_t *work_list;

	work_list = kmalloc(sizeof(phytium_scto_sm2_work_list_t), GFP_KERNEL);
	if (unlikely(work_list == NULL)) {
		pr_err("%s:kmalloc failed!\n", __func__);
		return -ENOMEM;
	}
	memset(work_list, 0, sizeof(phytium_scto_sm2_work_list_t));
	work_list->M = key_z;
	work_list->C = dst;
	work_list->decode_len = keylen;

	tfm = crypto_alloc_akcipher("sm2", 0, 0);
	if (IS_ERR(tfm)) {
		pr_err("crypto_alloc_akcipher failed!\n");
		kfree(work_list);
		return -1;
	}

	ret = -ENOMEM;
	req = akcipher_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_err("akcipher_request_alloc failed!\n");
		goto error_free_tfm;
	}

	sg_init_one(&sign_e_sg, null, 64);
	sg_init_one(&null_sg, null, 32);
	akcipher_request_set_crypt(req, &sign_e_sg, &null_sg, 0, 32);
	crypto_init_wait(&cwait);
	akcipher_request_set_callback(
		req, CRYPTO_TFM_REQ_MAY_BACKLOG | CRYPTO_TFM_REQ_MAY_SLEEP,
		crypto_req_done, &cwait);

	ret = crypto_wait_req(phytium_scto_sm2_exchangekey(req, work_list),
			      &cwait);
	if (ret)
		goto error_free_req;

error_free_req:
	akcipher_request_free(req);
error_free_tfm:
	crypto_free_akcipher(tfm);

	kfree(work_list);

	return ret;
}

static long _sm2_sign(u8 *priv_key, u8 *e, u8 *sign, u32 *sign_len)
{
	struct crypto_wait cwait;
	struct crypto_akcipher *tfm;
	struct akcipher_request *req;
	struct scatterlist e_sg, sign_sg;
	int ret;

	tfm = crypto_alloc_akcipher("sm2", 0, 0);
	if (IS_ERR(tfm)) {
		pr_err("crypto_alloc_akcipher failed!\n");
		return -1;
	}

	ret = -ENOMEM;
	req = akcipher_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_err("akcipher_request_alloc failed!\n");
		goto error_free_tfm;
	}

	ret = crypto_akcipher_set_priv_key(tfm, priv_key, 32);
	if (ret) {
		pr_err("crypto_akcipher_set_priv_key failed!\n");
		goto error_free_req;
	}

	sg_init_one(&e_sg, e, 32);
	sg_init_one(&sign_sg, sign, *sign_len);
	akcipher_request_set_crypt(req, &e_sg, &sign_sg, 32, *sign_len);
	crypto_init_wait(&cwait);
	akcipher_request_set_callback(
		req, CRYPTO_TFM_REQ_MAY_BACKLOG | CRYPTO_TFM_REQ_MAY_SLEEP,
		crypto_req_done, &cwait);

	ret = crypto_wait_req(crypto_akcipher_sign(req), &cwait);
	if (ret)
		goto error_free_req;

	*sign_len = req->dst_len;

error_free_req:
	akcipher_request_free(req);
error_free_tfm:
	crypto_free_akcipher(tfm);

	return ret;
}

static long _sm2_verify(u8 *pub_key, u8 *sign_e, u32 len)
{
	struct crypto_wait cwait;
	struct crypto_akcipher *tfm;
	struct akcipher_request *req;
	struct scatterlist sign_e_sg, null_sg;
	u8 null[64];
	int ret;

	tfm = crypto_alloc_akcipher("sm2", 0, 0);
	if (IS_ERR(tfm)) {
		pr_err("crypto_alloc_akcipher failed!\n");
		return -1;
	}

	ret = -ENOMEM;
	req = akcipher_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_err("akcipher_request_alloc failed!\n");
		goto error_free_tfm;
	}

	ret = crypto_akcipher_set_pub_key(tfm, pub_key, 65);
	if (ret) {
		pr_err("crypto_akcipher_set_pub_key failed!\n");
		goto error_free_req;
	}

	sg_init_one(&sign_e_sg, sign_e, len);
	sg_init_one(&null_sg, null, 32);
	akcipher_request_set_crypt(req, &sign_e_sg, &null_sg, len - 32, 32);
	crypto_init_wait(&cwait);
	akcipher_request_set_callback(
		req, CRYPTO_TFM_REQ_MAY_BACKLOG | CRYPTO_TFM_REQ_MAY_SLEEP,
		crypto_req_done, &cwait);

	ret = crypto_wait_req(crypto_akcipher_verify(req), &cwait);
	if (ret)
		goto error_free_req;

error_free_req:
	akcipher_request_free(req);
error_free_tfm:
	crypto_free_akcipher(tfm);

	return ret;
}

static long _sm2_enc(u8 *pub_key, u8 *src, u8 *dst, u32 src_len, u32 *dst_len)
{
	struct crypto_wait cwait;
	struct crypto_akcipher *tfm;
	struct akcipher_request *req;
	struct scatterlist src_sg, dst_sg;
	int ret;

	tfm = crypto_alloc_akcipher("sm2", 0, 0);
	if (IS_ERR(tfm)) {
		pr_err("crypto_alloc_akcipher failed!\n");
		return -1;
	}

	ret = -ENOMEM;
	req = akcipher_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_err("akcipher_request_alloc failed!\n");
		goto error_free_tfm;
	}

	ret = crypto_akcipher_set_pub_key(tfm, pub_key, 65);
	if (ret) {
		pr_err("crypto_akcipher_set_pub_key failed!\n");
		goto error_free_req;
	}

	sg_init_one(&src_sg, src, src_len);
	sg_init_one(&dst_sg, dst, *dst_len);
	akcipher_request_set_crypt(req, &src_sg, &dst_sg, src_len, *dst_len);
	crypto_init_wait(&cwait);
	akcipher_request_set_callback(
		req, CRYPTO_TFM_REQ_MAY_BACKLOG | CRYPTO_TFM_REQ_MAY_SLEEP,
		crypto_req_done, &cwait);

	ret = crypto_wait_req(crypto_akcipher_encrypt(req), &cwait);
	if (ret)
		goto error_free_req;

	*dst_len = req->dst_len;

error_free_req:
	akcipher_request_free(req);
error_free_tfm:
	crypto_free_akcipher(tfm);

	return ret;
}

static long _sm2_dec(u8 *priv_key, u8 *src, u8 *dst, u32 src_len, u32 *dst_len)
{
	struct crypto_wait cwait;
	struct crypto_akcipher *tfm;
	struct akcipher_request *req;
	struct scatterlist src_sg, dst_sg;
	int ret;

	tfm = crypto_alloc_akcipher("sm2", 0, 0);
	if (IS_ERR(tfm)) {
		pr_err("crypto_alloc_akcipher failed!\n");
		return -1;
	}

	ret = -ENOMEM;
	req = akcipher_request_alloc(tfm, GFP_KERNEL);
	if (!req) {
		pr_err("akcipher_request_alloc failed!\n");
		goto error_free_tfm;
	}

	ret = crypto_akcipher_set_priv_key(tfm, priv_key, 32);
	if (ret) {
		pr_err("crypto_akcipher_set_priv_key failed!\n");
		goto error_free_req;
	}

	sg_init_one(&src_sg, src, src_len);
	sg_init_one(&dst_sg, dst, *dst_len);
	akcipher_request_set_crypt(req, &src_sg, &dst_sg, src_len, *dst_len);
	crypto_init_wait(&cwait);
	akcipher_request_set_callback(
		req, CRYPTO_TFM_REQ_MAY_BACKLOG | CRYPTO_TFM_REQ_MAY_SLEEP,
		crypto_req_done, &cwait);

	ret = crypto_wait_req(crypto_akcipher_decrypt(req), &cwait);
	if (ret)
		goto error_free_req;

	*dst_len = req->dst_len;

error_free_req:
	akcipher_request_free(req);
error_free_tfm:
	crypto_free_akcipher(tfm);

	return ret;
}

struct sm2_info {
	u32 type;
	u32 src_len;
	u32 dst_len;
	u8 *src;
	u8 *dst;
	u8 *key;
};

long scto_sm2(unsigned int cmd, unsigned long arg)
{
	struct sm2_info info;
	long ret;
	u8 *src, *dst, *key;

	if (unlikely(cmd != SCTO_SM2)) {
		pr_err("cmd[0x%x] not support!\n", cmd);
		return -1;
	}

	ret = copy_from_user(&info, (void *)arg, sizeof(info));
	if (ret) {
		pr_err("copy_from_user[arg:0x%lx] failed!\n", arg);
		return ret;
	}

	switch (info.type) {
	case ALG_SM2_PUBKEY: {
		u8 prikey[32];
		u8 pubkey[65];

		ret = copy_from_user(prikey, info.key, 32);
		if (ret) {
			pr_err("copy_from_user[key:%p] failed!\n", info.key);
			return ret;
		}

		ret = _sm2_pubkey(prikey, pubkey);
		if (ret) {
			pr_err("pub key error\r\n");
			return ret;
		}

		ret = copy_to_user(info.key + 32, pubkey, 65);
		if (ret) {
			pr_err("copy_to_user[dst:%p] failed!\n", pubkey);

			return ret;
		}

		return ret;
	}
	case ALG_SM2_KEY: {
		u8 prikey[32];
		u8 pubkey[65];

		ret = _sm2_genkey(prikey, pubkey);
		if (ret) {
			pr_err("gen key error\r\n");
			return ret;
		}

		ret = copy_to_user(info.key, prikey, 32);
		if (ret) {
			pr_err("copy_to_user[dst:%p] failed!\n", prikey);

			return ret;
		}

		ret = copy_to_user(info.key + 32, pubkey, 65);
		if (ret) {
			pr_err("copy_to_user[dst:%p] failed!\n", pubkey);

			return ret;
		}

		return ret;
	}
	case ALG_SM2_SIGN:
		if (info.src_len != 32) {
			pr_err("ALG_SM2_SIGN srclen[0x%x] err!\n",
			       info.src_len);
			return -EINVAL;
		}
		src = kmalloc(info.src_len + info.dst_len + 256, GFP_KERNEL);
		if (src == NULL)
			return -ENOMEM;

		ret = copy_from_user(src, info.src, info.src_len);
		if (ret) {
			pr_err("copy_from_user[src:%p] failed!\n", info.src);
			kfree(src);
			return ret;
		}
		dst = src + info.src_len + 64;
		key = dst + info.dst_len + 64;
		ret = copy_from_user(key, info.key, 32);
		if (ret) {
			pr_err("copy_from_user[key:%p] failed!\n", info.key);
			kfree(src);
			return ret;
		}

		ret = _sm2_sign(key, src, dst, &info.dst_len);
		if (ret) {
			pr_err("sm2_sign[ret:0x%lx] failed!\n", ret);
			kfree(src);
			return ret;
		}

		ret = copy_to_user(info.dst, dst, info.dst_len);
		if (ret) {
			pr_err("copy_to_user[dst:%p] failed!\n", info.dst);
			kfree(src);
			return ret;
		}
		ret = copy_to_user((void *)arg, &info, sizeof(info));
		if (ret) {
			pr_err("copy_to_user[arg:0x%lx] failed!\n", arg);
			kfree(src);
			return ret;
		}
		kfree(src);
		return ret;
	case ALG_SM2_VERIFY:
		src = kmalloc(info.src_len + 256, GFP_KERNEL);
		if (src == NULL)
			return -ENOMEM;

		ret = copy_from_user(src, info.src, info.src_len);
		if (ret) {
			pr_err("copy_from_user[src:%p] failed!\n", info.src);
			kfree(src);
			return ret;
		}
		key = src + info.src_len + 64;
		ret = copy_from_user(key, info.key, 65);
		if (ret) {
			pr_err("copy_from_user[key:%p] failed!\n", info.key);
			kfree(src);
			return ret;
		}

		ret = _sm2_verify(key, src, info.src_len);
		if (ret) {
			pr_err("sm2_verify[ret:0x%lx] failed!\n", ret);
			kfree(src);
			return ret;
		}

		kfree(src);
		return ret;
	case ALG_SM2_ENC:
		src = kmalloc(info.src_len + info.dst_len + 256, GFP_KERNEL);
		if (src == NULL)
			return -ENOMEM;

		ret = copy_from_user(src, info.src, info.src_len);
		if (ret) {
			pr_err("copy_from_user[src:%p] failed!\n", info.src);
			kfree(src);
			return ret;
		}
		dst = src + info.src_len + 64;
		key = dst + info.dst_len + 64;
		ret = copy_from_user(key, info.key, 65);
		if (ret) {
			pr_err("copy_from_user[key:%p] failed!\n", info.key);
			kfree(src);
			return ret;
		}

		ret = _sm2_enc(key, src, dst, info.src_len, &info.dst_len);
		if (ret) {
			pr_err("sm2_enc[ret:0x%lx] failed!\n", ret);
			kfree(src);
			return ret;
		}

		ret = copy_to_user(info.dst, dst, info.dst_len);
		if (ret) {
			pr_err("copy_to_user[dst:%p] failed!\n", info.dst);
			kfree(src);
			return ret;
		}
		ret = copy_to_user((void *)arg, &info, sizeof(info));
		if (ret) {
			pr_err("copy_to_user[arg:0x%lx] failed!\n", arg);
			kfree(src);
			return ret;
		}

		kfree(src);
		return ret;
	case ALG_SM2_DEC:
		src = kmalloc(info.src_len + info.dst_len + 256, GFP_KERNEL);
		if (src == NULL)
			return -ENOMEM;

		ret = copy_from_user(src, info.src, info.src_len);
		if (ret) {
			pr_err("copy_from_user[src:%p] failed!\n", info.src);
			kfree(src);
			return ret;
		}
		dst = src + info.src_len + 64;
		key = dst + info.dst_len + 64;
		ret = copy_from_user(key, info.key, 32);
		if (ret) {
			pr_err("copy_from_user[key:%p] failed!\n", info.key);
			kfree(src);
			return ret;
		}

		ret = _sm2_dec(key, src, dst, info.src_len, &info.dst_len);
		if (ret) {
			pr_err("sm2_dec[ret:0x%lx] failed!\n", ret);
			kfree(src);
			return ret;
		}

		ret = copy_to_user(info.dst, dst, info.dst_len);
		if (ret) {
			pr_err("copy_to_user[dst:%p] failed!\n", info.dst);
			kfree(src);
			return ret;
		}
		ret = copy_to_user((void *)arg, &info, sizeof(info));
		if (ret) {
			pr_err("copy_to_user[arg:0x%lx] failed!\n", arg);
			kfree(src);
			return ret;
		}

		kfree(src);
		return ret;
	case ALG_SM2_EXCHANGEKEY: {
		/*check length*/
		if (info.src_len < 385) {
			pr_err("Invalid src len %u", info.src_len);
			return -EINVAL;
		}

		if (info.dst_len < 65) {
			pr_err("Invalid dst len %u", info.dst_len);
			return -EINVAL;
		}

		src = kmalloc(info.src_len, GFP_KERNEL);
		if (src == NULL)
			return -ENOMEM;

		dst = kmalloc(info.dst_len, GFP_KERNEL);
		if (dst == NULL) {
			kfree(src);
			return -ENOMEM;
		}

		ret = copy_from_user(src, info.src, info.src_len);
		if (ret) {
			pr_err("copy_from_user src failed!\n");
			kfree(src);
			kfree(dst);
			return ret;
		}

		ret = _sm2_exchangekey(src, dst, info.dst_len - 64);
		if (!ret) {
			ret = copy_to_user(info.dst, dst, info.dst_len);
			if (ret)
				pr_err("copy_to_user dst failed!\n");
		}

		kfree(src);
		kfree(dst);

		return ret;
	}
	default:
		return -1;
	}
}

#endif
